import numpy as np

from retrieval.instruments.spectral_imager import MultiSensorImager
import retrieval.instruments.sensors.sensor as sen
import retrieval.instruments.optics.optics as opt


class uCATSModel(MultiSensorImager):
    """
    Instrument model of micro-CATS, the Canadian Atmospheric Tomography System. 
    Designed to be a limb viewing instrument with 7 independent lines of sight. 
    Each line of sight can measure different wavelengths with variable point spread
    functions across the spectrum.
    
    The model is built on the :class:`~instrument.MultiSensorImager` class and uses
    basic optics, which assume gaussian point spread functions. If the lines of 
    sight are oversampled all samples are assumed to have equal weight.
    """
    def __init__( self, wavel_nm, los, vertical=None, obs=None ):
        """
        Parameterss
        ------
        wavel_nm : list(np.arrays)
            A list of wavelength (nanometers) arrays. One array per sensor.
        los : list(np.arrays)
            A list of line of sight arrays (nx3). One array per sensor where n
            is the number of sub exposures. Each sensor can have a different number
            of sub exposures.
        vert : list(np.arrays)
            Same format as los, describing the vertical orientation of the sensors
            in geographic coordinates (xyz) in case of spacial oversampling
        obs : list(np.arrays)
            Same format as los, describing the observer position. This is usefull
            when using the instrument to setup a radiative transfer model.
        """
        super().__init__(wavel_nm, los, vertical, obs)
        self.initialize_sensors_and_optics()

    def initialize_sensors_and_optics(self):
        """
        Sets up the sensors, optics and high resolution geoemetries
        """
        #TODO: find a better way to setup sensors, preferably with a variable number of them
        #setup blue-green sensors
        self.sensor[0],self.optics[0]  = self.uCATS_sensor(vertRes_km = 0.2, horRes_km = 5,  wavel_nm=np.array([0, 10000]), wavePSF = np.array([0.6, 0.6]))
        self.sensor[1],self.optics[1]  = self.uCATS_sensor(vertRes_km = 0.2, horRes_km = 9,  wavel_nm=np.array([0, 10000]), wavePSF = np.array([0.6, 0.6]))
        self.sensor[2],self.optics[2]  = self.uCATS_sensor(vertRes_km = 0.2, horRes_km = 13, wavel_nm=np.array([0, 10000]), wavePSF = np.array([0.6, 0.6]))
        self.sensor[3],self.optics[3]  = self.uCATS_sensor(vertRes_km = 0.2, horRes_km = 17, wavel_nm=np.array([0, 10000]), wavePSF = np.array([0.6, 0.6]))
        
        #setup green-red sensors
        self.sensor[4],self.optics[4]  = self.uCATS_sensor(vertRes_km = 0.2, horRes_km = 10, wavel_nm=np.array([0, 10000]), wavePSF = np.array([1.1, 1.1]))
        self.sensor[5],self.optics[5]  = self.uCATS_sensor(vertRes_km = 0.2, horRes_km = 17, wavel_nm=np.array([0, 10000]), wavePSF = np.array([1.1, 1.1]))
        self.sensor[6],self.optics[6]  = self.uCATS_sensor(vertRes_km = 0.2, horRes_km = 17, wavel_nm=np.array([0, 10000]), wavePSF = np.array([1.1, 1.1]))
        
        for i in range(0,self.num_sensors):
            self.num_pixels[i]     = self.sensor[i].num_pixels

        self.setup_instrument( )
        
    def uCATS_sensor(self, vertRes_km = 0.2, horRes_km = 10, wavel_nm=np.array([0, 10000]),wavePSF = np.array([0.6, 0.6])):
        """
        Create a CATS sensor. Assumed to have one line of sight
        
        Parameters
        ----------
        vertRes_km : double
            vertical resolution of a line of sight at the tangent point in kilometers
        horRes_km: double
            horizontal resolution of a line sight at the tangent point in kilometers
        wavel_nm : np.array
            an array of wavelengths in nanometers that the sensor measures
        wavePSF : np.array
            Gaussian point spread function of the instrument (standard deviation) at the wavelengths specified by wavel_nm
        
        Returns
        -------
        sensor : pointer to sensor class
        
        optics : pointer to optics class
        """
        Re         = 6372 #radius of the earth (approx)
        satAlt_km  = 600  #orbital altitude
        tanAlt_km  = 25   #tangent point altitude
        
        #convert the resolution from km to radians
        vertRes = np.arctan( vertRes_km/np.sqrt((Re+satAlt_km)**2-(Re+tanAlt_km)**2) )
        horRes  = np.arctan( horRes_km/ np.sqrt((Re+satAlt_km)**2-(Re+tanAlt_km)**2) )
            
        #setup the sensors and optics
        sensor = sen.NxMSensor(1,1,0,0)
        optics = opt.BasicOptics(vertRes=vertRes,horRes=horRes,wavel_nm=np.array([0, 10000]),wavePSF = np.array([0.35, 0.35]),vertSamples=self.vertSamples,horSamples=self.horSamples,waveSamples=self.waveSamples)
            
        return sensor, optics

class uCATSModel_scan(object):
    """
    A class designed to provide radiances of a set of uCATs observations in a 
    convenient, retrieval ready format. 
    """
    def __init__( self, wavel_nm, los, vertical=None, obs = None):
        """
        Inputs are the same as the uCATSModel, however now each uCATS observation 
        set is provided as a list. 
        
        Observation (list)
            -- Sensor (list)
                -- Line of sight (np.array)
        
        Parameters
        ------
        wavel_nm : list(list(np.arrays))
            A list of lists of wavelengths (nanometers) arrays. One array per sensor.
        los : list(list(np.arrays))
            A list of lists of line of sight arrays (nx3). One list per observation
            (exposure). One array per sensor where n is the number of sub exposures. 
            Each sensor can have a different number of sub exposures.
        vert : list(list(np.arrays))
            Same format as los, describing the vertical orientation of the sensors
            in geographic coordinates (xyz) in case of spacial oversampling
        obs : list(list(np.arrays))
            Same format as los, describing the observer position. This is usefull
            when using the instrument to setup a radiative transfer model.
        """
        #the number of observations
        self.num_exposures = len(los)
        
        self.los     = [0,]*self.num_exposures
        self.obs     = [0,]*self.num_exposures
        self.exp_idx = [0,]*self.num_exposures
        self.ucats   = [0,]*self.num_exposures
        
        #set vetical direction as default if it is not passed in
        if vertical is None:
            vertical = [0,]*self.num_exposures            
            for i in range(0,self.num_exposures):
                vertical[i] = None
                
        if obs is None:
            obs = [0,]*self.num_exposures            
            for i in range(0,self.num_exposures):
                obs[i] = None
        
        for i in range(0,self.num_exposures):
            self.ucats[i] = uCATSModel(wavel_nm, los[i], vertical[i], obs[i])
            self.ucats[i].initialize_sensors_and_optics()
            
    def convert_radiance_to_l1( self, radiance ):
        """
        Converts hi resolution radiances to a level one measurement.
        
        Paramters
        ---------
        radiance : np.array
            a line of sight x wavelength array        
            
        Returns
        -------
        measurement : np.array
            returns the measurement as a line of sight x wavelength array. Lines
            of sight are grouped by sensor. If a wavelength or line of sight
            is not measured by the sensor it is filled with a nan
        """
        self.measurement = [0,]*self.ucats[0].num_sensors
        self.tanalt      = [0,]*self.ucats[0].num_sensors
        self.pixel_los   = [0,]*self.ucats[0].num_sensors
        self.pixel_obs   = [0,]*self.ucats[0].num_sensors
        
        #initialize measurements
        wv_meas = np.array([])
        for s in range(0,self.ucats[0].num_sensors):
            self.measurement[s] = np.zeros((0,len(self.ucats[0].wavel_nm[s])))
            wv_meas = np.append(wv_meas,self.ucats[0].wavel_nm[s])
            
        wv_meas = np.unique(wv_meas)
        for i in range(0,self.num_exposures):   #find level1 data for each exposure
            meas = self.ucats[i].convert_radiance_to_l1(radiance[self.exp_idx[i]])
            for s in range(0,self.ucats[0].num_sensors):    #stack each sensors l1 data across exposures
                if i is 0:
                    self.pixel_los[s] = np.zeros((0,3))
                    self.pixel_obs[s] = np.zeros((0,3))
                    
                self.measurement[s] = np.vstack((self.measurement[s], meas[s]))
                self.pixel_los[s]   = np.vstack((self.pixel_los[s], self.ucats[i].pixel_los[s]))
                self.pixel_obs[s]   = np.vstack((self.pixel_obs[s], self.ucats[i].pixel_obs[s]))
        
        meas_array = np.zeros((0,len(wv_meas)))
        
        for s in range(0,self.ucats[0].num_sensors):
            good_wave = np.where(np.in1d(wv_meas,self.ucats[0].wavel_nm[s]))[0]
            temp = np.ones((np.size(self.measurement[s],0),len(wv_meas)))*np.nan
            temp[:,good_wave] = self.measurement[s]
            meas_array = np.vstack((meas_array,temp))
            
        los_array  = np.vstack(self.pixel_los)
        obs_array  = np.vstack(self.pixel_obs)
        
        #TODO: have a seperate call for measurement geometry
        return meas_array, los_array, obs_array, wv_meas
            
    def radtran_wavelengths(self):
        wv = self.ucats[0].radtran_wavelengths()
        return wv
        
    def radtran_geometry(self):
        
        #wv  = self.radtran_wavelengths()
        
        los = [0,]*self.num_exposures
        obs = [0,]*self.num_exposures
        mjd = [0,]*self.num_exposures
        for i in range(0,self.num_exposures):
            los[i],obs[i], mjd[i] = self.ucats[i].radtran_geometry()
            
        los_all = np.zeros((0,3))
        obs_all = np.zeros((0,3))
        mjd_all = np.zeros((0,1))
        self.exp_idx = [0,]*self.num_exposures
        start = 0
        for i in range(0,self.num_exposures):
            los_all = np.vstack((los_all, los[i]))
            obs_all = np.vstack((obs_all, obs[i]))
            mjd_all = np.vstack((mjd_all, mjd[i]))
            end = start+np.size(los[i],0)
            self.exp_idx[i] = np.arange(start,end)
            start = end
        
        
        return los_all, obs_all, mjd_all