import numpy as np
import xarray as xr

# Maximum altitude to return a cloud at
MAX_CLOULD_ALT_KM = 30


def vertical_log_gradient(heights_km: np.array, rad_array: np.array) -> np.array:
    """
    Calculates the vertical log gradient of a radiance profile
    Parameters
    ----------
    heights_km: np.array
        Vertical tangent heights in m
    rad_array: np.array
        Radiance

    Returns
    -------
    vertical_log_gradient
    """
    return np.gradient(np.log(rad_array), heights_km[1] - heights_km[0])


def vertical_log_gradient_difference(radiance: xr.DataArray, low_wavel_nm: float, high_wavel_nm: float) -> (
        np.array, np.array):
    """
    Returns the difference between the vertical log gradient of the radiance at a short wavelength (LOW_WAVEL_NM)
    and a high wavelength (HIGH_WAVEL_NM)
    Parameters
    ----------
    radiance: xarray.DataArray
        Radiance dataarray with coords 'tangent_height' and 'wavelength'
    low_wavel_nm: float
        Lower wavelength to calculate the vertical log gradient at
    high_wavel_nm: float
        Higher wavelength to calculate the vertical log gradient at

    Returns
    -------
    tangent_height
    vertical_log_gradient_difference

    """
    low_wavel_rad = radiance.sel(wavelength=low_wavel_nm, method='nearest')
    high_wavel_rad = radiance.sel(wavelength=high_wavel_nm, method='nearest')

    th = low_wavel_rad['tangent_height'].values

    v_log_grad_diff = vertical_log_gradient(th, low_wavel_rad.values) - vertical_log_gradient(th, high_wavel_rad.values)

    if np.all(np.diff(th) >= 0):
        return th, v_log_grad_diff
    elif np.all(np.diff(th) <= 0):
        return th[::-1], v_log_grad_diff[::-1]
    else:
        raise ValueError("Tangent heights are not monotonic")


def cloud_alt_km(radiance: xr.DataArray, threshold: float=0.15, low_wavel_nm: float = 674,
                 high_wavel_nm: float = 868) -> float:
    """
    Returns the highest altitude cloud in km.  Clouds above the global variable MAX_CLOUD_ALT_KM will not be
    returned.  The algorithm is taken from "A new algorithm for detecting cloud height using OMPS/LP measurements"
    (Chen 2016).  The default values for low/high wavelengths and the threshold are the same used in the paper.
    Parameters
    ----------
    radiance: xarray.DataArray
        Radiance dataarray with coords 'tangent_height' and 'wavelength'
    threshold: float
        Value the vertical log gradient must be larger than to detect a cloud
    low_wavel_nm: float
        Lower wavelength to calculate the vertical log gradient at.  Default is 674 nm.
    high_wavel_nm: float
        Upper wavelength to calculate the vertical log gradient at.  Default is 868 nm.

    Returns
    -------
    cloud_alt_km: float
        Altitude of the highest cloud found.  If no cloud is found this returns 0

    """
    heights_km, v_log_grad_diff = vertical_log_gradient_difference(radiance, low_wavel_nm, high_wavel_nm)

    above_threshold_idx = np.nonzero(v_log_grad_diff > threshold)[0]
    heights_above_threshold = heights_km[above_threshold_idx]

    # Iterate over the values above the threshold in reverse order
    for h, idx in zip(heights_above_threshold[::-1], above_threshold_idx[::-1]):
        if h < MAX_CLOULD_ALT_KM:
            return h

    # No cloud found, return 0
    return 0
