from omps.omps_1d_retrieval import OMPSRetrieval
import matplotlib.pyplot as plt
import utility.plot_results as pr
import omps.level2 as omps_l2
import time

import h5py
import numpy as np
import logging

class RetrievalOutputSSAI:
    """
    Class designed to handle the output from the OMPSRetrieval class.

        >>> omps = OMPSRetrieval( 12115, 40 )
        >>> output = omps.retrieve_profiles()
        >>> h5out = RetrievalOutput( output )
        >>> h5out.filename = 'filename.h5'
        >>> h5out.save_to_output_file()
    """
    def __init__(self, output):

        self.save_log_messages   = True
        self.save_meas_vecs      = True
        self.save_radiances      = True
        self.num_images_in_orbit = None
        self.num_slits           = 3
        self._image_offset       = 0
        self.output = output

        #variables for easier access
        self.ret_opt     = self.output['retrieval_info','ret_opt']
        self.imagenumber = self.output['retrieval_info','image_num']
        self.orbitnumber = self.output['retrieval_info','orbit_num']
        self.slitnumber  = self.output['retrieval_info','slit_num']

        self.filename   = None
        self.retinfo_filename = None
        self.folder     = None
        self._file      = None
        self._retfile   = None

        self.num_alts     = 100
        self.num_o3_meas  = 600
        self.num_aer_meas = 75
        self.num_alb_meas = 1

    @property
    def image_offset(self):
        return self._image_offset

    @image_offset.setter
    def image_offset(self, image_offset):
        self.imagenumber  = self.output['retrieval_info','image_num'] - image_offset
        self._image_offset = image_offset

    def save_to_output_file(self):
        """
        Appends the retrieved profiles to an orbit file. If the image already exists
        in the file it is overwritten
        """

        found_folder = False
        #check if folder exists
        if os.path.isdir( self.folder ):
            found_folder = True
        #if not check if its on the python path and make the full path
        else:
            user_paths = os.environ['PYTHONPATH'].split(os.pathsep)
            for i in range(len(user_paths)):
                if os.path.isdir( user_paths[i]+self.folder ):
                    found_folder = True
                    self.folder = user_paths[i]+self.folder
                    break

        if found_folder is False:
            logging.error('could not find output folder')

        if self.filename is None:
            self._file      = self.folder + 'OMPS_L2_' + str(self.orbitnumber) + '.h5'
        else:
            self._file    = self.folder + self.filename

        exists = os.path.isfile(self._file)
        if not exists:
            self._create_empty_hdf_file()

        self._save_to_hdf_file()

    def _create_empty_hdf_file(self):
        f = h5py.File( self._file, 'a' )

        if self.num_images_in_orbit is None:
            num_images_in_orbit = omps_l2.get_num_images_in_orbit(self.orbitnumber, self.ret_opt['omps_l1_folder'])
        else:
            num_images_in_orbit = self.num_images_in_orbit

        num_slits = self.num_slits

        profile = np.zeros((self.num_alts,    num_images_in_orbit,num_slits))*np.nan
        yo3     = np.zeros((self.num_o3_meas, num_images_in_orbit,num_slits))*np.nan
        yaer    = np.zeros((self.num_aer_meas,num_images_in_orbit,num_slits))*np.nan
        yalb    = np.zeros((self.num_alb_meas,num_images_in_orbit,num_slits))*np.nan
        value   = np.zeros((num_images_in_orbit,num_slits))*np.nan

        f.create_dataset('/Data/Ozone',   data=profile)
        f.create_dataset('/Data/Aerosol', data=profile)
        f.create_dataset('/Data/Albedo',  data=value)

        if self.save_log_messages:
            dt = h5py.special_dtype(vlen=str)
            f.create_dataset('/Log/logging_info',(num_images_in_orbit,num_slits),dtype=dt)

        f.create_dataset('/Geo/SZA',      data=value)
        f.create_dataset('/Geo/MJD',      data=value)
        f.create_dataset('/Geo/Latitude', data=value)
        f.create_dataset('/Geo/Longitude',data=value)

        f.create_dataset('/Geo/Image',    data =np.zeros(num_images_in_orbit))
        f.create_dataset('/Geo/Orbit',    data =self.orbitnumber)
        f.close()

    def _save_to_hdf_file(self):
        f = h5py.File( self._file, 'r+' )
        if any("ozone" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Ozone'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','o3_valid']

        if any("aerosol" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Aerosol'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','aero_valid']

        if any("albedo" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Albedo'][self.imagenumber,self.slitnumber] = self.output['retrieved','albedo']

        f['/Geo/Image'][self.imagenumber]                       = self.output['retrieval_info','image_num']
        f['/Geo/SZA'][self.imagenumber, self.slitnumber]        = self.output['omps_geometry','sza']
        f['/Geo/MJD'][self.imagenumber, self.slitnumber]        = self.output['omps_geometry','mjd']
        f['/Geo/Latitude'][self.imagenumber, self.slitnumber]   = self.output['omps_geometry','latitude']
        f['/Geo/Longitude'][self.imagenumber, self.slitnumber]  = self.output['omps_geometry','longitude']

        #ouput the log messages to the hdf5 file
        if self.save_log_messages:
            log = logging.getLogger()
            for hndl in log.handlers:
                if type(hndl) == logging.StreamHandler:
                    f['/Log/logging_info'][self.imagenumber, self.slitnumber] = hndl.stream.getvalue()

        f.close()

def driver(controlFile):

        print("controlFile: " + str(controlFile))

        # Preparations
        #1. setup logging format
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # Set default runtime parameters; if YAML file is not provided, this group of setting will be used.
        l1_folder     = r"/home/tzhu/OMPS/software/Repos_SasktranV3LinuxBuild/Repos_skPython/utility/data/omps_testing/level1_radiance/"
        l2_anc_folder = r"/home/tzhu/OMPS/software/Repos_SasktranV3LinuxBuild/Repos_skPython/utility/data/omps_testing/level1_ancillary/"
        l2_out_folder = r"/home/tzhu/OMPS/software/Repos_SasktranV3LinuxBuild/Repos_skPython/utility/data/omps_testing/level2_output/"
        log_file      = r"/home/tzhu/OMPS/software/Repos_SasktranV3LinuxBuild/Repos_skPython/utility/data/omps_testing/level2_output/"
        slits_to_process  = [1]      #[0,1,2] #[1],[0,1], etc
        orbits_to_process = [12112]  #can be a list of orbits
        szaCutoff = 88.0
        levelOfDetails = 2
        numthreads = 0

        # Return a dictionary if YAML file is provided; defualt runtime parameters will be overwritten.
        if controlFile != None:  # get the info from the input YAML control file

                print("This run uses runtime parameters specified in a YAML input file <" + str(controlFile) + ">.")
                content = ProcessYAML(controlFile)

                # Code modification needed to allow input and output file names to be defined by user:
                L1Gfn = content['Input Files']['LP-L1G-EV'][0]
                ANCfn = content['Input Files']['LP-L1-ANC'][0]
                L2fn  = content['Output Files']['LP-L2-USASK'][0]
                print("L1Gfn: " + L1Gfn)
                print("ANCfn: " + ANCfn)
                print("L2fn: " + L2fn)

                l1_folder         = content['Runtime Parameters']['L1Gfolder']
                l2_anc_folder     = content['Runtime Parameters']['ANCfolder']
                l2_out_folder     = content['Runtime Parameters']['L2folder']
                slits_to_process  = content['Runtime Parameters']['slits_to_process']
                orbits_to_process = [ int(content['Runtime Parameters']['OrbitNumber']) ]
                szaCutoff         = content['Runtime Parameters']['szaCutoff']
                levelOfDetails    = content['Runtime Parameters']['Maximum Level Of Detail']
                numthreads        = content['Runtime Parameters']['numthreads']
                log_file          = l2_out_folder  # the log file will go to the same directory as the output L2 H5 file

        print("Runtime parameters for this process:")
        print("    l1_folder: " + str(l1_folder))
        print("    l2_anc_folder: " + str(l2_anc_folder))
        print("    l2_out_folder: " + str(l2_out_folder))
        print("    log_file: " + log_file)
        print("    slits_to_process: " + str(slits_to_process))
        print("    orbits_to_process: " + str(orbits_to_process))
        print("    szaCutoff: " + str(szaCutoff))
        print("    levelOfDetails: " + str(levelOfDetails))
        print("    numthreads: " + str(numthreads))

        # Start retrieval process
        start = time.time()

        for orbit in orbits_to_process:
                num_images = omps_l2.get_num_images_in_orbit(orbit, l1_folder)
                testFrame = 90

                #setup/change log file
                fileh     = logging.FileHandler( log_file + str(orbit) + '.log' )
                fileh.setFormatter(formatter)
                log = logging.getLogger()
                for hdlr in log.handlers:
                        log.removeHandler(hdlr)
                log.addHandler(fileh)

                #for im_idx in range(num_images):
                for im_idx in range(testFrame,(testFrame+1)):    # for test only

                        for slit in slits_to_process:

                                log.info('\n#############################################################\n'
                                         'PROCESSING: orbit: ' + str(orbit) + ' image: ' + str(im_idx) + ' slit: ' + str(slit) +
                                         '\n#############################################################\n')
                                #get solar zenith angle
                                sza = omps_l2.get_image_solar_zenith_angle(orbit, im_idx, slit,l1_folder)

                                #only process sunlit scans
                                if sza < szaCutoff:

                                        omps = OMPSRetrieval( orbit, im_idx, slit )
                                        omps.ret_opt['retrieval_order']        = ['albedo','aerosol','albedo','aerosol','ozone']
                                        omps.ret_opt['max_iter']               = [2,5,2,5,7]
                                        omps.ret_opt['scatterorder']           = [10,10,10,10,30]
                                        omps.ret_opt['num_diffuse_profiles']   = 1
                                        omps.ret_opt['wavel_samples']          = [1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1,     1    ]
                                        omps.ret_opt['wavel_samples_nm']       = [292.43,302.17,306.06,310.70,315.82,322.00,331.09,350.31,470.00,543.84,602.39,678.85,750.00]
                                        omps.ret_opt['verbose_output']         = levelOfDetails
                                        omps.ret_opt['background']             = 'omps_anc'
                                        omps.ret_opt['omps_l1_folder']         = l1_folder
                                        omps.ret_opt['omps_anc_folder']        = l2_anc_folder
                                        omps.ret_opt['output_folder']          = l2_out_folder
                                        omps.ret_opt['diffuse_incoming_resolution'] = [6,8,10,12]
                                        omps.ret_opt['numthreads']             = numthreads # zero will use all threads
                                        output = omps.retrieve_profiles()

                                        h5out                       = RetrievalOutputSSAI( output )
                                        h5out.num_images_in_orbit   = endFrame - startFrame
                                        h5out.image_offset          = startFrame
                                        h5out.folder                = omps.ret_opt['output_folder']
                                        h5out.filename              = omps.ret_opt['output_file']
                                        h5out.save_to_output_file()


                                        if omps.ret_opt['verbose_output'] == 1:
                                                pr.make_plots(output['intermediate',3][1]['lsq_output'])
                                                pr.make_plots(output['intermediate',4][1]['lsq_output'])
                                                plt.show()

                                else:
                                        log.info('Scan not processed, solar zenith angle too large: ' + str(sza))

        end = time.time()
        ProcessingTime = (end - start)/60.0
        tt = "%10.2f" % ProcessingTime
        log.info('ProcessingTime: ' + tt + 'min')

def ProcessYAML(controlFile):

        # Return a dictionary which contains all settings in the YAML file
        import yaml
        import re

        #tells YAML parser to treat orbit number strictly as a string
        regex = re.compile('^\s*(OrbitNumber):\d*')

        newContent = ""
        with open(controlFile, 'r') as inputStream:
                for line in inputStream:
                        m = regex.match(line)
                        if m is not None:
                                newLine = re.sub(m.group(), m.group() + ' !!str ', line, 1)
                                newContent = newContent + newLine
                        else:
                                newContent = newContent + line

        content = yaml.load(newContent)
        topLevelKeys = content.keys()

        requiredKeys = ['Runtime Parameters', 'Input Files', 'Output Files']
        for k in requiredKeys:
                if k not in topLevelKeys:
                        print("[Quit]: Required key <" + k + "> missing in <" + controlFile + ">. Job terminated.")

        requiredParams = ['OrbitNumber','szaCutoff','Maximum Level Of Detail','numthreads','L1Gfolder','ANCfolder','L2folder','slits_to_process']
        paramDict = content['Runtime Parameters']
        for k in requiredParams:
                if k not in paramDict:
                        print("[Quit]: Required runtime parameter <" + k + "> missing in <" + controlFile + ">. Job terminated.")

        return content

if __name__ == "__main__":

        from optparse import OptionParser
        import os
        import sys

        parser = OptionParser(description=__doc__)

        parser.add_option("-i","--input",dest="YAMLFilename",type="string",action="store",help="YAML Input Filename.")

        (options,args) = parser.parse_args()

        driver(options.YAMLFilename)