import h5py
import numpy as np
import utility.date_functions as df
from astropy.time import Time
from datetime import datetime
from os import listdir
import os
import re
import logging

def load(filename):
    """
    Loads an OMPS L2 Ozone File
    
    Parameters
    ----------
    filename : String
        raw string of the filename to open
    
    Returns
    -------
    out
        Data structure
    """
    output = dict()
    f = h5py.File( filename, 'r' )
    #output['o3uv'] = f['DataFields/O3UvValue'][:]
    #output['o3vis'] = f['DataFields/O3VisValue'][:]
    output['slit'] = f['DataFields/SlitNumber'][:]
    output['orbit'] = f['GeolocationFields/OrbitNumber'][:]
    output['height'] = f['GeolocationFields/HeightScale'][:]
    output['aero'] = f['DataFields/aerosolExtinctionValue'][:]
    output['lat'] = f['GeolocationFields/Latitude']
    
    return output
    
def load_ozone_profile(target_mjd, slit):
    """
    Loads an OMPS L2 Ozone profile
    
    Parameters
    ----------
    filename : String
        raw string of the filename to open
    
    target_mjd : double
        loads the closest scan to this time
    slit : int (1, 2, 3)
        loads the profile from the left (1), middle (2) or right slit (3)
        
    Returns
    -------
    out
        Data structure
    """
    
   #target_time = Time(target_mjd)
    
    folder = r'//utls/Users/scratch/OMPS_L2/'
    files =  listdir(folder)
    files_aer = []      
    found_mjd = False
    
    target_time = Time(target_mjd, format='mjd')
    
    for i in range(0,len(files)):
        if files[i][15:17] == 'O3':
            files_aer.append(files[i])

    #TOOD: should probably read dates in file names as a course check
    for i in range(0,len(files_aer)):
        f = h5py.File(folder+files_aer[i], 'r')
        date = np.double(f['GeolocationFields/Date'])
        time = np.double(f['GeolocationFields/Time'][:])
        
        year    = np.floor(date/10000);
        month   = np.floor( (date-year*10000)/100)
        day     = date- year*10000 - month*100
        hour    = np.floor(  time/60/60)
        minute  = np.floor( (time - hour*60*60)/60)
        sec     = np.floor(  time - hour*60*60 - minute*60)
        usec    = np.floor( (time - hour*60*60 - minute*60 - sec)*1e6 )
        t_start = Time(datetime.datetime(np.int(year),np.int(month),np.int(day),np.int(hour[0]), np.int(minute[0]), np.int(sec[0]), np.int(usec[0])),  'utc')
        t_end   = Time(datetime.datetime(np.int(year),np.int(month),np.int(day),np.int(hour[-1]), np.int(minute[-1]), np.int(sec[-1]), np.int(usec[-1])), 'utc')
        
        if target_time > t_start and target_time < t_end:
            found_mjd = True
            output = dict()
        
            date =  np.double(f['/GeolocationFields/Date'])
            slit_num = f['/DataFields/SlitNumber'][:]
        
            time = np.double(f['/GeolocationFields/Time'][:])
            
            year = np.floor(date/10000);
            month = np.floor( (date-year*10000)/100)
            day = date- year*10000 - month*100
            
            hour = np.floor(time/60/60)
            minute  = np.floor( (time - hour*60*60)/60)
            sec  = time - hour*60*60 - minute*60
        
            good = slit_num==slit
            mjd = df.date2mjd(year,month,day,hour,minute,sec);
            mjd[~good] = -999     #find only the mjd from the slit we want
            
            idx = np.argmin( np.abs(mjd-target_mjd) )
            
            output['o3'] = f['DataFields/O3CombinedValue'][idx,:]
            output['o3_uv'] = f['DataFields/O3UvValue'][idx,:]
            output['o3_vis'] = f['DataFields/O3VisValue'][idx,:]
            output['height'] = f['GeolocationFields/HeightScale'][:]
            output['latitude']= f['GeolocationFields/Latitude'][idx]
            output['longitude']= f['GeolocationFields/Longitude'][idx]
            output['trop'] = f['AncillaryData/TropopauseAltitude'][idx]
            output['albedo'] = f['DataFields/SurfaceReflectance'][idx]
            output['mjd']= mjd[idx]
            
            break
            
    if not found_mjd:
        output = []
    
    return output

def load_aerosol_profile(target_mjd, slit):
    """
    Loads an OMPS L2 Aerosol profile closest to the target mjd
    
    Parameters
    ----------
    target_mjd : double
        loads the closest scan to this time
    slit : int (1, 2, 3)
        loads the profile from the left (1), middle (2) or right slit (3)
        
    Returns
    -------
    out
        Data structure
    """

    folder = r'//utls/Users/scratch/OMPS_L2/'
    files =  listdir(folder)
    files_aer = []      
    found_mjd = False
    
    target_time = Time(target_mjd, format='mjd')
    
    for i in range(0,len(files)):
        if files[i][15:18] == 'AER':
            files_aer.append(files[i])

    #TOOD: should probably read dates in file names as a course check
    for i in range(0,len(files_aer)):
        f = h5py.File(folder+files_aer[i], 'r')
        date = np.double(f['GeolocationFields/Date'])
        time = np.double(f['GeolocationFields/Time'][:])
        
        year    = np.floor(date/10000);
        month   = np.floor( (date-year*10000)/100)
        day     = date- year*10000 - month*100
        hour    = np.floor(  time/60/60)
        minute  = np.floor( (time - hour*60*60)/60)
        sec     = np.floor(  time - hour*60*60 - minute*60)
        usec    = np.floor( (time - hour*60*60 - minute*60 - sec)*1e6 )
        t_start = Time(datetime.datetime(np.int(year),np.int(month),np.int(day),np.int(hour[0]), np.int(minute[0]), np.int(sec[0]), np.int(usec[0])),  'utc')
        t_end   = Time(datetime.datetime(np.int(year),np.int(month),np.int(day),np.int(hour[-1]), np.int(minute[-1]), np.int(sec[-1]), np.int(usec[-1])), 'utc')
        
        if target_time > t_start and target_time < t_end:
            found_mjd = True
            output = dict()
        
            date =  np.double(f['/GeolocationFields/Date'])
            slit_num = f['/DataFields/SlitNumber'][:]
        
            time = np.double(f['/GeolocationFields/Time'][:])
            
            year = np.floor(date/10000);
            month = np.floor( (date-year*10000)/100)
            day = date- year*10000 - month*100
            
            hour = np.floor(time/60/60)
            minute  = np.floor( (time - hour*60*60)/60)
            sec  = time - hour*60*60 - minute*60
        
            good = slit_num==slit
            mjd = df.date2mjd(year,month,day,hour,minute,sec);
            mjd[~good] = -999     #find only the mjd from the slit we want
            
            idx = np.argmin( np.abs(mjd-target_mjd) )
            
            output['height'] = f['GeolocationFields/HeightScale'][:]
            output['aero'] = f['DataFields/aerosolExtinctionValue'][:][idx,:,:]
            output['aero_wavel'] = f['DataFields/aerosolWavelength'][:][idx,:]
            output['latitude'] = f['GeolocationFields/Latitude'][:][idx]
            output['longitude']= f['GeolocationFields/Longitude'][idx]
            output['trop'] = f['AncillaryData/TropopauseAltitude'][idx]
            output['albedo'] = f['DataFields/SurfaceReflectance'][idx]
            output['mjd']= mjd[idx]
            
            break
        
    if not found_mjd:
        output = []
        
    return output
    
def load_aerosol_orbit(orbit_number, slit, folder=r'//utls/Users/scratch/OMPS_L2/'):
    """
    Loads an OMPS L2 Aerosol profile for one orbit
    
    Parameters
    ----------
    orbit_number : scalar
        Orbit number to load
    slit : int (1,2,3)
        loads the profile from the left (1), middle (2) or right slit (3)
        
    Returns
    -------
    out
        Data structure
    """

    files =  listdir(folder)
    files_aer = []      
    
    found_orbit = False
    
    for i in range(0,len(files)):
        if files[i][15:18] == 'AER':
            files_aer.append(files[i])
    
    for i in range(0,len(files_aer)):
        f = h5py.File(folder+files_aer[i], 'r')
        orbit = np.unique(f['GeolocationFields/OrbitNumber'][:])
        if np.size(np.where(orbit==orbit_number)) == 1:
            found_orbit = True
            output = dict()
            omps_slit = f['DataFields/SlitNumber'][:]
            omps_orbit = f['GeolocationFields/OrbitNumber'][:]
            good = (omps_orbit == orbit_number) & (slit == omps_slit)
            output['height'] = f['GeolocationFields/HeightScale'][:]
            output['aero'] = f['DataFields/aerosolExtinctionValue'][:][good,:,:]
            output['aero_wavel'] = f['DataFields/aerosolWavelength'][:][good,:]
            output['latitude'] = f['GeolocationFields/Latitude'][:][good]
            output['longitude']= f['GeolocationFields/Longitude'][good]
            output['trop'] = f['AncillaryData/TropopauseAltitude'][good]
            output['albedo'] = f['DataFields/SurfaceReflectance'][good]
            break
        
    if not found_orbit:
        output = []
        
    return output
    
def load_ozone_orbit(orbit_number, slit, folder=r'//utls/Users/scratch/OMPS_L2/'):
    """
    Loads an OMPS L2 Ozone profile for one orbit
    
    Parameters
    ----------
    orbit_number : scalar
        Orbit number to load
    slit : int (1,2,3)
        loads the profile from the left (1), middle (2) or right slit (3)
        
    Returns
    -------
    out
        Data structure
    """

    files =  listdir(folder)
    files_o3 = []      
    
    found_orbit = False
    
    for i in range(0,len(files)):
        if files[i][15:17] == 'O3':
            files_o3.append(files[i])
    
    mjd = np.zeros(0)
    for i in range(0,len(files_o3)):
        f = h5py.File(folder+files_o3[i], 'r')
        orbit = np.unique(f['GeolocationFields/OrbitNumber'][:])
        if np.size(np.where(orbit==orbit_number)) == 1:
            found_orbit = True
            output = dict()
            omps_slit = f['DataFields/SlitNumber'][:]
            omps_orbit = f['GeolocationFields/OrbitNumber'][:]
            good = (omps_orbit == orbit_number) & (slit == omps_slit)
            output['height'] = f['GeolocationFields/HeightScale'][:]
            output['latitude'] = f['GeolocationFields/Latitude'][:][good]
            output['longitude']= f['GeolocationFields/Longitude'][good]
            output['o3'] = f['DataFields/O3CombinedValue'][good,:]
            output['o3_uv'] = f['DataFields/O3UvValue'][good,:]
            output['o3_vis'] = f['DataFields/O3VisValue'][good,:]
            output['trop'] = f['AncillaryData/TropopauseAltitude'][good]
            output['albedo'] = f['DataFields/SurfaceReflectance'][good]
            
            date =  np.double(f['GeolocationFields/Date'])
            time = np.double(f['GeolocationFields/Time'][:][good])
            
            year = np.floor(date/10000);
            month = np.floor( (date-year*10000)/100)
            day = date- year*10000 - month*100
            
            hour = np.floor(time/60/60)
            minute  = np.floor( (time - hour*60*60)/60)
            sec  = time - hour*60*60 - minute*60
        
            for j in range(len(time)):
#                mjd = np.concatenate((mjd, 1));
                mjd = np.concatenate((mjd, np.array([df.date2mjd(year,month,day,hour[j],minute[j],sec[j])])));


    if not found_orbit:
        output = []
        print('could not find orbit in folder')
    else:
        output['mjd'] = mjd
        
    return output


def load_l2_from_orbit(orbit,folder='Z:\\scratch\\OMPS_L2_OSIRIS\\',truncate=True,load_meas_vecs=False):
    """
    A function for loading level 2 data processed using the class omps_1d_retrieval.py

    :param orbit: orbit number (int)
    :param folder: optional argument for loading data from different sources
    :param truncate: Truncate the profiles above and below retrieval limits, truncation is
            useful for comparisons, but bad if we want to model radiances using l2
    :param load_meas_vecs: include measurement vectors in the returned data structure
    :return:
    """
    data = dict()

    if folder is 'OMPS_L2':
        pass

        o3_omps  = load_ozone_orbit(orbit,2)
        aer_omps = load_aerosol_orbit(orbit,2)

        data['o3'] = o3_omps['o3'].transpose()
        data['aer'] = aer_omps['aero'][:,3,:].transpose()
        data['alb'] = aer_omps['albedo']
        data['lat'] = o3_omps['latitude']
        data['lon'] = o3_omps['longitude']
        data['alt'] = np.arange(0.5,100.5)

        #add some zeros
        data['aer'] = np.vstack((data['aer'], np.zeros((59, np.shape(data['aer'])[1] ))))
        data['o3']  = np.vstack((data['o3'],  np.zeros((39, np.shape(data['aer'])[1] ))))

        #set the zeros to nan if we are truncating the profiles
        if truncate:
            data['aer'][data['aer'] <= 0]   = np.nan
            data['o3'][ data['o3']  <= 0]   = np.nan
            data['alb'][data['alb']  < 0]   = np.nan

    else:
        file   = folder + 'OMPS_L2_' + str(orbit) + '.h5'

        f           = h5py.File(file)
        data['o3']  = f['/Data/Ozone'][:]
        data['aer'] = f['/Data/Aerosol'][:]*1.246e-5
        data['alb'] = f['/Data/Albedo'][:]
        data['lat'] = f['/Geo/Latitude'][:]
        data['lon'] = f['/Geo/Longitude'][:]

        if load_meas_vecs:
            data['o3_meas'] = f['/Data/Ozone_ymeas'][:]
            data['o3_ret'] = f['/Data/Ozone_yret'][:]
            data['aer_meas'] = f['/Data/Aerosol_ymeas'][:]
            data['aer_ret'] = f['/Data/Aerosol_yret'][:]
            data['alb_meas'] = f['/Data/Albedo_ymeas'][:]
            data['alb_ret'] = f['/Data/Albedo_yret'][:]

        data['lon'][data['lon'] > 180] = data['lon'][data['lon'] > 180] - 360
        data['sza'] = f['/Geo/SZA'][:]
        data['alt'] = np.arange(0.5,100.5)

        data['o3'][:,data['sza']>88] = np.nan
        data['aer'][:,data['sza']>88] = np.nan
        data['alb'][data['sza']>88] = np.nan

        if truncate:
            if np.any(data['o3']==0):
                data['o3'][data['o3']==0]     = np.nan


            data['aer'][data['aer']<1e-6] = np.nan

            #remove altitudes with saturated aerosol
            data['aer'][data['aer']>1.24e-2] = np.nan
            data['o3'][data['aer']>1.24e-2] = np.nan

            omps_l2 = load_aerosol_orbit(orbit,2)

            trop = omps_l2['trop']
            for i,t in enumerate(trop):
                data['o3'][data['alt']<trop[i],i] = np.nan
                data['aer'][data['alt']<trop[i],i] = np.nan

        f.close()

#    if orbit==12113:
#        data['o3'][:,0:180] = np.nan

    return data

def get_num_images_in_orbit( orbit_number, folder = r"//UTLS/Users/scratch/OMPS_L1/" ):
    """

    :param orbit_number: int
    :param folder: str
        name of the folder containing the level 1 data files
    :return: int
        the number of radiance profiles in an orbit
    """
    filename = l1_filename_for_orbit( orbit_number, folder, 'L1G')
    f = h5py.File(filename,'r')

    return np.shape(f['/GRIDDED_DATA/Radiance'][:])[0]

def get_image_solar_zenith_angle( orbit_number, image_number, slit_number, folder = r"//UTLS/Users/scratch/OMPS_L1/" ):
    filename = l1_filename_for_orbit( orbit_number, folder, 'L1G')
    f = h5py.File(filename,'r')

    return f['/GEOLOCATION_DATA/SolarZenithAngle_35km'][image_number, slit_number]

def get_orbit_solar_zenith_angle( orbit_number,  slit_number, folder = r"//UTLS/Users/scratch/OMPS_L1/" ):
    filename = l1_filename_for_orbit( orbit_number, folder, 'L1G')
    f = h5py.File(filename,'r')

    return f['/GEOLOCATION_DATA/SolarZenithAngle_35km'][:, slit_number]

def get_image_geolocation_data( orbit_number, image_number, slit_number, folder = r"//UTLS/Users/scratch/OMPS_L1/" ):
    filename = l1_filename_for_orbit( orbit_number, folder, 'ANC')
    f = h5py.File(filename,'r')
    lat = f['/GEOLOCATION_DATA/Latitude'][image_number, slit_number]
    lon = f['/GEOLOCATION_DATA/Longitude'][image_number, slit_number]

    return lat, lon

def get_orbit_geolocation_data( orbit_number, slit_number, folder = r"//UTLS/Users/scratch/OMPS_L1/" ):
    filename = l1_filename_for_orbit( orbit_number, folder, 'ANC')
    f = h5py.File(filename,'r')
    lat = f['/GEOLOCATION_DATA/Latitude'][:, slit_number]
    lon = f['/GEOLOCATION_DATA/Longitude'][:, slit_number]

    return lat, lon

def get_obrits_in_mjd_range( min_mjd, max_mjd, folder=r'//UTLS/Users/scratch/OMPS_L2_ANC/'):

    if not os.path.isdir( folder ):
        logging.error('could not find ancillary folder')

    files = listdir(folder)
    keyword = 'ANC'
    orbits = []

    for f in files:
        m = re.search(keyword, f)
        if m is not None:
            m = re.search( 'o([0-9]*)', f )
            orbit = int(m.group(1))
            f = h5py.File(folder + f)
            date = f['/GEOLOCATION_DATA/Date'][0,0]
            time = f['/GEOLOCATION_DATA/Time'][0,0]

            year = np.floor(date/10000);
            month = np.floor( (date-year*10000)/100)
            day = date- year*10000 - month*100

            hour = np.floor(time/60/60)
            minute  = np.floor( (time - hour*60*60)/60)
            sec  = time - hour*60*60 - minute*60

            t = Time(datetime(int(year), int(month), int(day), int(hour), int(minute), int(sec)), scale='utc')
            mjd = t.mjd

            if (mjd > min_mjd) & (mjd < max_mjd):
                orbits.append(orbit)

    return orbits

def l1_filename_for_orbit( orbitnumber, folder, keyword ):
    """
    Parses through the files in folder to find the orbit in question and returns
    the name of the file.

    :param orbitnumber: int
    :param folder: str
    :param keyword : str
    :return: str
    """

    found_folder = False
    #check if folder exists
    if os.path.isdir( folder ):
        files = listdir(folder)
        found_folder = True
    #if not check if its on the python path
    else:
        try:
            user_paths = os.environ['PYTHONPATH'].split(os.pathsep)
            for i in range(len(user_paths)):
                if os.path.isdir( user_paths[i]+folder ):
                    found_folder = True
                    folder = user_paths[i]+folder
                    files = listdir(folder)
                    break
        except KeyError:
            pass

    if found_folder is False:
        logging.error('could not find the data folder')
        raise Exception('could not find the data folder')

    #TODO: there has got to be a better way to parse files
    found_file = False
    for f in files:
        m = re.search(keyword, f)
        if m is not None:
            m = re.search( 'o([0-9]*)', f )
            orbit = int(m.group(1))
            if orbit == orbitnumber:
                found_file = True
                return folder + f

    if not found_file:
        logging.error('Orbit does not exist in folder')
        raise Exception('Orbit does not exist in folder')

