#import retrieval classes
import retrieval.core as retrieval
import retrieval.method.lsq as lsq
import retrieval.method.rodgers as rodgers
import radtran.sasktran as radtran
import retrieval.species_o3 as species_o3
import retrieval.species_aerosol as species_aer
import retrieval.species_aerosol_particlesize as species_aer_ps
import retrieval.species_albedo as species_alb

#import radiative transfer module
import sasktranif.sasktranif as skif
import radtran.options as opt
import radtran.atmosphere as atmosphere
import radtran.climatologies.background as background

#import climatology and geometry utilities
import radtran.climatologies.ozone as ozone
import radtran.climatologies.no2 as no2
import radtran.climatologies.aerosol as aerosol
import radtran.climatologies.utility as clim_util

#import classes to load omps level2 data
import omps.level2 as omps_l2

#import instrument classes
import retrieval.instruments.limb_profiler as instrument
import omps.omps_model as omps

#import various helper libraries
import numpy as np
import os
from itertools import chain
from os import listdir
import copy as copy
import re
import logging
import h5py
import utility as utility


class OMPSRetrieval(object):
    """
    Class which handles the full retrieval (using simulatied or real data) for a 
    single OMPS image
    
    Typically, the class is constructed wtih an OMPS orbit and image number. The user
    then changes any desired options, and finally :func:`~OMPSRetrieval.retrieve_profiles`
    is called, starting the retrieval.
    
    The option do_simulation, allows the user to switch between retrievals using
    real or simulated OMPS data.
    
    Parameters
    ----------
    orbitnumber : int
        The OMPS orbit number to retrieve from
    imagenumber : int
        The OMPS image number to retrieve from
    slitnumber : int
        the OMPS slit to retrieve from

        * 0 - left slit
        * 1 - center slit
        * 2 - right slit

    Examples
    --------
    A simple retrieval with default settings
    
    >>> omps = OMPSRetrieval( 12115, 40 )
    >>> output = omps.retrieve_profile()
    >>> plt.plot(output['retrieved','aero'], np.arange(0.5, 100.5))
    
    Retrieve only ozone with single scatter only
    
    >>> omps = OMPSRetrieval( 12115, 40 )
    >>> omps.ret_opt['retrieval_order'] = ['ozone']
    >>> omps.ret_opt['max_iter'] = [5]
    >>> omps.ret_opt['scatterorder'] = [1]
    >>> output = omps.retrieve_profiles()

    Options
    -------
    Options are set by accessing the ret_opt dictionary within the class,
    e.g.,
    
    >>> omps = OMPSRetrieval( 12115, 40 )
    >>> omps.ret_opt['retrieval_order'] = ['albedo', 'aerosol', 'ozone']
    >>> omps.ret_opt['scatterorder'] = [10,10,30]

    All options are optional
    
    scatterorder : list of ints
        The number of orders of scatter to use in the radiative transfer model. Each
        scalar corresponds to the string in the list given to retrieval_order. If
        a single scalar is given all retrievals will use this value
        Default is 15
    
    retrieval_order : list of strings
        A combination of the strings 'albedo', 'aerosol', 'ozone' which determines
        the number of retrievals to run and their order.  The default is 
        ['albedo', 'aerosol', 'ozone', 'albedo', 'ozone']
    
    max_iter : list of scalars
        The numbe of iterations to do for each of the retrievals, should be the
        same length as retrieval_order if set. Default is 5.
    
    verbose_output : int
        If true, then intermediate results such as the measurement vectors will
        be saved in the output structure. Default is 0
        
        * 0 - No intermediate values are stored
        * 1 - Values are stored after each species retrieval
        * 2 - Values at every iteration are stored
    
    omps_l1_folder : string
        The folder name where the OMPS L1 files are stored, default is
        r"\\utls\\users\\scratch\\OMPS_L1\\"
        
    omps_l2_folder : string
        The folder name where the OMPS L2 files are stored, default is
        r"\\utls\\users\\scratch\\OMPS_L2\\"  The level 2 files are only necessary
        if the retrieval options 'use_omps_l2_as_apriori' is set to True (default is False)

    omps_anc_folder : string
        The folder name where the OMPS Ancillary files are stored, default is
        r"\\utls\\users\\scratch\\OMPS_L2\\".

    num_images_in_orbit : int
        Number of images that will be processed for this orbit. This changes the
        size of the output h5 file to be (num_alts x num_images_in_orbit x num_slits)
        Default is None, and the number of images will be determined from the number
        of radiance profiles in the level 1 gridded data files. Note: this will only take
        effect if the h5 file does not yet exist. If it does this setting will be ignored
        and the existing tables will be populated.
        
    wavel_samples : int or array
        The number of wavelength samples per pixel. If an integer is given
        all pixels have the same oversampling. If an array is given the
        number of samples is interpolated as a function of wavelength,
        given by wavel_samples_nm.

    wavel_samples_nm : array
        Same size as wavel_samples. If wavel_samples is an integer this entry
        is ignored.
        
    vert_samples : int
        Number of lines of sight per pixel.
        Default value is one. Cannot be changed between retrievals
        
    add_no2 : bool
        Include no2 in the atmospheric state. Default value is True

    background : string
        Background pressure, temperature and number density profiles
        to be used in the retrieval

        * ecmwf - Uses the European Centre for Medium-Range Weather
        Forecasts operation product. MSIS-90 is merged in above
        the highest provided pressure level.
        * msis - Uses the MSIS-90 atmospheric model
        * omps_anc - Uses the atmospheric state provided in the
        OMPS level 2 ancillary files. Altitudes above 80km are scaled
        to MSIS-90

    ozone_apriori : str
        Initial guess for ozone. Default is labow

        * labow
        * omps_l2

    aerosol_apriori : str
        Inital guess for aerosol. Default is a user defined climatology

        * user_defined
        * omps_l2

    no2_apriori : str
        Initial guess for NO2. Default is pratmo

        * pratmo

    albedo_apriori : str or float
        Initial guess for albedo. Default is 0.3

        * scalar
        * omps_l2
        * None - sets albedo to 0.3

    use_wf : bool
        Use the analytic weighting function provided by SasktranHR.

    num_diffuse_profiles : int
        Number of diffuse profiles in the radiative transfer model. Default is 5.

    diffuse_point_altitudes : str, 1d array
        Altitude of diffuse points on each diffuse profile. Can be specified as a
         1d numpy array, or as a string. Default placement is 1 diffuse point
         every 1km. The following strings are accepted:

        * verylow : places 1 diffuse point at the ground, then every 5km up to 100km
        * low : places 1 diffuse point at the ground, then every 3km up to 100km
        * moderate : places 1 diffuse at the ground, then every 1km until 60km,
            every 3km until 80, and every 5km until 100km.

    ozone_triplets : dict
        Defines the triplets used in the ozone retrieval. The dictionary has the following
        structure:

        input['wavelengths'] = [[wavel_1,  wavel_2,  ...], [wavel_3,  wavel_4,  ...], ...]
        input['weights']     = [[weight_1, weight_2, ...], [weight_3, weight_4, ...], ...]

        Due to the formulation of the Wijk matrix, this option is only currently supported
        with the 'LM' retrieval method. The default value is None, and will use the

    aerosol_triplets : dict
        Defines the triplets used in the ozone retrieval. Dictionary has the following
        structure:

        input['wavelengths'] = [[wavel_1,  wavel_2,  ...], [wavel_3,  wavel_4,  ...], ...]
        input['weights']     = [[weight_1, weight_2, ...], [weight_3, weight_4, ...], ...]

        Due to the formulation of the Wijk matrix, this option is only currently supported
        with the 'LM' retrieval method.

    The following options are also available for the simulated data. These affect
    the radiative transfer model used to generate the 'measured' radiance, but 
    will not affect the retrieval.
    
    * scatterorder : list(int) with length one
    * wavel_samples : int or array
    * vert_samples : int
    * add_no2 : bool
    * use_wf : bool
    * num_diffuse_profiles : int
    
    These are accessed in the same way as retrieval options, except using the 
    sim_opt dictionary, for example
    
    >>> omps = OMPSRetrieval( 12115, 40 )
    >>> omps.sim_opt['scatterorder'] = [30]
    >>> omps.sim_opt['vert_samples'] = 10

    Output
    ------
    The output structure is a dictionary, typically indexed by a tuple of two strings. Here
    we list the keys and the corresponding item in the dictionary.  E.g.
    
    >>> output = omps.retrieve_profiles()
    >>> sza = output['omps_geometry', 'sza']

    ('omps_geometry','image_index') : int
        The indicies of the images used in the retrieval, this is typically something
        like 10:170, since images at the end are thrown out due to the SZA being too
        high
    ('omps_geometry', 'sza') : double
        Solar Zenith Angle reported from the OMPS L1 file.  This is the same size
        as ('omps_geometry', 'image_index')
    ('omps_geometry', 'longitude') : double
        Longitude as reported from the OMPS L1 file.  Same size as ('omps_geometry', 'image_index')
    ('omps_geometry', 'latitude') : double
        Latitude as reported from the OMPS L1 file.  Same size as ('omps_geometry', 'image_index')
    ('omps_measurement', 'radiance') : 2d numpy array
        Shape (los,wavel) of radiances measured by OMPS.  In the standard
        photons/cm^2/s/sterad/nm
    ('omps_measurement', 'wavel_nm') : 1d numpy array
        Wavelengths corresponding to the OMPS measurements
    ('omps_measurement', 'psf') : 1d numpy array
        Point spread function of the omps measurements assuming a gaussian shape
    ('retrieval_geometry', 'altitude') : 1d numpy array
        Altitudes in m of the retrieval grid
    ('retrieval_geometry', 'latitude') : 1d numpy array
        Latitude of the retrieval locations
    ('retrieval_geometry', 'longitude') : 1d numpy array
        Longitude of the retrieval locations
    ('retrieval_geometry', 'location_xyz') : 2d numpy array
        Array of shape(numretprof,3) of the xyz locations of the retrieval
        points
    ('retrieved', 'o3') : 2d numpy array
        Retrieved ozone.  If no ozone was retrieved then this is the apriori.
        Shape (numret, 100)
    ('retrieved', 'aero') : 2d numpy array
        Retrieved aerosol.  If no aerosol was retrieved then this is the apriori
        shape (numret, 100)
    ('retrieved', 'albedo') : 2d numpy array
        Retrieved albedo.  If no albedo was retrieved then this is the apriori
  
    """
    def __init__(self, orbitnumber, imagenumber, slitnumber=1, do_simulation=False):
        self.orbitnumber = orbitnumber
        self.imagenumber = imagenumber
        self.slitnumber  = slitnumber
        self.ret_opt     = self.retrieval_options_default()
        self.sim_opt     = self.simulation_options_default()
        self.filename    = None #self.l1_filename_for_orbit( orbitnumber, self.ret_opt['omps_l1_folder'] )
        
        self.cloud_height = 0.0
        
        self.ret_func_dict = { 'ozone'        : lambda a: self.retrieve_ozone(a),
                               'aerosol'      : lambda a: self.retrieve_aerosol(a),
                               'albedo'       : lambda a: self.retrieve_albedo(a),
                               'particlesize' : lambda a: self.retrieve_particlesize(a), 
        }
                
        self.output = dict()
        self.do_simulation = do_simulation
        self.method = None

        self.software_version    = '1.01'

    def version(self):
        """
        :return: the software version number
        """
        return self.software_version

    def retrieval_options_default(self):
        ret_opt = dict()

        #Radiative transfer settings
        ret_opt['scatterorder']    = [15, 15, 15, 15, 15]
        ret_opt['wavel_samples']   = 1
        ret_opt['wavel_samples_nm']= None
        ret_opt['wavel_int_type']  = 'linear'
        ret_opt['wavel_num_stdev'] = 3
        ret_opt['vert_samples']    = 1 #cannot be changed between species at present
        ret_opt['num_diffuse_profiles'] = 5
        ret_opt['diffuse_incoming_resolution'] = [6,8,10,12]
        ret_opt['diffuse_point_altitudes'] = 'default'
        ret_opt['numthreads']      = 0
        ret_opt['ms_wavel_space']  = 0.5
        ret_opt['do_ss_approx']    = False

        #input/output settings
        ret_opt['omps_l1_folder']  = r"\\utls\\users\\scratch\\OMPS_L1\\"
        ret_opt['omps_l2_folder']  = r"\\utls\\users\\scratch\\OMPS_L2_OSIRIS\\"
        ret_opt['omps_anc_folder'] = r"\\utls\\users\\scratch\\OMPS_L2_ANC\\"
        ret_opt['output_folder']   = r'\\utls\\users\\scratch\\OMPS_L2_OSIRIS\\'
        ret_opt['output_file']     = None
        ret_opt['num_images_in_orbit'] = None

        #atmosphere settings
        ret_opt['add_no2']         = True
        ret_opt['ozone_apriori']   = 'labow'
        ret_opt['aerosol_apriori'] = 'user_defined'
        ret_opt['aerosol_apriori_filename'] = 'aerosol_apriori.txt'
        ret_opt['no2_apriori']     = 'pratmo'
        ret_opt['albedo_apriori']  = 0.3
        ret_opt['background']      = 'ecmwf'
        ret_opt['use_wf']          = True

        #retrieval parameters
        ret_opt['retrieval_order']  = ['albedo', 'aerosol', 'albedo', 'aerosol', 'ozone']
        ret_opt['max_iter']         = [2,5,2,5,7]
        ret_opt['verbose_output']   = 0
        ret_opt['retrieval_method'] = 'LM'
        ret_opt['aerosol_vectors']  = None
        ret_opt['ozone_vectors']    = None

        return ret_opt
        
    def simulation_options_default(self):
        sim_opt = dict()
        sim_opt['scatterorder']    = [30]   
        sim_opt['wavel_samples']   = 1
        sim_opt['wavel_samples_nm']= None
        sim_opt['vert_samples']    = 1
        sim_opt['add_no2']         = True
        sim_opt['use_wf']          = False
        sim_opt['num_diffuse_profiles'] = 5
        sim_opt['diffuse_incoming_resolution'] = [6,8,10,12]
        return sim_opt
        
    def check_retrieval_options(self):
        """
        Simple check to ensure that the retrieval options match the expected size.
        If they do not a warning is issued and the defaults are used.
        """
        
        num_spec = len(self.ret_opt['retrieval_order'])

        #check max iterations
        if len(self.ret_opt['max_iter']) != num_spec:
            logging.warning('max_iter wrong size, setting to: ', self.ret_opt['max_iter'][0])
            self.ret_opt['max_iter'] = [self.ret_opt['max_iter'][0] for i in range(num_spec)]

        #check numordersofscatter
        if len(self.ret_opt['scatterorder']) != num_spec:
            logging.warning('scatterorder wrong size, setting to: ', self.ret_opt['scatterorder'][0])
            self.ret_opt['scatterorder'] = [self.ret_opt['scatterorder'][0] for i in range(num_spec)]

        #check retrieval method
        if isinstance(self.ret_opt['retrieval_method'],list):
            if len(self.ret_opt['retrieval_method']) != num_spec:
                logging.warning('retrieval_method wrong size, setting to: LM')
                self.ret_opt['retrieval_method'] = ['LM' for i in range(num_spec)]
        else:
            retmethod = self.ret_opt['retrieval_method']
            self.ret_opt['retrieval_method'] = [retmethod for i in range(num_spec)]

        #check/setup diffuse profile
        if (self.ret_opt['diffuse_point_altitudes'] is None) | (self.ret_opt['diffuse_point_altitudes'] == 'default'):
            self.ret_opt['diffuse_point_altitudes'] = np.double(np.concatenate((np.array([1e-6]), np.arange(500.0, 100000.0, 1000.0))))
        elif self.ret_opt['diffuse_point_altitudes'] == 'moderate':
            self.ret_opt['diffuse_point_altitudes'] = np.double(np.concatenate((np.array([1e-6]),
                                                                np.arange(500.0,  60000.0, 1000.0),
                                                                np.arange(60500.0,80000.0, 3000.0),
                                                                np.arange(80500.0,100000.0,5000.0))))
        elif self.ret_opt['diffuse_point_altitudes'] == 'low':
            self.ret_opt['diffuse_point_altitudes'] = np.double(np.concatenate((np.array([1e-6]), np.arange(500.0, 60000.0, 3000.0), np.arange(62500.0, 100000.0, 5000.0))))
        elif self.ret_opt['diffuse_point_altitudes'] == 'verylow':
            self.ret_opt['diffuse_point_altitudes'] = np.double(np.concatenate((np.array([1e-6]), np.arange(500.0, 100000.0, 5000.0))))
        elif self.ret_opt['diffuse_point_altitudes'] == 'version_1_00':
            self.ret_opt['diffuse_point_altitudes'] = np.double(np.concatenate((np.array([1e-6]), np.arange(500.0, 24500.0, 2000), np.arange(24500.0, 100000.0, 5000.0))))
        
    def HR_options(self, hr_opt):
        options = opt.HROptions()
        options.add_option('numordersofscatter', hr_opt['scatterorder'][0] )
        options.add_option('raytracingshells', 1000)
        options.add_option('numdiffuseprofilesinplane', hr_opt['num_diffuse_profiles'])
        options.add_option('diffuseplacementtype', 1)
        options.add_option('solarraytracingshells', 1000)
        options.add_option('integrationtechnique', 0 )
        options.add_option('wfprecision', 1)
        options.add_option('diffuseincomingresolution',hr_opt['diffuse_incoming_resolution'])
        options.add_option('numthreads', int(hr_opt['numthreads']))
        options.add_option('ManualDiffuseHeights',hr_opt['diffuse_point_altitudes'])
        #options.add_option('numthreads', 7)
        options.add_option('calcwf', np.int(hr_opt['use_wf'])) 
            
        return options

    def make_meas_inst(self):
        """
        Make the instrument that generates the measured level 1 data. Data is 
        loaded from the gridded level 1B product provided by NASA-GSFC.
        
        Returns
        -------
        meas_inst : :class:`~instruments.omps_model.OMPS`
        """
        filename  = omps_l2.l1_filename_for_orbit( self.orbitnumber, self.ret_opt['omps_l1_folder'], 'L1G' )
        meas_inst = omps.OMPS( filename, self.imagenumber, self.slitnumber )
        
        self.output['omps_geometry', 'image_index'] = self.imagenumber
        self.output['omps_geometry', 'sza'] = meas_inst.sza
        self.output['omps_geometry', 'mjd'] = meas_inst.mjd
        self.output['omps_geometry', 'latitude'] = meas_inst.lat
        self.output['omps_geometry', 'longitude'] = meas_inst.lon
        
        self.output['omps_measurement', 'radiance'] = meas_inst.rad
        self.output['omps_measurement', 'wavel_nm'] = meas_inst.wavel_nm
        self.output['omps_measurement', 'psf'] = meas_inst.bandpass
        
        return meas_inst
        
    def make_retrieval_inst(self):
        """
        Make the instrument that generates the simulated level 1 data for the
        retrieval. This is used to generate the measurement vector, f.
        
        Returns
        -------
        ret_inst : :class:`~instruments.limb_profiler.LimbProfiler_scan`
        """
        los, obs, mjd = self.meas_inst.radtran_geometry()
    
        #setup the simulated omps instrument
        ret_inst = instrument.LimbProfiler_scan( np.array([300]), los, obs,  obs, mjd, ideal=False )
        ret_inst.psf_wavel_nm    = self.meas_inst.psf_wavel_nm
        ret_inst.psf             = self.meas_inst.bandpass
        ret_inst.waveSamples     = self.ret_opt['wavel_samples']
        ret_inst.waveSamples_nm  = self.ret_opt['wavel_samples_nm']
        ret_inst.wavel_int_type  = self.ret_opt['wavel_int_type']
        ret_inst.wavel_num_stdev = self.ret_opt['wavel_num_stdev']
        
        #calculate the vertical resolution based on a FWHM of 1.6km (approximate only)
        #this only affects retreival if ret_opt['vert_samples'] > 1
        Re      = 6378.0
        x       = np.sqrt((Re+self.meas_inst.satalt/1000)**2-Re**2)
        theta   = np.arctan((1.6/2.3548)/x)   #convert FWHM to gaussian standard deviation
        
        #setup the instrument vertical resoluiton
        ret_inst.vertRes     = theta
        ret_inst.vertSamples = self.ret_opt['vert_samples']
        
        #initialze the instrument
        ret_inst.initialize_sensors_and_optics()
    
        return ret_inst
        
    def make_simulation_inst(self):
        """
        Make the instrument that generates the simulated level 1 data.
        
        Returns
        -------
        meas_inst : :class:`~instruments.limb_profiler.LimbProfiler_scan`
        """
        #TODO: Make this code that simulation settings can be adjusted by the calling
        #script
        
        #make a real omps instrument to get the geometries
        omps_inst = self.make_meas_inst()        
        los, obs, mjd = omps_inst.radtran_geometry()
        
        #setup the simulated omps instrument           
        meas_inst = instrument.LimbProfiler_scan( np.array([300]), los, obs,  obs, mjd, ideal=False )
        meas_inst.psf_wavel_nm = omps_inst.wavel_nm
        meas_inst.psf          = omps_inst.bandpass
        meas_inst.waveSamples  = self.sim_opt['wavel_samples']
        meas_inst._nm  = self.sim_opt['wavel_samples_nm']
        
        meas_inst.satalt   = omps_inst.satalt
        meas_inst.bandpass = omps_inst.bandpass
    
        #calculate the vertical resolution based on a FWHM of 1.6km (approximate only)
        #this only affects retreival if ret_opt['vert_samples'] > 1
        Re      = 6378.0
        x       = np.sqrt((Re+omps_inst.satalt/1000)**2-Re**2)
        theta   = np.arctan((1.6/2.3548)/x)   #convert FWHM to gaussian standard deviation
        
        #setup the instrument vertical resoluiton
        meas_inst.vertRes     = theta
        meas_inst.vertSamples = self.sim_opt['vert_samples']

        #initialze the instrument
        meas_inst.initialize_sensors_and_optics()
        meas_inst.psf_wavel_nm = omps_inst.wavel_nm
        
        return meas_inst
        
    def generate_measured_radiances(self, radtran):
        """
        Generate the simulated measurements used in the retrieval. Radiances are 
        simulated at the wavelengths needed by the retrieval species.

        This is only used in simulated retrievals
        """
        
        #TODO: add if statement to check which retrievals are needed
        #TODO: adjust wavelengths to actual OMPS values
        #get all the wavelengths used by the retrievals
        meas_wavel_nm = []
        meas_wavel_nm.append(self.species_o3.get_required_wavelengths())
        meas_wavel_nm.append(self.species_albedo.get_required_wavelengths())
        meas_wavel_nm.append(self.species_aero.get_required_wavelengths())
        meas_wavel_nm.append(self.species_aero_ps.get_required_wavelengths())
        meas_wavel_nm = np.unique(np.hstack(meas_wavel_nm))
        
        #setup the instrument with the rtm and the needed wavelengths
        self.meas_inst.setup_radtran_model(radtran)
        self.meas_inst.set_wavelengths(meas_wavel_nm)
        
        #do the calculation
        self.meas_inst.calculate_radiance()

    def setup_simulation_model(self, sim_options ):
        """
        Creates a radiative transfer model with the desired options and atmospheric state
        
        Parameters
        ----------
        
        sim_opt : dictionary
            Contains the sasktran HR settings, ie numdiffuseprofiles, numordersofscatter, etc.
            
        atmospheric_state : dictionary
            Contains the type of climatologies used for ozone, no2 and aerosol
            (labow, pratmo, userdefined, etc) and, if userdefined, their values. 
            
        Returns
        -------
        radiative transfer model : :class:`~radtran.SasktranHR`
            Pointer to the radiative transfer model instantiated in c++
            
        atmospheric state : :class:`~atmosphere.Atmosphere`
            Instance of an Atmosphere, which holds pointers to the optical properties
            and climatologies
        
        """
        #setup atmospheric species climatologies, optical properties and identifiers (guids)
        clim_aer, optprop_aer, guid_aer  = self.aerosol_apriori(        self.ret_opt['aerosol_apriori'], aer_filename=self.ret_opt['aerosol_apriori_filename'] )
        clim_o3,  optprop_o3,  guid_o3   = self.ozone_apriori(          self.ret_opt['ozone_apriori']   )
        clim_no2, optprop_no2, guid_no2  = self.no2_apriori(            self.ret_opt['no2_apriori']     )
        clim_bck, optprop_bck, guid_bck  = self.background_atmosphere(  self.ret_opt['background']      )
        albedo                           = self.albedo_apriori(         self.ret_opt['albedo_apriori']  )

        #Compute the amount of pre-convolution to apply to the ozone/no2 cross sections
        #at present ozone and no2 are the same
        ozone_psf_wavel, ozone_psf = self.ozone_convolution_parameters( self.ret_inst.psf_wavel_nm,
                                                                        self.ret_inst.psf,
                                                                        self.ret_opt['wavel_samples_nm'],
                                                                        self.ret_opt['wavel_samples'])

        if np.any(np.array(self.ret_opt['wavel_samples']) > 1):
            optprop_o3  = ozone.convolved_dbm_optical_property(    ozone_psf_wavel, ozone_psf, output_spacing=0.01  )
            optprop_no2 = no2.convolved_vandaele_optical_property( ozone_psf_wavel, ozone_psf, output_spacing=0.01  )
        else:
            optprop_o3  = ozone.convolved_dbm_optical_property(    ozone_psf_wavel, ozone_psf )
            optprop_no2 = no2.convolved_vandaele_optical_property( ozone_psf_wavel, ozone_psf )
            
        #setup apriori profiles
        apriori_alts        = np.arange(500.0, 100500.0, 1000)
        o3_apriori_profile  = clim_util.height_profile_from_clim(clim_o3, guid_o3,   apriori_alts,
                                                                 self.output['omps_geometry', 'latitude'], 
                                                                 self.output['omps_geometry', 'longitude'], 
                                                                 self.output['omps_geometry', 'mjd'] )
                                                                 
        aer_apriori_profile = clim_util.height_profile_from_clim(clim_aer, guid_aer,  apriori_alts, 
                                                                 self.output['omps_geometry', 'latitude'], 
                                                                 self.output['omps_geometry', 'longitude'], 
                                                                 self.output['omps_geometry', 'mjd'] )
        
        #setup atmosphere and add species
        atmo_ret = atmosphere.Atmosphere()
        atmo_ret.add_species( clim_bck, optprop_bck, guid_bck )
        atmo_ret.backgroundclim = clim_bck

        if sim_options['add_no2']:
            atmo_ret.add_species( clim_no2, optprop_no2, guid_no2 )
        atmo_ret.add_user_defined( guid_aer, optprop_aer, apriori_alts, copy.copy(aer_apriori_profile), 'limb_aerosol', doPiecewiseLinear=1 )
        atmo_ret.add_user_defined( guid_o3,  optprop_o3,  apriori_alts, copy.copy(o3_apriori_profile),  'limb_ozone' )
        atmo_ret.set_wf_species( optprop_aer )
        atmo_ret.albedo = albedo
        
        #setup radiative transfer model
        # sktran_sim   = radtran.SasktranHR( atmo_ret, self.HR_options(sim_options)  )
        sktran_sim   = radtran.SasktranHRMSApprox( atmo_ret, self.HR_options(sim_options), ms_wavel_space=self.ret_opt['ms_wavel_space']  )
        sktran_sim.do_ss_approx = self.ret_opt['do_ss_approx']
        
        return sktran_sim, atmo_ret

    def ozone_apriori(self, ozone_clim='labow', filename=None, o3_profile=None, o3_heights=None, xsec='O3_DBM' ):

        if ozone_clim == 'labow':
            clim, optprop,   guid   = ozone.labow(xsec)
        elif ozone_clim == 'user_defined':
            clim, optprop,   guid   = ozone.userdefined(o3_profile, o3_heights, cross_section=xsec)
        elif ozone_clim == 'omps_l2':
            data = omps_l2.load_l2_from_orbit(orbit=self.orbitnumber,truncate=False,folder=self.ret_opt['omps_l2_folder'])
            clim, optprop,   guid   = ozone.userdefined(np.double(data['o3'][:,self.imagenumber]), np.double(data['alt']*1000), cross_section=xsec)
        else:
            logging.warning('that ozone option is not supported, using labow')
            clim, optprop,   guid   = ozone.labow('O3_DBM')

        return clim, optprop, guid

    def aerosol_apriori(self, aer_clim='user_defined', aer_filename='aerosol_apriori.txt', partsize_clim='user_defined',partesize_filename='particlesize_apriori.txt' ):
        user_paths = utility.data_folder()

        #load in the aerosol climatology
        aero = 0
        if aer_clim == 'user_defined':
            aero = np.loadtxt(user_paths+aer_filename)
        elif aer_clim == 'omps_l2':
            data = omps_l2.load_l2_from_orbit(orbit=self.orbitnumber,truncate=False,folder=self.ret_opt['omps_l2_folder'])
            aero = np.zeros((100, 2))
            aero[:,0] = data['alt']*1000
            aero[:,1] = data['aer'][:,self.imagenumber]/1.246e-5
        else:
            aero = np.zeros((100, 2))
            aero[:,0] = np.arange(0, 100500, 1000)
            logging.warning('no aerosol climatology present, setting profile to zero')
        if isinstance(aero, int):
            logging.warning('could not find aerosol apriori file')
        aer_heights = aero[:,0]
        aer_profile = aero[:,1]

        #load in the particle size climatology
        if partsize_clim == 'user_defined':
            ps = np.loadtxt(user_paths + partesize_filename)
            ps_heights  = ps[:,0]
            rg_profile  = ps[:,1]
            mg_profile  = ps[:,2]
        else:
            ps_heights  = np.arange(500, 100500, 1000)
            rg_profile  = np.ones(np.shape(ps_heights))*0.08
            mg_profile  = np.ones(np.shape(ps_heights))*1.6
            logging.warning('no particle size climatology present, setting mode radius to 80nm, mode width to 1.6')

        clim, optprop,  guid  = aerosol.from_arrays( aer_profile, aer_heights,
                                                                 mode_radius = rg_profile,
                                                                 mode_width  = mg_profile,
                                                                 particlesize_heights = ps_heights,
                                                                 DoLogInterpolation=0, DoPiecewiseLinear=1)

        return clim, optprop, guid

    def no2_apriori(self, no2_clim='pratmo', xsec='NO2_VANDAELE1998'):
        if no2_clim == 'pratmo':
            clim, optprop,  guid  = no2.pratmo(xsec)
        else:
            logging.warning('only pratmo is currently supported')
            clim, optprop,  guid  = no2.pratmo(xsec)

        return clim, optprop, guid

    def background_atmosphere(self, background_clim='ecmwf'):

        if background_clim == 'ecmwf':
            clim, optprop, guid = background.ecmwf()
        elif background_clim == 'msis':
            clim, optprop, guid = background.msis()
        elif background_clim == 'omps_anc':
            clim, optprop, guid = background.omps_anc(self.orbitnumber,self.imagenumber,self.slitnumber,self.ret_opt['omps_anc_folder'], keyword='ANC')
        else:
            logging.warning('only ecmwf, msis and omps_anc are currently supported, setting to ecmwf')
            clim, optprop, guid = background.ecmwf()

        return clim, optprop, guid

    def albedo_apriori(self, albedo_clim=None):

        if albedo_clim is None:
            albedo = 0.3
        elif isinstance(albedo_clim, float):
            albedo = albedo_clim
        elif albedo_clim == 'omps_l2':
            data = omps_l2.load_l2_from_orbit(orbit=self.orbitnumber,truncate=False,folder=self.ret_opt['omps_l2_folder'])
            albedo = data['alb'][self.imagenumber]
        else:
            logging.warning('albedo climatology not recognized, setting to 0.3')
            albedo = 0.3

        return albedo

    def retrieve_profiles(self):
        """
        The high level retrieval script that puts together the instruments, atmosphere,
        radiative transfer model and method to retrieve the desired species.
        """
        #set variables to defaults if required
        self.check_retrieval_options()
        
        #setup the instrument used to generate measurements
        if self.do_simulation:
            self.meas_inst = self.make_simulation_inst()
        else:
            self.meas_inst = self.make_meas_inst()
        
        #make the instrument used to simulate measurements
        self.ret_inst = self.make_retrieval_inst()     
        
        #setup the atmosphere and radiative transfer model
        self.sktran_ret, self.atmo_ret = self.setup_simulation_model(self.ret_opt)
        
        #attach radiative transfer model to instrument
        self.ret_inst.setup_radtran_model(self.sktran_ret)
        
        #Create the retrieval method attach instruments to it
        self.ret = retrieval.RetrievalLimb( self.meas_inst, self.ret_inst )
        
        #setup retrieval species
        self.species_aero    = species_aer.AerosolLimbProfile( self.atmo_ret )
        if self.ret_opt['aerosol_vectors'] is not None:
            self.species_aero.triplet = []
            self.species_aero.normalt = []
            for wvl,wght,ref in zip(self.ret_opt['aerosol_vectors']['wavelengths'], self.ret_opt['aerosol_vectors']['weights'], self.ret_opt['aerosol_vectors']['ref_alts']):
                self.species_aero.triplet.append((wvl, wght))
                self.species_aero.normalt.append((ref))

        self.species_o3      = species_o3.OzoneLimbProfile(    self.atmo_ret )
        if self.ret_opt['ozone_vectors'] is not None:
            self.species_o3.triplet = []
            self.species_o3.normalt = []
            for wvl,wght,ref in zip(self.ret_opt['ozone_vectors']['wavelengths'], self.ret_opt['ozone_vectors']['weights'], self.ret_opt['ozone_vectors']['ref_alts']):
                self.species_o3.triplet.append((wvl, wght))
                self.species_o3.normalt.append((ref))
            self.species_o3.wik_file = self.ret_opt['ozone_vectors']['wik_file']

        self.species_albedo  = species_alb.AlbedoLimbProfile(  self.atmo_ret )
        #self.species_aero_ps = species_aer_ps.AerosolPartSizeLimbProfile( self.atmo_ret, apriori_atmo )
          
        #if we are using a simulated instrument instead of real data create it here.
        if self.do_simulation:
            #TODO: setup atmopsheric state with input files for simulations
            self.sktran_sim, self.atmo_sim = self.setup_simulation_model(self.sim_opt)
            self.generate_measured_radiances(self.sktran_sim)
            self.output['omps_measurement', 'radiance'] = self.meas_inst.rad
        
        #setup the retrieval method (max iter is later over-written with user-specified value)
        save_output = False
        if self.ret_opt['verbose_output'] > 1:
            save_output = True

        #do each retrieval
        for idx, (spec,o) in enumerate(zip(self.ret_opt['retrieval_order'], self.ret_opt['max_iter'])):

            if self.ret_opt['retrieval_method'][idx] == 'MART':
                self.method  = lsq.MART(numiter=7, save_output=save_output)
            elif self.ret_opt['retrieval_method'][idx] == 'LM':
                self.method  = lsq.LevenbergMarquardt(damping=0.1, maxiter=7, save_output=save_output,test_convergence=True)
            elif self.ret_opt['retrieval_method'][idx].lower() == 'rodgers':
                self.method = rodgers.Rodgers()
            else:
                logging.error('unknown retrieval method')

            #update instrument with wavelength oversampling
            self.ret_inst.waveSamples  = self.ret_opt['wavel_samples']
            self.ret_inst.waveSamples_nm = self.ret_opt['wavel_samples_nm']
            self.ret_inst.initialize_sensors_and_optics()
            self.sktran_ret.set_num_orders_of_scatter( self.ret_opt['scatterorder'][idx] )
            #do the retrieval
            data = self.ret_func_dict[spec](o)
            #store the debug data
            self.output['intermediate', idx] = data

        #store some more data
        self.output['radiances','meas_rad']       = self.meas_inst.rad
        self.output['radiances','meas_wavel']     = self.meas_inst.wavel_nm
        self.output['retrieval_info','image_num'] = self.imagenumber
        self.output['retrieval_info','orbit_num'] = self.orbitnumber
        self.output['retrieval_info','slit_num']  = self.slitnumber
        self.output['retrieval_info','ret_opt']   = self.ret_opt
        self.output['retrieval_info','version']   = self.version()
        return self.output
        
    def retrieve_ozone(self, maxiter):
        """
        Performs an ozone retrieval using the method, radiative transfer model,
        and species defined in retrieve_profiles(). Retrieval details are contained
        in the species class.
        
        """
        
        #set the number of iterations
        self.method.maxiter = maxiter
        
        #Tell the radiative transfer model what species we are retrieving so it
        #can calculate the appropriate weighting functions
        self.atmo_ret.set_wf_species( self.atmo_ret.user_optprop['limb_ozone'] )
        
        #set the 'cloud top height' by finding the highest altitude that exceeds
        #the allowed max aerosol value
        high_aer = self.species_aero.get_atmospheric_profile() >= self.species_aero.max_ext_val
        if np.any(high_aer):
            self.cloud_height = np.max(self.species_aero.profile_alts[high_aer])
        else:
            self.cloud_height = 0.0
        
        #load tropopause height (in km) and set as tropopause height (in meters)
        anc_file = omps_l2.l1_filename_for_orbit( self.orbitnumber, self.ret_opt['omps_anc_folder'], 'ANC')
        f = h5py.File(anc_file)
        trop_height = f['/GRIDDED_DATA/TropopauseAltitude'][self.imagenumber, self.slitnumber]*1000
        
        #set the minimum altitude to the higher of cloud and 2km below tropopause
        #add 500m to altitude to ensure last point is above this
        self.species_o3.min_alt = np.max( np.array([self.cloud_height, trop_height-2000])+500)
        
        #do the retrieval
        self.ret.retrieve_species( self.species_o3, self.method, do_error=True )
        
        #compute the measured/retrieved measurement vectors for storage
        self.ymeas_o3,self.dymeas_o3   = self.species_o3.measurement_vector(self.ret.meas_inst, do_error=True)
        self.yret_o3,e    = self.species_o3.measurement_vector(self.ret.ret_inst)
        
        rad_ret   = self.ret.ret_inst.rad
        wavel_ret = self.ret.ret_inst.wavel_nm

        #create the debugging output variables
        debug = dict()
        debug['ymeas']     = self.ymeas_o3
        debug['yret']      = self.yret_o3
        debug['rad_ret']   = rad_ret
        debug['wavel_ret'] = wavel_ret

        measrad = np.zeros(self.ret_inst.rad.shape)
        for idx, w in enumerate(self.ret_inst.wavel_nm[0]):
            widx = np.argmin(np.abs(self.meas_inst.wavel_nm - w))
            measrad[:,idx] = self.meas_inst.rad[:,widx]

        self.output['radiances', 'o3_measrad']   = measrad

        if self.ret_opt['verbose_output'] > 1:
            debug['lsq_output'] = self.method.output
            self.ozone_residual = self.method.output['residual'][-1]

        #save output
        self.output['retrieved', 'o3']       = self.species_o3.output_data()
        self.output['retrieved', 'o3_valid'] = self.species_o3.output_data_bounded()
        self.output['retrieved', 'o3_ymeas'] = self.ymeas_o3
        self.output['retrieved', 'o3_yret']  = self.yret_o3
        self.output['retrieved', 'o3_res']   = self.ozone_residual
        self.output['radiances', 'o3_retrad']= self.ret_inst.rad
        self.output['radiances', 'o3_wavel'] = self.ret_inst.wavel_nm[0]

        return self.species_o3.output_data(), debug
    
    def retrieve_aerosol(self, maxiter):
        
        self.method.maxiter = maxiter
        self.atmo_ret.set_wf_species( self.atmo_ret.user_optprop['limb_aerosol'] )
        self.ret.retrieve_species( self.species_aero, self.method, do_error=True )
        
        self.ymeas_aer,self.dymeas_aer   = self.species_aero.measurement_vector(self.ret.meas_inst, do_error=True)
        self.yret_aer,e    = self.species_aero.measurement_vector(self.ret.ret_inst)
        
        rad_ret   = self.ret.ret_inst.rad
        wavel_ret = self.ret.ret_inst.wavel_nm

        debug = dict()
        debug['ymeas']     = self.ymeas_aer
        debug['yret']      = self.yret_aer
        debug['rad_ret']   = rad_ret
        debug['wavel_ret'] = wavel_ret

        measrad = np.zeros(self.ret_inst.rad.shape)
        for idx, w in enumerate(self.ret_inst.wavel_nm[0]):
            widx = np.argmin(np.abs(self.meas_inst.wavel_nm - w))
            measrad[:,idx] = self.meas_inst.rad[:,widx]

        self.output['radiances', 'aer_measrad']   = measrad

        if self.ret_opt['verbose_output'] > 1:
            debug['lsq_output'] = self.method.output
            self.aerosol_residual = debug['lsq_output']['residual'][-1]

        self.output['retrieved', 'aero']       = self.species_aero.output_data()
        self.output['retrieved', 'aero_valid'] = self.species_aero.output_data_bounded()
        self.output['retrieved', 'aer_ymeas']  = self.ymeas_aer
        self.output['retrieved', 'aer_dymeas'] = self.dymeas_aer
        self.output['retrieved', 'aer_yret']   = self.yret_aer
        self.output['retrieved', 'aer_ref_alts'] = np.asarray(self.species_aero.tanalts)[np.asarray(self.species_aero.ref_idx)]
        self.output['retrieved', 'aer_res']    = self.aerosol_residual
        self.output['radiances', 'aer_retrad'] = self.ret_inst.rad
        self.output['radiances', 'aer_wavel']  = self.ret_inst.wavel_nm[0]


        return self.species_aero.output_data(), debug
    
    def retrieve_albedo(self, maxiter):
        
        self.method.maxiter = maxiter
        self.ret.retrieve_species( self.species_albedo, self.method, do_error=True )
        
        self.ymeas_alb,self.dymeas_alb   = self.species_albedo.measurement_vector(self.ret.meas_inst)
        self.yret_alb,e    = self.species_albedo.measurement_vector(self.ret.ret_inst)
        
        rad_ret   = self.ret.ret_inst.rad
        wavel_ret = self.ret.ret_inst.wavel_nm

        debug = dict()
        debug['ymeas']     = self.ymeas_alb
        debug['yret']      = self.yret_alb
        debug['rad_ret']   = rad_ret
        debug['wavel_ret'] = wavel_ret

        measrad = np.zeros(self.ret_inst.rad.shape)
        for idx, w in enumerate(self.ret_inst.wavel_nm[0]):
            widx = np.argmin(np.abs(self.meas_inst.wavel_nm - w))
            measrad[:,idx] = self.meas_inst.rad[:,widx]

        self.output['radiances', 'alb_measrad']   = measrad

        if self.ret_opt['verbose_output'] > 1:
            debug['lsq_output'] = self.method.output
            self.albedo_residual = debug['lsq_output']['residual'][-1]

        #save output
        self.output['retrieved', 'albedo']     = self.species_albedo.output_data()
        self.output['retrieved', 'alb_ymeas']  = self.ymeas_alb
        self.output['retrieved', 'alb_yret']   = self.yret_alb
        self.output['retrieved', 'alb_res']    = self.albedo_residual
        self.output['radiances', 'alb_retrad']    = self.ret_inst.rad
        self.output['radiances', 'alb_wavel']  = self.ret_inst.wavel_nm[0]
                
        return self.species_albedo.output_data(), debug
    
    def retrieve_particlesize(self, maxiter):
        
        self.method.maxiter = maxiter
        self.atmo_ret.set_wf_species( self.atmo_ret.user_optprop['limb_aerosol'] )
        self.ret.retrieve_species( self.species_aero_ps, self.method )       
        
        self.ymeas_ps,e   = self.species_aero_ps.measurement_vector(self.ret.meas_inst)
        self.yret_ps,e    = self.species_aero_ps.measurement_vector(self.ret.ret_inst)
        
        rad_ret   = self.ret.ret_inst.rad
        wavel_ret = self.ret.ret_inst.wavel_nm

        debug = dict()
        debug['ymeas']     = self.ymeas_ps
        debug['yret']      = self.yret_ps
        debug['rad_ret']   = rad_ret
        debug['wavel_ret'] = wavel_ret

        self.output['retrieved', 'particlesize'] = self.species_aero_ps.output_data()
        if self.ret_opt['verbose_output'] > 1:
            debug['lsq_output'] = self.method.output       
            
        return self.species_aero_ps.output_data(), debug        

    def ozone_convolution_parameters( self, psf_wavel_nm, psf, samples_wavel_nm, samples):
        """
        Determines the gaussian stand deviation used to convolve the ozone cross-section before the radiative
        transfer calculation. Strictly speaking this should be zero, however pre-convolution
        can greatly speed of radiative transfer calculations.

        Returns the gaussian std dev as a function of wavelength based on instrument point spread
        function and whether or not the radiances are being oversampled.
        """

        #interpolate the number of samples onto the wavelength grid
        if isinstance(samples,int):
            conv_psf_wavel_nm = psf_wavel_nm
            samples_int = np.zeros(np.shape(conv_psf_wavel_nm)) + samples
            conv_psf    = np.interp( conv_psf_wavel_nm, psf_wavel_nm, psf )
        else:
            #create a grid containing both the point spread function wavelengths
            #and the sampling wavelengths to avoid interpolation errors
            conv_psf_wavel_nm = np.unique(np.concatenate( (psf_wavel_nm, samples_wavel_nm)))

            samples_int = np.round(np.interp(conv_psf_wavel_nm, samples_wavel_nm, samples), decimals=0).astype(int)
            #make sure its an odd number
            samples_int += (1-(samples_int % 2))
            conv_psf    = np.interp( conv_psf_wavel_nm, psf_wavel_nm, psf )

        # if the radiances are  being oversampled, do not convolve the cross-section
        conv_psf[ samples_int > 1 ] = 0

        return conv_psf_wavel_nm, conv_psf
        
    def model_full_spectrum(self):
        """
        Compute the modelled radiance at all wavelengths in the measurement instrument
        This should be called after a retrieval has been performed.
        
        Returns
        -------
        rad : 2darray
            An array containg wavelength x radiance
        """        
        
        self.ret_inst.set_wavelengths(self.meas_inst.wavel_nm)
        self.ret_inst.calculate_radiance()      
        
        return self.ret_inst.rad
        
    def model_full_spectrum_using_L2(self, wavelengths=None, save_to_file=False):
        """
        Compute the modelled radiance at all wavelengths in the measurement instrument
        using ozone, aerosol and albedo values retrieved previously and stored in an
        output hdf5 file.
        
        Returns
        -------
        rad : 2darray
            An array containg wavelength x radiance
        """        
        
        #set variables to defaults if required
        self.check_retrieval_options()
        
        #setup the instrument used to generate measurements
        self.meas_inst = self.make_meas_inst()
        
        #make the instrument used to simulate measurements
        self.ret_inst = self.make_retrieval_inst()     
        self.sktran_ret, self.atmo_ret = self.setup_simulation_model(self.ret_opt)
        
        #attach radiative transfer model to instrument
        self.ret_inst.setup_radtran_model(self.sktran_ret)

        #use the OMPS instrument wavelengths unless otherwise specified
        if wavelengths is None:
            self.ret_inst.set_wavelengths(self.meas_inst.wavel_nm)
        else:
            self.ret_inst.set_wavelengths(wavelengths)

        self.ret_inst.calculate_radiance()

        if save_to_file:
            folder = r'\\utls\\users\\scratch\\OMPS_L2_OSIRIS\\'
            file   = folder + 'OMPS_L2_radiance_' + str(self.orbitnumber) + '_' + str(self.imagenumber) + '.h5'

            f = h5py.File( file, 'a' )
            f.create_dataset('/Data/Wavelength',   data=self.meas_inst.wavel_nm)
            f.create_dataset('/Data/PSF',          data=self.meas_inst.bandpass)
            f.create_dataset('/Data/MeasRad',      data=self.meas_inst.rad)
            f.create_dataset('/Data/ModRad',       data=self.ret_inst.rad)
            f.close()
        
        return self.ret_inst.rad