import os

import numpy as np
import h5py
import logging

from astropy.time import Time
import scipy.constants as const

from retrieval.instruments.spectral_imager import SpectralImager
import retrieval.instruments.sensors.sensor as sensor
import retrieval.instruments.optics.optics as optics
import radtran.geometry as geometry


class OMPSModel(SpectralImager):
    #TODO: get rid of the text file once level one serivices is up and running
    def __init__(self, wavel_nm, los, vertical=None, obs=None, ideal=True, psf_file='instruments\\osiris_psf.txt' ):
        super().__init__()
        
        if ideal:
            self.sensor         = sensor.NxMSensor(100,3,0,0)
            self.optics         = optics.BasicOptics(0,0,np.array([0, 10000]),np.array([0.35, 0.35]),1,1,1)
        else:
            path = os.environ['PYTHONPATH'].split(os.pathsep)
            file_name = path[0]+'\\'+psf_file
            t = np.loadtxt(file_name)
            t = np.vstack((t,np.array([10000, t[-1,1]])))   #extend wavelengths
            t = np.vstack((np.array([0, t[1,1]]),t))
            self.sensor         = sensor.NxMSensor(1,1,0,0)
            self.optics         = optics.BasicOptics(3.54e-4,0.0141,t[:,0],t[:,1]/2.35482,3,3,11)
        
        self.wavel_nm       = wavel_nm
        self.los            = los
        self.mjd            = []
        self.rad            = []
        self.pixel_los      = []
        
        self.hires_wavel    = []
        self.hires_los      = []
        self.los_weights    = []
        self.wavel_weights  = []
        
        self.setup_instrument( wavel_nm, los, vertical)
        
#this is currently a combination instrument/platform, maybe not ideal
class OMPS(object):
    
    def __init__(self, filename, scan_idx, sensor_idx):
        
        self.max_alt = 75000
        f = h5py.File(r''+filename,'r')

        # in the NOAA input files GRIDDED_DATA_ORIGINAL is used instead of GRIDDED_DATA
        if 'GRIDDED_DATA_ORIGINAL' in f:
            gridded_group_name = 'GRIDDED_DATA_ORIGINAL'
        else:
            gridded_group_name = 'GRIDDED_DATA'

        # Somewhat of a hack, v2.0-2.4 use a Wavelength field that may vary, but 2.5 removed this
        # In 2.5 the wavelengthgrid field is also length 266, while the dimension in the radiance stays the same so we
        # have to cutoff the fields based on how long the wavelength array is
        try:
            self.wavel_nm    = (f['/{}/Wavelength'.format(gridded_group_name)][scan_idx, sensor_idx, :])*1000    #convert to nanometers
        except:
            self.wavel_nm = (f['/{}/WavelengthGrid'.format(gridded_group_name)][:]) * 1000
        self.tanalts     = (f['/{}/TangentHeight'.format(gridded_group_name)][scan_idx, sensor_idx, :])*1000 #altitude in meters
        self.rad         = f['/{}/Radiance'.format(gridded_group_name)][scan_idx, sensor_idx, :, :len(self.wavel_nm)]           #radiance in W/m^2/nm/sr
        self.snr         = f['/{}/SNR'.format(gridded_group_name)][scan_idx, sensor_idx, :, :len(self.wavel_nm)]
        self.bandpass    = f['/{}/Bandpass'.format(gridded_group_name)][scan_idx, sensor_idx, :len(self.wavel_nm)]*1000/2.3548        #gaussian standard deviation [nm]
        
        #get rid of nonsensical data
        good          = self.wavel_nm>0
        self.wavel_nm = self.wavel_nm[good]
        self.psf_wavel_nm = self.wavel_nm
        self.bandpass = self.bandpass[good]
        self.rad      = self.rad[:,good]
        self.snr      = self.snr[:,good]  
        
        self.rad      = self.rad[self.tanalts < self.max_alt,:]
        self.snr      = self.snr[self.tanalts < self.max_alt,:]
        self.tanalts  = self.tanalts[self.tanalts < self.max_alt]

        if np.any(self.tanalts == -999000):
            logging.error('at least one tangent altitude is invalid')
        
        #select good radiances and conver to photons/cm^2/s/nm/sr
        self.rad[ self.rad < 0] = np.nan
        E = const.h*const.c/self.wavel_nm*1e9 #energy per photon       
        self.rad = self.rad/E/(100**2)
        # self.rad_error   = self.rad / self.snr  #not recommended for scientific use
        self.rad_error     = self.rad * 0.01      # assume 1% radiance uncertainty

        #tangent point geographic info
        self.lat  = f['/GEOLOCATION_DATA/Latitude_35km'][scan_idx, sensor_idx]
        self.lon  = f['/GEOLOCATION_DATA/Longitude_35km'][scan_idx, sensor_idx]
        self.sza  = f['/GEOLOCATION_DATA/SolarZenithAngle_35km'][scan_idx, sensor_idx]
        self.saa  = f['/GEOLOCATION_DATA/SolarAzimuth_35km'][scan_idx, sensor_idx]
        
        #satelite geographic info
        self.time    = f['/{}/DateTimeUTC'.format(gridded_group_name)][scan_idx]
        self.satlat  = f['/GEOLOCATION_DATA/SpacecraftLatitude'][scan_idx]
        self.satlon  = f['/GEOLOCATION_DATA/SpacecraftLongitude'][scan_idx]
        self.sata    = f['/GEOLOCATION_DATA/SatelliteAzimuth_35km'][scan_idx, sensor_idx]
        self.satalt  = (f['/GEOLOCATION_DATA/SpacecraftAltitude'][scan_idx, sensor_idx])*1000
        
        t            = Time(self.time.decode("utf-8")[0:-2])
        self.mjd     = t.mjd
                
    def calculate_radiance(self):
        return self.rad
        
        
    def radtran_geometry(self):
        #TODO: most of this should probably be in an instrument class
        
        geo = geometry.skif.ISKGeodetic()
        geo2 = geometry.skif.ISKGeodetic()
        geo.SetProperty('WGS84', 0)
        geo2.SetProperty('WGS84', 0)
            
        if self.satlon < 0:
            satlon = self.satlon + 360
        else:
            satlon = self.satlon            
            
        geo.SetLocationLatLonAlt(self.satlat,satlon, np.double(self.satalt) )
        sat_xyz = geo.GetLocationXYZ()
        
        if self.lon < 0:
            lon = self.lon + 360
        else:
            lon = self.lon
        
        geo.SetLocationLatLonAlt( np.double(self.lat), np.double(lon), 0)
        tp_xyz = geo.GetLocationXYZ()

        boresight = sat_xyz-tp_xyz  #vector from satellite to tangent point at 35km
        boresight = boresight/np.sqrt(np.sum(boresight**2))
        
        obs = [0,]*len(self.tanalts)
        los = [0,]*len(self.tanalts)
        mjd = [0,]*len(self.tanalts)
        for i in range(len(self.tanalts)):
            ok, nxlos = geo.SetLocationFromTangentAltitude(np.double(self.tanalts[i]), sat_xyz, boresight)

            geo2.SetLocationFromTangentPoint(sat_xyz, nxlos)
            if np.linalg.norm(geo2.GetLocationXYZ() - tp_xyz ) > 200000:
                #REDO with negative boresight
                ok, nxlos = geo.SetLocationFromTangentAltitude(np.double(self.tanalts[i]), sat_xyz, -1*boresight)
                geo2.SetLocationFromTangentPoint( sat_xyz, nxlos)
                if np.linalg.norm(  geo2.GetLocationXYZ() - tp_xyz ) > 200000:
                    logging.error('could not replicate satellite geometry')
                    raise
            
            obs[i] = sat_xyz
            los[i] = nxlos
            mjd[i] = np.array([self.mjd])
            
        return los, obs, mjd

