import sqlite3
import numpy as np
import omps.level2 as l2
import logging
from filelock import FileLock

class OMPS_jobs:
    def __init__(self, database = 'omps_jobs.db', table = 'jobs', db_folder = '\\\\UTLS\\Users\\scratch\\OMPS_L2_Testing\\'):

        self.database = db_folder + database
        self.database_lock = db_folder + database + '.lock'
        self.table    = table
        self.timeout  = 60

    def make_new_table(self):

        conn = sqlite3.connect(self.database)
        c = conn.cursor()

        c.execute("CREATE TABLE " + str(self.table) + " (profileID PRIMARY KEY, isReady, isProcessing, isFinished, failed, priority DEFAULT(10))")

        conn.commit()
        conn.close()

    def add_orbit_to_table(self, orbit, slits=[1], priority=10,folder = r"//UTLS/Users/scratch/OMPS_L1/"):

        try:
            num_images = l2.get_num_images_in_orbit(orbit,folder)
            for image in range(num_images):
                for slit in slits:
                    self.add_entry_to_table(orbit, image, slit, priority)
        except:
            print('could not add orbit ' + str(orbit) + ' to the database')

    def add_entry_to_table(self, orbit, image, slit, priority=10):

        if hasattr(orbit,'len') & ~hasattr(orbit,'len'):
            priority = np.ones(np.shape(orbit))*priority

        conn = sqlite3.connect(self.database)
        c = conn.cursor()

        if hasattr(orbit,'len'):
            for o,i,s,p in zip(orbit,image,slit,priority):
                job = (int(o*10000+i*10+s), 1, 0, 0, 0, int(p))
                c.execute('INSERT OR IGNORE INTO ' + self.table + ' VALUES (?,?,?,?,?,?)',job)
        else:
            job = (int(orbit*10000+image*10+slit), 1, 0, 0, 0, int(priority))
            c.execute('INSERT OR IGNORE INTO ' + self.table + ' VALUES (?,?,?,?,?,?)',job)
        conn.commit()
        conn.close()

    def add_entry_to_table_from_txt(self, filename, all_slits=False):

        temp = np.ndfromtxt(filename,dtype=int)

        if all_slits is True:
            slits = [0,1,2]
        else:
            slits = [1]

        if len(temp.shape) == 1:
            p = np.arange(np.shape(temp)[0],0,-1)   #order priority same as text file
            for idx,orbit in enumerate(temp):
                self.add_orbit_to_table(orbit,slits, priority=p[idx])
        elif np.shape(temp)[1] == 3:
            for idx,t in enumerate(temp):
                self.add_entry_to_table(t[0],t[1],t[2])
        elif np.shape(temp)[1] == 4:
            for idx,t in enumerate(temp):
                self.add_entry_to_table(t[0],t[1],t[2],t[3])
        elif np.shape(temp)[1] == 2:
            for t in temp:
                self.add_orbit_to_table(t[0],slits, priority=t[1])
        elif np.shape(temp)[1] == 1:
            p = np.arange(np.shape(temp)[0],0,-1)   #order priority same as text file
            for idx,orbit in enumerate(temp):
                self.add_orbit_to_table(orbit,slits, priority=p[idx])

    def get_next_entry(self):

        temp = []
        try:
            conn = sqlite3.connect(self.database, timeout=self.timeout)
            c = conn.cursor()
            data = c.execute('SELECT * FROM ' + self.table + ' WHERE isReady=1 AND isProcessing=0 AND isFinished=0 ORDER BY priority DESC LIMIT 1')
            temp = data.fetchall()
            conn.close()
        except sqlite3.Error as e:
            logging.error("An error occurred accessing the sqlite database : ", e.args[0])

        if len(temp) == 0:
            orbit = None
            image = None
            slit  = None
        else:
            temp = temp[0]
            orbit = int(np.floor(temp[0]/10000))
            image = int(np.floor((temp[0]-orbit*10000)/10))
            slit  = int(temp[0]- orbit*10000-image*10)

        return orbit, image, slit

    def db_execute_and_commit(self, conn, com, com_var=None):

        #lock the database before commiting
        lock = FileLock(self.database_lock,timeout=60)

        with lock:
            try:
                c = conn.cursor()
                if com_var is None:
                    c.execute(com)
                else:
                    c.execute(com, com_var)
                conn.commit()
            except:
                logging.error('database_commit_failed')


    def set_entry_as_processing(self, orbit, image, slit):

        conn = sqlite3.connect(self.database, timeout=self.timeout)
        self.db_execute_and_commit(conn,'UPDATE ' + self.table + ' SET isProcessing=1 WHERE profileID=?',(orbit*10000 + image*10 + slit,))
        conn.close()

    def set_entry_as_finished(self, orbit, image, slit, failed=0):

        conn = sqlite3.connect(self.database, timeout=self.timeout)
        self.db_execute_and_commit(conn,'UPDATE ' + self.table + ' SET isFinished=?, isProcessing=?, failed=? WHERE profileID=?', (1,0,failed,orbit*10000 + image*10 + slit))
        conn.close()

    def reset_processing_entries_as_ready(self):

        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        c.execute('UPDATE ' + self.table + ' SET isProcessing=0 WHERE isProcessing=1 AND isFinished=0')
        conn.commit()
        conn.close()

    def reset_all_entries(self):

        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        c.execute('UPDATE ' + self.table + ' SET isProcessing=0, isFinished=0, failed=0')
        conn.commit()
        conn.close()

    def view_table(self):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        #print the last 100 to be processed
        for row in c.execute('SELECT * FROM (SELECT * FROM ' + self.table + ' WHERE isFinished=1 OR isProcessing=1 ORDER BY priority ASC LIMIT 100) ORDER BY priority DESC LIMIT 100'):
            print(row)

        #print the next 100 to be processed
        for row in c.execute('SELECT * FROM ' + self.table + ' WHERE isFinished=0 AND isProcessing=0 ORDER BY priority DESC LIMIT 100'):
            print(row)

        conn.close()

class OMPS_jobs_orbit(OMPS_jobs):

    def add_entry_to_table(self, orbit, slit, priority=10):

        if hasattr(orbit,'len') & ~hasattr(orbit,'len'):
            priority = np.ones(np.shape(orbit))*priority

        conn = sqlite3.connect(self.database)
        c = conn.cursor()

        if hasattr(orbit,'len'):
            for o,p in zip(orbit,slit, priority):
                job = (int(o*10 + slit), 1, 0, 0, 0, int(p))
                c.execute('INSERT OR IGNORE INTO ' + self.table + ' VALUES (?,?,?,?,?,?)',job)
        else:
            job = (int(orbit*10+slit), 1, 0, 0, 0, int(priority))
            c.execute('INSERT OR IGNORE INTO ' + self.table + ' VALUES (?,?,?,?,?,?)',job)
        conn.commit()
        conn.close()

    def get_next_entry(self):

        temp = []
        try:
            conn = sqlite3.connect(self.database, timeout=self.timeout)
            c = conn.cursor()
            data = c.execute('SELECT * FROM ' + self.table + ' WHERE isReady=1 AND isProcessing=0 AND isFinished=0 ORDER BY priority DESC LIMIT 1')
            temp = data.fetchall()
            conn.close()
        except sqlite3.Error as e:
            logging.error("An error occurred accessing the sqlite database : ", e.args[0])

        if len(temp) == 0:
            orbit = None
            slit  = None
        else:
            temp = temp[0]
            orbit = int(np.floor(temp[0]/10))
            slit  = int(temp[0]- orbit*10)

        return orbit, slit

    def set_entry_as_processing(self, orbit, slit):

        conn = sqlite3.connect(self.database, timeout=self.timeout)
        self.db_execute_and_commit(conn,'UPDATE ' + self.table + ' SET isProcessing=1 WHERE profileID=?',(orbit*10 + slit,))
        conn.close()

    def set_entry_as_finished(self, orbit, slit, failed=0):

        conn = sqlite3.connect(self.database, timeout=self.timeout)
        self.db_execute_and_commit(conn,'UPDATE ' + self.table + ' SET isFinished=?, isProcessing=?, failed=? WHERE profileID=?', (1,0,failed,orbit*10 + slit))
        conn.close()

    def view_table(self):
        conn = sqlite3.connect(self.database)
        c = conn.cursor()
        #print the last 100 to be processed
        for row in c.execute('SELECT * FROM (SELECT * FROM ' + self.table + ' WHERE isFinished=1 OR isProcessing=1 ORDER BY priority ASC LIMIT 100) ORDER BY priority DESC LIMIT 100'):
            print(row)

        #print the next 100 to be processed
        for row in c.execute('SELECT * FROM ' + self.table + ' WHERE isFinished=0 AND isProcessing=0 ORDER BY priority DESC LIMIT 100'):
            print(row)

        conn.close()