import numpy as np
import h5py
import os
import astropy.time as at
from datetime import datetime as dt
import sqlite3
import copy
from utility.timing import Timer
import omps.level2 as omps_l2
import pyodbc
from itertools import groupby
import omps.level2 as ol2

class LoadDataFromHDF:

    def __init__(self, folder='\\\\UTLS\\Users\\data\\OMPS\\OMPS_L2_USASK\\'):

        self.l2_folder  = folder
        self.anc_folder = '\\\\UTLS\\Users\\data\\OMPS\\OMPS_ANC\\'

        self.slit_num   = 1
        self.altitude   = np.arange(0.5,100,1)
        self.ozone      = np.zeros((100,0))
        self.aerosol    = np.zeros((100,0))
        self.albedo     = np.zeros(0)
        self.latitude   = np.zeros(0)
        self.longitude  = np.zeros(0)
        self.mjd        = np.zeros(0)
        self.time       = np.zeros(0)
        self.orbit      = np.zeros(0)
        self.sza        = np.zeros(0)
        self.trop_height= np.zeros(0)

        self.t_start = None
        self.t_end   = None

        self._needs_update = True

        #smoothing/binning parameters
        self._mjd_res        = 1
        self._lat_res        = 1
        self._lon_res        = 5
        self._mjd_smoothing  = 1
        self._lat_smoothing  = 2
        self._lon_smoothing  = 10

        #smoothed data
        self.do_ozone       = True
        self.do_aerosol     = True
        self.do_albedo      = True
        self.do_ozone_col   = True
        self.do_aerosol_col = True
        self.lat_smooth     = None
        self.lon_smooth     = None
        self.mjd_smooth     = None
        self.ozone_smooth   = None
        self.aerosol_smooth = None
        self.albedo_smooth  = None
        self.o3_col_smooth  = None
        self.aer_col_smooth = None

        #binned_data
        self.lat_binned     = None
        self.lon_binned     = None
        self.mjd_binned     = None
        self.ozone_binned   = None
        self.aerosol_binned = None
        self.albedo_binned  = None

        #zonal data
        self.lat_zonal     = None
        self.mjd_zonal     = None
        self.ozone_zonal   = None
        self.aerosol_zonal = None
        self.albedo_zonal  = None

        self.database = 'orbit_info.db'
        self.db_folder = '\\\\UTLS\\Users\\data\\OMPS\\'
        self.db_table = 'header'

    @property
    def mjd_res(self):
        return self._mjd_res

    @mjd_res.setter
    def mjd_res(self, value):
        self._mjd_res = value
        self._needs_update = True

    @property
    def lat_res(self):
        return self._lat_res

    @lat_res.setter
    def lat_res(self, value):
        self._lat_res = value
        self._needs_update = True

    @property
    def lon_res(self):
        return self._lon_res

    @lon_res.setter
    def lon_res(self, value):
        self._lon_res = value
        self._needs_update = True

    @property
    def mjd_smoothing(self):
        return self._mjd_smoothing

    @mjd_smoothing.setter
    def mjd_smoothing(self, value):
        self._mjd_smoothing = value
        self._needs_update = True

    @property
    def lat_smoothing(self):
        return self._lat_smoothing

    @lat_smoothing.setter
    def lat_smoothing(self, value):
        self._lat_smoothing = value
        self._needs_update = True

    @property
    def lon_smoothing(self):
        return self._lon_smoothing

    @lon_smoothing.setter
    def lon_smoothing(self, value):
        self._lon_smoothing = value
        self._needs_update = True

    @staticmethod
    def convert_date_object_to_astropy( date, dateformat=None ):
        """
        convert the date into an astropy.time object
        :param date:
        :param dateformat:
        :return: date
        """
        if isinstance(date, at.core.Time):
            t = date
        elif isinstance(date,tuple):
            a = np.zeros(7,dtype=int)
            a[1] = 1
            a[2] = 1
            for i in range(7):
                if i < len(date):
                    a[i] = date[i]

            t = at.Time(dt(a[0],a[1],a[2],a[3],a[4],a[5],a[6]))
        elif isinstance(date, dt):
            t = at.Time(date)
        else:
            t = at.Time( date, format=dateformat )

        return t

    def get_orbits_in_date_range(self, begin, end, dateformat=None):

        t_start = self.convert_date_object_to_astropy(begin, dateformat)
        t_end   = self.convert_date_object_to_astropy(end,   dateformat)

        conn = sqlite3.connect(self.db_folder + self.database)
        c = conn.cursor()
        data = c.execute('SELECT orbit FROM ' + self.db_table + ' WHERE start_mjd > ? AND end_mjd < ?', (t_start.mjd, t_end.mjd))
        orbits = data.fetchall()
        conn.close()

        return np.asarray(orbits).flatten()

    def load_data_from_date_range(self, begin, end, dateformat=None):

        self._needs_update = True
        self.t_start = self.convert_date_object_to_astropy(begin, dateformat)
        self.t_end   = self.convert_date_object_to_astropy(end,   dateformat)

        orbits = self.get_orbits_in_date_range( self.t_start, self.t_end )

        self.load_data_from_orbits( orbits )

    def load_data_from_orbits(self, orbits):

        self._needs_update = True
        self.ozone      = np.zeros((100,0))
        self.aerosol    = np.zeros((100,0))
        self.albedo     = np.zeros(0)
        self.latitude   = np.zeros(0)
        self.longitude  = np.zeros(0)
        self.mjd        = np.zeros(0)
        self.orbit      = np.zeros(0)
        self.sza        = np.zeros(0)
        self.trop_height= np.zeros(0)

        for orbit in orbits:
            if os.path.isfile(self.l2_folder + 'OMPS_L2_' + str(orbit) + '.h5'):
                try:
                    f = h5py.File(self.l2_folder + 'OMPS_L2_' + str(orbit) + '.h5')
                    good = ~np.isnan(f['/Geo/Latitude'][:,self.slit_num])
                    self.ozone      = np.hstack((self.ozone,     f['/Data/Ozone'][:,good,self.slit_num]))
                    self.aerosol    = np.hstack((self.aerosol,   f['/Data/Aerosol'][:,good,self.slit_num]))
                    self.albedo     = np.concatenate((self.albedo,    f['/Data/Albedo'][good,self.slit_num]))
                    self.latitude   = np.concatenate((self.latitude,  f['/Geo/Latitude'][good,self.slit_num]))
                    self.longitude  = np.concatenate((self.longitude, f['/Geo/Longitude'][good,self.slit_num]))
                    self.mjd        = np.concatenate((self.mjd,       f['/Geo/MJD'][good,self.slit_num]))
                    self.orbit      = np.concatenate((self.orbit,     np.ones(len(f['/Geo/MJD'][good,self.slit_num]))*orbit))
                    self.sza        = np.concatenate((self.sza,       f['/Geo/SZA'][good,self.slit_num]))

                    #get the tropopause data
                    try:
                        anc_file = omps_l2.l1_filename_for_orbit( orbit, folder='\\\\UTLS\\Users\\data\\OMPS\\OMPS_ANC\\', keyword='ANC')
                        f = h5py.File(anc_file)
                        self.trop_height = np.concatenate((self.trop_height, f['/GRIDDED_DATA/TropopauseAltitude'][good, self.slit_num]))
                    except:
                        self.trop_height = np.concatenate((self.trop_height, f['/Geo/MJD'][good,self.slit_num]*np.nan))
                except:
                    print('failed to load orbit: ' + str(orbit))

        self.mjd[np.isnan(self.mjd)] = 0.0
        self.time = at.Time(self.mjd, format='mjd')
        self.mjd[self.mjd==0] = np.nan

        self.t_start = self.convert_date_object_to_astropy(self.mjd[~np.isnan(self.mjd)][0], dateformat='mjd')
        self.t_end   = self.convert_date_object_to_astropy(self.mjd[~np.isnan(self.mjd)][-1], dateformat='mjd')

    def load_retinfo_from_orbits(self, orbits):
        self.aerosol_y = np.zeros((75,0))
        self.aerosol_f = np.zeros((75,0))
        self.ozone_y   = np.zeros((8,75,0))
        self.ozone_f   = np.zeros((8,75,0))
        self.albedo_y  = np.zeros(0)
        self.albedo_f  = np.zeros(0)
        self.ozone_wavl = np.zeros((11,0))
        self.ozone_rrad = np.zeros((75,11,0))
        self.ozone_mrad = np.zeros((75,11,0))
        self.aerosol_wavel = np.zeros(0)
        self.aerosol_rrad = np.zeros((75,0))
        self.aerosol_mrad = np.zeros((75,0))

        for orbit in orbits:
            if os.path.isfile(self.l2_folder + 'OMPS_L2_RetInfo_' + str(orbit) + '.h5'):
                f  = h5py.File(self.l2_folder + 'OMPS_L2_' + str(orbit) + '.h5')
                fi = h5py.File(self.l2_folder + 'OMPS_L2_RetInfo_' + str(orbit) + '.h5')
                good = ~np.isnan(f['/Geo/Latitude'][:,self.slit_num])

                self.aerosol_y = np.hstack((self.aerosol_y,     fi['/RetInfo/Aerosol_ymeas'][:,good,self.slit_num]))
                self.aerosol_f = np.hstack((self.aerosol_f,     fi['/RetInfo/Aerosol_yret'][:,good,self.slit_num]))
                self.ozone_y   = np.concatenate((self.ozone_y,      np.reshape(fi['/RetInfo/Ozone_ymeas'][:,good,self.slit_num],(8,75,-1))),axis=2)
                self.ozone_f   = np.concatenate((self.ozone_f,      np.reshape(fi['/RetInfo/Ozone_yret' ][:,good,self.slit_num],(8,75,-1))),axis=2)
                self.albedo_y  = np.concatenate((self.albedo_y, fi['/RetInfo/Albedo_ymeas'][0,good,self.slit_num]))
                self.albedo_f  = np.concatenate((self.albedo_f, fi['/RetInfo/Albedo_yret'][0,good,self.slit_num]))
                # self.latitude   = np.concatenate((self.latitude,  f['/Geo/Latitude'][good,self.slit_num]))
                # self.longitude  = np.concatenate((self.longitude, f['/Geo/Longitude'][good,self.slit_num]))
                # self.mjd        = np.concatenate((self.mjd,       f['/Geo/MJD'][good,self.slit_num]))
                # self.orbit      = np.concatenate((self.orbit,     np.ones(len(f['/Geo/MJD'][good,self.slit_num]))*orbit))
                # self.sza        = np.concatenate((self.sza,       f['/Geo/SZA'][good,self.slit_num]))


    def zonal_ave(self):

        if (self.ozone_zonal is None) | self._needs_update:
            self.lat_zonal = np.arange(-90, 90+self.lat_res, self.lat_res)
            self.mjd_zonal = np.arange(np.floor(np.nanmin(self.mjd)), np.ceil(np.nanmax(self.mjd)), self.mjd_res)

            self.ozone_zonal   = np.zeros((self.ozone.shape[0],  len(self.lat_zonal),len(self.mjd_zonal)))*np.nan
            self.aerosol_zonal = np.zeros((self.aerosol.shape[0],len(self.lat_zonal), len(self.mjd_zonal)))*np.nan
            self.albedo_zonal  = np.zeros((len(self.lat_zonal),len(self.mjd_zonal)))*np.nan
            self.trop_zonal    = np.zeros((len(self.lat_zonal),len(self.mjd_zonal)))*np.nan

            max_o3_idx = 60
            max_aer_idx = 40

            #get rid of nans for now to speed up computation
            trop_zeros  = np.isnan(self.trop_height)
            ozone_zeros = np.isnan(self.ozone)
            aer_zeros   = np.isnan(self.aerosol)
            alb_zeros   = np.isnan(self.albedo)
            self.ozone[ozone_zeros]      = 0.0
            self.aerosol[aer_zeros]      = 0.0
            self.albedo[alb_zeros]       = 0.0
            self.trop_height[trop_zeros] = 0.0
            for lidx,lat in enumerate(self.lat_zonal):
                for midx,mjd in enumerate(self.mjd_zonal):

                    lat_dist = np.abs(self.latitude-lat)
                    mjd_dist = np.abs(self.mjd - mjd)

                    w = np.exp( -(lat_dist**2/(2*self._lat_smoothing**2) + mjd_dist**2/(2*self._mjd_smoothing**2)))
                    w[w<1e-2] = 0.0

                    good = w>0.0
                    #dont interpolate outside of the latitude limits
                    if np.any(good) & (lat > np.nanmin(self.latitude)) & (lat < np.nanmax(self.latitude)):
                        #ozone
                        wt = np.tile(w[good],(max_o3_idx,1))
                        wt[ozone_zeros[0:max_o3_idx,good]] = 0.0
                        self.ozone_zonal[0:max_o3_idx,lidx,midx] = np.sum(self.ozone[0:max_o3_idx,good]*wt,axis=1)/np.sum(wt,axis=1)

                        #aerosol
                        wt = np.tile(w[good],(max_aer_idx,1))
                        wt[aer_zeros[0:max_aer_idx,good]] = 0.0
                        self.aerosol_zonal[0:max_aer_idx,lidx,midx] = np.sum(self.aerosol[0:max_aer_idx,good]*wt,axis=1)/np.sum(wt,axis=1)

                        #albedo
                        wt = w[good]
                        wt[alb_zeros[good]] = 0.0
                        self.albedo_zonal[lidx,midx] = np.sum(self.albedo[good]*wt)/np.sum(wt)

                        #trop
                        wt = w[good]
                        wt[trop_zeros[good]] = 0.0
                        self.trop_zonal[lidx,midx]  = np.sum(self.trop_height[good]*wt)/np.sum(wt)

    def bin_data(self):

        if (self.ozone_binned is None) | self._needs_update:
            self._needs_update = False

            #edges of mjd/lat/lon bins
            self.lat_binned = np.arange(-90, 90+self.lat_res, self.lat_res)
            self.lon_binned = np.arange(-180,180+self.lon_res,self.lon_res)
            self.mjd_binned = np.arange(np.floor(np.nanmin(self.mjd)), np.ceil(np.nanmax(self.mjd)), self.mjd_res)

            #speed up the computations a bit by throwing out high altitudes
            max_o3_idx  = 60
            max_aer_idx = 40

            self.ozone_binned   = np.zeros((self.ozone.shape[0],    len(self.mjd_binned)-1, len(self.lat_binned)-1, len(self.lon_binned)-1))*np.nan
            self.aerosol_binned = np.zeros((self.aerosol.shape[0],  len(self.mjd_binned)-1, len(self.lat_binned)-1, len(self.lon_binned)-1))*np.nan
            self.albedo_binned  = np.zeros((len(self.mjd_binned)-1, len(self.lat_binned)-1, len(self.lon_binned)-1))*np.nan
            print('smoothing the data, this might take a while...')

            #get rid of nans for now to speed up computation
            ozone    = copy.copy(self.ozone)
            aerosol  = copy.copy(self.aerosol)
            albedo   = copy.copy(self.albedo)

            ozone[ozone > 1e13]  = np.nan

            for m,min_mjd,max_mjd in zip(np.arange(0,len(self.mjd_binned[0:-1])),self.mjd_binned[0:-1], self.mjd_binned[1::]):
                print('done : ' + str(m/len(self.mjd_binned)*100) + '%')
                for i,min_lon,max_lon in zip(np.arange(0,len(self.lon_binned[0:-1])),self.lon_binned[0:-1], self.lon_binned[1::]):
                    for j,min_lat,max_lat in zip(np.arange(0,len(self.lat_binned[0:-1])),self.lat_binned[0:-1], self.lat_binned[1::]):

                        good = (self.latitude > min_lat) & (self.latitude < max_lat) & (self.mjd > min_mjd) & (self.mjd < max_mjd) & (self.longitude > min_lon) & (self.longitude < max_lon)

                        if np.any(good):
                            self.ozone_binned[:,m,j,i]   = np.nanmean( ozone[:,good],   axis=1)
                            self.aerosol_binned[:,m,j,i] = np.nanmean( aerosol[:,good], axis=1)
                            self.albedo_binned[m,j,i]    = np.nanmean( albedo[good])

            #report center of bins
            self.lat_binned = self.lat_binned[0:-1] + np.diff(self.lat_binned)/2
            self.lon_binned = self.lon_binned[0:-1] + np.diff(self.lon_binned)/2
            self.mjd_binned = self.mjd_binned[0:-1] + np.diff(self.mjd_binned)/2


    def smooth_data(self):

        if (self.ozone_smooth is None) | self._needs_update:
            self._needs_update = False
            #create a list of latitude/longitude grid points
            self.lat_smooth = np.arange(-90, 90+self.lat_res, self.lat_res)
            self.lon_smooth = np.arange(-180,180+self.lon_res,self.lon_res)
            self.mjd_smooth = np.arange(np.floor(np.nanmin(self.mjd)), np.ceil(np.nanmax(self.mjd)), self.mjd_res)
            lon_grid,lat_grid,mjd_grid = np.meshgrid(self.lon_smooth,self.lat_smooth,self.mjd_smooth)
            lat_grid = lat_grid.flatten()
            lon_grid = lon_grid.flatten()
            mjd_grid = mjd_grid.flatten()

            #speed up the computations a bit by throwing out high altitudes
            max_o3_idx = 60
            max_aer_idx = 40

            if self.do_ozone:
                self.ozone_smooth   = np.zeros((self.ozone.shape[0],  len(lat_grid)))*np.nan
                ozone    = copy.copy(self.ozone)
                ozone_zeros = np.isnan(ozone) | (ozone > 1e13)
                ozone[ozone_zeros] = 0.0
            if self.do_aerosol:
                self.aerosol_smooth = np.zeros((self.aerosol.shape[0],len(lat_grid)))*np.nan
                aerosol  = copy.copy(self.aerosol)
                aer_zeros   = np.isnan(aerosol)
                aerosol[aer_zeros] = 0.0
            if self.do_albedo:
                self.albedo_smooth  = np.zeros(len(lat_grid))*np.nan
                albedo   = copy.copy(self.albedo)
                alb_zeros   = np.isnan(albedo)
                albedo[alb_zeros]  = 0.0
            if self.do_ozone_col:
                self.o3_col_smooth     = np.zeros(len(lat_grid))*np.nan
                ozone    = copy.copy(self.ozone)
                ozone_zeros = np.isnan(ozone) | (ozone > 1e13)
                ozone[ozone_zeros] = 0.0
                isStrat            = np.ones((len(self.altitude), len(self.latitude)))*self.altitude[:,np.newaxis] > np.ones((len(self.altitude), len(self.latitude)))*self.trop_height[np.newaxis,:]
                dob                = np.sum(ozone*isStrat, axis=0)*1e5/2.69e16
            if self.do_aerosol_col:
                self.aer_col_smooth     = np.zeros(len(lat_grid))*np.nan
                aerosol    = copy.copy(self.aerosol)
                aeroosl_zeros = np.isnan(aerosol) | (aerosol > 0.01)
                aerosol[aeroosl_zeros] = 0.0
                isStrat            = np.ones((len(self.altitude), len(self.latitude)))*self.altitude[:,np.newaxis] > np.ones((len(self.altitude), len(self.latitude)))*self.trop_height[np.newaxis,:]+2
                aod                = np.sum(aerosol*isStrat, axis=0)

            # self.aod_smooth     = np.zeros(len(lat_grid))*np.nan
            print('smoothing the data, this might take a while...')
            perc_done_idxs = np.floor(np.linspace(0,len(lat_grid),11)[1::])

            #get rid of nans for now to speed up computation

            for j in range(len(lat_grid)):

                if np.any(j == perc_done_idxs):
                    print(str((np.where(j==perc_done_idxs)[0][0]+1)*10) + '%')

                lat_dist = np.abs(self.latitude-lat_grid[j])
                # lat_dist[np.isnan(lat_dist)] = 1000
                lon_dist = np.abs(self.longitude-lon_grid[j])
                # lon_dist[np.isnan(lon_dist)] = 1000
                lon_dist[lon_dist>180] = np.abs(lon_dist[lon_dist>180]-360)
                mjd_dist = np.abs(self.mjd - mjd_grid[j])
                # mjd_dist[np.isnan(mjd_dist)] = 1000

                good = (np.abs(lat_dist) < 3*self._lat_smoothing) & (np.abs(lon_dist) < 3*self._lon_smoothing) & (np.abs(mjd_dist) < 3*self._mjd_smoothing)
                good = good & (lat_grid[j] > np.min(self.latitude)) & (lat_grid[j] < np.max(self.latitude))

                w = np.exp( -(lat_dist[good]**2/(2*self._lat_smoothing**2) + lon_dist[good]**2/(2*self._lon_smoothing**2) + mjd_dist[good]**2/(2*self._mjd_smoothing**2)))
                if np.any(good):
                    if self.do_ozone:
                        wt = np.tile(w,(max_o3_idx,1))
                        wt[ozone_zeros[0:max_o3_idx,good]] = 0.0
                        self.ozone_smooth[0:max_o3_idx,j] = np.sum(np.nan_to_num(self.ozone[0:max_o3_idx,good])*wt,axis=1)/np.sum(wt,axis=1)

                    if self.do_aerosol:
                        wt = np.tile(w,(max_aer_idx,1))
                        wt[aer_zeros[0:max_aer_idx,good]] = 0.0
                        self.aerosol_smooth[0:max_aer_idx,j] = np.sum(np.nan_to_num(self.aerosol[0:max_aer_idx,good])*wt,axis=1)/np.sum(wt,axis=1)

                    if self.do_albedo:
                        self.albedo_smooth[j] = np.sum(self.albedo[good]*w)/np.sum(w)

                    if self.do_ozone_col:
                        self.o3_col_smooth[j] = np.dot(dob[good],w)/np.sum(w)
                        # self.o3_col_smooth[j] = np.sum(dob[good]*w)/np.sum(w)

                    if self.do_aerosol_col:
                        self.aer_col_smooth[j] = np.dot(aod[good],w)/np.sum(w)

            if self.do_ozone:
                self.ozone_smooth = np.reshape(self.ozone_smooth,(self.ozone.shape[0],len(self.mjd_smooth),len(self.lon_smooth),len(self.lat_smooth)),order='F').transpose(0,1,3,2)
                self.ozone_smooth[self.ozone_smooth==0] = np.nan

            if self.do_aerosol:
                self.aerosol_smooth = np.reshape(self.aerosol_smooth,(self.ozone.shape[0],len(self.mjd_smooth),len(self.lon_smooth),len(self.lat_smooth)),order='F').transpose(0,1,3,2)
                self.aerosol_smooth[self.aerosol_smooth==0] = np.nan

            if self.do_albedo:
                self.albedo_smooth = np.reshape(self.albedo_smooth,(len(self.mjd_smooth), len(self.lon_smooth),len(self.lat_smooth)),order='F').transpose(0,2,1)
                self.albedo_smooth[self.albedo_smooth==0] = np.nan

            if self.do_ozone_col:
                self.o3_col_smooth = np.reshape(self.o3_col_smooth,(len(self.mjd_smooth), len(self.lon_smooth),len(self.lat_smooth)),order='F').transpose(0,2,1)
                self.o3_col_smooth[self.o3_col_smooth==0] = np.nan

            if self.do_aerosol_col:
                self.aer_col_smooth = np.reshape(self.aer_col_smooth,(len(self.mjd_smooth), len(self.lon_smooth),len(self.lat_smooth)),order='F').transpose(0,2,1)
                self.aer_col_smooth[self.aer_col_smooth==0] = np.nan

class LoadDataFromHDF_MLS(LoadDataFromHDF):
    def __init__(self):
        super().__init__()

    def load_data_from_date_range(self, begin, end, dateformat=None):

        self._needs_update = True
        self.t_start = self.convert_date_object_to_astropy(begin, dateformat)
        self.t_end   = self.convert_date_object_to_astropy(end,   dateformat)



        self.ozone      = np.zeros((100,0))
        self.aerosol    = np.zeros((100,0))
        self.albedo     = np.zeros(0)
        self.latitude   = np.zeros(0)
        self.longitude  = np.zeros(0)
        self.mjd        = np.zeros(0)
        self.orbit      = np.zeros(0)
        self.sza        = np.zeros(0)

        conn = pyodbc.connect(r'DRIVER={SQL Server Native Client 11.0};SERVER=UTLS;DATABASE=L2;Trusted_Connection=yes')

        cur = conn.cursor()
        mjd_min = self.t_start.mjd
        mjd_max = self.t_end.mjd
        cur.execute('select h.id, p.O3NumberDensity, p.Altitude, h.latitude, h.longitude, h.MJD from l2.dbo.MLS_AltitudeProfileData p with (READUNCOMMITTED) join l2.dbo.MLS_Header h on h.id = p.id where h.MJD between ' + str(mjd_min) + ' and ' + str(mjd_max) + ' ORDER BY h.id ASC')

        results = cur.fetchall()
        results = np.array(results).transpose()
        id = np.squeeze(results[0, :])
        unique_id = np.unique(id)

        o3 = np.zeros((len(unique_id), 100))
        lat = np.zeros(len(unique_id))
        lon = np.zeros(len(unique_id))
        mjd = np.zeros(len(unique_id))

        for idx, (k, g) in enumerate(groupby(results.transpose(), lambda r: r[0])):
            dat = np.array([x for x in g])
            prof = dat[:, 1].astype(float)
            alt = dat[:, 2].astype(float)
            lat[idx] = dat[0, 3]
            lon[idx] = dat[0, 4]
            mjd[idx] = dat[0, 5]
            prof = np.interp(np.linspace(500, 99500, 100), alt, prof)
            o3[idx, :] = prof
        self.ozone = o3.transpose()
        self.aerosol = np.zeros(np.shape(self.ozone))
        self.mjd = mjd
        self.albedo = np.zeros(np.shape(self.mjd))
        self.trop_height = np.zeros(np.shape(self.mjd))
        self.latitude = lat
        self.longitude = lon
        self.mjd[np.isnan(self.mjd)] = 0.0
        self.time = at.Time(self.mjd, format='mjd')
        self.mjd[self.mjd==0] = np.nan

class LoadDataFromHDF_TomographyV02(LoadDataFromHDF):
    def __init__(self, folder):
        super().__init__(folder)
        self.raw_folder = self.l2_folder +'raw\\'
        self.l2_folder += 'L2\\'

    def load_data_from_orbits(self, orbits):

        self.ozone      = np.zeros((100,0))
        self.aerosol    = np.zeros((100,0))
        self.albedo     = np.zeros(0)
        self.latitude   = np.zeros(0)
        self.longitude  = np.zeros(0)
        self.mjd        = np.zeros(0)
        self.orbit      = np.zeros(0)
        self.sza        = np.zeros(0)

        for orbit in orbits:
            if os.path.isfile(self.l2_folder + 'o' + str(orbit) + '.h5'):
                f = h5py.File(self.l2_folder + 'o' + str(orbit) + '.h5')
                self.ozone      = np.hstack((self.ozone,          f['/Data/OzoneNumDen'][:,:]))
                self.aerosol    = np.hstack((self.aerosol,        f['/Data/AerosolNumDen'][:,:]))
                self.albedo     = np.concatenate((self.albedo,    f['/Data/Albedo'][:]))
                self.latitude   = np.concatenate((self.latitude,  f['/Geo/Latitude'][:]))
                self.longitude  = np.concatenate((self.longitude, f['/Geo/Longitude'][:]))
                self.mjd        = np.concatenate((self.mjd,       f['/Geo/MJD'][:][:]))
                self.orbit      = np.concatenate((self.orbit,     np.ones(len(f['/Geo/MJD'][:]))*orbit))

        self.aerosol *= 1.27e-5 # convert to extinction
        self.trop_height = np.zeros(self.mjd.shape)*np.nan
        self.mjd[np.isnan(self.mjd)] = 0.0
        self.trop_height = np.zeros(np.shape(self.mjd))
        self.time = at.Time(self.mjd, format='mjd')
        self.mjd[self.mjd==0] = np.nan

    def load_data_from_date_range(self, begin, end, dateformat=None):

        self._needs_update = True
        self.t_start = self.convert_date_object_to_astropy(begin, dateformat)
        self.t_end   = self.convert_date_object_to_astropy(end,   dateformat)

        orbits = self.get_orbits_in_date_range( self.t_start, self.t_end )

        self.load_data_from_orbits(orbits)

    def load_retinfo_from_orbits(self, orbits):
        self.aerosol_y = np.zeros((75,0))
        self.aerosol_f = np.zeros((75,0))
        self.ozone_y   = np.zeros((8,75,0))
        self.ozone_f   = np.zeros((8,75,0))
        self.albedo_y  = np.zeros(0)
        self.albedo_f  = np.zeros(0)
        self.ozone_wavl = np.zeros((11,0))
        self.ozone_rrad = np.zeros((75,11,0))
        self.ozone_mrad = np.zeros((75,11,0))
        self.aerosol_wavel = np.zeros(0)
        self.aerosol_rrad = np.zeros((75,0))
        self.aerosol_mrad = np.zeros((75,0))

        for orbit in orbits:
            if os.path.isfile(self.raw_folder + 'RetInfo_' + str(orbit) + '.h5'):
                f  = h5py.File(self.l2_folder + 'o' + str(orbit) + '.h5')
                fi = h5py.File(self.raw_folder + 'RetInfo_' + str(orbit) + '.h5')
                good = ~np.isnan(f['/Geo/Latitude'][:])

                self.aerosol_y = np.hstack((self.aerosol_y,     fi['/RetInfo/Aerosol_ymeas'][good,:].T))
                self.aerosol_f = np.hstack((self.aerosol_f,     fi['/RetInfo/Aerosol_yret'][good,:].T))
                self.ozone_y   = np.concatenate((self.ozone_y,      fi['/RetInfo/Ozone_ymeas'][good,:].flatten().reshape((8, -1, 75)).transpose((0,2,1))),axis=2)
                self.ozone_f   = np.concatenate((self.ozone_f,      fi['/RetInfo/Ozone_yret' ][good,:].flatten().reshape((8, -1, 75)).transpose((0,2,1))),axis=2)
                self.albedo_y  = np.concatenate((self.albedo_y, fi['/RetInfo/Albedo_ymeas'][good,0]))
                self.albedo_f  = np.concatenate((self.albedo_f, fi['/RetInfo/Albedo_yret'][good,0]))


class LoadDataFromHDF_TomographyV10(LoadDataFromHDF):
    def __init__(self, folder):
        super().__init__(folder)
        self.raw_folder = self.l2_folder +'raw\\'
        self.l2_folder += 'L2\\'

    def load_data_from_orbits(self, orbits):

        self.ozone      = np.zeros((100,0))
        self.aerosol    = np.zeros((100,0))
        self.albedo     = np.zeros(0)
        self.latitude   = np.zeros(0)
        self.longitude  = np.zeros(0)
        self.mjd        = np.zeros(0)
        self.trop_height = np.zeros(0)
        self.orbit      = np.zeros(0)
        self.sza        = np.zeros(0)

        for orbit in orbits:
            if os.path.isfile(self.l2_folder + 'o' + str(orbit) + '.h5'):
                f = h5py.File(self.l2_folder + 'o' + str(orbit) + '.h5')
                self.ozone      = np.hstack((self.ozone,          f['/DataUnbounded/OzoneNumDen'][:,:]))
                self.aerosol    = np.hstack((self.aerosol,        f['/DataUnbounded/AerosolExtinction'][:,:]))
                self.albedo     = np.concatenate((self.albedo,    f['/Data/Albedo'][:]))
                self.latitude   = np.concatenate((self.latitude,  f['/Geo/Latitude'][:]))
                self.trop_height = np.concatenate((self.trop_height, f['/Data/TropopauseHeight'][:]))
                self.longitude  = np.concatenate((self.longitude, f['/Geo/Longitude'][:]))
                self.mjd        = np.concatenate((self.mjd,       f['/Geo/MJD'][:][:]))
                self.orbit      = np.concatenate((self.orbit,     np.ones(len(f['/Geo/MJD'][:]))*orbit))

        self.mjd[np.isnan(self.mjd)] = 0.0
        self.time = at.Time(self.mjd, format='mjd')
        self.mjd[self.mjd==0] = np.nan

    def load_data_from_date_range(self, begin, end, dateformat=None):

        self._needs_update = True
        self.t_start = self.convert_date_object_to_astropy(begin, dateformat)
        self.t_end   = self.convert_date_object_to_astropy(end,   dateformat)

        orbits = self.get_orbits_in_date_range( self.t_start, self.t_end )

        self.load_data_from_orbits(orbits)

    def load_retinfo_from_orbits(self, orbits):
        self.aerosol_y = np.zeros((75,0))
        self.aerosol_f = np.zeros((75,0))
        self.ozone_y   = np.zeros((8,75,0))
        self.ozone_f   = np.zeros((8,75,0))
        self.albedo_y  = np.zeros(0)
        self.albedo_f  = np.zeros(0)
        self.ozone_wavl = np.zeros((11,0))
        self.ozone_rrad = np.zeros((75,11,0))
        self.ozone_mrad = np.zeros((75,11,0))
        self.aerosol_wavel = np.zeros(0)
        self.aerosol_rrad = np.zeros((75,0))
        self.aerosol_mrad = np.zeros((75,0))

        for orbit in orbits:
            if os.path.isfile(self.raw_folder + 'RetInfo_' + str(orbit) + '.h5'):
                f  = h5py.File(self.l2_folder + 'o' + str(orbit) + '.h5')
                fi = h5py.File(self.raw_folder + 'RetInfo_' + str(orbit) + '.h5')
                good = ~np.isnan(f['/Geo/Latitude'][:])

                self.aerosol_y = np.hstack((self.aerosol_y,     fi['/RetInfo/Aerosol_ymeas'][good,:].T))
                self.aerosol_f = np.hstack((self.aerosol_f,     fi['/RetInfo/Aerosol_yret'][good,:].T))
                self.ozone_y   = np.concatenate((self.ozone_y,      fi['/RetInfo/Ozone_ymeas'][good,:].flatten().reshape((8, -1, 75)).transpose((0,2,1))),axis=2)
                self.ozone_f   = np.concatenate((self.ozone_f,      fi['/RetInfo/Ozone_yret' ][good,:].flatten().reshape((8, -1, 75)).transpose((0,2,1))),axis=2)
                self.albedo_y  = np.concatenate((self.albedo_y, fi['/RetInfo/Albedo_ymeas'][good,0]))
                self.albedo_f  = np.concatenate((self.albedo_f, fi['/RetInfo/Albedo_yret'][good,0]))

class LoadDataFromHDF_OMPS_L2(LoadDataFromHDF):
    def __init__(self, folder=r'\\utls\\users\\data\\OMPS\\OMPS_L2\\'):
        self.folder = folder
        super().__init__()

    def load_data_from_date_range(self, begin, end, dateformat=None):

        self._needs_update = True
        self.t_start = self.convert_date_object_to_astropy(begin, dateformat)
        self.t_end   = self.convert_date_object_to_astropy(end,   dateformat)

        orbits = self.get_orbits_in_date_range( self.t_start, self.t_end )

        ozone      = np.zeros((61,0))
        aerosol    = np.zeros((41,0))
        self.albedo     = np.zeros(0)
        self.latitude   = np.zeros(0)
        self.longitude  = np.zeros(0)
        self.mjd        = np.zeros(0)
        self.orbit      = np.zeros(0)
        self.sza        = np.zeros(0)
        self.trop       = np.zeros(0)

        for orbit in orbits:
            data_ozone = ol2.load_ozone_orbit(orbit, 2, folder=self.folder)
            data_aerosol = ol2.load_aerosol_orbit(orbit, 2, folder=self.folder)
            if len(data_ozone) > 1:

                if data_aerosol['aero_wavel'].shape[1] != 5:
                    print('unexpected wavelengths')

                aer_lat = data_aerosol['latitude']
                o3_lat  = data_ozone['latitude']

                if len(aer_lat) != len(o3_lat):
                    print('aerosol adn ozone different lengths in orbit : ' + str(orbit))
                else:
                    ozone      = np.hstack((ozone,          data_ozone['o3'].transpose()))
                    aerosol    = np.hstack((aerosol,        data_aerosol['aero'][:, 3, :].transpose()))

                    self.albedo     = np.concatenate((self.albedo,    data_ozone['albedo']))
                    self.latitude   = np.concatenate((self.latitude,  data_ozone['latitude']))
                    self.longitude  = np.concatenate((self.longitude, data_ozone['longitude']))
                    mjd = data_ozone['mjd'][range(data_ozone['o3'].shape[0])]
                    self.mjd        = np.concatenate((self.mjd,       mjd))
                    self.sza        = np.concatenate((self.sza,       np.ones(len(mjd))))
                    self.orbit      = np.concatenate((self.orbit,     np.ones(len(mjd))*orbit))
                    self.trop       = np.concatenate((self.trop,      data_ozone['trop']*1000))

        self.ozone = np.zeros((100, len(self.mjd)))*np.nan
        self.aerosol = np.zeros((100, len(self.mjd)))*np.nan

        self.ozone[:61, :] = ozone
        self.aerosol[:41]  = aerosol
        self.ozone[self.ozone < 0] = np.nan
        self.aerosol[self.aerosol < 0] = np.nan

        self.mjd[np.isnan(self.mjd)] = 0.0
        self.trop_height = np.zeros(np.shape(self.mjd))
        self.time = at.Time(self.mjd, format='mjd')
        self.mjd[self.mjd==0] = np.nan