import h5py
import os
import logging
import omps.level2 as omps_l2
import numpy as np
import matplotlib.pyplot as plt
import filelock
import time

class RetrievalOutput:
    """
    Class designed to handle the output from the OMPSRetrieval class.

        >>> omps = OMPSRetrieval( 12115, 40 )
        >>> output = omps.retrieve_profiles()
        >>> h5out = RetrievalOutput( output )
        >>> h5.filename = 'filename.h5'
        >>> h5.save_to_output_file()
    """
    def __init__(self, output):

        self.save_log_messages   = True
        self.save_meas_vecs      = True
        self.save_radiances      = True
        self.num_images_in_orbit = None
        self.num_slits           = 3
        self._image_offset       = 0

        self.output = output

        #variables for easier access
        self.ret_opt     = self.output['retrieval_info','ret_opt']
        self.imagenumber = self.output['retrieval_info','image_num']
        self.orbitnumber = self.output['retrieval_info','orbit_num']
        self.slitnumber  = self.output['retrieval_info','slit_num']

        self.filename   = None
        self.retinfo_filename = None
        self.folder     = None
        self._file      = None
        self._retfile   = None
        self._lock_file = None

        self.num_alts     = 100
        self.num_o3_meas  = 600
        self.num_aer_meas = 75
        self.num_alb_meas = 1

    @property
    def image_offset(self):
        return self._image_offset

    @image_offset.setter
    def image_offset(self, image_offset):
        self.imagenumber  = self.output['retrieval_info','image_num'] - image_offset
        self._image_offset = image_offset

    def save_to_output_file(self):
        """
        Appends the retrieved profiles to an orbit file. If the image already exists
        in the file it is overwritten
        """

        found_folder = False
        #check if folder exists
        if os.path.isdir( self.folder ):
            found_folder = True
        #if not check if its on the python path and make the full path
        else:
            user_paths = os.environ['PYTHONPATH'].split(os.pathsep)
            for i in range(len(user_paths)):
                if os.path.isdir( user_paths[i]+self.folder ):
                    found_folder = True
                    self.folder = user_paths[i]+self.folder
                    break

        if found_folder is False:
            logging.error('could not find output folder')

        if self.filename is None:
            self._file      = self.folder + 'OMPS_L2_' + str(self.orbitnumber) + '.h5'
            self._retfile   = self.folder + 'OMPS_L2_RetInfo_' + str(self.orbitnumber) + '.h5'
        else:
            self._file    = self.folder + self.filename
            self._retfile = self.folder + self.retinfo_filename

        self._lock_file  = self._file + '.lock'
        #lock the file and create/save
        lock = filelock.FileLock(self._lock_file)
        with lock.acquire(timeout=60):
            exists = os.path.isfile(self._file)
            if not exists:
                self._create_empty_hdf_file()

            self._save_to_hdf_file()

    def _create_empty_hdf_file(self):
        f = h5py.File( self._file, 'a' )

        if self.num_images_in_orbit is None:
            num_images_in_orbit = omps_l2.get_num_images_in_orbit(self.orbitnumber, self.ret_opt['omps_l1_folder'])
        else:
            num_images_in_orbit = self.num_images_in_orbit

        num_slits = self.num_slits

        profile = np.zeros((self.num_alts,    num_images_in_orbit,num_slits))*np.nan
        yo3     = np.zeros((self.num_o3_meas, num_images_in_orbit,num_slits))*np.nan
        yaer    = np.zeros((self.num_aer_meas,num_images_in_orbit,num_slits))*np.nan
        yalb    = np.zeros((self.num_alb_meas,num_images_in_orbit,num_slits))*np.nan
        value   = np.zeros((num_images_in_orbit,num_slits))*np.nan

        f.create_dataset('/Data/Ozone',   data=profile)
        f.create_dataset('/Data/Aerosol', data=profile)
        f.create_dataset('/Data/Albedo',  data=value)

        if self.save_log_messages:
            dt = h5py.special_dtype(vlen=str)
            f.create_dataset('/Log/logging_info',(num_images_in_orbit,num_slits),dtype=dt)

        if self.save_meas_vecs:
            f.create_dataset('/Data/Ozone_ymeas',   data=yo3)
            f.create_dataset('/Data/Ozone_yret',    data=yo3)
            f.create_dataset('/Data/Aerosol_ymeas', data=yaer)
            f.create_dataset('/Data/Aerosol_dymeas',data=yaer)
            f.create_dataset('/Data/Aerosol_yret',  data=yaer)
            f.create_dataset('/Data/Albedo_ymeas',  data=yalb)
            f.create_dataset('/Data/Albedo_yret',   data=yalb)

            f.create_dataset('/Data/Ozone_residual',   data=value)
            f.create_dataset('/Data/Aerosol_residual', data=value)
            f.create_dataset('/Data/Albedo_residual',  data=value)

        f.create_dataset('/Geo/SZA',      data=value)
        f.create_dataset('/Geo/MJD',      data=value)
        f.create_dataset('/Geo/Latitude', data=value)
        f.create_dataset('/Geo/Longitude',data=value)

        f.create_dataset('/Geo/Image',    data =np.zeros(num_images_in_orbit))
        f.create_dataset('/Geo/Orbit',    data =self.orbitnumber)
        f.close()

    def _save_to_hdf_file(self):
        f = h5py.File( self._file, 'r+' )
        if any("ozone" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Ozone'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','o3_valid']
            if self.save_meas_vecs:
                f['/Data/Ozone_ymeas'][:,self.imagenumber,self.slitnumber]  = self.output['retrieved','o3_ymeas']
                f['/Data/Ozone_yret'][:,self.imagenumber,self.slitnumber]   = self.output['retrieved','o3_yret']
                f['/Data/Ozone_residual'][self.imagenumber,self.slitnumber] = self.output['retrieved','o3_res']

        if any("aerosol" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Aerosol'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','aero_valid']
            if self.save_meas_vecs:
                f['/Data/Aerosol_ymeas'][:,self.imagenumber,self.slitnumber]  = self.output['retrieved','aer_ymeas']
                f['/Data/Aerosol_dymeas'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','aer_dymeas']
                f['/Data/Aerosol_yret'][:,self.imagenumber,self.slitnumber]   = self.output['retrieved','aer_yret']
                f['/Data/Aerosol_residual'][self.imagenumber,self.slitnumber] = self.output['retrieved','aer_res']

        if any("albedo" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Albedo'][self.imagenumber,self.slitnumber] = self.output['retrieved','albedo']
            if self.save_meas_vecs:
                f['/Data/Albedo_ymeas'][:,self.imagenumber,self.slitnumber]  = self.output['retrieved','alb_ymeas']
                f['/Data/Albedo_yret'][:,self.imagenumber,self.slitnumber]   = self.output['retrieved','alb_yret']
                f['/Data/Albedo_residual'][self.imagenumber,self.slitnumber] = self.output['retrieved','alb_res']

        f['/Geo/Image'][self.imagenumber]                       = self.output['retrieval_info','image_num']
        f['/Geo/SZA'][self.imagenumber, self.slitnumber]        = self.output['omps_geometry','sza']
        f['/Geo/MJD'][self.imagenumber, self.slitnumber]        = self.output['omps_geometry','mjd']
        f['/Geo/Latitude'][self.imagenumber, self.slitnumber]   = self.output['omps_geometry','latitude']
        f['/Geo/Longitude'][self.imagenumber, self.slitnumber]  = self.output['omps_geometry','longitude']

        #ouput the log messages to the hdf5 file
        if self.save_log_messages:
            log = logging.getLogger()
            for hndl in log.handlers:
                if type(hndl) == logging.StreamHandler:
                    f['/Log/logging_info'][self.imagenumber, self.slitnumber] = hndl.stream.getvalue()

        f.close()

    def plot_output(self, species, iteration=-1, plot_error=True):
        """
        Plots the results of the retrieval.

        :param species: name of the species being retrieved, eg. 'aerosol', 'ozone'
        :param iteration: iteration number if the species was retrieved more than once, default is the last iteration
        (1 is the first iteration)
        :param plot_error: also plot the relative difference between the measured and retrieved vectors
        """
        num_subplots = 2
        if plot_error:
            num_subplots = 3

        apriori_alts = np.arange(0.5,100.5)

        #determine what species and iteration we want
        found = False
        iter  = 0
        for idx, spec in enumerate(self.ret_opt['retrieval_order']):
            if spec == species:
                iter += 1   #increment each time we find the species

            if iteration ==-1:  # if we want the last one don't bother checking iteration #
                if spec == species:
                    found = True
                    index = idx
            else:
                if (iteration == iter) & (spec == species): #get the correct species and iteration
                    found = True
                    index = idx

        if not found:
            raise Exception('species and/or iteration not found')

        lsq_output = self.output['intermediate',index][1]['lsq_output']

        #make some plots
        plt.figure()
        plt.subplot(1,num_subplots,1)
        plt.grid(color=[0.8,0.8,0.8], linestyle='-')
        line1, = plt.plot(lsq_output['apriori'],apriori_alts, color=(0.7,0.4,0.2))
        for i in range(0,len(lsq_output['profile'])):
            line2, = plt.plot(lsq_output['profile'][i],apriori_alts,color=(0.2, 0.4, 0.7,0.5))

        line3, = plt.plot(lsq_output['profile'][-1],apriori_alts,color=(0.2, 0.4, 0.7), linewidth=2)
        plt.ylim([5, 60])
        plt.ylabel('Altitude [km]')
        plt.xlabel('Number Density')
        num_vec = int(len(lsq_output['y_meas'])/len(lsq_output['tanalts']))

        plt.legend([line1,line2,line3],['Apriori','Iterations','Retrieved'])

        plt.subplot(1,num_subplots,2)
        plt.grid(color=[0.8,0.8,0.8], linestyle='-')
        for i in range(num_vec):
            line1, =plt.plot( np.reshape(lsq_output['y_meas'],(num_vec,-1))[i,:], np.asarray(lsq_output['tanalts'])/1000, color='black',marker="o" )
            plt.fill_betweenx(np.asarray(lsq_output['tanalts'])/1000,  np.reshape(lsq_output['y_meas']-lsq_output['y_meas_error'],(num_vec,-1))[i,:],  np.reshape(lsq_output['y_meas'],(num_vec,-1))[i,:], color='black', alpha=0.1)
            plt.fill_betweenx(np.asarray(lsq_output['tanalts'])/1000,  np.reshape(lsq_output['y_meas']+lsq_output['y_meas_error'],(num_vec,-1))[i,:],  np.reshape(lsq_output['y_meas'],(num_vec,-1))[i,:], color='black', alpha=0.1)

        for i in range(0,len(lsq_output['y_ret'])):
            for j in range(num_vec):
                line2, = plt.plot( np.reshape(lsq_output['y_ret'][i],(num_vec,-1))[j,:], np.asarray(lsq_output['tanalts'])/1000, color=(0.2, 0.4, 0.7,0.5) )

        for i in range(num_vec):
                line3, = plt.plot( np.reshape(lsq_output['y_ret'][-1],(num_vec,-1))[i,:], np.asarray(lsq_output['tanalts'])/1000, color=(0.2, 0.4, 0.7),linewidth=2 )

        plt.ylim([5, 60])
        plt.legend((line1,line2,line3),('Measured','Iterations','Retrieved'))

        if num_vec == 8:
            mylegend = ['292','302','306','310','315','322','331','600']
        elif num_vec == 7:
            mylegend = ['292','302','306','315','322','331','600']
        else:
            mylegend = ['750']

        if plot_error:
            plt.subplot(1,num_subplots,3)
            for i in range(num_vec):
                f = np.reshape(lsq_output['y_ret'][-1],(num_vec,-1))[i,:]
                y = np.reshape(lsq_output['y_meas'],(num_vec,-1))[i,:]
                error = (y-f)/y*100
                plt.plot(error, np.asarray(lsq_output['tanalts'])/1000, label=mylegend[i])
                plt.xlabel('Percent Error in Measurement Vectors')
                plt.grid(color=[0.8,0.8,0.8], linestyle='-')

            plt.legend()
            plt.ylim([5, 60])

class RetrievalOutputUSASK(RetrievalOutput):
    """
    Class designed to save a few additional output fields in addition to the base class. Uses GZIP compression
    """
    def _create_empty_hdf_file(self):
        f  = h5py.File( self._file,    'a' )
        fr = h5py.File( self._retfile, 'a' )

        if self.num_images_in_orbit is None:
            num_images_in_orbit = omps_l2.get_num_images_in_orbit(self.orbitnumber, self.ret_opt['omps_l1_folder'])
        else:
            num_images_in_orbit = self.num_images_in_orbit

        num_slits = self.num_slits

        profile = np.zeros((self.num_alts,    num_images_in_orbit,num_slits))*np.nan
        yo3     = np.zeros((self.num_o3_meas, num_images_in_orbit,num_slits))*np.nan
        yaer    = np.zeros((self.num_aer_meas,num_images_in_orbit,num_slits))*np.nan
        yalb    = np.zeros((self.num_alb_meas,num_images_in_orbit,num_slits))*np.nan
        value   = np.zeros((num_images_in_orbit,num_slits))*np.nan

        f.create_dataset('/Software/version', data=self.output['retrieval_info','version'])
        f.create_dataset('/Software/processing_date', data=time.asctime())
        f.create_dataset('/Software/processing_location', data='University of Saskatchewan')
        f.create_dataset('/Data/Altitude',    data=np.arange(0.5,100,1))
        f.create_dataset('/Data/Ozone',       data=profile)
        fr.create_dataset('/RetInfo/Atmosphere/Ozone_rtm',   data=profile)
        f.create_dataset('/Data/Aerosol',     data=profile)
        fr.create_dataset('/RetInfo/Atmosphere/Aerosol_rtm', data=profile)
        f.create_dataset('/Data/Albedo',      data=value)
        fr.create_dataset('/RetInfo/Atmosphere/Albedo_rtm',  data=value)

        if self.save_log_messages:
            dt = h5py.special_dtype(vlen=str)
            fr.create_dataset('/Log/logging_info',(num_images_in_orbit,num_slits),dtype=dt)

        if self.save_meas_vecs:
            fr.create_dataset('/RetInfo/Ozone_ymeas',   data=yo3, compression="gzip")
            fr.create_dataset('/RetInfo/Ozone_yret',    data=yo3, compression="gzip")
            fr.create_dataset('/RetInfo/Aerosol_ymeas', data=yaer, compression="gzip")
            fr.create_dataset('/RetInfo/Aerosol_dymeas',data=yaer, compression="gzip")
            fr.create_dataset('/RetInfo/Aerosol_yret',  data=yaer, compression="gzip")
            fr.create_dataset('/RetInfo/Albedo_ymeas',  data=yalb, compression="gzip")
            fr.create_dataset('/RetInfo/Albedo_yret',   data=yalb, compression="gzip")

            fr.create_dataset('/RetInfo/Ozone_residual',   data=value, compression="gzip")
            fr.create_dataset('/RetInfo/Aerosol_residual', data=value, compression="gzip")
            fr.create_dataset('/RetInfo/Albedo_residual',  data=value, compression="gzip")

        if self.save_radiances:
            fr.create_dataset('/RetInfo/Radiances/Ozone_retrad',    data=np.zeros((self.output['radiances','o3_retrad'].shape[0],  self.output['radiances','o3_retrad'].shape[1] , num_images_in_orbit,num_slits)), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Aerosol_retrad',  data=np.zeros((self.output['radiances','aer_retrad'].shape[0], self.output['radiances','aer_retrad'].shape[1], num_images_in_orbit,num_slits)), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Albedo_retrad',   data=np.zeros((self.output['radiances','alb_retrad'].shape[0], self.output['radiances','alb_retrad'].shape[1], num_images_in_orbit,num_slits)), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Ozone_measrad',   data=np.zeros((self.output['radiances','o3_measrad'].shape[0], self.output['radiances','o3_measrad'].shape[1], num_images_in_orbit,num_slits)), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Aerosol_measrad', data=np.zeros((self.output['radiances','aer_measrad'].shape[0],self.output['radiances','aer_measrad'].shape[1],num_images_in_orbit,num_slits)), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Albedo_measrad',  data=np.zeros((self.output['radiances','alb_measrad'].shape[0],self.output['radiances','alb_measrad'].shape[1],num_images_in_orbit,num_slits)), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Ozone_wavel',     data=np.zeros((self.output['radiances','o3_wavel'].shape[0],num_images_in_orbit,num_slits) ), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Aerosol_wavel',   data=np.zeros((self.output['radiances','aer_wavel'].shape[0],num_images_in_orbit,num_slits)), compression="gzip" )
            fr.create_dataset('/RetInfo/Radiances/Albedo_wavel',    data=np.zeros((self.output['radiances','alb_wavel'].shape[0],num_images_in_orbit,num_slits)), compression="gzip" )

        f.create_dataset('/Geo/SZA',      data=value)
        f.create_dataset('/Geo/MJD',      data=value)
        f.create_dataset('/Geo/Latitude', data=value)
        f.create_dataset('/Geo/Longitude',data=value)

        f.create_dataset('/Geo/Image',    data =np.zeros(num_images_in_orbit)*np.nan)
        f.create_dataset('/Geo/Orbit',    data =self.orbitnumber)
        f.close()

    def _save_to_hdf_file(self):
        f = h5py.File( self._file, 'r+' )
        fr = h5py.File( self._retfile, 'r+' )
        if any("ozone" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Ozone'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','o3_valid']
            fr['/RetInfo/Atmosphere/Ozone_rtm'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','o3']
            if self.save_meas_vecs:
                fr['/RetInfo/Ozone_ymeas'][:,self.imagenumber,self.slitnumber]  = self.output['retrieved','o3_ymeas']
                fr['/RetInfo/Ozone_yret'][:,self.imagenumber,self.slitnumber]   = self.output['retrieved','o3_yret']
                fr['/RetInfo/Ozone_residual'][self.imagenumber,self.slitnumber] = self.output['retrieved','o3_res']

            if self.save_radiances:
                fr['/RetInfo/Radiances/Ozone_retrad'][:,:,self.imagenumber,self.slitnumber]  = self.output['radiances','o3_retrad']
                fr['/RetInfo/Radiances/Ozone_measrad'][:,:,self.imagenumber,self.slitnumber] = self.output['radiances','o3_measrad']
                fr['/RetInfo/Radiances/Ozone_wavel'][:,self.imagenumber,self.slitnumber]     = self.output['radiances','o3_wavel']

        if any("aerosol" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Aerosol'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','aero_valid']
            fr['/RetInfo/Atmosphere/Aerosol_rtm'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','aero']
            if self.save_meas_vecs:
                fr['/RetInfo/Aerosol_ymeas'][:,self.imagenumber,self.slitnumber]  = self.output['retrieved','aer_ymeas']
                fr['/RetInfo/Aerosol_dymeas'][:,self.imagenumber,self.slitnumber] = self.output['retrieved','aer_dymeas']
                fr['/RetInfo/Aerosol_yret'][:,self.imagenumber,self.slitnumber]   = self.output['retrieved','aer_yret']
                fr['/RetInfo/Aerosol_residual'][self.imagenumber,self.slitnumber] = self.output['retrieved','aer_res']

            if self.save_radiances:
                fr['/RetInfo/Radiances/Aerosol_retrad'][:,:,self.imagenumber,self.slitnumber]  = self.output['radiances','aer_retrad']
                fr['/RetInfo/Radiances/Aerosol_measrad'][:,:,self.imagenumber,self.slitnumber] = self.output['radiances','aer_measrad']
                fr['/RetInfo/Radiances/Aerosol_wavel'][:,self.imagenumber,self.slitnumber]     = self.output['radiances','aer_wavel']

        if any("albedo" in s for s in self.ret_opt['retrieval_order']):
            f['/Data/Albedo'][self.imagenumber,self.slitnumber] = self.output['retrieved','albedo']
            fr['/RetInfo/Atmosphere/Albedo_rtm'][self.imagenumber,self.slitnumber] = self.output['retrieved','albedo']
            if self.save_meas_vecs:
                fr['/RetInfo/Albedo_ymeas'][:,self.imagenumber,self.slitnumber]  = self.output['retrieved','alb_ymeas']
                fr['/RetInfo/Albedo_yret'][:,self.imagenumber,self.slitnumber]   = self.output['retrieved','alb_yret']
                fr['/RetInfo/Albedo_residual'][self.imagenumber,self.slitnumber] = self.output['retrieved','alb_res']

            if self.save_radiances:
                fr['/RetInfo/Radiances/Albedo_retrad'][:,:,self.imagenumber,self.slitnumber]  = self.output['radiances','alb_retrad']
                fr['/RetInfo/Radiances/Albedo_measrad'][:,:,self.imagenumber,self.slitnumber] = self.output['radiances','alb_measrad']
                fr['/RetInfo/Radiances/Albedo_wavel'][:,self.imagenumber,self.slitnumber]     = self.output['radiances','alb_wavel']

        f['/Geo/Image'][self.imagenumber]                       = self.output['retrieval_info','image_num']
        f['/Geo/SZA'][self.imagenumber, self.slitnumber]        = self.output['omps_geometry','sza']
        f['/Geo/MJD'][self.imagenumber, self.slitnumber]        = self.output['omps_geometry','mjd']
        f['/Geo/Latitude'][self.imagenumber, self.slitnumber]   = self.output['omps_geometry','latitude']
        f['/Geo/Longitude'][self.imagenumber, self.slitnumber]  = self.output['omps_geometry','longitude']

        #ouput the log messages to the hdf5 file
        if self.save_log_messages:
            log = logging.getLogger()
            for hndl in log.handlers:
                if type(hndl) == logging.StreamHandler:
                    fr['/Log/logging_info'][self.imagenumber, self.slitnumber] = hndl.stream.getvalue()

        f.close()

