from radtran.sasktran import SasktranHR, SasktranHRMSApprox
from radtran.scan import Scan
from radtran.options import GenericOptions
import matplotlib.pyplot as plt
import numpy as np
from radtran.atmosphere import Atmosphere
from radtran.climatologies import background, ozone, no2, aerosol
from utility.timing import Timer
import scipy.signal as sp

atmo = Atmosphere()

#create a background climatology from msis90 and add it to the atmosphere
clim, optprop, guid = background.msis()
atmo.add_species( clim, optprop, guid )

#set this at the background climatology, other species will use the pressure
#and temperature from the background to determine cross sections
atmo.backgroundclim = clim

#add ozone, no2 and aerosol to the model
clim, optprop, guid = ozone.labow('O3_DBM')
atmo.add_species(clim, optprop, guid)

clim, optprop, guid = no2.pratmo()
atmo.add_species(clim, optprop, guid)

clim, optprop, guid = aerosol.apriori()
atmo.add_species(clim,optprop, guid)

scan = Scan()
altitudes = np.arange(0.5,100,1)
scan.make_scan_saa(sza=70, saa=0, lat=0, lon=0, tanalts_km=altitudes, mjd=54372.5, locallook=0)

#setup some radiative transfer options
hr_options = GenericOptions()
hr_options.add_option('setsun', scan.sun)
hr_options.add_option('numordersofscatter',10)   #case sensitive
hr_options.add_option('ManualDiffuseHeights', np.concatenate((np.array([1e-6]), np.arange(500.0, 100000.0, 5000.0))))

#create the engine, add the atmosphere and the options
hr_engine   = SasktranHR(atmo, hr_options)
hrms_engine = SasktranHRMSApprox(atmo, hr_options, ms_wavel_space=0.5)

#add the instrument lines of sight
for obs, los, mjd in zip(scan.obs, scan.look, scan.mjd):
    hr_engine.add_line_of_sight(obs, los, mjd)
    hrms_engine.add_line_of_sight(obs, los, mjd)

#set the wavelengths and calculate radiance
wavelengths = np.arange(288, 360, 0.1)
hr_engine.set_wavelengths(wavelengths)
hrms_engine.set_wavelengths(wavelengths)
with Timer():
    radhr = hr_engine.calculate_radiance()[0]
with Timer():
    radhrms = hrms_engine.calculate_radiance()[0]

plt.plot(wavelengths, radhr[20,:])
plt.plot(wavelengths, radhrms[20,:])
plt.show()