import sasktranif.sasktranif as skif
import numpy

#------------------------------------------------------------------------------
#           class Atmosphere
#------------------------------------------------------------------------------

class Atmosphere:
    """
    A class which holds the atmosphere for an engine within the SASKTRAN
    framework, i.e. that follows the climatology, optical property, guid
    model.  
    
    There are two primary methods of adding species to the atmosphere, one is the
    default way of specifying a climatology/optical property/guid, and the other is
    by specifying a user defined profile.
    
    Climatology Optical Property Guid:
        Species can be added through :func:`~Atmosphere.add_species`, this function takes in a climatology,
        optical property, and guid from sasktranif and keeps a reference to them
        inside the class.  Once the species is added it cannot be modified/removed
        directly through the :class:`~Atmosphere` class.  This is the preferred
        method for adding species that wont be modified later in some retrieval
        process.
    
    
    User Defined Profile:
        The other method of adding a species is with :func:`~Atmosphere.add_user_defined`
        here we add a species by specifying the optical property and guid (from sasktranif)
        and an array of heights/profile values.  The atmosphere class automatically
        creates a ISKClimatology('USERDEFINED_PROFILE') object and stores the species
        internally.  Later the climatology can be modified through the functions
        :func:`~Atmosphere.modify_user_defined` and :func:`~Atmosphere.get_user_defined`.  Use this method
        of adding a species when you are going to be modifying it later.
    
    Albedo:
        Albedo is assumed to be constant both spatially and spectrally.  It can be
        set through the albedo member variable
    
    Background Atmosphere:
        As of July 23 2015, all SASKTRAN engines appear to have a bug where the
        default pressure/temperature climatology (MSIS) is not initialized properly for
        the first wavelength.  Because of this we explicitly call :func:`~ISKEngine.SetAtmosphericState`
        with a background climatology.  By default this is ISKCLIMATOLOGY('ECMWF'),
        but can be changed by modifying the backgroundclim member variable.
    
    """
    def __init__(self):
        self.clim = []
        self.optprop = []
        self.guid = []
        self.albedo = None
        self.wfoptprop = None
        self.backgroundclim = None
        
        # hold a map of user defined profiles
        self.user_clim = dict()
        self.user_profile = dict()
        self.user_profile_height_m = dict()
        self.user_optprop = dict()
        self.user_guid = dict()
        
        
    #------------------------------------------------------------------------------
    #               def add_species
    #------------------------------------------------------------------------------

    def add_species( self,
                     clim    : skif.ISKClimatology,
                     optprop : skif.ISKOpticalProperty,
                     guid    : str ) ->None :
        """
        Adds a species to the atmosphere
        
        Parameters
        ----------
        clim : 
            ISKClimatology  
        optprop : 
            ISKOpticalProperty            
        guid :
            guid from sasktranif
        """
        self.clim.append( clim )
        self.optprop.append( optprop )
        self.guid.append( guid )

    #------------------------------------------------------------------------------
    #           add_user_defined
    #------------------------------------------------------------------------------

    def add_user_defined( self,
                          guid    : str,
                          optprop : skif.ISKOpticalProperty,
                          heights : numpy.ndarray,
                          profile : numpy.ndarray,
                          ident   : str,
                          doLogInterpolation : bool = False,
                          doPiecewiseLinear  : bool = False ) -> None:
        """
        Adds a user defined species (of type ISKClimatology('USERDEFINED_PROFILE')) 
        to the atmosphere
        
        Parameters
        ----------
        guid :
            guid from sasktranif
        optprop : 
            ISKOpticalProperty
        heights : numpy array
            Heights in meters in ascending order
        profile : numpy array
            values for the user defined profile, same size and shape as heights
        ident : string
            A user specified identifier for the profile, for example 'o3'
        """
        self.user_guid[ident] = guid
        self.user_clim[ident] = skif.ISKClimatology('USERDEFINED_PROFILE')
        self.user_optprop[ident] = optprop
        
        self.user_clim[ident].SetProperty('Heights', heights)
        self.user_clim[ident].SetProperty('DoLogInterpolation', 1 if doLogInterpolation else 0 )
        self.user_clim[ident].SetProperty('DoPiecewiseLinear',  1 if doPiecewiseLinear  else 0 )
        self.user_clim[ident].SetPropertyUserDefined(guid, profile)

        self.user_profile[ident] = profile
        self.user_profile_height_m[ident] = heights

    #------------------------------------------------------------------------------
    #           modify_user_defined
    #------------------------------------------------------------------------------

    def modify_user_defined(self, ident : str, profile : numpy.ndarray) -> None:
        """
        Modifies a user defined profile that has already been added to the
        atmosphere
        
        Parameters
        ----------
        ident : string
            The user specified identifier for the profile to be modified
        profile : numpy array
            The new value for the profile
        
        Raises
        ------
        KeyError
            If a profile matching ident does not exist in the atmosphere
        """
        self.user_clim[ident].SetPropertyUserDefined(self.user_guid[ident], profile)
        self.user_profile[ident] = profile

    #------------------------------------------------------------------------------
    #           get_user_defined
    #------------------------------------------------------------------------------

    def get_user_defined(self, ident : str ):
        """
        Gets a user defined profile that is present in the Atmosphere
        
        Parameters
        ----------
        ident : string
            The user specified identifer for the profile to be retrieved
        
        Returns
        -------
        tuple
            (profile_heights_m, profile_values)
        
        Raises
        ------
        KeyError
            If there is no profile specified by ident
        """
        return self.user_profile_height_m[ident], self.user_profile[ident]

    #------------------------------------------------------------------------------
    #           set_wf_species
    #------------------------------------------------------------------------------

    def set_wf_species( self, optprop : skif.ISKOpticalProperty ) -> None:
        """
        Sets the species to calculate weighting functions for
        
        Parameters
        ----------
        optprop : ISKOpticalProperty
            Optical Property from sasktranif           
        """
        self.wfoptprop = optprop

    #------------------------------------------------------------------------------
    #           add_wf_species_to_engine
    #------------------------------------------------------------------------------

    def add_wf_species_to_engine( self, engine : skif.ISKEngine ) -> None:
        """
        Adds the weighting function species to the engine
        
        Parameters
        ----------
        engine : ISKEngine
            
        """
        if self.wfoptprop != None:
            engine.SetProperty( 'wfspecies', self.wfoptprop )

    #------------------------------------------------------------------------------
    #           add_albedo_to_engine
    #------------------------------------------------------------------------------

    def add_albedo_to_engine(self, engine: skif.ISKEngine) ->None:
    
        if self.albedo != None:
            engine.SetAlbedo( self.albedo )

    #------------------------------------------------------------------------------
    #               def add_to_engine( self, engine :"ISKEngine", addbackground=True ):
    #------------------------------------------------------------------------------

    def add_to_engine( self,
                       engine        : skif.ISKEngine,
                       addbackground : bool = True )-> None:
        """
        Adds the atmosphere to the engine
        
        Parameters
        ----------
        engine : ISKEngine
        
        """
        for clim, optprop, guid in zip( self.clim, self.optprop, self.guid ):
            engine.AddSpecies( guid, clim, optprop )
            
        for clim, optprop, guid in zip( self.user_clim.items(), self.user_optprop.items(), self.user_guid.items() ):
            engine.AddSpecies( guid[1], clim[1], optprop[1] )

        if addbackground:
            if self.backgroundclim is None:
                raise RuntimeError('Background climatology is required but not defined, try setting the backgroundclim member variable')
            engine.SetAtmosphericState( self.backgroundclim )
        
        self.add_albedo_to_engine(engine)
        
        self.add_wf_species_to_engine( engine )

    #------------------------------------------------------------------------------
    #               def set_albedo
    #------------------------------------------------------------------------------

    def set_albedo( self,
                    albedo : float,
                    engine : skif.ISKEngine = None ) -> None:
        self.albedo = albedo
        if (engine is not None): engine.SetAlbedo( self.albedo )
            
        
    
        