import sasktranif.sasktranif as skif
import radtran.climatologies.background as background
import numpy as np
from numpy import matlib
import os
import utility as utility
from scipy import io


def get_aerosol_cross_section(wavel_nm, mode_radius, mode_width):
    """
    Compute the scattering cross section of a lognormal particle size distribution

    Parameters
    ----------
    wavel_nm : wavelength in namometers
    mode_radius : mode radius of the lognormal distribution (microns)
    mode_width : mode width of the lognormal distribution (unitless, > 1)

    Returns
    -------
    scattering cross section : scattering cross in units of (cm^2)
    """
    optprop = optical_property([mode_radius, mode_radius], [mode_width, mode_width], [0, 100000], doLogInterpolation=0,
                               doPiecewiseLinear=1)

    air_clim, aer_opt, air_guid = background.msis()
    optprop.SetAtmosphericState(air_clim)
    optprop.SetLocation([0, 0, 500, 53000])

    return optprop.CalculateCrossSections(1e7 / wavel_nm)[2]


def get_angstrom_exponent(wavel_nm_1, wavel_nm_2, mode_radius, mode_width):
    xsec_1 = get_aerosol_cross_section(wavel_nm_1, mode_radius, mode_width)
    xsec_2 = get_aerosol_cross_section(wavel_nm_2, mode_radius, mode_width)

    return np.log(xsec_1 / xsec_2) / np.log(wavel_nm_1 / wavel_nm_2)


def get_phase_function(wavel_nm, mode_radius, mode_width,
                       cache_folder=r'C:\LocalAppDataFolder\SkOpticalProperties\mie_aerosolcache\mie_aerosol\h2so4\lognormal',
                       vector_phase_function=False):
    """
    Compute the scattering cross section of a lognormal particle size distribution

    Parameters
    ----------
    wavel_nm : wavelength in namometers
    mode_radius : mode radius of the lognormal distribution (microns)
    mode_width : mode width of the lognormal distribution (unitless, > 1)
    cache_folder : location of the aerosol cache files produced by sasktran
    vector_phase_function : return the full 4x4 phase matrix (default: False)

    Returns
    -------
    angles : angles in degrees
    phase function : The scalar component of the aerosol phase function p_11(theta)
    """

    filename = os.path.join(cache_folder, 'ln_' + str(int(np.round(mode_radius * 10000))).zfill(5) + '_' + str(
        int(np.round(mode_width * 1000))).zfill(5) + '_mie_' + str(int(np.round(wavel_nm * 10))).zfill(5) + '.dat')

    # generate the file if it doesn't exist
    if ~os.path.isfile(filename):
        get_aerosol_cross_section(wavel_nm, mode_radius, mode_width)
    with open(filename, 'rb') as file:
        buffer = file.read()

    data = np.frombuffer(buffer, dtype=float, count=4)
    wavel = 1e7 / data[0]
    kabs = data[1]
    kext = data[2]
    kscat = data[3]
    num_scatter = np.frombuffer(buffer[32::], dtype=np.uint32, count=1)[0]

    phase_function = np.reshape(np.frombuffer(buffer, dtype=float, count=num_scatter * 16, offset=36),
                                (4, 4, num_scatter), order='F')
    angles = np.rad2deg(np.arccos(np.linspace(-1, 1, num_scatter)))

    if not vector_phase_function:
        phase_function = phase_function[0, 0, :]

    return angles, phase_function


def from_text_file(fname, mode_radius=(0.08, 0.08), mode_width=(1.6, 1.6), particlesize_heights=(0, 100000),
                   DoLogInterpolation=0, DoPiecewiseLinear=0):
    clim = skif.ISKClimatology('USERDEFINED_PROFILE')
    guid = skif.SKCLIMATOLOGY_AEROSOL_CM3

    aero = np.loadtxt(fname)

    clim.SetProperty('Heights', aero[:, 0])
    clim.SetPropertyUserDefined(guid, aero[:, 1])
    clim.SetProperty('DoLogInterpolation', DoLogInterpolation)
    clim.SetProperty('DoPiecewiseLinear', DoPiecewiseLinear)

    optprop = optical_property(mode_radius, mode_width, particlesize_heights)

    return clim, optprop, guid


def apriori():
    user_paths = utility.data_folder()
    aer_filename = 'aerosol_apriori.txt'

    clim = None
    optprop = None
    guid = None
    if os.path.isfile(user_paths + aer_filename):
        clim, optprop, guid = from_text_file(user_paths + aer_filename, DoLogInterpolation=0, DoPiecewiseLinear=1)
    else:
        raise FileNotFoundError("Could not find aerosol apriori file")

    return clim, optprop, guid


def from_arrays(aero, aero_heights, mode_radius=(0.08, 0.08), mode_width=(1.6, 1.6), particlesize_heights=(0, 100000),
                DoLogInterpolation=0, DoPiecewiseLinear=0, mode=1):
    clim = skif.ISKClimatology('USERDEFINED_PROFILE')

    if mode == 1:
        guid = 'SKCLIMATOLOGY_AEROSOL_CM3'
    elif mode == 2:
        guid = 'SKCLIMATOLOGY_AEROSOL_CM3_MODE2'

    clim.SetProperty('Heights', aero_heights)
    clim.SetProperty('DoLogInterpolation', DoLogInterpolation)
    clim.SetProperty('DoPiecewiseLinear', DoPiecewiseLinear)
    clim.SetPropertyUserDefined(guid, aero)

    optprop = optical_property(mode_radius, mode_width, particlesize_heights)

    return clim, optprop, guid


def optical_property(mode_radius, mode_width, particlesize_heights, doLogInterpolation=0, doPiecewiseLinear=0):
    """
    Returns the aerosol optical property for MIEAEROSOL_H2SO4
    
    Parameters
    ----------
    mode_radius : list or numpy array
        Mode radii at different altitudes, same size as particlesize_heights
    mode_width : list or numpy array
        Mode width at different altitudes, same size as particlesize_heights
    particlesize_heights : list or numpy array
        Heights in m where the mode radii and mode width are specified
        
    Returns
    -------
    optprop : ISKOpticalProperty
        Aerosol optical property
    """
    optprop = skif.ISKOpticalProperty('MIEAEROSOL_H2SO4')

    particlesize = skif.ISKClimatology('USERDEFINED_PROFILE')
    particlesize.SetProperty('Heights', particlesize_heights)
    if isinstance(mode_radius, np.ndarray):
        mode_radius = mode_radius.tolist()
    if isinstance(mode_width, np.ndarray):
        mode_width = mode_width.tolist()
    if isinstance(particlesize_heights, np.ndarray):
        particlesize_heights = particlesize_heights.tolist()
    particlesize.SetPropertyUserDefined(skif.SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS, mode_radius)
    particlesize.SetPropertyUserDefined(skif.SKCLIMATOLOGY_LOGNORMAL_MODEWIDTH, mode_width)
    particlesize.SetProperty('DoLogInterpolation', doLogInterpolation)
    particlesize.SetProperty('DoPiecewiseLinear', doPiecewiseLinear)

    optprop.SetProperty('SetParticleSizeClimatology', particlesize)

    return optprop


def simulated_aerosol_plume(tp, look, plume_angle_middle, plume_width_angle, enhancement_factor, background_heights,
                            background_profile, plume_low_alt, plume_high_alt):
    """
    Creates a two dimensional plane climatology representing a simulated aerosol plume

    Parameters
    ----------
    tp : numpy array (3,)
        Reference tangent point
    look : numpy array (3,)
        Average look direction
    plume_angle_middle : scalar
        The angle in degrees from the tangent point where the plume should be centred.  Negative angles are closer
        to the observer
    plume_width_angle : scalar
        The width of the plume in degrees
    enhancement_factor : scalar
        The number to multiply the background profile by to create the plume
    background_heights : numpy array (n,)
        Heights of the background profile to use in [m]
    background_profile : numpy array (n,)
        Number density of the background aerosol profile in [/cm3]
    plume_low_alt : scalar
        Lower boundary of the plume in [m]
    plume_high_alt : scalar
        Upper boundary of the plume in [m]

    Returns
    -------
    clim : ISKClimatology
    optprop : ISKOpticalProperty
    guid : GUID
    normalandreference : numpy array (6,)
        Normal and referenced used for the plane climatology, intended to be passed back to the engine
    """
    clim = skif.ISKClimatology('USERDEFINED_PROFILE_PLANE')
    optprop = optical_property([0.08, 0.08], [1.6, 1.6], [0, 100000])
    guid = skif.SKCLIMATOLOGY_AEROSOL_CM3

    angles = np.array(
        [-15, -plume_width_angle / 2 + plume_angle_middle - 1.0, -plume_width_angle / 2 + plume_angle_middle,
         plume_width_angle / 2 + plume_angle_middle, plume_width_angle / 2 + plume_angle_middle + 1.0, 15])

    clim.SetProperty('Angles', angles)
    clim.SetProperty('Heights', background_heights)

    aer = matlib.repmat(background_profile, len(angles), 1)
    alt_idx = (background_heights > plume_low_alt) & (background_heights < plume_high_alt)
    aer[2:4, alt_idx] *= enhancement_factor

    reference = tp / np.linalg.norm(tp)
    normal = np.cross(reference, look)
    normal /= np.linalg.norm(normal)
    normalandreference = np.concatenate((normal, reference))

    clim.SetProperty('normalandreference', normalandreference)
    clim.SetPropertyUserDefined(guid, aer.flatten())

    return clim, optprop, guid, normalandreference


def test_scenario_elash2016(ext: str, fine_mode: str, coarse_mode_fraction: float):
    """
    Test aerosol scanarios used for Elash et al (2016) in JQRST

    Parameters
    ----------
    ext : str
        'small' to use background aerosol extinction profile, 'large' to use profile after nabro
    fine_mode : str
        'small' to use mode radius of 0.04, mode width of 1.8, 'large' to use mode radius of 0.12, mode width of 1.25
    coarse_mode_fraction : float
        Fraction of coarse mode to include. Coarse mode has mode radius of 0.3, mode with of 1.15
    Returns
    -------
    [(clim_fine, clim_coarse), (optprop_fine, optprop_coarse), (guid_fine, guid_coarse)]
        Note that if coarse_mode_fraction = 0 then the return will be [[clim_fine), (optprop_fine), (guid_fine)] and
        conversely if coarse_mode_fraction = 1

    """
    data_file = os.path.join(utility.data_folder(), 'aerosol', 'aerosol_scenarios.mat')

    data = io.loadmat(data_file)

    rg_coarse = data['rg_coarse'][:, 0]
    mw_coarse = data['mg_coarse'][:, 0]

    if ext.lower() == 'small':
        kext = data['kext'][:, 0]
    elif ext.lower() == 'large':
        kext = data['kext'][:, 1]
    else:
        raise ValueError('ext should be one of large or small')
    kext[np.isnan(kext)] = 0

    if fine_mode.lower() == 'small':
        rg_fine = data['rg_fine'][:, 0]
        mw_fine = data['mg_fine'][:, 0]
    elif fine_mode.lower() == 'large':
        rg_fine = data['rg_fine'][:, 1]
        mw_fine = data['mg_fine'][:, 1]
    else:
        raise ValueError('fine_mode should be one of large or small')

    guids = []
    optprops = []
    clims = []

    alt_km = data['alts'].T[:, 0]

    if coarse_mode_fraction < 1:
        fine_xs = get_aerosol_cross_section(750, rg_fine[0], mw_fine[0])
        fine_n = (1 - coarse_mode_fraction) * kext / fine_xs / 100000

        clim, optprop, guid = from_arrays(fine_n, alt_km * 1000, (rg_fine[0], rg_fine[0]),
                                          (mw_fine[0], mw_fine[0]))
        clims.append(clim)
        optprops.append(optprop)
        guids.append(guid)

    if coarse_mode_fraction > 0:
        coarse_xs = get_aerosol_cross_section(750, rg_coarse[0], mw_coarse[0])

        coarse_n = kext * coarse_mode_fraction / coarse_xs / 100000

        clim, optprop, guid = from_arrays(coarse_n, alt_km * 1000, (rg_coarse[0], rg_coarse[0]),
                                          (mw_coarse[0], mw_coarse[0]), mode=2)

        clims.append(clim)
        optprops.append(optprop)
        guids.append(guid)

    return clims, optprops, guids

