import datetime
import logging
import os
import re
from os import listdir

import h5py
import numpy as np
import sasktranif.sasktranif as skif
from astropy.time import Time

import radtran.climatologies.utility as clim_util


def ecmwf():
    """
    Returns back the ecmwf background climatology
    
    Returns
    -------
    clim : ISKClimatology('ECMWF')
    optprop : ISKOpticalProperty('RAYLEIGH')
    guid : SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3
    """
    
    return skif.ISKClimatology('ECMWF'), skif.ISKOpticalProperty('RAYLEIGH'), skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3

def osiris_ecmwf():
    """
    Returns back the ecmwf background climatology

    Returns
    -------
    clim : ISKClimatology('ECMWF')
    optprop : ISKOpticalProperty('RAYLEIGH')
    guid : SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3
    """

    return skif.ISKClimatology('OSIRIS_ECMWF'), skif.ISKOpticalProperty('RAYLEIGH'), skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3

def msis():
    """
    Returns back the msis background climatology

    Returns
    -------
    clim : ISKClimatology('MSIS90')
    optprop : ISKOpticalProperty('RAYLEIGH')
    guid : SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3
    """
    return skif.ISKClimatology('MSIS90'), skif.ISKOpticalProperty('RAYLEIGH'), skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3


def omps_anc_profile(orbit, image, slit, folder=r"\\utls\\users\scratch\\OMPS_L2_ANC\\", keyword='ANC', filename=None):
    """
    Returns back a 1-D climatology from the OMPS level 2 ancillary
    files, using scaled MSIS-90 above 80km

    Parameters
    ----------
    orbit : int
    image : int
    slit : int

        * 0 - left
        * 1 - middle
        * 2 - right

    Returns
    -------
    heights : Array (n,) of heights in m
    temperature : Array (n,) of temperature in K
    pressure : Array (n,) of pressure in hPa
    air_numberdensity : Array (n,) of number density in /cm^3
    """
    if filename is None:
        found_folder = False
        #check if folder exists
        if os.path.isdir( folder ):
            files = listdir(folder)
            found_folder = True
        #if not check if its on the python path
        else:
            user_paths = os.environ['PYTHONPATH'].split(os.pathsep)
            for i in range(len(user_paths)):
                if os.path.isdir( user_paths[i]+folder ):
                    found_folder = True
                    folder = user_paths[i]+folder
                    files = listdir(folder)
                    break

        if found_folder is False:
            logging.error('Could not find the omps ancillary folder')
            raise Exception('Could not find the omps ancillary folder')

        filename = None
        for f in files:
                m = re.search(keyword, f)
                if m is not None:
                    m = re.search( 'o([0-9]*)', f )
                    orbitnum = int(m.group(1))
                    if orbitnum == orbit:
                        filename = folder + f
                        break

    if filename is None:
        logging.error('Could not find the omps ancillary file')
        raise Exception('Could not find the omps ancillary file')
    else:
        #load in the atmospheric data from the OMPS ancillary files
        f = h5py.File(r''+filename,'r')
        altitude    = f['GEOLOCATION_DATA/Altitude'][:]*1000          #altitude in meters
        omps_temp   = f['/GRIDDED_DATA/Temperature'][image, slit, :]  #temperautre in K
        omps_pres   = f['/GRIDDED_DATA/Pressure'][image, slit, :]*100 #pressure in pascals
        omps_aird   = f['/GRIDDED_DATA/AirDensity'][image, slit, :]   #density in cm^-3
        latitude    = float(f['/GEOLOCATION_DATA/Latitude'][image, slit])
        longitude   = float(f['/GEOLOCATION_DATA/Longitude'][image, slit])
        date        = f['/GEOLOCATION_DATA/Date'][:]
        time        = f['/GEOLOCATION_DATA/Time'][:]
        # In ANC version 2.5 date and time fields were changed from (image, slit) to just slit
        if len(np.shape(date)) == 2:
            date = date[image, slit]
            time = time[image, slit]
        else:
            date = date[image]
            time = time[image]

        #determine the mjd
        year    = int(np.floor(date/10000))
        month   = int(np.floor( (date-year*10000)/100))
        day     = int(date- year*10000 - month*100)
        hour    = int(np.floor(  time/60/60))
        minute  = int(np.floor( (time - hour*60*60)/60))
        sec     = int(np.floor(  time - hour*60*60 - minute*60))
        usec    = int(np.floor( (time - hour*60*60 - minute*60 - sec)*1e6 ))
        t = Time(datetime.datetime(year,month,day,hour, minute, sec, usec),  'utc')
        mjd = t.mjd

        #create an msis-90 climatology and sample it at the OMPS locations
        clim_msis, opt_msis, guid_msis = msis()
        alts = np.arange(500.0,101500.0, 1000.0, dtype=float)
        alts = np.concatenate((np.array([0.0]), alts))
        msis_temp = clim_util.height_profile_from_clim(clim_msis, skif.SKCLIMATOLOGY_TEMPERATURE_K,        alts, latitude, longitude, mjd )
        msis_pres = clim_util.height_profile_from_clim(clim_msis, skif.SKCLIMATOLOGY_PRESSURE_PA,          alts, latitude, longitude, mjd )
        msis_aird = clim_util.height_profile_from_clim(clim_msis, skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, alts, latitude, longitude, mjd )

        #scale msis to match omps and use it above 79.5, also add ground point
        temperature = np.concatenate((np.array([msis_temp[0] * omps_temp[0]/msis_temp[1]]), omps_temp, msis_temp[82::]*omps_temp[80]/msis_temp[81] ))
        pressure    = np.concatenate((np.array([msis_pres[0] * omps_pres[0]/msis_pres[1]]), omps_pres, msis_pres[82::]*omps_pres[80]/msis_pres[81] ))
        air_density = np.concatenate((np.array([msis_aird[0] * omps_aird[0]/msis_aird[1]]), omps_aird, msis_aird[82::]*omps_aird[80]/msis_aird[81] ))

        return alts, temperature, pressure, air_density


def omps_anc(orbit, image, slit, folder=r"\\utls\\users\\scratch\\OMPS_L2_ANC\\", keyword='ANC'):
    """
    Returns back a 1-D climatology from the OMPS level 2 ancillary
    files, using scaled MSIS-90 above 80km

    Parameters
    ----------
    orbit : int
    image : int
    slit : int

        * 0 - left
        * 1 - middle
        * 2 - right

    Returns
    -------
    clim : ISKClimatology('USER_DEFINED')
    optprop : ISKOpticalProperty('RAYLEIGH')
    guid : SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3
    """

    alts, temperature, pressure, air_density = omps_anc_profile(orbit, image, slit, folder, keyword)

    # create the background climatology
    clim = skif.ISKClimatology('USERDEFINED_PROFILE')

    # temperature climatology
    clim.SetProperty( 'Heights', alts )
    clim.SetProperty('DoLogInterpolation',0)
    clim.SetProperty('DoPiecewiseLinear',1)
    clim.SetPropertyUserDefined( skif.SKCLIMATOLOGY_TEMPERATURE_K, temperature)

    # pressure climatology
    clim.SetProperty( 'Heights', alts )
    clim.SetProperty('DoLogInterpolation',1)
    clim.SetProperty('DoPiecewiseLinear',1)
    clim.SetPropertyUserDefined( skif.SKCLIMATOLOGY_PRESSURE_PA, pressure)

    # air number density climatology
    clim.SetProperty( 'Heights', alts )
    clim.SetProperty('DoLogInterpolation',1)
    clim.SetProperty('DoPiecewiseLinear',1)
    clim.SetPropertyUserDefined( skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, air_density)

    return clim, skif.ISKOpticalProperty('RAYLEIGH'), skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3
