import sasktranif.sasktranif as skif
import numpy as np

def generic_gaussian_profile( cloud_top, cloud_thickness, height_res, num_sigma ):
    std_dev = cloud_thickness / (2.0*np.sqrt(2.0*np.log(2.0)))
    mean_dist = cloud_top - cloud_thickness / 2.0
    
    num_heights = np.floor( 2.0*num_sigma*std_dev / height_res ) + 1
    h_index = np.arange(0, num_heights)
    
    height_m = height_res*np.round( (mean_dist - num_sigma*std_dev + h_index * height_res)/height_res )
    num_den = np.exp( -(height_m - mean_dist)**2 / (2.0*std_dev**2) )
    
    return num_den, height_m

def configure_profile_to_tau( height_m, num_den, cross_sect, desired_tau ):
    height_res = height_m[1] - height_m[0]
    
    current_tau = height_res * 100 * np.sum( num_den * cross_sect )
    
    return ( num_den * desired_tau / current_tau )

def one_dim_cloud( cloud_top, cloud_thickness, height_res, num_sigma, desired_tau, wavel_for_tau, effective_size):
    optprop = skif.ISKOpticalProperty('BAUM_ICECRYSTALS')
    optprop.SetProperty( 'EffectiveSizeMicrons', effective_size )
    
    # Neutral climatology and geo_point have no effect on ice crystal
    # cross sections
    neutral_clim = skif.ISKClimatology( 'MSIS90' )
    geo_instant = [0, 0, 1000, 5372]
    optprop.SetAtmosphericState(neutral_clim)
    optprop.SetLocation(geo_instant)
    
    wave_number = 10.0**7 / wavel_for_tau
    
    cross_sect = optprop.CalculateCrossSections( wave_number )[3]
    
    num_den, height_m = generic_gaussian_profile( cloud_top, cloud_thickness, height_res, num_sigma )
    num_den = configure_profile_to_tau( height_m, num_den, cross_sect, desired_tau )
    
    clim = skif.ISKClimatology( 'USERDEFINED_PROFILE' )
    guid = skif.SKCLIMATOLOGY_ICE_CM3
    clim.SetProperty( 'Heights', height_m )
    clim.SetPropertyUserDefined(  guid, num_den )

    return clim, optprop, guid, height_m    
    
def two_dim_cloud( cloud_top, cloud_thickness, height_res, num_sigma, desired_tau, wavel_for_tau, effective_size, tp, look, cloud_angle_middle, cloud_width_angle ):
    guid = skif.SKCLIMATOLOGY_ICE_CM3
    optprop = skif.ISKOpticalProperty('BAUM_ICECRYSTALS')
    optprop.SetProperty( 'EffectiveSizeMicrons', effective_size )
    
    # Neutral climatology and geo_point have no effect on ice crystal
    # cross sections
    neutral_clim = skif.ISKClimatology( 'MSIS90' )
    geo_instant = [0, 0, 1000, 5372]
    optprop.SetAtmosphericState( neutral_clim, geo_instant )
    
    wave_number = 10.0**7 / wavel_for_tau
    
    cross_sect = optprop.CalculateCrossSections( wave_number )[3]
    
    num_den, height_m = generic_gaussian_profile( cloud_top, cloud_thickness, height_res, num_sigma )
    num_den = configure_profile_to_tau( height_m, num_den, cross_sect, desired_tau )
    
    clim = skif.ISKClimatology( 'USERDEFINED_PROFILE_PLANE')
    angles = [-15, -cloud_width_angle/2.0 + cloud_angle_middle - 1.0, -cloud_width_angle/2.0 + cloud_angle_middle, cloud_width_angle/2.0 + cloud_angle_middle, cloud_width_angle/2.0 + cloud_angle_middle + 1.0, 15];
    
    clim.SetProperty('Angles', angles )
    clim.SetProperty('Heights', height_m )
    
    reference = tp / np.linalg.norm( tp )
    normal    = np.cross( reference, look )
    normal = normal / np.linalg.norm( normal )
    
    clim.SetProperty('normalandreference', normal.tolist() + reference.tolist())
    cloud = np.zeros([len(angles), np.size(height_m)])
    
    cloud[2,:] = num_den
    cloud[3,:] = num_den
    
    clim.SetPropertyUserDefined( guid, cloud.flatten() )
    
    return clim, optprop, guid, height_m

def two_dim_cloud_boundary( cloud_top, cloud_thickness, height_res, num_sigma, desired_tau, wavel_for_tau, effective_size, tp, look, cloud_left_angle, cloud_right_angle ):
    guid = skif.SKCLIMATOLOGY_ICE_CM3
    optprop = skif.ISKOpticalProperty('BAUM_ICECRYSTALS')
    optprop.SetProperty( 'EffectiveSizeMicrons', effective_size )
    
    # Neutral climatology and geo_point have no effect on ice crystal
    # cross sections
    neutral_clim = skif.ISKClimatology( 'MSIS90' )
    geo_instant = [0, 0, 1000, 5372]
    optprop.SetAtmosphericState( neutral_clim, geo_instant )
    
    wave_number = 10.0**7 / wavel_for_tau
    
    cross_sect = optprop.CalculateCrossSections( wave_number )[3]
    
    num_den, height_m = generic_gaussian_profile( cloud_top, cloud_thickness, height_res, num_sigma )
    num_den = configure_profile_to_tau( height_m, num_den, cross_sect, desired_tau )
    
    clim = skif.ISKClimatology( 'USERDEFINED_PROFILE_PLANE')
    angles = [-15, cloud_left_angle - 1.0, cloud_left_angle, cloud_right_angle, cloud_right_angle + 1.0, 15];
    
    clim.SetProperty('Angles', angles )
    clim.SetProperty('Heights', height_m )
    
    reference = tp / np.linalg.norm( tp )
    normal    = np.cross( reference, look )
    normal = normal / np.linalg.norm( normal )
    
    clim.SetProperty('normalandreference', normal.tolist() + reference.tolist())
    cloud = np.zeros([len(angles), np.size(height_m)])
    
    cloud[2,:] = num_den
    cloud[3,:] = num_den
    
    clim.SetPropertyUserDefined( guid, cloud.flatten() )
    
    return clim, optprop, guid, height_m