import numpy as np
import scipy.io.netcdf as netcdf
import scipy.interpolate as interpolate
import warnings
import sasktranif.sasktranif as skif

def mjd_to_cmam_time( mjd ):
    # Formula given on the wikipedia page
    y = 4716
    j = 1401
    m = 2
    n = 12
    r = 4
    p = 1461
    v = 3
    u = 5
    s = 153
    w = 2
    B = 274277
    C = -38
    
    J = mjd + 2400000.5
    
    f = J + j + ((4*J + B) // 146097) // 4 + C
    e = r*f+v
    g = (e % p) // r
    h = u*g+w
    D = (h % s) // u + 1
    M = (h // s + m) % n + 1
    Y = e // p - y + (n + m - M) // n
    
    H = (np.floor( J ) - J)*24
    
    a = np.floor( (14-1) / 2)
    y = Y + 4800 - a
    m = 1 + 12*a - 3
    
    day_number = D + np.floor((153*m+2)/5) + 365*y + np.floor(y/4) - np.floor(y/100) + np.floor(y/400) - 32045
    days_elapsed = J - day_number
    
    t = (Y-1900)*365*24 + days_elapsed*24 + H
    
    return t
    
def phi_to_z( phi, lat ):
    a = 6378137
    b = 6356750
    
    grav = 9.780327*(1+0.0053024 * np.sin( np.deg2rad(lat) )**2 - 0.0000058*np.sin( np.deg2rad(2*lat))**2)
    
    reff = np.sqrt( 1 / (np.cos(np.deg2rad(lat))**2/a**2 + np.sin(np.deg2rad(lat)**2/b**2) ) )
    
    z = phi * reff / (grav * reff / 9.80665 - phi)
    
    return z


# Custom simple binary search returning the index of the first element
# greater than val
# np.searchsorted loads the entire data array into memory which is not good
# for memory mapped netcdf variables
def binary_search( data, val, start = 0, end = None ):
    if( end == None):
        end = len(data)
    
    mid = int(np.floor( (start+end)/2 ))
    if val > data[mid]:
        if( mid == start):
            return end
        return binary_search( data, val, mid, end )
    else:
        if( mid == end ):
            return end
        return binary_search( data, val, start, mid)

def get_cmam_ozone_profile( mjd ):
    time = mjd_to_cmam_time( mjd )
    
    ozone_fname = r'\\utls\users\djz828\CMAM_DATA\o3_smf_cmam20_r02f_200507_200512_gp6_o3_numden_-_1_over_cm3.nc'
    phi_fname   = r'\\utls\users\djz828\CMAM_DATA\phi_sm_cmam20_r02f_199001_201012_gp6_gz.nc'
    
    f_o3 = netcdf.netcdf_file( ozone_fname, 'r' )
    f_phi = netcdf.netcdf_file( phi_fname, 'r' )
    
    t_o3 = f_o3.variables['time']
    t_phi = f_phi.variables['time']
    o3_full = f_o3.variables['o3']
    phi_full = f_phi.variables['phi']
    
    lon_o3 = f_o3.variables['lon']
    lat_o3 = f_o3.variables['lat']

    ind_o3 = binary_search( t_o3[:], time )
    ind_phi = binary_search( t_phi[:], time )

    if( ind_o3 == len( t_o3[:] ) ):
        ind_o3 = ind_o3 - 1
    
    # TODO: interpolate
    o3 = o3_full[ind_o3].copy()
    phi = phi_full[ind_phi].copy()

    z = phi
    for i in range(len(lat_o3[:])):
        z[:,i,:] = phi_to_z( phi[:,i,:] / 10, lat_o3[i] )
    
    heights = np.linspace(500, 99500, 100)
    o3_interp = np.zeros( (len(heights), len(lat_o3[:]), len(lon_o3[:]) + 1) )
    for i in range(len(lat_o3[:])):
        for j in range(len(lon_o3[:])):
            o3_interp[:,i,j] = interpolate.interp1d( z[:,i,j], o3[:,i,j], bounds_error = False, fill_value = 0  ) ( heights )
    
    o3_interp[:,:,-1] = o3_interp[:,:,-2]
    # cleanup netcdf mapped variables
    t_o3 = None
    t_phi = None
    o3_full = None
    phi_full = None
    
    lon = np.zeros(len(lon_o3[:])+1)
    lon[:-1] = lon_o3[:].copy()    
    lon[-1] = 360
    
    lat = lat_o3[:].copy()
    
    return o3_interp, lat, lon, heights
    
def get_cmam_climatology( mjd ):
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        o3,lat,lon,heights = get_cmam_ozone_profile( mjd )
    
    guid = skif.SKCLIMATOLOGY_O3_CM3
    optprop = skif.ISKOpticalProperty('O3_OSIRISRES')
    clim = skif.ISKClimatology('USERDEFINED_PROFILE3D_LATLONHEIGHT')
    
    clim.SetProperty('Heights', heights)
    clim.SetProperty('Longitudes', lon)
    clim.SetProperty('Latitudes', np.linspace(np.min(lat), np.max(lat), len(lat) ))
    o3 = o3.swapaxes(1, 2)
    clim.SetPropertyUserDefined( guid, o3.flatten('F') )
        
    return clim, optprop, guid
    
    
    