import numpy as np
import sasktranif.sasktranif as skif
from scipy import interpolate

def pressure_vmr_to_altitude_nd(lat, lon, mjd, vmr, pressure, alt):
    """
    Converts a single profile on a vmr/pressure grid to a numberdensity/altitude
    grid.  
    
    Parameters
    ----------
    lat : scalar
        Latitude in degrees in range (-90,90)
    lon : scalar
        Longitude in degrees in range(0,360) or (-180,180)
    mjd : scalar
        Modified julian date
    vmr : numpy array
        Volume mixing ratio of the profile
    pressure : numpy array
        Pressure levels the profile is on, in Pa
    alt : numpy array
        Altitudes to calculate the resulting number density on in m
    
    """
    ecmwf = skif.ISKClimatology('ECMWF')
    
    geo = [lat, lon, 0, mjd]
       
    p  = np.zeros(np.shape(alt))
    background = np.zeros(np.shape(alt))
    for idx, h in enumerate(alt):
        geo[2] = h
        
        ok, background[idx] = ecmwf.GetParameter( skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, geo )
        ok, p[idx] = ecmwf.GetParameter( skif.SKCLIMATOLOGY_PRESSURE_PA, geo )
    
    
    # Convert to vmr/altitude grid
    # Assume a constant vmr below/above the limits
    vmr_alt = interpolate.interp1d( pressure, vmr, bounds_error=False, fill_value=np.nan ) ( p )
    
    ind = np.where(~np.isnan(vmr_alt))[0]
    first, last = ind[0], ind[-1]
    vmr_alt[:first] = vmr_alt[first]
    vmr_alt[last+1:] = vmr_alt[last]
    
    return vmr_alt * background
    
    
def pressure_vmr_to_altitude_nd_gridded(lat, lon, mjd, vmr, pressure, alt):
    """
    Converts a grid of profiles on a vmr/pressure grid to a numberdensity/altitude
    grid.  
    
    Parameters
    ----------
    lat : scalar
        Latitude in degrees in range (-90,90)
    lon : scalar
        Longitude in degrees in range(0,360) or (-180,180)
    mjd : scalar
        Modified julian date
    vmr : numpy array
        Volume mixing ratio of the profile
    pressure : numpy array
        Pressure levels the profile is on, in Pa
    alt : numpy array
        Altitudes to calculate the resulting number density on in m
    
    """
    ecmwf = skif.ISKClimatology('ECMWF')
    
    geo = [0, 0, 0, mjd]
    n = np.zeros( (len(lat), len(lon), len(alt) ) )
    
    for idx, la in enumerate(lat):
        geo[0] = la
        print(idx)
        for idy, lo in enumerate(lon):
            geo[1] = lo
               
            p  = np.zeros(np.shape(alt))
            background = np.zeros(np.shape(alt))
            for idz, h in enumerate(alt):
                geo[2] = h
                ok, background[idz] = ecmwf.GetParameter( skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, geo )
                ok, p[idz] = ecmwf.GetParameter( skif.SKCLIMATOLOGY_PRESSURE_PA, geo )
            
            
            # Convert to vmr/altitude grid
            # Assume a constant vmr below/above the limits
            vmr_alt = interpolate.interp1d( np.log(pressure), vmr[:,idx,idy], bounds_error=False, fill_value=np.nan ) ( np.log(p) )
            
            ind = np.where(~np.isnan(vmr_alt))[0]
            first, last = ind[0], ind[-1]
            vmr_alt[:first] = vmr_alt[first]
            vmr_alt[last+1:] = vmr_alt[last]
            
            n[idx,idy,:] = vmr_alt * background
    return n
    