import numpy as np
from scipy import signal

import sasktranif.sasktranif as skif


def pratmo( cross_section='NO2_OSIRISRES'):
    """
    Initializes the NO2 climatology using PRATMO
    
    Returns
    -------
    clim: pointer to the NO2 climatology
    optprop: pointer to NO2 optical properties
    guid: unique identifier to no2
    """
    guid = skif.SKCLIMATOLOGY_NO2_CM3
    clim = skif.ISKClimatology('NO2PRATMO')
    optprop = skif.ISKOpticalProperty(cross_section)
    
    return clim, optprop, guid

def pratmo_box_model():
    guid = skif.SKCLIMATOLOGY_NO2_CM3
    clim_pratmo = skif.ISKClimatology('PRATMOBOXMODEL')
#    clim_pratmo.SetProperty('LatitudeResolution', 0)
#    clim_pratmo.SetProperty('NumLatitudeProfiles', 1)
    
    optprop = skif.ISKOpticalProperty('NO2_OSIRISRES')
    return clim_pratmo, optprop, guid

def convolved_vandaele_optical_property( wavel, psf, wavel_spacing = 0.01, num_stdev = 4, output_spacing=None  ):
    """
    Convolves down the Vandaele NO2 cross sections and creates a user defined
    optical property
    
    Parameters
    ----------
    wavel : numpy array
        Wavelengths to calculate the new cross section for in nm
    psf : numpy array
        Standard deviations of the gaussian to convolve by for each wavelength
        Can be the same size as wavel, or size 1 in which case the psf is assumed
        to be the same for all wavelengths
    wavel_spacing : scalar, optional
        Spacing to calculate the highresolution wavelengths on, default is the 
        Vandaele resolution
    num_stdev : scalar, optional
        Number of standard deviations to use in the gaussian for convolution,
        default is 4
    output_spacing : scalar,string optional
        Spacing to calculate the convolved cross section on. Either a scalar to
        use a fixed spacing, or None to use the wavelengths of
        the point spread function (wavel parameter)
        
    Returns
    -------
    optprop : ISKOpticalProperty
        Convolved optical property
    """
    optprop = skif.ISKOpticalProperty('USERDEFINED_TABLES')
    try:
        optprop.SetProperty('WavelengthTruncation', 1)
    except skif._sasktranif.functionfail:
        # This function mistakingly returns fail in older versions of SASKTRANIF so we supress the error
        # (It still works even if it fails)
        pass
    # If we only have one psf value, copy it for every wavelength
    if len(psf) == 1:
        psf = np.repeat(psf, len(wavel))
    # Temperatures of the DBM xsect    
    temp = [220, 294]
    # Ranges in whick each temperature has data
    temp_range = [ [238.0, 1000.0], [238.0, 1000.0] ]
    dbm = skif.ISKOpticalProperty('NO2_VANDAELE1998')
    clim_temp = skif.ISKClimatology('USERDEFINED_PROFILE')
    clim_temp.SetProperty('Heights', [0, 100000])
    geo = [0, 0, 1000, 54372]
    for T, ra in zip(temp, temp_range):
        # Get the high resolution DBM cross section
        clim_temp.SetPropertyUserDefined(skif.SKCLIMATOLOGY_TEMPERATURE_K, [T,T])

        dbm.SetAtmosphericState( clim_temp )
        dbm.SetLocation( geo )
        
        wavel_van = np.linspace( ra[0], ra[1], np.int(np.floor((ra[1] - ra[0] )/wavel_spacing)))
        xsect = np.zeros(np.shape(wavel_van))
        #TODO: Change this to calculatecrossectionsarray, but it is broken in the
        # interface right now
        for idx, w in enumerate(wavel_van):
            xsect[idx] = dbm.CalculateCrossSections(1e7/w)[1]
        
        # Only do the convolution for wavelenghts in the temperature range
        if output_spacing is not None:
            wavel_to_calc_for = np.linspace( ra[0], ra[1], np.floor((ra[1] - ra[0] )/output_spacing) )
            psf_to_calc_for = np.interp( wavel_to_calc_for, wavel, psf )
        else:
            good = (wavel > ra[0]) & (wavel < ra[1])
            wavel_to_calc_for = wavel[good]
            psf_to_calc_for = psf[good]

        wavelconv = np.zeros(np.shape(wavel_to_calc_for))
        xsectconv = np.zeros(np.shape(wavel_to_calc_for))

        for widx, (w, p) in enumerate( zip( wavel_to_calc_for, psf_to_calc_for) ):
            if (w > (wavel_van[0]+1*num_stdev * p)) | (w < wavel_van[-1]-1*num_stdev * p):

                good = (wavel_van > w-num_stdev*p) & (wavel_van < w+num_stdev*p)
                if not np.any(good):
                    wavelconv[widx] = w
                    xsectconv[widx] = np.interp(w, wavel_van,xsect)
                else:
                    wavel_filter = wavel_van[good]-w
                    f = np.exp( -0.5* (wavel_filter / p)**2 )
                    f = f / np.sum(f)

                    wavelconv[widx] = np.dot(wavel_van[good],f)
                    xsectconv[widx] = np.dot(xsect[good],f)

        good = xsectconv > 0
        full_xsect = np.interp( wavel_to_calc_for, wavelconv[good], xsectconv[good], left=np.nan, right=np.nan )

        if len(wavel_to_calc_for[~np.isnan(full_xsect)]) == 0:
            continue
        optprop.AddUserDefined(T, wavel_to_calc_for[~np.isnan(full_xsect)], full_xsect[~np.isnan(full_xsect)])
        
    return optprop

def userdefined( profile, heights, cross_section='NO2_OSIRISRES'):
    """
    Parameters
    ----------
    profile : 1d array
        one dimensional array containing the ozone profile [cm^-3]

    heights : 1d array
        height in meters of the profile

    cross_section : string
        name of the ozone cross section to be used
    """
    guid = skif.SKCLIMATOLOGY_NO2_CM3
    clim = skif.ISKClimatology('USERDEFINED_PROFILE')

    clim.SetProperty('Heights',heights)
    clim.SetPropertyUserDefined(guid, profile)

    optprop = skif.ISKOpticalProperty(cross_section)

    return clim, optprop, guid