import numpy as np
from scipy import signal

import sasktranif.sasktranif as skif
import radtran.climatologies.utility as utility
from utility.timing import Timer



def labow( cross_section='O3_OSIRISRES'):
    """
    Parameters
    ----------
    cross_section : string
        name of the ozone cross section to be used. defualt is O3_OSIRISRES

    Returns
    -------
    clim    : ISKClimatology
        labow ozone climatology

    optprop : ISKOpticalProperty
        ozone optical property

    guid : sastkran guid
        ozone unique identifier
    """
    guid = skif.SKCLIMATOLOGY_O3_CM3
    clim = skif.ISKClimatology('O3LABOW')
    optprop = skif.ISKOpticalProperty(cross_section)
    
    return clim, optprop, guid
    
    
def userdefined(profile, heights, cross_section='O3_OSIRISRES', do_log_interpolation=0, do_piecewise_linear=0):
    """
    Parameters
    ----------
    profile : 1d array
        one dimensional array containing the ozone profile [cm^-3]

    heights : 1d array
        height in meters of the profile
        
    cross_section : string
        name of the ozone cross section to be used

    do_log_interpolation : int, 0 or 1
        Set to 1 to use log of the profile in the interpolation.  Default 0

    do_piecewise_linear : int, 0 or 1
        Set to 1 to use linear interpolation rather than a spline. Default 0
    """
    guid = skif.SKCLIMATOLOGY_O3_CM3
    clim = skif.ISKClimatology('USERDEFINED_PROFILE')
    
    clim.SetProperty('Heights', heights)
    clim.SetProperty('DoLogInterpolation', do_log_interpolation)
    clim.SetProperty('DoPiecewiseLinear', do_piecewise_linear)
    clim.SetPropertyUserDefined(guid, profile)
    
    optprop = skif.ISKOpticalProperty(cross_section)
    
    return clim, optprop, guid

# Returns back clim,optprop,guid
def simulated_ozone_hole( ozone_hole_frac, start_lat, lat_res, refgeo ):
    clim = skif.ISKClimatology('USERDEFINED_PROFILE3D_LATLONHEIGHT')
    
    backgroundclim, optprop, guid = labow()
    
    lon = np.linspace( 0, 360, 12 )
    
    heights = np.linspace( 500, 99500, 100)
    lat = np.arange(-90, 90, lat_res)
    lat = np.append( lat, 90 )

    o3 = utility.height_profile_from_clim( backgroundclim, guid, heights, refgeo[0], refgeo[1], refgeo[3] )
    
    o3 = np.tile(o3, (len(lat), len(lon), 1) )
    o3[lat < start_lat,:,:] = o3[lat < start_lat, :, :] * ozone_hole_frac
    
    
    clim.SetProperty( 'Latitudes', lat )
    clim.SetProperty( 'Longitudes', lon )
    clim.SetProperty( 'Heights', heights )
    clim.SetPropertyUserDefined( guid, o3.flatten() )
    
    return clim, optprop, guid
    

def convolved_dbm_optical_property( wavel, psf, wavel_spacing = 0.01, num_stdev = 4, output_spacing=None ):
    """
    Convolves down the DBM ozone cross sections and creates a user defined
    optical property
    
    Parameters
    ----------
    wavel : numpy array
        Wavelengths to calculate the new cross section for in nm
    psf : numpy array
        Standard deviations of the gaussian to convolve by for each wavelength
        Can be the same size as wavel, or size 1 in which case the psf is assumed
        to be the same for all wavelengths
    wavel_spacing : scalar, optional
        Spacing to calculate the highresolution wavelengths on, default is the 
        DBM resolution
    num_stdev : scalar, optional
        Number of standard deviations to use in the gaussian for convolution,
        default is 4
    output_spacing : scalar,string optional
        Spacing to calculate the convolved cross section on. Either a scalar to
        use a fixed spacing, or None to use the wavelengths of
        the point spread function (wavel parameter). Default is none.
        
    Returns
    -------
    optprop : ISKOpticalProperty
        Convolved optical property
    """
    optprop = skif.ISKOpticalProperty('USERDEFINED_TABLES')
    try:
        optprop.SetProperty('WavelengthTruncation', 1)
    except skif._sasktranif.functionfail:
        # This function mistakingly returns fail in older versions of SASKTRANIF so we supress the error
        # (It still works even if it fails)
        pass
    # If we only have one psf value, copy it for every wavelength
    if len(psf) == 1:
        psf = np.repeat(psf, len(wavel))
    # Temperatures of the DBM xsect    
    temp = [218, 228, 243, 273, 295]
    # Ranges in whick each temperature has data
    temp_range = [ [194.50, 650.01], [194.50, 520.01], [194.50, 519.01], [299.50, 520.01], [195.00, 830.00] ]
    dbm = skif.ISKOpticalProperty('O3_DBM')
    clim_temp = skif.ISKClimatology('USERDEFINED_PROFILE')
    clim_temp.SetProperty('Heights', [0, 100000])
    geo = [0, 0, 1000, 54372]
    for T, ra in zip(temp, temp_range):
        # Get the high resolution DBM cross section
        clim_temp.SetPropertyUserDefined(skif.SKCLIMATOLOGY_TEMPERATURE_K, [T,T])
        dbm.SetAtmosphericState( clim_temp )
        dbm.SetLocation( geo )

        wavel_dbm = np.linspace( ra[0], ra[1], np.int(np.floor((ra[1] - ra[0] )/wavel_spacing)))
        xsect = np.zeros(np.shape(wavel_dbm))
        #TODO: Change this to calculatecrossectionsarray, but it is broken in the
        # interface right now
        for idx, w in enumerate(wavel_dbm):
            xsect[idx] = dbm.CalculateCrossSections(1e7/w)[1]
        
        # Only do the convolution for wavelenghts in the temperature range
        if output_spacing is not None:
            wavel_to_calc_for = np.linspace( ra[0], ra[1], np.floor((ra[1] - ra[0] )/output_spacing) )
            psf_to_calc_for = np.interp( wavel_to_calc_for, wavel, psf )
        else:
            good = (wavel > ra[0]) & (wavel < ra[1])
            wavel_to_calc_for = wavel[good]
            psf_to_calc_for = psf[good]
        
        wavelconv = np.zeros(np.shape(wavel_to_calc_for))
        xsectconv = np.zeros(np.shape(wavel_to_calc_for))

        for widx, (w, p) in enumerate( zip( wavel_to_calc_for, psf_to_calc_for) ):
            if (w > (wavel_dbm[0]+1*num_stdev * p)) | (w < wavel_dbm[-1]-1*num_stdev * p):

                good = (wavel_dbm > w-num_stdev*p) & (wavel_dbm < w+num_stdev*p)
                if not np.any(good):
                    wavelconv[widx] = w
                    xsectconv[widx] = np.interp(w, wavel_dbm,xsect)
                else:
                    wavel_filter = wavel_dbm[good]-w
                    f = np.exp( -0.5* (wavel_filter / p)**2 )
                    f = f / np.sum(f)

                    wavelconv[widx] = np.dot(wavel_dbm[good],f)
                    xsectconv[widx] = np.dot(xsect[good],f)

        good = xsectconv > 0
        full_xsect = np.interp( wavel_to_calc_for, wavelconv[good], xsectconv[good], left=np.nan, right=np.nan )
        
        if len(wavel_to_calc_for[~np.isnan(full_xsect)]) == 0:
            continue
        optprop.AddUserDefined(T, wavel_to_calc_for[~np.isnan(full_xsect)], full_xsect[~np.isnan(full_xsect)])
        
    return optprop
            
            