import numpy as np
import sasktranif.sasktranif as skif

def height_profile_from_clim( climatology, guid, heights, lat, lon, mjd ):
    """
    Get the height profile from a climatology at the specified location

    Parameters
    ----------
    climatology: ISKClimatology object
    guid:        ISKGuid object
    heights:     array
        an array of heights in meters
    lat: double
        latitude
    lon: double
        longitude
    mjd: double
        modified julian date

    Returns
    -------
    profile: numpy.1darray
        array the same size as heights containing the profile values.
    """

    numden = np.zeros(len(heights))
    
    geo = [lat, lon, 0, mjd]
    for i in range(len(heights)):
        geo[2] = heights[i]
        ok,numden[i] = climatology.GetParameter( guid, geo)
        
    return numden
    



def extinction_to_number_density( heights_m, kext, wavel, optprop, neutral = None, lat = 0, lon = 0, mjd = 54372 ):
    """
    Converts a (height,extinction) profile to (height,numberdensity)
    
    Parameters
    ----------
    heights_m : numpy array
        Heights in m for the extinction profile
    kext : numpy array
        Same shape as heights_m, extinction in per cm
    wavel : scalar
        Wavelength in which the extinction is specified
    optprop : ISKOpticalProperty
        Optical property to use to calculate cross sections
    neutral : ISKClimatology (optional)
        Climatology to use for the neutral background, if not specified ECMWF is used
    lat : scalar, optional
        Latitude, defaults to 0
    lon : scalar, optional
        Longitude, defaults to 0
    mjd : scalar, optional
        modified julian date, defaults to 54372
        
    Returns
    -------
    n : numpy array
        Number density specified at heights_m in /cm3
    """
    if neutral is None:
        neutral = skif.ISKClimatology('ECMWF')
    
    n = np.zeros(np.shape(kext))    
    
    geo = [lat, lon, 0, mjd]
    optprop.SetAtmosphericState( neutral )
    for idx, (h,k) in enumerate(zip(heights_m, kext)):
        geo[2] = h
        optprop.SetLocation( geo )
        ok, sigma_abs, sigma, sigma_scat = optprop.CalculateCrossSections( 1e7 / wavel )
        n[idx] = k / sigma
    
    return n
    
def number_density_to_extinction( heights_m, n_cm3, wavel, optprop, neutral = None, lat = 0, lon = 0, mjd = 54372 ):
    """
    Converts a (height,number density) profile to (height,extinction)
    
    Parameters
    ----------
    heights_m : numpy array
        Heights in m for the extinction profile
    n_cm3 : numpy array
        Same shape as heights_m, aerosol number density in particles per cm^3
    wavel : scalar
        Wavelength in which the extinction is specified
    optprop : ISKOpticalProperty
        Optical property to use to calculate cross sections
    neutral : ISKClimatology (optional)
        Climatology to use for the neutral background, if not specified ECMWF is used
    lat : scalar, optional
        Latitude, defaults to 0
    lon : scalar, optional
        Longitude, defaults to 0
    mjd : scalar, optional
        modified julian date, defaults to 54372
        
    Returns
    -------
    k : numpy array
        extinction specified in per km
    """
    if neutral is None:
        neutral = skif.ISKClimatology('ECMWF')
    
    k = np.zeros(np.shape(n_cm3))    
    geo = np.double([lat, lon, 0, mjd])
    for idx, (h,n) in enumerate(zip(heights_m, n_cm3)):
        geo[2] = h
        optprop.SetAtmosphericState( neutral )
        optprop.SetLocation( geo )
        ok, sigma_abs, sigma, sigma_scat = optprop.CalculateCrossSections( 1e7 / wavel )
        k[idx] = n * sigma * 1e5
    
    return k