"""
Radiative transfer engine interface designed for use in the retrieval module.
Currently only interfaces to the SASKTRAN framework are implemented.

SASKTRAN Classes
-------
    :class:`~retrieval.radtran.SasktranGeneric`
        Should be the default class to use.  Supports any engine under the
        SASKTRAN framework (currently 'MC', 'HR', 'SO')
    :class:`~retrieval.radtran.SasktranHR`
        A specialized class that only supports the 'HR' engine.  Use this class
        if you need to calculate weighting functions for the species
    :class:`~retrieval.radtran.SasktranHRPreCacheWavelengths`
        Another specialized version of the HR engine which caches optical 
        cross sections at all wavelengths before performing the calculation.
        This can speed up the calculation significantly when modelling an atmosphere
        containing HITRAN species, at the expense of more RAM used.  
        Only use this class if you have HITRAN species in the atmosphere and
        don't mind the extra RAM.  For a one dimensional calculation, 320
        wavelengths uses approximately 1.5gb of memory, and should scale linearly
        with the number of wavelengths.
    :class:`~retrieval.radtran.SasktranHRTwoDim`
        Specialized version of HR designed for two dimensional radiative transfer
        calculations, specifically for tomography along an orbit track.  Since
        a single radiative transfer calculation occurs at one instant in time,
        the calculation must be split to handle primarily movement of the sun.
        This class automatically does this splitting and handles the translation
        of an orbit track to local planes.  Do not use this class for general
        two/three dimensional radiative transfer calculations with HR, instead use 
        :class:`~retrieval.radtran.SasktranHR`  .  This class should only be used
        for tomographic retrievals.
        
Example
-------
Required imports

>>> import radtran.atmosphere at atmosphere
>>> import radtran.sasktran as radtran
>>> import radtran.climatologies as clim
>>> from radtran.scan import Scan
>>> import numpy as np
>>> import radtran.options as opt

For setup, we need both an atmosphere and a geometry.  Start with the
atmosphere

>>> atmo = atmosphere.Atmosphere()

Add in background rayleigh atmosphere

>>> clim, optprop, guid = clim.background.ecmwf()
>>> atmo.add_species( clim, optprop, guid )

And some ozone

>>> clim, optprop, guid = clim.ozone.labow()
>>> atmo.add_species( clim, optprop, guid )

Set the albedo to something we want

>>> atmo.albedo = 0.3

More information on the atmosphere can be found in :class:`~utility.sasktran.atmosphere`
Species can also be added manually using the sasktranif module, more information can
be found in the sasktranif external documentation.

Now, set up the geometry,

>>> scan = Scan()
>>> scan.make_scan_saa( 20, 60, 80, 20, np.arange(12,42), 57174, 0 )

Make our wavelengths,

>>> wavel_nm = np.array([280, 350, 600])

We can optionally set up radiative transfer options,

>>> options = opt.GenericOptions()
>>> options.add_option('setsun', scan.sun )

If we dont set the sun manually the sun position will be automatically calculated
through the mjd.  Documentation for options can be found in the external sasktranif
document, options for the HR engine are also found under :class:`~utility.sasktran.options.HROptions`

Now we can make the radiative transfer engine,

>>> engine = radtran.SasktranGeneric( atmo, 'HR', options )

And add our lines of sight and wavelengths,

>>> for obs, look, mjd in zip(scan.obs, scan.look, scan.mjd):
>>>     engine.add_line_of_sight( obs, look, mjd )
>>> engine.set_wavelengths( wavel_nm )

And perform the radiative transfer calculation, we take the first return value
since weighting functions are returned as the second return value

>>> rad = engine.calculate_radiance()[0]
"""

import copy

import numpy as np

from radtran.radtranbase import RadTranWrapper
import sasktranif.sasktranif as skif
import radtran.options as skopt
from scipy import interpolate


class SasktranGeneric(RadTranWrapper):
    """ Generic wrapper class for SASKTRAN engines.  Currently supports
        HR,MC and SO.  More complicated features such as polarization or
        weighting functions are not supported through this class
    """
    
    def __init__(self, atmosphere, enginetype, options=None):
        """
        Creates the SasktranGeneric class
        
        Parameters
        ----------
        atmosphere :
            Atmosphere to add to the engine, must support add_to_engine
        enginetype : String
            Any engine that ISKEngine supports, currently can be 'HR', 'MC', or 'SO'
        options : optional
            Options to add to the engine, must support add_to_engine
        """
        self.engine = skif.ISKEngine(enginetype)
        self.wavel_nm = None
        self.options = options
        self.atmosphere = atmosphere
        self.numlos = 0
        atmosphere.add_to_engine( self.engine )
        self.is_engine_initialized = False
    
    def add_line_of_sight(self, observer, look, mjd):
        """
        Adds a single line of sight to the engine
        
        Parameters
        ----------
        observer : numpy array, or nxvector
            Observer position in (x,y,z) format
        look : numpy array, or nxvector
            Unit look direction in (x,y,z) format
        mjd : scalar
            Modified julian date
        """
        self.engine.AddLineOfSight(mjd, observer, look)
        self.numlos += 1
    
    def set_wavelengths(self, wavel_nm):
        """
        Sets the wavelengths for the radiative transfer calculation
        
        Parameters
        ----------
        wavel_nm : numpy array
            Wavelengths in nm 
        """
        self.wavel_nm = wavel_nm
    
    def calculate_radiance(self):
        """
        Calculates the radiance
        
        Returns
        -------
        tuple
            (rad,zeros) radiance is a numpy array in the shape (line_of_sight, wavelength)
        """
        self.atmosphere.add_albedo_to_engine(self.engine)
        if not self.is_engine_initialized:
            if self.options is not None:
                self.options.add_to_engine(self.engine)
            self.engine.InitializeModel()
            self.is_engine_initialized = True
        
        self.engine.SetWavelengths( self.wavel_nm )
        ok, rad = self.engine.CalculateRadiance()
        return rad.transpose(), np.zeros(np.shape(rad))


class SasktranHR(SasktranGeneric):
    def __init__(self, atmosphere, options=None):
        """ 
        Specialized class for calculations with the SASKTRAN HR engine.
        Supports the calculation of weighting functions
        
        Parameters
        ----------
        atmosphere :
            Atmosphere for the engine, needs to support add_to_engine and add_wf_species_to_engine
        options : optional
            Options for the engine, needs to support add_to_engine
        """
        self.options = options
        if self.options is None:
            self.options = skopt.HROptions()
            self.calcwf = True
        else:
            self.calcwf = False
            if 'calcwf' in self.options.optmap:
                if self.options.optmap['calcwf'] != 0:
                    self.calcwf = True
        self.engine = skif.ISKEngine('HR')
        self.atmosphere = atmosphere
        self.options = options
        self.is_engine_initialized = False
        self.applied_options = None

        self.los = []
        self.mjd = []
        self.obs = []

    def add_line_of_sight( self, observer, look, mjd ):
        self.obs.append(observer)
        self.los.append(look)
        self.mjd.append(mjd)
        self.is_engine_initialized = False

    def calculate_radiance(self):
        """
        Runs the engine and returns back the radiance and weighting functions
        Weighting functions are only calculated if calcwf is set to something
        other than 0
        
        Returns
        -------
        tuple
            (radiance, weighting_function) radiance is of shape (line of sight, wavelength)
            and weighting_function is of shape (wavelength, line of sight, perturbation)
            the number and nature of the perturbations is given by the options used
            to set up the engine
        """
        stokes, wf = self.calculate_polarized_radiance(0)
        rad = np.zeros(np.shape(stokes))
        for i in range(len(stokes)):
            for j in range(len(stokes[i])):
                rad[i, j] = stokes[i, j].I()

        return rad, wf

    def calculate_polarized_radiance(self, polarizationtype=None):
        """
        Parameters
        ----------
        polarizationtype : scalar
            - 0 - No polarization
            - 1 - Exact to scatter order 1
            - 2 - Exact to scatter order 2
            - 3 - Exact to scatter order 3
            - 4 - Exact to scatter order 3, assumes higher orders are strongly polarized
            - 5 - Exact to scatter order 4, assumes higher orders are strongly polarized
            - 6 - Exact to scatter order 4, assumed higher orders are weakly polarized

        Returns
        -------
        radiance : numpy array shape (num_los, num_wavel)

        wf : numpy array shape(wavelength, line_of_sight, perturbation)

        stokes_vector : numpy array shape (num_line_of_sight, num_wavelength)  of type StokesVector
        """
        if polarizationtype is not None:
            self.options.add_option('polarizationtype', polarizationtype)

        if self.applied_options != self.options.optmap:
            self.engine = None
            self.engine = skif.ISKEngine('HR')
            self.atmosphere.add_to_engine(self.engine)
            self.is_engine_initialized = False

        if not self.is_engine_initialized:
            self.options.add_to_engine(self.engine)
            self.numlos = 0
            for mjd, obs, los in zip(self.mjd, self.obs, self.los):
                super().add_line_of_sight(obs, los, mjd)
            self.applied_options = copy.copy(self.options.optmap)
            self.is_engine_initialized = True
        stokes_vector = np.empty((len(self.wavel_nm), self.numlos), dtype=StokesVector)
        self.atmosphere.add_wf_species_to_engine(self.engine)
        self.atmosphere.add_albedo_to_engine(self.engine)
        self.engine.SetWavelengths(self.wavel_nm)

        ok, stokes = self.engine.CalculateStokesVector()
        ok, wf = self.engine.GetWeightingFunctions()
        for idx in range(len(self.wavel_nm)):
            for idy in range(self.numlos):
                basis = np.zeros((3, 3))
                basis[0, :] = stokes[idx, idy].propagation_direction()
                basis[1, :] = stokes[idx, idy].theta_direction()
                basis[2, :] = stokes[idx, idy].phi_direction()
                v = np.array([stokes[idx, idy].I(), stokes[idx, idy].Q(), stokes[idx, idy].U(), stokes[idx, idy].V()])
                stokes_vector[idx, idy] = StokesVector(v, basis)
        return stokes_vector.transpose(), wf

    def set_num_orders_of_scatter(self, num_scatters):
        self.engine.SetProperty('numordersofscatter', num_scatters)


class SasktranHRPrecacheWavelengths(SasktranHR):
    """
    A specialized version of the SASKTRAN HR engine which supports precaching 
    a set of wavelengths.  This can be particularly useful when modelling HITRAN
    species over a large number of wavelengths.  This class also supports
    calculation of weighting functions
    
    This causes the calculation to use significantly more RAM
    """
    
    def __init__(self, atmosphere, options = None):
        super().__init__(atmosphere, options)
    
    def set_wavelengths(self, wavel_nm):
        if isinstance(wavel_nm, np.ndarray):
            self.wavel_nm = wavel_nm
        else:
            self.wavel_nm = np.array(wavel_nm)
    
    def set_wf_alts( self, alts_m ):
        #TODO: actually use the wf alts
        self.wf_alts = alts_m
    
    def calculate_polarized_radiance(self, polarizationtype=None):
        """
        Parameters
        ----------
        polarizationtype : scalar
            - 0 - No polarization
            - 1 - Exact to scatter order 1
            - 2 - Exact to scatter order 2
            - 3 - Exact to scatter order 3
            - 4 - Exact to scatter order 3, assumes higher orders are strongly polarized
            - 5 - Exact to scatter order 4, assumes higher orders are strongly polarized
            - 6 - Exact to scatter order 4, assumed higher orders are weakly polarized

        Returns
        -------
        radiance : numpy array shape (num_line_of_sight, num_wavelength)

        wf : numpy array shape(wavelength, line_of_sight, perturbation)

        stokes_vector : numpy array shape (num_line_of_sight, num_wavelength)  of type StokesVector
        """
        # All we do here is make sure that the engine is deleted inbetween calculations

        # add user options to the engine
        if self.options is not None:
            self.options.add_option('precachewavel', self.wavel_nm.tolist())

        self.is_engine_initialized = False
        self.applied_options = None
        stokes_vector, wf = super().calculate_polarized_radiance(polarizationtype)

        return stokes_vector, wf


class SasktranHRMSApprox(SasktranHR):
    """
    Specialization of the SASKTRAN HR model which calculates the full multiple scatter radiance field on a coarse
    resolution grid, and then interpolates onto a finer resolution single scatter calculation grid.

    """
    def __init__(self, atmo, opt, ms_wavel_space=10):
        super().__init__(atmo, opt)
        self.wavel_nm = None
        self.ms_wavel = None
        self.ms_wavel_idx = None
        self.ms_wavel_space = ms_wavel_space
        self.do_ss_approx = True
        if 'numordersofscatter' in self.options.optmap:
            self.numordersofscatter = self.options.optmap['numordersofscatter']

    def set_multiscatter_wavelength_indicies(self, ms_wavel_idx):
        """
        Sets the wavelength grid in which to calculate the multiple scatter signal on.
        Parameters
        ----------
        ms_wavel_idx : list or numpy array
            Indicies to perform the multiple scatter calculation on.  self.wavel_nm[ms_wavel_idx] are the wavelengths
            in which the calculation is performed.  If ms_wavel_idx is None, then the single scatter approximation is
            turned off, i.e., all calculations are done with the full multi scatter.

        Returns
        -------
        None
        """
        if ms_wavel_idx is None:
            self.do_ss_approx = False
        else:
            self.ms_wavel_idx = np.array(ms_wavel_idx)
            self.do_ss_approx = True
            self.ms_wavel = self.wavel_nm[self.ms_wavel_idx]
            if len(self.ms_wavel) == len(self.wavel_nm):
                if all(np.isclose(self.ms_wavel, self.wavel_nm)):
                    self.do_ss_approx = False

    def __create_default_multiscatter_wavelength_indicies(self):
        # Figure out which wavelengths we want to perform the full MS calculation on
        self.ms_wavel = []
        self.ms_wavel_idx = []
        cur_wavel = self.wavel_nm[0]
        if not all(x<y for x, y in zip(self.wavel_nm, self.wavel_nm[1:])):
            raise ValueError('Wavelengths should be entered in ascending order')
        self.ms_wavel_idx.append(0)
        for idx, w in enumerate(self.wavel_nm):
            wavel_diff = np.abs(cur_wavel - w)
            if wavel_diff > self.ms_wavel_space:
                self.ms_wavel_idx.append(idx)
                cur_wavel = w
        self.ms_wavel_idx.append(len(self.wavel_nm)-1)
        self.ms_wavel_idx = np.unique(self.ms_wavel_idx)
        self.ms_wavel = self.wavel_nm[self.ms_wavel_idx]

    def calculate_polarized_radiance(self, polarizationtype=None):
        if self.do_ss_approx is False:
            return super().calculate_polarized_radiance(polarizationtype)
        if self.ms_wavel is None:
            self.__create_default_multiscatter_wavelength_indicies()
        all_wavel = self.wavel_nm
        # First calculate the SS radiance
        self.options.add_option('numordersofscatter', 1)
        self.options.add_option('wfprecision', 1)

        super().set_num_orders_of_scatter(1)
        stokesSS, wfSS = super().calculate_polarized_radiance(polarizationtype)

        # Now Reset to MS
        self.wavel_nm = self.ms_wavel
        self.options.add_option('numordersofscatter', self.numordersofscatter)
        self.options.add_option('wfprecision', 0)
        super().set_num_orders_of_scatter(self.numordersofscatter)
        stokesMS, wfMS = super().calculate_polarized_radiance(polarizationtype)

        # Set wavelengths back
        self.wavel_nm = all_wavel

        # Get the scalar radiance components
        radSS = np.zeros(np.shape(stokesSS))
        for i in range(len(stokesSS)):
            for j in range(len(stokesSS[i])):
                radSS[i, j] = stokesSS[i, j].I()
        radMS = np.zeros(np.shape(stokesMS))
        for i in range(len(stokesMS)):
            for j in range(len(stokesMS[i])):
                radMS[i, j] = stokesMS[i, j].I()

        # Get the MS Components
        ms_component = copy.copy(radMS / radSS[:, self.ms_wavel_idx])
        # For every LOS construct the interpolating function
        for i in range(len(stokesSS)):
            for j in range(len(stokesSS[0])):
                stokesSS[i, j].stokes *= np.interp(all_wavel[j], self.ms_wavel, ms_component[i, :], left=ms_component[i, 0], right=ms_component[i, -1])

        if self.calcwf:
            with np.errstate(invalid='ignore'):
                ms_component_wf = copy.copy(wfMS - wfSS[self.ms_wavel_idx, :, :])
            for i in range(np.size(wfSS, axis=0)):
                for j in range(np.size(wfSS, axis=1)):
                    for k in range(np.size(wfSS, axis=2)):
                        wfSS[i, j, k] += np.interp(all_wavel[i], self.ms_wavel, ms_component_wf[:, j, k], left=ms_component_wf[0, j, k], right=ms_component_wf[-1, j, k])


        return stokesSS, wfSS

    def set_wavelengths(self, wavel_nm):
        """
        Sets the wavelengths for the radiative transfer calculation

        Parameters
        ----------
        wavel_nm : numpy array
            Wavelengths in nm
        """

        if self.wavel_nm is None:
            self.ms_wavel = None
        elif not np.all(np.asarray(wavel_nm) == self.wavel_nm):
            self.ms_wavel = None

        self.wavel_nm = wavel_nm


class StokesVector(object):
    """
    Defines a stokes vector with its associated basis.

    Parameters
    ----------
    stokes : numpy array shape (4,)
        The stokes vector [I, Q, U, V]

    basis : numpy array shape (3,3)
        Coordinate basis the stokes vector is defined in.  basis[0, :] is the propagation direction, basis[1, :] is
        the theta direction, and basis[2, :] is the phi direction.  Directions are specified in ECEF coordinates.
        The basis must be constructed such that basis[1, :] cross basis[2, :] is equal to basis[0, :]

    Raises
    ------
    ValueError if the basis is not correctly constructed

    """
    def __init__(self, stokes, basis):
        if not np.isclose(basis[0, :].dot(np.cross(basis[1, :], basis[2, :])), 1):
            raise ValueError('The propagation direction should be equal to the cross product of the theta and phi' +
                             'directions')
        self.stokes = copy.copy(stokes)
        self.basis = copy.copy(basis)

    def to_new_basis(self, new_basis):
        """
        Converts the stokes vector to a new basis. A rotation matrix between the new basis and old basis is constructed
        and applied to the stokes vector.  Note that this process overrides the old basis

        Parameters
        ----------
        new_basis : numpy array shape (3,3)
            The new basis.  See the class constructor documentation for the format
        """

        if not np.isclose(new_basis[0, :].dot(self.basis[0, :]), 1):
            raise ValueError('new_basis and current basis should have the same propagation direction')
        if not np.isclose(new_basis[0, :].dot(np.cross(new_basis[1, :], new_basis[2, :])), 1):
            raise ValueError('The propagation direction should be equal to the cross product of the theta and phi' +
                             'directions for the new basis')

        cos_eta = self.basis[1, :].dot(new_basis[1, :])
        sin_eta = -1*self.basis[1, :].dot(new_basis[2, :])

        cos_two_eta = cos_eta**2 - sin_eta**2
        sin_two_eta = 2.0*cos_eta*sin_eta

        rot = np.array([[1, 0, 0, 0], [0, cos_two_eta, -sin_two_eta, 0],
                        [0, sin_two_eta, cos_two_eta, 0], [0, 0, 0, 1]])

        self.stokes = rot.dot(self.stokes)
        self.basis = copy.copy(new_basis)

    def propagation_direction(self):
        """
        Returns the propagation direction of the current basis in ECEF coordinates
        """
        return self.basis[0, :]

    def theta_direction(self):
        """
        Returns the theta direction of the current basis in ECEF coordinates
        """
        return self.basis[1, :]

    def phi_direction(self):
        """
        Returns the phi direction of the current basis in ECEF coordinates
        """
        return self.basis[2, :]

    def I(self):
        return self.stokes[0]

    def Q(self):
        return self.stokes[1]

    def U(self):
        return self.stokes[2]

    def V(self):
        return self.stokes[3]
