import sasktranif.sasktranif as skif
import numpy as np
import logging
import radtran.sasktran as sasktran


def sun_location(mjd):
    engine = skif.ISKEngine('HR')
    
    los = np.array([-1, 0, 0])
    obs = 6371*1000*np.array([1, 0, 1])
    
    engine.AddLineOfSight(mjd, obs, los)
    
    engine.InitializeModel()
    ok,sun = engine.GetProperty('sun')
    
    return sun


class Scan:
    """
    Utility class for creating vertical scans from a satellite platform.  
    Meant for quick radiative transfer simulations, for retrievals it is recommended
    to use one of the classes in instruments or platforms.  
    """
    def __init__(self):
        self.obs = []           # Array of nxVector types
        self.look = []          # Array of nxVector types
        self.wavelnm = []       # numpy array of wavelengths in nm
        self.sun = []           # numpy array containing sun unit vector
        self.mjd = []           # numpy array of sun directions
        self.reftp = []         # Reference TP of scan as numpy array
        self.reflook = []       # Reference look direction as numpy array
        
        self.lon = []
        self.lat = []
        self.sza = []
        self.saa = []
        self.ssa = []
        self.refmjd = []
        self.rebobs = None

    def make_scan_saa(self, sza, saa, lat, lon, tanalts_km, mjd, locallook, sataltkm=600, refaltkm=20 ):
        """
        Sets up the scan geometry.  The resulting scan wont have the exact specified
        parameters, but the exact values are calculated afterwards and stored in
        scan.lon, scan.lat, scan.sza, scan.saa, scan.ssa
        
        Parameters
        ----------
        sza : scalar 
            Solar zenith angle in degrees
        saa : scalar
            Solar geographic azimuth angle in degrees at geographic location lat,lon. 0 is north, 90 is east
        lat : scalar
            Latitude in degrees (-90 to 90)
        lon : scalar
            Longitude in degrees, either (0,360) or (-180,180)
        tanalts_km : numpy array
            Array of tangent altitudes in km 
        mjd : scalar
            Modified julian date
        locallook : scalar
            Direction to look on the surface of the earth in degrees. 0 is local
            north, 90 is local east.  In range (0,360)
        sataltkm : scalar, optional
            Satellite altitude in km, default is 600
        refaltkm : scalar, optional
            The tangent altitude where sza,saa,lat,lon are valid.  Default is
            20km
        """
        geo = skif.ISKGeodetic()
        
        geo.SetLocationLatLonAlt(lat, lon, refaltkm * 1000)
        
        self.reftp = geo.GetLocationXYZ()
        west = geo.GetLocalWest()
        south = geo.GetLocalSouth()
        up = geo.GetLocalUp()
        
        sunh = -np.cos(np.deg2rad(saa)) * south - np.sin(np.deg2rad(saa))*west
        self.sun = np.cos(np.deg2rad(sza)) * up + np.sin(np.deg2rad(sza))*sunh
        
        self.reflook = -np.cos( np.deg2rad( locallook ) )*south - np.sin( np.deg2rad( locallook ) )*west
        
        a = (np.linalg.norm( self.reftp ) + sataltkm*1000)**2
        
        s = -np.dot( self.reftp, -self.reflook ) + np.sqrt( np.dot( self.reftp, -self.reflook)**2 - np.linalg.norm( self.reftp )**2 + a )
        
        obs = self.reftp - s*self.reflook
        self.refobs = self.reftp - s*self.reflook
        for alt in tanalts_km:
            ok, look = geo.SetLocationFromTangentAltitude(alt*1000.0, obs, self.reflook)
            self.obs.append(obs)
            self.look.append(look)
            self.mjd.append(mjd)
        
        self.calc_real_parameters()

    def calc_real_parameters(self):
        geo = skif.ISKGeodetic()
        geo.SetLocationFromTangentPoint(self.refobs, self.reflook)
        self.reftp = geo.GetLocationXYZ()

        self.lon = geo.GetLongitude()
        self.lat = geo.GetLatitude()
        
        # SSA is sun dot look
        self.ssa = np.rad2deg(np.arccos(np.dot(self.sun, self.reflook)))
        # SZA is sun dot tp
        self.sza = np.rad2deg(np.arccos(np.dot(self.sun, self.reftp / np.linalg.norm(self.reftp))))
        # TODO: calc saa
        self.saa = None
        self.refmjd = np.mean(self.mjd)

    def make_scan_ssa( self, sza, ssa, lat, lon, tanalts_km, mjd, locallook, sataltkm=600, refaltkm=20 ):
        """
        Same as make_scan_saa, except the scattering angle is specified rather than the
        solar azimuth angle. Since not all scattering angles are allowed for
        a specific geometry, if the scattering angle is too low it it set
        to the lowest value allowed and a warning is displayed.
        """
        sza_r = np.deg2rad( sza )
        ssa_r = np.deg2rad( ssa )        
        
        if np.cos( ssa_r ) / np.sin( sza_r ) > 1:
            logging.warning('Unphysical geometry, forcing scattering angle to be the maximum allowed, %f', np.rad2deg( np.arccos( np.sin( sza_r ) ) ))
            saa =  locallook
        else:
            saa =  np.rad2deg( np.arccos( np.cos( ssa_r ) / np.sin( sza_r ) ) ) + locallook
        
        self.make_scan_saa( sza, saa, lat, lon, tanalts_km, mjd, locallook, sataltkm, refaltkm )

    def make_scan_sun_fixed_mjd( self, sza, ssa, tanalts_km, mjd, locallook, sataltkm=600, refaltkm=20 ):
        """
        Currently buggy, do not use expecting good results
        """
        sun = sun_location( mjd )
        
        sza_r = np.deg2rad( sza )
        ssa_r = np.deg2rad( ssa )
        
        if np.cos( ssa_r ) / np.sin( sza_r ) > 1:
            logging.warning('Unphysical geometry, forcing scattering angle to be the maximum allowed, %f', np.rad2deg( np.arccos( np.sin( sza_r ) ) ))
            saa =  locallook
        else:
            saa =  np.rad2deg( np.arccos( np.cos( ssa_r ) / np.sin( sza_r ) ) ) + locallook
        
        saa_r = np.deg2rad( saa )
        
        A = -1 * np.sin(sza_r) * np.cos( saa_r )
        B = -1 * np.sin(sza_r) * np.sin( saa_r )
        C = np.cos(sza_r)
        
        theta = np.arccos( sun[2] / np.sqrt(A**2 + C**2) ) - np.arctan2( A, C )
        
        lat = np.pi/2 - theta
        lon = np.arccos( C * np.sin(theta) + A * np.cos(theta) ) / np.sqrt(sun[0]**2 + sun[1]**2) - np.arctan2( -sun[1], sun[0])
        
        h = np.array([np.cos(lon), np.sin(lon), 0])
        z = np.array([0,0,1])
        th = theta
        up = np.sin(th)*h + np.cos(th)*z
        south = np.cos(th)*h - np.sin(th)*z
        west = np.cross(south,up)
        newsun = A*south+B*west+C*up
        
        if np.linalg.norm(newsun - sun) > 1e-10:
            lon = -1*np.arccos( C * np.sin(theta) + A * np.cos(theta) ) / np.sqrt(sun[0]**2 + sun[1]**2) - np.arctan2( -sun[1], sun[0])
            h = np.array([np.cos(lon), np.sin(lon), 0])
            up = np.sin(th)*h + np.cos(th)*z
            south = np.cos(th)*h - np.sin(th)*z
            west = np.cross(south,up)
            newsun = A*south+B*west+C*up
        
        lon = np.mod(lon+np.pi,2*np.pi) - np.pi
        if theta < 0:
            theta = np.abs(theta)
            lat = np.pi/2 - theta
            lon = lon - np.pi
        
        geo = skif.ISKGeodetic()
        geo.SetLocationLatLonAlt( np.rad2deg(lat), np.rad2deg(lon), refaltkm*1000)
        self.reftp = geo.GetLocationXYZ()
        west    = geo.GetLocalWest()
        south   = geo.GetLocalSouth()
        up      = geo.GetLocalUp()
        
        sunh = -np.cos( np.deg2rad( saa ) ) * south - np.sin( np.deg2rad( saa ) )*west
        self.sun = np.cos( np.deg2rad( sza ) ) * up + np.sin( np.deg2rad( sza ) )*sunh
        
        self.reflook = -np.cos( np.deg2rad( locallook ) )*south - np.sin( np.deg2rad( locallook ) )*west
        
        a = (np.linalg.norm( self.reftp ) + sataltkm*1000)**2
        
        s = -np.dot( self.reftp, -self.reflook ) + np.sqrt( np.dot( self.reftp, -self.reflook)**2 - np.linalg.norm( self.reftp )**2 + a )
        
        obs = self.reftp - s*self.reflook

        for alt in tanalts_km:
            ok, look = geo.SetLocationFromTangentAltitude(alt*1000.0, obs, self.reflook)
            self.obs.append(obs)
            self.look.append(look)
            self.mjd.append(mjd)
        
        self.calc_real_parameters()

    def make_scan_from_scia_angles(self, saa_tp, sza_tp, tan_alts, mean_lat, mean_lon, mean_mjd, sat_alt=790,
                                   earth_rad=6378):
        '''
        Sets up an approximate sciamachy geometry. Latitude is only approximate, and is shifted so that the solar zenith and
        solar azimuth angles match the desired values.

        Parameters
        ----------
        saa_tp : solar azimuth angle - angle betweeen the sun and look vector at the tangent point in the tangent plane
        sza_tp : solar zenith angle - angle between the vertical and the sun
        tan_alts : tangent altitudes in km
        mean_lat : average latitude of the scan
        mean_lon : average longitude of the scan
        mean_mjd : mean mjd of the scan - used to set the sun direction
        sat_alt : satellite altitude in km
        earth_rad : earth radius in km - used in conjunction with the satellite altitude to determine satellite position (optional)

        '''
        self.sun = sun_location(mean_mjd)

        self.sza = np.ones(len(tan_alts))
        self.saa = np.ones(len(tan_alts))
        self.lat = np.ones(len(tan_alts))
        self.lon = np.ones(len(tan_alts))
        lat_ret = np.zeros(len(tan_alts))  # retrieved latitudes

        for lidx, alt, saa, sza in zip(np.arange(0, len(tan_alts)), tan_alts * 1000, saa_tp * np.pi / 180,
                                       sza_tp * np.pi / 180):
            geo = skif.ISKGeodetic()

            n = 11
            sza_arr = np.zeros(n)
            lat_arr = np.linspace(mean_lat - 2, mean_lat + 2, n)
            for idx, lati in enumerate(lat_arr):
                geo.SetLocationLatLonAlt(lati, mean_lon, alt)
                up = geo.GetLocalUp()
                sunp = np.cross(up, np.cross(self.sun, up))
                sunp = sunp / np.sqrt(np.sum(sunp ** 2))

                l = sunp * np.cos(saa) + np.cross(up, sunp) * np.sin(saa) + up * np.dot(up, sunp) * (1 - np.cos(saa))

                x = np.sqrt((sat_alt + earth_rad) ** 2 - (alt / 1000 + earth_rad) ** 2)  # sat distance from TP in kilometers
                o = geo.GetLocationXYZ() - l * x * 1000

                geo.SetLocationFromTangentPoint(o, l)
                test_up = geo.GetLocalUp()
                sza_arr[idx] = np.arccos(np.dot(self.sun, test_up))

            # x values must be increasing
            if sza_arr[0] > sza_arr[-1]:
                lat_ret[lidx] = np.interp(sza, sza_arr[::-1], lat_arr[::-1])
            else:
                lat_ret[lidx] = np.interp(sza, sza_arr, lat_arr)

            geo.SetLocationLatLonAlt(lat_ret[lidx], mean_lon, alt)
            up = geo.GetLocalUp()
            sunp = np.cross(up, np.cross(self.sun, up))
            sunp = sunp / np.sqrt(np.sum(sunp ** 2))

            self.look.append(
                sunp * np.cos(saa) + np.cross(up, sunp) * np.sin(saa) + up * np.dot(up, sunp) * (1 - np.cos(saa)))

            x = np.sqrt(
                (sat_alt + earth_rad) ** 2 - (alt / 1000 + earth_rad) ** 2)  # sat distance from TP in kilometers
            self.obs.append(geo.GetLocationXYZ() - self.look[lidx] * x * 1000)
            self.mjd.append(mean_mjd)

            geo.SetLocationFromTangentPoint(self.obs[lidx], self.look[lidx])
            self.reftp.append(geo.GetLocationXYZ())
            test_up = geo.GetLocalUp()
            self.sza[lidx] = np.arccos(np.dot(self.sun, test_up)) * 180 / np.pi
            self.saa[lidx] = -np.arccos(np.dot(sunp, self.look[lidx])) * 180 / np.pi  # how scia defines saa
            self.lat[lidx] = geo.GetLatitude()
            self.lon[lidx] = geo.GetLongitude()

    def add_to_engine(self, engine):
        """
        Initializes an engine based on the geometry inside of the class.  All lines of sight are added to the engine
        and inside the engines options the sun is set to the appropriate values.

        engine: skif.ISKEngine or radtran.sasktran.SasktranGeneric
        """
        if isinstance(engine, skif.ISKEngine):
            for mjd, obs, look in zip(self.mjd, self.obs, self.look):
                engine.AddLineOfSight(mjd, obs, look)

            if self.wavelnm is not []:
                engine.SetWavelengths(self.wavelnm)
            engine.SetProperty('SetSun', self.sun)
        elif isinstance(engine, sasktran.SasktranGeneric):
            for o, l, m in zip(self.obs, self.look, self.mjd):
                engine.add_line_of_sight(o, l, m)
                engine.options.add_option('SetSun', self.sun)
        else:
            raise TypeError('Unsupported engine type')
