
class Retrieval(object):
    
    def retrieve_species( self, species, method ):
        raise NotImplementedError()
        

class RetrievalLimb(Retrieval):
    """
    See the example for a quick working explanation    
    
    The class which handles all limb retrievals. The core concept is we have
    measurements **y** and a model capable of calculating **F** and we want
    to minimize the difference between the two.  In order to do this, we need
    four things:
    
    meas_inst
        This is a class capable of providing the information required to
        calculate **y**.  It can either be a simulated instrument or a container
        for data for a real instrument
    
    ret_inst
        This class provides the information to calculate **F**.  Typically
        ret_inst will support :func:`calculate_radiance()` which uses a radiative
        transfer model to get the information required for **F**
        
    species
        species is a class capable of taking in an instrument (either meas_inst,
        or ret_inst) and converting the raw data (usually radiance) to a measurement
        vector.  The species and instrument type are loosly coupled together,
        for example, the :class:`H2OShow` species is unlikely to be compatible
        with the OSIRIS instrument.  However, it is possible to write species which
        are compatible with more than one instrument.
        
        When the retrieval starts, :func:`species.configure_instrument_and_method`
        is called, which gives the species a chance to configure the instruments
        and method for the specific retrieval task at hand.
        
        The species is also responsible for updating the atmosphere, typically
        through an :func:`update_profile` call
    
    method
        The iteration method used to push the solution in a good direction.  Typically
        the general algorithm is as follows
            - call :func:`ret_inst.calculate_radiance` to prepare the instrument 
            - call :func:`species.measurement_vector` to get **F**
            - Compare **F** and **y** to figure how much to update the profile by
              (This is usually where the key differences between methods are)
            - Tell the species how much to change the profile
            - Repeat
        
        The method is very loosely coupled to the species.  Methods like Levenberg-Marquardt
        require that the Jacobian of the measurement vector be calculated, and the species
        will need to be capable of that for Levenberg-Marquardt to work.  However, usually
        the same method can be reused for many species
    
    
    The techniques these four objects use to accomplish their tasks has been
    abstracted out of the retrieval, this is to ensure that this class will be
    compatible for essentially any limb retrieval. 
    
    The current way that all retrievals are done is with the help of two classes,
    
    radtran
        A wrapper to a radiative transfer model.  Basically the radiative transfer wrapper
        just has :func:`add_line_of_sight` and :func:`calculate_radiance`.  ret_inst
        holds a copy of this class and uses it to calculate radiances required
    
    atmosphere
        A wrapper class for specifying the atmosphere.  For SASKTRAN radiative
        transfer models this is typically just a list of climatologies, opticalproperties,
        and guids.  Usually this class has a function :func:`add_to_engine` which
        adds the current atmosphere to the radiative transfer model.
        Copies of this class are held in the radtran class, and also the species class,
        which uses it to update the atmosphere.
    
    Example
    -------
    Here we set up, explain, and perform a full retrieval for the SHOW instrument.  Lots 
    of these operations such as setting up the apriori are usually done in 
    separate functions, but we expand them out here for clarity.
    
    Required imports
    
    >>> from retrieval import instruments as show
    >>> import radtran.scan as scan
    >>> import radtran.atmosphere as atmosphere
    >>> import radtran.climatologies.water as water
    >>> import sasktranif.sasktranif as skif
    >>> import radtran.options as options
    >>> import retrieval.method.lsq as lsq
    >>> import retrieval.species as species
    >>> import radtran.sasktran as radtran
    >>> import retrieval.core as retrieval
    
    Start by picking a geometry, this can be done any way, as long as we get
    los, obs, mjd in the end
    
    >>> sim_scan = scan.Scan()
    >>> sim_scan.make_scan_saa( 30, 60, 40, 20, np.arange(12,40,1), 54372, 90 )
    >>> los = sim_scan.look
    >>> obs = sim_scan.obs
    
    Make the instrument we are going to simulate with
    
    >>> wavel_nm_sim = np.linspace(1362.5, 1366.5, 320)
    >>> meas_inst = show.SHOWModel(  wavel_nm_sim, los, obs, sim_scan.mjd )
    
    And the instrument we are going to retrieve with.  The wavelengths we enter
    here do not matter, as the species will overwrite them with the correct 
    ones in the retrieval
    
    >>> ret_inst = show.SHOWModel(  wavel_nm_sim, los, obs, sim_scan.mjd )
    
    Now let's make the atmosphere we are going to simulate with, we will only
    add background atmosphere and water
    
    >>> atmo_sim = atmosphere.Atmosphere()
    >>> atmo_sim.add_species( skif.ISKClimatology('ECMWF'), skif.ISKOpticalProperty('Rayleigh'), skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3 )
    >>> clim, optprop, guid = water.from_text_file( FILENAME_TO_WATER_PROFILE_HERE )
    >>> atmo_sim.add_species( clim, optprop, guid ) 
    >>> atmo_sim.albedo = 0.3
    
    We can make our retrieval atmosphere the same way, except we add the water profile
    as a user defined profile so we can modify it later.  Here we interpolate the
    profile onto a uniform grid that we want to retrieve on
    
    >>> atmo_ret = atmosphere.Atmosphere()
    >>> atmo_ret.add_species( skif.ISKClimatology('ECMWF'), skif.ISKOpticalProperty('Rayleigh'), skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3 )
    >>> atmo_ret.albedo = 0.3
    >>> water_apriori = np.loadtxt( WATER_APRIORI_FILENAME )
    >>> ret_alts = np.linspace(0, 100000, 101)
    >>> water_ret = np.exp( np.interp( ret_alts, water_apriori[:,0], np.log(water_apriori[:,1]) ) )
    >>> atmo_ret.add_user_defined( guid, optprop, ret_alts, water_ret, 'show_water' )
    
    Now we make the radiative transfer models, first the simulation model.  
    We use SasktranHRPrecacheWavelengths which is a specialized version of the HR model
    which caches optical properties at all wavelengths, this way HITRAN cross
    sections are not recomputed every wavelength
    
    >>> sim_options = options.HROptions()
    >>> sim_options.add_option('numordersofscatter', 1)
    >>> sim_options.add_option('calcwf', 1 )
    >>> sim_options.add_option('setsun', sim_scan.sun.tolist() )
    >>> sktran_sim = radtran.SasktranHRPrecacheWavelengths( atmo_sim, sim_options )
    >>> meas_inst.setup_radtran_model( sktran_sim )
    
    And the retrieval model 
    
    >>> ret_options = options.HROptions()
    >>> ret_options.add_option('calcwf', 1 )
    >>> ret_options.add_option('numordersofscatter', 1)
    >>> ret_options.add_option('setsun', sim_scan.sun.tolist() )
    >>> sktran_ret = radtran.SasktranHRPrecacheWavelengths( atmo_ret, ret_options )
    >>> ret_inst.setup_radtran_model( sktran_ret )
    
    Now we can do the simulation
    
    >>> meas_inst.calculate_radiance()
    
    And create our retrieval objects

    >>> ret = retrieval.RetrievalLimb( meas_inst, ret_inst )
    >>> method = lsq.LevenbergMarquardt(maxiter=5, damping=0.1, dampingpower = 10)
    >>> ret_species = species.H2OShow( atmo_ret )

    And finally perform the retrieval

    >>> output = ret.retrieve_species( ret_species, method )

    """
    def __init__( self, meas_inst, ret_inst ):
        self.meas_inst = meas_inst
        self.ret_inst = ret_inst
        
    def retrieve_species( self, species, method, do_error=False ):
        species.configure_instrument_and_method( self.ret_inst, self.meas_inst, method )    
        
        y_meas, y_meas_error = species.measurement_vector( self.meas_inst, do_error )
        
        method.retrieve( y_meas, y_meas_error, self.ret_inst, species )
        
        return species.output_data()