import numpy as np
import copy
import os
from scipy import io
from retrieval.instruments.spectral_imager import MultiSensorImager
import retrieval.instruments.sensors.sensor as sen
import retrieval.instruments.optics.optics as opt
import utility.solar_spectrum as sol

class LimbProfiler_scan(MultiSensorImager):
    """
    Class used to combine measurements from instruments such OSIRIS, OMPS, SCHIAMACHY
    into a single measurement. Wavelengths are assumed not to change between
    exposures.
    Measurements are grouped into line of a sight x wavelength array.

    Parameters
    ----------
    wavel_nm : np.ndarray
        an array of wavelengths

    los : np.ndarray
        a list of nx3 arrays. Each list element contains the optical axis lines
        of sight for an exposure with n subexposures. Specificies in geodetic x,y,z
        coordinate.

    vertical : np.ndarray
        a list in the same format and size as los describing the vertical
        orientation of the sensor. If the sensor has more than one pixel,
        or the pixel line of sight is oversamples this should be set.

    obs : np.ndarray
        a list in the same format and size as los describing the observer position.
        This should be set if the instrument will be used to setup the radiative
        transfer model.

    mjd : np.ndarray
        a list in the same format and size as los describing the modified
        julian date of each line of sight. This should be set if the
        instrument will be used to setup the radiative transfer model.
"""
    #TODO: get rid of the text file once level one serivices is up and running
    def __init__(self, wavel_nm, los, vertical=None,obs=None,mjd=None, ideal=True ):
        """

            
        """
        
        if isinstance(wavel_nm, np.ndarray):
            temp_wave= wavel_nm
            wavel_nm = [0,]*len(los)
            for i in range(0,len(los)):
                wavel_nm[i] = temp_wave
                
                
        super().__init__(wavel_nm, los, vertical, obs, mjd)
        self.ideal = ideal
        self.solar_spectrum = None
        self.solar_spectrum_wavel = None
        self.measurement = None
        self.multiply_by_solar_spectra = True

    def get_meas_geometry(self):
        """

        Provides the lines of sight and observer positions of the instrument measurements
        
        Returns
        -------
        pixel_los : np.array
            array of exposure lines of sight in geographic coordinates
        pixel_obs : np.array
            array of exposure locations in geographic coordinates
        """
        
        return np.vstack(self.pixel_los), np.vstack(self.pixel_obs)        
        
    def initialize_sensors_and_optics(self):
        if self.ideal:
            for i in range(0,self.num_sensors):
                self.sensor[i]         = sen.NxMSensor(1,1,0,0)
                self.optics[i]         = opt.BasicOptics(0,0,np.array([0, np.inf]),np.array([0.35, 0.35]),1,1,1)
                self.num_pixels[i]     = 1
        else:
            for i in range(0,self.num_sensors):
                self.sensor[i]         = sen.NxMSensor(1,1,0,0)
                self.optics[i]         = opt.BasicOptics(self.vertRes,self.horRes,self.psf_wavel_nm,self.psf,self.vertSamples,self.horSamples,self.waveSamples,self.waveSamples_nm)
                self.optics[i].integration_type = self.wavel_int_type
                self.optics[i].wavel_num_stdev  = self.wavel_num_stdev
                self.num_pixels[i]     = 1
        
        self.setup_instrument( )
        
    def set_wavelengths( self, wavel_nm ):
        """
        Sets up the instrument with the measurement wavelengths
        
        Parameters
        ----------
        wavel_nm: np.array
            an array of wavelengths 
        """
        #setup the lo-res wavelengths
        if isinstance(wavel_nm, np.ndarray):
            temp_wave= wavel_nm
            for i in range(self.num_sensors):
                self.wavel_nm[i] = temp_wave
                
         #setup the hi-res wavelengths
        for i in range(self.num_sensors):
            self.set_sensor_wavelengths( self.wavel_nm[i], i )
                
    def convert_radiance_to_l1(self, radiance):
        
        #computer the measuerments using the base class
        super().convert_radiance_to_l1(radiance)   
        
        #rearrange them into a matrix (los x wavelength)
        meas = np.zeros((0,len(self.wavel_nm[0])))
        for i in range(0,len(self.measurement)):
            meas = np.vstack((meas, self.measurement[i]))
        
        #Multiply by the radiance by the solar spectrum (assumes the wavelength
        # doesn't change between exposure)
        if self.multiply_by_solar_spectra:
            if not np.all(np.array(self.solar_spectrum_wavel) == self.wavel_nm[0]):
                self.solar_spectrum = sol.convolved_solar_spectrum( self.wavel_nm[0], self.psf_wavel_nm, self.psf )
                self.solar_spectrum_wavel = copy.copy(self.wavel_nm[0])

            self.measurement = meas*self.solar_spectrum[np.newaxis, :]
        else:
            self.measurement = copy.copy(meas)
        #self.measurement = meas
        return self.measurement


    def calculate_radiance( self ):
        """
        Calculates the instrument measurement and weighting function.
        """
        all_wave = self.radtran_wavelengths()
        self.radtran.set_wavelengths( all_wave )

        rad, wf  = self.radtran.calculate_radiance()
        self.rad = self.convert_radiance_to_l1( rad )

        if np.size(wf)!=0:
            self.wf = np.zeros( (np.size(self.wavel_nm[0]), len(self.los), np.shape(wf)[2]) )

            #integrate analytic weighting functions only if required (it can be slow compared to a
            #single scatter calculation)
            if np.any(np.asarray(self.waveSamples) > 1) | np.any(np.asarray(self.vertSamples) > 1) | np.any(np.asarray(self.horSamples) > 1) | np.any(np.asarray(self.num_exposures) > 1):
                for idx in range(np.shape(wf)[2]):
                    self.wf[:,:,idx] = np.transpose( self.convert_radiance_to_l1( np.transpose( wf[:,:,idx] ) ) )
            else:
                for idx in range(np.shape(wf)[2]):
                    self.wf[:,:,idx] = wf[:,:,idx]

                if self.multiply_by_solar_spectra:
                    self.wf *= np.interp( self.wavel_nm[0], self.solar_spectrum_wavel, self.solar_spectrum)[:,np.newaxis ,np.newaxis]
