import numpy as np
import logging

class Optics(object):

    def __init__(self):
        pass
    
    def LinesOfSightFromOpticsAxis(self, lookdir, orientup=[0,0,0] ):
        raise NotImplementedError()
        
    def RecommendedRTMWavelengths(self, pixelWavelengths ):
        raise NotImplementedError()
        
    def RecommendedRTMParameters(self, pixelLOS ):
        raise NotImplementedError()
        
class IdealOptics(Optics):
    #passes wavelength and line of sight straight through
    def __init__(self):
        super().__init__()
        self.vertRes = 0
        self.horRes  = 0
        self.waveRes = 0
        
        self.vertSamples = 1  #1 vertical samples for each pixel
        self.horSamples  = 1  #1 horizontal samples for each pixel
        self.waveSamples = 1  #1 wavelength sample for each wavelength
        
    def RecommendedRTMWavelengths(self, pixelWavelengths):
        weights = np.ones(np.size(pixelWavelengths))
        return pixelWavelengths, weights
        
    def RecommendedRTMParameters(self, pixelLOS ):
        weights = 1
        return pixelLOS, weights
        
class BasicOptics(Optics):
    """
    Class which handles spatial and spectral integration parameters. Given a
    line of sight and wavelength, the optics tells the instrument what is needed
    to model the measurement and how to do the integration.
    """ 

    def __init__(self, vertRes=3.5e-4,horRes=1e-3,wavel_nm=np.array([0, 10000]),wavePSF = np.array([0.35, 0.35]),vertSamples=1,horSamples=1,waveSamples=(1,1),waveSamples_nm=(0,np.Inf), opticsParams=None):
        """
        
        Parameters
        ----------
        All parameters are optional.
        
        vertRes: double
            vertical resolution of the line of sight in radians
            
        horRes: double
            horizontal resoliton of the line of sight in radians
            
        wavel_nm: 1d array(double)
            an array of wavelengths at which the point spread function is provided
            
        wavePSF: 1d array(double)
            an array of point spread function values. Given as the standard deviation
            of a gaussian.
            
        vertSamples: int
            How many times each line of sight should be oversampled in the 
            vertical direction.
            
        horSamples: int
            How many times each line of sight should be oversampled in the 
            horizontal direction.
            
        waveSamples: list or int
            How many times each wavelength should be oversampled. This is an array of numbers the
            same length as waveSamples_nm, or a single value. If it is a single value, all wavelengths
            will have the same number of samples.

        waveSamples_nm: list
            The wavelengths in nanometers corresponding to the waveSamples array. If waveSamples is an integer
            this variable is ignored.

        opticsParams: class
            An class which specifies the above parameters.
        """
        super().__init__()
        
        self.equal_weights = False
        self.integration_type = 'linear'
        self.wavel_num_stdev  = 3
        #can be setup using individual parameters or an opticsParameters class
        if opticsParams is None:        
            self.vertRes = vertRes          #vertical resolution of line of sight in radians
            self.horRes  = horRes           #horizontal resolution of line of sight in radians        
            self.wavelen = wavel_nm         #wavelengths at which resolution in calculated in nm
            self.wavePSF = wavePSF          #standard deviation of point spread fucntion in nm (assumed to be guassian)
                
            self.vertSamples   = vertSamples  #number of vertical samples per pixel
            self.horSamples    = horSamples   #number of horizontal samples per pixel
            self.waveSamples   = waveSamples  #maximum wavelength samples for each pixel (depending on pixel spacing the actual number may be less)
            self.waveSamples_nm = waveSamples_nm  #maximum wavelength samples for each pixel (depending on pixel spacing the actual number may be less)
        else:
            self.vertRes = opticsParams.vertRes          #vertical resolution in radians
            self.horRes  = opticsParams.horRes           #horizontal resolution in radians        
            self.wavelen = opticsParams.wavel_nm         #wavelengths at which resolution in calculated in nm
            self.wavePSF = opticsParams.wavePSF          #standard deviation of point spread fucntion in nm (assumed to be guassian)
                
            self.vertSamples = opticsParams.vertSamples  #vertical samples for each pixel
            self.horSamples  = opticsParams.horSamples   #horizontal samples for each pixel
            self.waveSamples = opticsParams.waveSamples  #maximum wavelength samples-1 for each pixel (depending on pixel spacing the actual number may be less)

    def RecommendedRTMWavelengths(self, pixelWavelengths):
        """
        Determines the wavelengths needed for modelling and their weights.        
        
        Parameters
        ---------
        pixelWavelengths: 1d array(double)
            An array containting the measured wavelengths.
        sampling: str
            Indicates the type of sampling used

            * linear - equally spaced samples
            * exponential - samples are exponentially spaced
        
        Returns
        -------
        wavelengths: 1d array(double)
            The high resolution wavelengths that should be modelled in order to
            properly compute the integrated measurement
            
        weights: 1d array(double)
            The integration weights for the hi-resolution wavelengths. Weights 
            are guaranteed to sum to one.
        """

        if np.size(pixelWavelengths) == 1:
            pixelWavelengths = [pixelWavelengths]
                
        #create a list with all the oversampled wavelengths. 
        wl      = list(pixelWavelengths)
        weights = list(pixelWavelengths)
        
        #number of standard deviations to sample from the central wavelength
        num_sigma = self.wavel_num_stdev
        
        psf = np.interp(pixelWavelengths, self.wavelen, self.wavePSF)
        for i in range(0,len(pixelWavelengths)):
            
            #determine how many samples for this pixel, make sure its an odd number
            if hasattr(self.waveSamples, '__len__'):
                waveSamples = int(np.round(np.interp( pixelWavelengths[i], self.waveSamples_nm, self.waveSamples),decimals=0))
            else:
                waveSamples = self.waveSamples
                
            #evenly space wavelengths across num-sigma range
            if (waveSamples == 1) | (psf[i] == 0):
                wl[i] = pixelWavelengths[i]
                weights[i] = 1.0
            elif self.integration_type is 'linear':
                wl[i] = np.linspace(pixelWavelengths[i]-psf[i]*num_sigma,pixelWavelengths[i]+psf[i]*num_sigma,waveSamples)    #wavelengths
                weights[i] = np.exp( -(pixelWavelengths[i]-wl[i])**2/(2*psf[i]**2)   )
                weights[i] = weights[i]/np.sum(weights[i])
            elif self.integration_type is 'trapezoidal':
                wl[i] = np.linspace(pixelWavelengths[i]-psf[i]*num_sigma,pixelWavelengths[i]+psf[i]*num_sigma,waveSamples)    #wavelengths
                gauss = np.exp( -(pixelWavelengths[i]-wl[i])**2/(2*psf[i]**2)   )
                weights[i]     = np.zeros(np.shape(wl[i])) + 2
                weights[i][0]  = 1
                weights[i][-1] = 1
                weights[i]    *= gauss
                weights[i]    /= np.sum(weights[i])
            elif self.integration_type is 'simpsons':
                if waveSamples % 2 == 0:    #must have an odd number of samples
                    waveSamples += 1
                wl[i] = np.linspace(pixelWavelengths[i]-psf[i]*num_sigma,pixelWavelengths[i]+psf[i]*num_sigma,waveSamples)    #wavelengths
                gauss = np.exp( -(pixelWavelengths[i]-wl[i])**2/(2*psf[i]**2)   )
                weights[i] = np.zeros(np.shape(wl[i]))
                weights[i][1::2] = 4
                weights[i][0::2] = 2
                weights[i][0]    = 1
                weights[i][-1]   = 1
                weights[i]      *= gauss
                weights[i]      /= np.sum(weights[i])
            elif self.integration_type is 'gauss_hermite':
                if waveSamples > 100:
                    logging.warning('hermite polynomials are untested for n>100')
                [xi, wi] = np.polynomial.hermite.hermgauss(waveSamples)
                wl[i] = xi*np.sqrt(2)*psf[i] + pixelWavelengths[i]
                weights[i] = wi/np.sum(wi)

        
        #TODO: remove unecessary wavlengths and improve spacing/weights
        
        #to use a matrix instead of lists
        #weights = np.exp( -(pixelWavelengths[:,np.newaxis]-wl)**2/(2*psf[:,np.newaxis]**2)   )
        #weights = weights/np.sum(weights,1)[:,np.newaxis]  #normalze sum of weights in each row to one
        return wl,weights
            
    def RecommendedRTMParameters(self, pixelLOS, vertical=np.array([0,0,1]) ):
        """
        Parameters
        ----------
        pixelLOS: 3 vector
            A 3 element array containing the line of sight direction
            
        vertical: 3 vector (optional)
            A 3 element array containing the local vertical direction. This direction
            is used to define the vertical and horizontal field of view.
        
        Returns
        -------
        lines of sight: 2d array(double)
            A nx3 array of dictating the high resolution lines of sight that are 
            required to model each measured line of sight.
            
        weights: 1d array(double)
            The weights of each line of sight for integration
        """
        if self.equal_weights:   #equally weighted lines of sight
            #sample center of n bins for averaging
            if self.vertSamples is 1:
                thetav = np.array([0])
            else:
                thetav = np.linspace(-1*self.vertRes,  self.vertRes,  self.vertSamples+1);
                thetav = thetav[0:-1]+np.diff(thetav)/2
                
            if self.horSamples is 1:
                thetah = np.array([0])
            else:
                thetah =  np.linspace(-1*self.horRes,  self.horRes,  self.horSamples+1);
                thetah = thetah[0:-1]+np.diff(thetah)/2
            
            #equally weight all lines of sight
            weights = np.ones(self.vertSamples*self.horSamples)
            
        else:                   #gaussian weighted lines of sight
            #sample center of n bins for averaging
            if self.vertSamples is 1:
                thetav = np.array([0])
            else:
                thetav = np.linspace(-2*self.vertRes,  2*self.vertRes,  self.vertSamples);
                
            if self.horSamples is 1:
                thetah = np.array([0])
            else:
                thetah =  np.linspace(-1*self.horRes,  self.horRes,  self.horSamples+1);
                thetah = thetah[0:-1]+np.diff(thetah)/2
            
            #gaussian weight lines of sight  
            ctr = 0
            weights =np.zeros(self.vertSamples*self.horSamples)
            for i in range(0,len(thetav)):
                for j in range(0,len(thetah)):
                    if self.vertRes == 0:
                        weights[ctr] = 1
                    else:
                        weights[ctr] = np.exp(-(thetav[i]**2)/(2*self.vertRes**2))
                    ctr+=1
        
        weights = weights/np.sum(weights)
        
        k = np.cross( pixelLOS, vertical )  #vertical rotation axis
        k = k/np.sqrt(np.sum(k**2))         #make unit vector
        v = pixelLOS
        p = vertical                        #horizontal rotation axis
            
        #raster scan the sampled area with verSamples*horSamples
        los = np.zeros((self.vertSamples*self.horSamples,3))
        ctr = 0
        for i in range(0,len(thetav)):
            #rotate up/down
            u = v*np.cos(thetav[i])+np.cross(k,v)*np.sin(thetav[i])+k*np.dot(k,v)*(1-np.cos(thetav[i]))
            for j in range(0,len(thetah)):
                #rotate left/right
                los[ctr,:] = u*np.cos(thetah[j]) +np.cross(p,u)*np.sin(thetah[j]) +p*np.dot(p,u)*(1-np.cos(thetah[j]));
                ctr+=1
                
        return los, weights
        
        
class BasicOpticsParams(object):
    """
    Class designed to hold optics parameters for an instrument.
    """
    def __init__(self,vertRes=3.5e-4,horRes=1e-3,wavel_nm=np.array([0, 10000]),wavePSF = np.array([0.35, 35]),vertSamples=3,horSamples=3,waveSamples=11):
        
            self.vertRes = vertRes          #vertical resolution in radians
            self.horRes  = horRes           #horizontal resolution in radians        
            self.wavelen = wavel_nm         #wavelengths at which resolution in calculated in nm
            self.wavePSF = wavePSF          #standard deviation of point spread fucntion in nm (assumed to be guassian)
                
            self.vertSamples = vertSamples  #3 vertical samples for each pixel
            self.horSamples  = horSamples   #3 horizontal samples for each pixel
            self.waveSamples = waveSamples  #maximum wavelength samples-1 for each pixel (depending on pixel spacing the actual number may be less)
        
    def setOpticsAsIdeal(self):
        
            self.vertRes = 0          #vertical resolution in radians
            self.horRes  = 0           #horizontal resolution in radians        
            self.wavelen = np.array([0,np.Inf])     #wavelengths at which resolution in calculated in nm
            self.wavePSF = np.array([0,0])          #standard deviation of point spread fucntion in nm (assumed to be guassian)
                
            self.vertSamples = 1  #3 vertical samples for each pixel
            self.horSamples  = 1   #3 horizontal samples for each pixel
            self.waveSamples = 1  #maximum wavelength samples-1 for each pixel (depending on pixel spacing the actual number may be less)
        