import numpy as np

class Sensor(object):

    def __init__(self):
        pass
    
    def LinesOfSightFromOpticalAxis(self, lookdir, vertical ):
        raise NotImplementedError()
        
class PointSensor(Sensor):
    
    def __init__(self):
        super().__init__()
        
        self.vertPixels = 1
        self.horPixels  = 1
        self.vertFOV    = 0
        self.horFov     = 0
        
    def LinesOfSightFromOpticalAxis(self, lookdir, vertical=np.array([0,0,1])):
        return lookdir
        
class NxMSensor(Sensor):
    """
    A rectangular imaging sensor. Given an optical axis this class tells user the line of 
    sight for each pixel in the sensor.    
    """
    def __init__(self, numVertPixels=10, numHorPixels=10, vertFOV_deg=1, horFOV_deg=1):
        """
        
        Parameters
        ----------
        numVertPixels: int
            The number of pixels in the vertical direction.
            
        numHorPixels: int
            The number of pixels in the horizontal direction.
            
        vertFOV_deg: double
            The vertical field of view of the sensor in degrees.
            
        horFOV_deg: double
            The horizontal field of view of the sensor in degrees
        """        
        
        super().__init__()
        
        self.vertPixels = numVertPixels
        self.horPixels  = numHorPixels
        self.vertFOV    = vertFOV_deg*np.pi/180
        self.horFOV     = horFOV_deg*np.pi/180
        self.num_pixels = numVertPixels*numHorPixels
        
    def LinesOfSightFromOpticalAxis(self, lookdir, vertical=np.array([0,0,1])):
        
         """
         Parameters
         ----------
         lookdir: 3 vector
            A three element array containing the line of sight direction of the
            optical axis. 
            
         vertical: 3 vector
            A three element array containing the vertical orientation of the sensor.
            Vertical columns of pixels are aligned with this dimension, horizontal
            rows of pixels are perpendicular to it.
            
            
         Returns
         -------
         lines of sight: 2d array(doubles)
             An num pixels x 3 array containing the look direction of each pixel.
             Row major ordering.
         """
        
         #determine vertical field of view
         if self.vertPixels>1:
              thetav = np.linspace(-1*self.vertFOV/2, self.vertFOV/2, self.vertPixels );
         else:
              thetav = np.array([0]);
                
         #determine horizontal field of view
         if self.horFOV is not 0:
              thetah = np.linspace(-1*self.horFOV/2, self.horFOV/2, self.horPixels );
         else:
              thetah = np.array([0]);
                
         #find rotation axis
         k = np.cross( lookdir, vertical );
         k /= np.linalg.norm(k)
         los = np.zeros((self.vertPixels*self.horPixels,3));
         los_idx = 0;
                
         #raster scan lines of sight, creating one for each pixel
         for i in range(0,self.vertPixels):
             u = lookdir*np.cos(thetav[i])+np.cross(k,lookdir)*np.sin(thetav[i])+k*np.dot(k,lookdir)*(1-np.cos(thetav[i]))
             for j in range(0,self.horPixels):
                 los[los_idx,:] = u*np.cos(thetah[j])+np.cross(vertical,u)*np.sin(thetah[j])+vertical*np.dot(vertical,u)*(1-np.cos(thetah[j]))
                 los_idx += 1

         return los
        
        
        
        