import numpy as np

from retrieval.instruments.instrument import Instrument
import radtran.sasktran as radt
import retrieval.instruments.sensors.sensor as sen
import retrieval.instruments.optics.optics as opt


#An instrument whose sensor contains one or more pixels, each with one or more wavelengths
#Takes a radiance field and produced a level one product

#TODO: Add integration over an exposure


class SpectralImager(Instrument):
    # Setting the radiance is optional, if it is not set we will assume this
    # instrument is meant for simulation or retrieval
    def __init__( self, wavel_nm, los, vertical=None, obs=None, mjd=None ):
                
        self.num_exposures  = np.int(np.size(los)/3)
            
        #put vertical and observer into correct format
        self.vertical = self.define_vertical_axis( vertical )
        self.obs      = self.define_vertical_axis( obs )
        
        #no spacial or wavelength oversampling by default
        self.vertSamples  = 1
        self.horSamples   = 1
        self.waveSamples  = 1
        
        self.vertRes = 1e-4
        self.horRes  = 1e-3
        self.psf_wavel_nm = np.array([0,np.inf])
        self.psf          = np.array([0.35, 0.35])
        
        self.sensor     = []
        self.optics     = []
        self.num_pixels = None
            
        self.wavel_nm       = wavel_nm
        self.los            = los
        self.mjd            = mjd
        self.rad            = []
        self.pixel_los      = []
        self.pixel_obs      = []
        
        self.hires_wavel    = []
        self.hires_los      = []
        self.los_weights    = []
        self.wavel_weights  = []
        
        self.radtran        = None
    
    def initialize_sensors_and_optics(self):
        
        self.sensor         = sen.NxMSensor(1,1,0,0)
        self.optics         = opt.BasicOptics(self.vertRes,self.horRes,self.psf_wavel_nm,self.psf,self.vertSamples,self.horSamples,self.waveSamples)
        self.num_pixels     = self.sensor.num_pixels
        
        self.setup_instrument()

    def set_optics_as_ideal(self):
        self.vertSamples  = 1
        self.horSamples   = 1
        self.waveSamples  = 1
        
        self.vertRes = 0
        self.horRes  = 0
        self.psf_wavel_nm = np.array([0,np.inf])
        self.psf          = np.array([0, 0])
        
        
    def convert_radiance_to_l1( self, radiance ):
        
        all_wave = self.radtran_wavelengths()
        
        #initialize measurement
        self.measurement = np.zeros((self.num_pixels,np.size(self.wavel_nm)))        
        
        #do the averageing
        for i in range(0,self.num_pixels):
            for j in range(0,len(self.wavel_nm)):
                good_wave = np.where(np.in1d(all_wave,self.hires_wavel[j]))[0]  #find the wavelengths in this pixel
                temp = np.sum( radiance[np.ix_(self.pixel_los_idx[i],good_wave)]*np.reshape(self.los_weights[i],(np.size(self.los_weights[i]),-1)),0 )/self.num_exposures
                self.measurement[i,j] = np.sum( temp*self.wavel_weights[j] )
                    
        return self.measurement
    
    def set_num_orders_of_scatter(self, num_orders):        
        self.radtran.set_num_orders_of_scatter(num_orders)     
        
    def setup_instrument( self ):
        #setup the pixel lines of sight, observer and mjd, these are the 
        #geometries values associated with each integrated radiance       
        #also setup the high resolution lines of sight for spacial oversampling       
       
        #self.pixel_exp_idx = [0,]*self.num_sensors
        self.pixel_los   = np.zeros((self.num_pixels,3))
        self.pixel_obs   = np.zeros((self.num_pixels,3))
        self.pixel_mjd   = np.zeros((self.num_pixels,1))
        self.hires_los   = [0,]*self.num_pixels  
        self.los_weights = [0,]*self.num_pixels

        #Setup the Lines of sight for each pixel and sensor
        for j in range(0,self.num_exposures): #loop over exposures
            pixel_los_expj = self.sensor.LinesOfSightFromOpticalAxis(np.reshape(self.los,(-1,3))[j,:], np.reshape(self.vertical,(-1,3))[j,:])
            self.pixel_los += pixel_los_expj/self.num_exposures #find the average los for that pixel (assumes equally weighted exposures)
            self.pixel_obs += np.tile(self.obs[j,:],(np.size(pixel_los_expj,0),1))/self.num_exposures
            self.pixel_mjd += np.tile(self.mjd[j],  (np.size(pixel_los_expj,0),1))/self.num_exposures
                #Setup the hi resolution lines of sight and weights
            for k in range(0,self.num_pixels):  #loop over pixels
                if j is 0:  #initialize lines of sights and weights
                    self.hires_los[k] = np.zeros((0,3))
                    self.los_weights[k] = np.zeros((0))
                        
                temp_los, temp_w = self.optics.RecommendedRTMParameters(np.reshape(pixel_los_expj,(-1,3))[k,:],np.reshape(self.vertical,(-1,3))[j,:])
                self.hires_los[k]   = np.vstack((self.hires_los[k],temp_los))
                self.los_weights[k] = np.concatenate((self.los_weights[k],temp_w))
               
        self.set_wavelengths(self.wavel_nm)
        
    def get_meas_geometry(self):
        
        return self.pixel_los, self.pixel_obs
    
    def set_wavelengths( self, wavel_nm ):
        self.hires_wavel,self.wavel_weights = self.optics.RecommendedRTMWavelengths(wavel_nm)
    
    def radtran_wavelengths(self):

        wl = np.hstack(self.hires_wavel)
        all_wl = np.unique(np.hstack(wl))
        return all_wl
        
    def radtran_geometry(self):
        
        l = []
        hires_obs = []
        self.sensor_los_idx = [] #stores what lines of sight belong to what sensor
        self.pixel_los_idx  = [] #stores what lines of sight belong to what pixel
        start = 0
        sp = 0

        l = np.zeros((0,3))
        self.pixel_los_idx = [0,]*self.num_pixels
        for j in range(0,self.num_pixels):     #loop over pixels
            ep = sp+np.size(self.hires_los[j],0)
            l = np.vstack((l,self.hires_los[j]))
            self.pixel_los_idx[j] = np.arange(sp,ep)
            sp = ep
                
        hires_obs = np.tile(self.obs,(np.size(l,0),1))
        end = start+np.size(l,0)
        self.sensor_los_idx = np.arange(start,end)
        start = end
            
        all_los = np.vstack(l)
        all_obs = np.vstack(hires_obs)
       # wl = self.radtran_wavelengths()
        return all_los, all_obs

    def define_vertical_axis( self, vertical ):
        
        if vertical is None:
            vertical = np.zeros((self.num_exposures,3))
            for j in range(0,self.num_exposures):
                vertical[j,:] = np.array([0,0,1])
    
        return vertical
        
    def define_mjd( self, mjd ):
        if mjd is None:
            mjd = np.zeros((self.num_exposures,0))
            for j in range(0,self.num_exposures):
                mjd[j] = np.array([0])
    
        return mjd
#TODO: Add integration over an exposure
class MultiSensorImager(Instrument):
    """
        Class that holds multiple independent sensors. Useful for grouping multiple
        sensors into a single instrument, or multiple independent exposures into a single
        measurement
        
        Methods
        -------
 
        initialize_sensors_and_optics :
            Setup the sensors and optics classes. This shoul be called after
            setup but before using the class to calculate radiances or convert 
            level 1 data.
        
        convert_radiance_to_l1 :
            Takes in radiances calculated using an external radiadive transfer
            model and integrates over the wavelengths, lines of sight and exposures
            to produce a level one product.
        
        calculate_radiance :
            A function to calculate level one data using an internal radiative 
            transfer model passed to instrument. This function is provided to
            simplify retrieval code.
            
        setup_radtran_model :
            Sets up the internal radiative transfer model with lines of sight 
            specified by the instrument.
            
        get_meas_geometry :
            Returns the obs, los and mjd of each level one measurement produced
            by the instrument. This can be useful when using a high-res instrument
            to setup a low-res instrument in a retrieval.
            
        set_sensor_wavelengths :
            Set/change the wavelengths of one of the sensors. This is meant
            primarily as an internal function.
            
        radtran_wavelengths :
            Returns the oversampled wavelengths recommended for use in a radiative
            transfer calculation. The convert_radiance_to_l1 function expects
            radiances at these wavelengths.
            
        radtran_geometry :
            Returns the oversampled wavelengths recommended for use in a radiative
            transfer calculation. The convert_radiance_to_l1 function expects
            radiances at these geometries.


        Parameters
        ----------
        wavel_nm :
            a list of arrays, one for each sensors
        los :
            a list of nx3 arrays. Each list element contains the optical axis lines
            of sight for a sensor with n subexposures. Specificies in geodetic x,y,z
            coordinate.
        vertical :
            a list in the same format and size as los describing the vertical
            orientation of the sensor. If the sensor has more than one pixel,
            or the pixel line of sight is oversamples this should be set.
        obs :
            a list in the same format and size as los describing the observer position.
            This should be set if the instrument will be used to setup the radiative
            transfer model.
        mjd :
            a list in the same format and size as los describing the modified
            julian date of each line of sight. This should be set if the
            instrument will be used to setup the radiative transfer model.
        
        Examples
        --------
        
            Ex. 1 Create an instrument with the following properties
                1. 2 sensors
                2. Each sensor has one line sight
                3. Each line of sight has 2 subexposures
                4. Each sensor measures 3 wavelengths
                5. Wavelengths and line of sight are not oversampled
            
            >>> from retrieval import instruments as instrument
            >>> import numpy as np
            >>> wavel = [np.array([300,400,500]),     np.array([450, 600, 800])]            
            >>> los   = [np.array([[0,1,0],[0,1,0]]), np.array([[1,0,0],[1,0,0]])]
            >>> inst = instrument.MultiSensorImager(wavel, los)
            >>> inst.initialize_sensors_and_optics()
            
            Ex. 2 Create the same sensor, but oversample each wavelength 5 times
            
            >>> from retrieval import instruments as instrument
            >>> import numpy as np
            >>> wavel = [np.array([300,400,500]),     np.array([450, 600, 800])]            
            >>> los   = [np.array([[0,1,0],[0,1,0]]), np.array([[1,0,0],[1,0,0]])]
            >>> inst = instrument.MultiSensorImager(wavel, los)
            >>> inst.waveSamples = 5
            >>> inst.psf_wavel_nm = np.array([100 200 1000])
            >>> inst.psf          = np.array([0.2 0.3 1.0])
            >>> inst.initialize_sensors_and_optics()
    """
    def __init__( self, wavel_nm, los, vertical=None, obs=None, mjd=None):
        
        """
        """
        
        self.num_sensors = len(los)  #number of senors must be equal to the number of lines of sight provided 
        self.num_exposures = [0,]*self.num_sensors
        for i in range(0,self.num_sensors):
            self.num_exposures[i]  = np.int(np.size(los[i])/3)
            
        #put vertical and observer into correct format
        self.vertical = self.define_vertical_axis( vertical )
        self.obs      = self.define_vertical_axis( obs )
        
        self.mjd      = self.define_mjd( mjd )
        
        #Each sensor contains one line of sight by default
        self.vertPixels   = 1   #number of vertical pixels
        self.horPixels    = 1   #number of horizontal pixels
        self.vertFOV      = 0   #vertical field of view in radians
        self.horFOV       = 0   #vertical field of view in radians

        #no spacial or wavelength oversampling of the pixels by default
        self.vertSamples     = 1               #number of times each pixel is sampled vertically
        self.horSamples      = 1               #number of times each pixel is sampled horizontally
        self.waveSamples     = [1,1]           #number of wavelengths used for each pixel as a function of wavelength
        self.waveSamples_nm  = [0,np.Inf]      #wavelength corresponding to oversampling parameter self.waveSamples
        self.vertRes         = 1e-4            #vertical sampling resolution in radians
        self.horRes          = 1e-4            #horizontal sampling resolution in radians
        self.wavel_int_type  = 'linear'        #wavelength integration type
        self.wavel_num_stdev = 3               #number of standard deviations used for wavelenth integration
        
        # wavelengths at which point spread function is known
        self.psf_wavel_nm = np.array([0,np.inf])  
        #point spread function at psf_wavele_nm. The standard deviation of a guassian distribution in nanometers
        self.psf          = np.array([0.35, 0.35])
        
        #variable initialization
        self.sensor     = [0,]*self.num_sensors
        self.optics     = [0,]*self.num_sensors
        self.num_pixels = [0,]*self.num_sensors
        
        self.wavel_nm       = wavel_nm
        self.los            = los
        self.rad            = []
        self.wf             = []
        self.pixel_los      = [0,]*self.num_sensors
        self.pixel_obs      = [0,]*self.num_sensors
        self.pixel_mjd      = [0,]*self.num_sensors
        
        self.hires_wavel    = [0,]*self.num_sensors
        self.hires_los      = [0,]*self.num_sensors
        self.los_weights    = [0,]*self.num_sensors
        self.wavel_weights  = [0,]*self.num_sensors

        self.radtran        = None
    def initialize_sensors_and_optics(self):
    
        for i in range(0,self.num_sensors):
            self.sensor[i]         = sen.NxMSensor(self.vertPixels,self.horPixels,self.vertFOV,self.horFOV)
            self.optics[i]         = opt.BasicOptics(self.vertRes,self.horRes,self.psf_wavel_nm,self.psf,self.vertSamples,self.horSamples,self.waveSamples,self.waveSamples_nm)
            self.num_pixels[i]     = self.sensor[i].num_pixels

            self.optics[i].integration_type = self.wavel_int_type
            
        self.setup_instrument()

    # @profile
    def convert_radiance_to_l1( self, radiance ):
        """        
        Parameters
        ----------
        radiance : np.array
            an nxm array where n is the number of lines of sight and m is the 
            number of wavelenghts.
            
        Returns
        -------
        measurement : list[np.arrays]
            a list of 2 dimensional arrays, one for each sensor. 2 dimensional
            arrays have dimensions of lines of sight x wavelengths.
        """


        #get the wavelengths used by the radtran engine
        all_wave = self.radtran_wavelengths()

        #initialize measurement
        self.measurement =[0,]*self.num_sensors
        for s in range(0,self.num_sensors):
            self.measurement[s] = np.zeros((self.num_pixels[s],np.size(self.wavel_nm[s])))

        for s in range(0,self.num_sensors):
            for i in range(0,self.num_pixels[s]):
                for j in range(0,len(self.wavel_nm[s])):
                    good_wave = np.where(np.in1d(all_wave,self.hires_wavel[s][j]))[0]  #find the wavelengths in this pixel
                    temp = np.sum( radiance[np.ix_(self.pixel_los_idx[s][i],good_wave)]*np.reshape(self.los_weights[s][i],(np.size(self.los_weights[s][i]),-1)),0 )/self.num_exposures[s]
                    self.measurement[s][i,j] = np.sum( temp*self.wavel_weights[s][j] )

        return self.measurement

    # @profile
    def calculate_radiance( self ):
        """        
        Calculates the instrument measurement and weighting function.
        """
        all_wave = self.radtran_wavelengths()
        self.radtran.set_wavelengths( all_wave )
        
        rad, wf  = self.radtran.calculate_radiance()
        self.rad = self.convert_radiance_to_l1( rad )

        if np.size(wf)!=0:
            self.wf = np.zeros( (np.size(self.wavel_nm[0]), len(self.los), np.shape(wf)[2]) )
            temp    = np.zeros( (np.size(self.wavel_nm[0]), len(self.los), np.shape(wf)[2]) )
            #integrate analytic weighting functions only if required (its slow compared to a
            #single scatter calculation)
            for idx in range(np.shape(wf)[2]):
                self.wf[:,:,idx] = np.transpose( self.convert_radiance_to_l1( np.transpose( wf[:,:,idx] ) ) )

    def setup_radtran_model( self, radtran ):
        """        
        Sets up the radiative transfer model lines of sight and wavelengths
        
        Parameters
        ----------
        radtran : pointer to a radtran class
            See :mod:`~retrieval.radtran` for details
        """
        
        self.radtran = radtran
        los,obs,mjd = self.radtran_geometry()
        for l,o,m in zip(los,obs,mjd):
            self.radtran.add_line_of_sight( o, l, m[0])
        
        radtran.set_wavelengths( self.radtran_wavelengths() )
        
    
    def setup_instrument( self ):
        """        
        Initializes the instrument geometry including spacial oversampling, 
        this is typically called from  :func:`~instruments.spectral_imager.MultiSensorImager.initialize_sensors_and_optics`
        
        Creates the high resolution lines of sight and weighting factors for each pixel using the instrument optics
        """

        #self.pixel_exp_idx = [0,]*self.num_sensors
        for i in range(0,self.num_sensors):              #compute lines of sight for sensor
            self.pixel_los[i]   = np.zeros((self.num_pixels[i],3))
            self.pixel_obs[i]   = np.zeros((self.num_pixels[i],3))
            self.pixel_mjd[i]   = np.zeros((self.num_pixels[i],1))
            self.hires_los[i]   = [0,]*self.num_pixels[i]   
            self.los_weights[i] = [0,]*self.num_pixels[i]

            #Setup the Lines of sight for each pixel and sensor
            for j in range(0,self.num_exposures[i]): #loop over exposures
                pixel_los_expj = self.sensor[i].LinesOfSightFromOpticalAxis(np.reshape(self.los[i],(-1,3))[j,:], np.reshape(self.vertical[i],(-1,3))[j,:])
                self.pixel_los[i] += pixel_los_expj/self.num_exposures[i] #find the average los for that pixel (assumes equally weighted exposures)
                self.pixel_obs[i] += np.tile(np.reshape(self.obs[i],(-1,3))[j,:],(np.size(pixel_los_expj,0),1))/self.num_exposures[i]
                self.pixel_mjd[i] += np.tile(self.mjd[i][j],(np.size(pixel_los_expj,0),1))/self.num_exposures[i]
                #Setup the hi resolution lines of sight and weights
                for k in range(0,self.num_pixels[i]):  #loop over pixels
                    if j is 0:  #initialize lines of sights and weights
                        self.hires_los[i][k] = np.zeros((0,3))
                        self.los_weights[i][k] = np.zeros((0))
                        
                    temp_los, temp_w = self.optics[i].RecommendedRTMParameters(np.reshape(pixel_los_expj,(-1,3))[k,:],np.reshape(self.vertical[i],(-1,3))[j,:])
                    self.hires_los[i][k]   = np.vstack((self.hires_los[i][k],temp_los))
                    self.los_weights[i][k] = np.concatenate((self.los_weights[i][k],temp_w))
               
            self.set_sensor_wavelengths(self.wavel_nm[i], i)
    
    def get_meas_geometry(self):
        """        
        Provides the lines of sight and observer positions of the instrument measurements
        
        Returns
        ----------
        pixel_los : np.array
            list of pixel lines of sight in geographic coordinates
        pixel_obs : np.array
            list of pixel locations in geographic coordinates
        """
        return self.pixel_los, self.pixel_obs
        
    def set_sensor_wavelengths( self, wavel_nm, sensor_number ):
        """        
        Sets the low resolution wavelengths of a sensor. Typically this is used
        only internally.
        
        Params
        ----------
        wavel_nm : np.array
            an array of wavelengths for the center of each pixel
        sensor_number : int
            the internal index used for the sensor
        """
        
        self.hires_wavel[sensor_number],self.wavel_weights[sensor_number] = self.optics[sensor_number].RecommendedRTMWavelengths(wavel_nm)

    # @profile
    def radtran_wavelengths(self):
        """        
        Provides the high resolution wavelengths that the radiative transfer 
        calculation should be done at.
        
        Returns
        ----------
        wavelength : np.array
            an array of high resolution wavelengths for the radiative transfer
            transfer calculation
        """
        
        wl = [0,]*self.num_sensors
        for i in range(0,self.num_sensors):
            wl[i] = np.hstack(self.hires_wavel[i])

        all_wl = np.unique(wl)

        # all_wl = np.unique(np.hstack(self.hires_wavel))
        return all_wl
        
    def radtran_geometry(self):
        """        
        Provides the high resolution geometries that the radiative transfer 
        calculation should be done at.
        
        Returns
        ----------
        line of sight : np.array
            an array of lines of sight (nx3)
        observer : np.array
            an array of observer positions (nx3)
        line of sight : np.array
            an array of modified julian dates, one for each los/observer
        """
        
        l = [0,]*self.num_sensors
        hires_mjd = [0,]*self.num_sensors
        hires_obs = [0,]*self.num_sensors
        self.sensor_los_idx = [0,]*self.num_sensors #stores what lines of sight belong to what sensor
        self.pixel_los_idx  = [0,]*self.num_sensors #stores what lines of sight belong to what pixel
        start = 0
        sp = 0
        for i in range(0,self.num_sensors):                     #loop over sensors
            l[i] = np.zeros((0,3))
            self.pixel_los_idx[i] = [0,]*self.num_pixels[i]
            for j in range(0,self.num_pixels[i]):     #loop over pixels
                ep = sp+np.size(self.hires_los[i][j],0)
                l[i] = np.vstack((l[i],self.hires_los[i][j]))
                self.pixel_los_idx[i][j] = np.arange(sp,ep)
                sp = ep
                
            hires_obs[i] = np.tile(self.obs[i],(np.size(l[i],0),1))
            hires_mjd[i] = np.tile(self.mjd[i],(np.size(l[i],0),1))
            end = start+np.size(l[i],0)
            self.sensor_los_idx[i] = np.arange(start,end)
            start = end
        
        all_mjd = np.vstack(hires_mjd)
        all_los = np.vstack(l)
        all_obs = np.vstack(hires_obs)
       # wl = self.radtran_wavelengths()
        return all_los, all_obs, all_mjd
        
        
    def define_vertical_axis( self, vertical ):
        """
        Arranges the vertical orientation into the proper format if none is given
        """
        if vertical is None:
            vertical = [0,]*self.num_sensors
            for i in range(0,self.num_sensors):
                vertical[i] = np.zeros((self.num_exposures[i],3))
                for j in range(0,self.num_exposures[i]):
                    vertical[i][j,:] = np.array([0,0,1])
    
        return vertical
        
    def define_mjd( self, mjd):
        """
        Arranges the mjd into the proper format if none is given
        """
        if mjd is None:
            mjd = [0,]*self.num_sensors
            for i in range(0,self.num_sensors):
                mjd[i] = np.zeros((self.num_exposures[i],1))
                for j in range(0,self.num_exposures[i]):
                    mjd[i][j] = np.array([0])
                    
        return mjd
        
    def set_num_orders_of_scatter(self, num_orders):
        
        self.radtran.set_num_orders_of_scatter(num_orders)     
        
        