import numpy as np
import copy as copy
import scipy as sp
import logging


class LeastSquares(object):
    def needs_jacobian(self):
        raise NotImplementedError()
        
    def retrieve(self, y_meas, y_meas_error, ret_inst, species):
        raise NotImplementedError()

        
class LevenbergMarquardt(LeastSquares):
    def __init__(self, damping=0.1, maxiter=5, tolerence=1e-6, dampingpower=1.5, save_output=False, weight_error=False,
                 user_defined_weight=None, test_convergence=False):
        self.type = 'LM'

        self.damping = damping
        self.maxiter = maxiter
        self.tolerence = tolerence
        self.dampingpower = dampingpower
        
        self.save_output = save_output
        self.weight_error = weight_error
        self.user_defined_weight = False
        
        # user defined weights should be an array the same length measurement vector
        if user_defined_weight is not None:
            self.user_defined_weight = True
            self.W = user_defined_weight
        
        # Tikhonov regularization matrix
        self.gamma = np.zeros(1)
        
        if self.save_output:      
            self.output = self.init_output()

        self.test_convergence = test_convergence

    @staticmethod
    def init_output():
        output = dict()
        output['y_meas']       = []
        output['y_meas_error'] = []
        output['y_ret']        = []
        output['residual']     = []
        output['profile']      = []
        output['apriori']      = []
        output['jacobian']     = []
        output['W']            = []
        output['profile_alts'] = []
        output['tanalts']      = []
        output['rad_ret']      = []

        return output

    @staticmethod
    def allowed_to_step_back(current_residual, previous_residual, current_iter):
        """
        Added this virtual method to allow LM to step back if we are over 10 iterations. The default implementation
        in this class is replicates the original behaviour so we dont break existing code. A derived classes can override this method and provide their
        own behaviour. The defualt behaviour is a little quirky and should be overridden by other users.

        Parameters
        ----------
        current_residual :  The current residual
        previous_residual : the previous residual
        current_iter      : the current iteration

        Returns
        -------
        True if the LM algorithm is allowed to step back at this time.

        """
        return current_iter > 10

    def needs_jacobian(self):
        return True

    def retrieve(self, y_meas, y_meas_error, ret_inst, species):
        if self.save_output:      
            self.output = self.init_output()
            
        current_iter = 1
        
        previous_residual = 1e10
        current_residual  = 1000
        previous_x = None
        previous_y_ret = None
        J = None
        jjt = None
        y_difference_old = None
        converged = False
        
        good = ~np.isnan(y_meas)
            
        if self.weight_error:
            W = np.diag(1/(y_meas_error[good]**2))
        elif self.user_defined_weight:
            if len(np.shape(self.W)) == 1:
                W = np.diag(self.W[good])
            else:
                W = sp.sparse.diags(self.W.data[0][good], 0)
        else:
            W = np.eye(len(y_meas[good]))
            
        if self.save_output:
            self.output['y_meas'] = y_meas
            self.output['y_meas_error'] = y_meas_error
            self.output['apriori'] = copy.copy(species.get_atmospheric_profile())
            self.output['W'] = W
            self.output['tanalts'] = np.asarray(species.tanalts)
            # TODO: figure out a better way to get the y coordinates
            self.output['profile_alts'] = np.arange(0.5, 100.5)

        while (current_iter <= self.maxiter) & (not converged):
            logging.info(' LM ' + species.ident + ' iteration '+str(current_iter)+' of ' + str(self.maxiter))
            # First calculate f( species.get_x() )
            ret_inst.calculate_radiance()
            y_ret, y_ret_error = species.measurement_vector( ret_inst )
            y_difference = y_meas[good] - y_ret[good]
            # Calculate the residual
            # This is the residual of y - f( species.get_x() )
            current_residual = np.sum(W.dot(y_difference)**2)/np.sum(W.dot(y_meas[good])**2)
            # Previous residual holds the residual of y - f( previous_x )
            if current_residual > previous_residual and self.allowed_to_step_back(current_residual, previous_residual, current_iter):
                # Our old solution was better, go back to it and increase the
                # damping, we still have J and jjt stored from the old profile
                
                # only do this if this isnt our last iteration
                if current_iter != self.maxiter:
                    species.set_atmospheric_profile(previous_x)
                    self.damping *= self.dampingpower
                    delta = np.linalg.solve(jjt + self.damping * np.diag(np.diag(jjt)) + self.gamma.transpose().dot(self.gamma), J.transpose().dot(W.dot(y_difference_old)))
                    delta = np.real(delta)
                    species.update_profile(delta)
                    logging.info('\tSolution getting worse, setting damping to ' + str(self.damping))

            else:
                # Our solution got better, so we can keep going
                # Store the old profile
                previous_x = species.get_atmospheric_profile()
                
                #compute the jacobian
                J     = species.measurement_vector_jacobian( ret_inst )
                
                #remove any nans from measurements
                J = J[good, :]
                
                # And calculate how much to update our new profile by
                jjt = J.transpose().dot(W.dot(J))
                if sp.sparse.issparse(jjt):
                    jjt = jjt.toarray()
                if jjt.shape[0] != jjt.shape[1]:
                    logging.error("jjt variable from Jacobian matrix is not square. Expect an error...")

                delta = np.linalg.solve( jjt + self.damping * np.diag( np.diag(jjt) ) + self.gamma.transpose().dot( self.gamma ), J.transpose().dot( W.dot(y_difference)))
                delta = np.real( delta )
                species.update_profile( delta )
                previous_residual = current_residual

                # store our y_difference in another variable, incase this
                # solution is worse and we need to go back to it
                y_difference_old = y_difference

                if self.save_output:
                    self.output['y_ret'].append(y_ret)
                    self.output['profile'].append(copy.copy(species.get_atmospheric_profile()))
                    self.output['residual'].append(current_residual)
                    self.output['jacobian'].append(J)
                    self.output['rad_ret'].append(ret_inst.rad)

                if self.test_convergence:
                    converged = species.check_convergence(y_ret, previous_y_ret)
                previous_y_ret = y_ret

            logging.info('\tResidual: ' + str(current_residual))
   
            current_iter += 1


class MART(LeastSquares):
    """
    Implementation of the MART technique.
    Profile is updated by x_{i+1} = x_i + y_meas / y_ret * x_i
    """
    def __init__(self, numiter=5, W = None, save_output=False):
        self.maxiter = numiter
        self.W = W
        self.type = 'MART'
        self.save_output = save_output

        self.damping  = None

        if save_output:
            self.output = self.init_output()

    def init_output(self):

        output = dict()
        output['y_meas']       = []
        output['y_meas_error'] = []
        output['y_ret']        = []
        output['residual']     = []
        output['profile']      = []
        output['apriori']      = []
        output['jacobian']     = []
        output['W']            = []
        output['profile_alts'] = []
        output['tanalts']      = []
        output['rad_ret']      = []

        return output

    def needs_jacobian(self):
        return False

    def retrieve( self, y_meas, y_meas_error, ret_inst, species ):
        # we will be dividing by zero when we re-normalize the weighting matrix

        if self.save_output:
            self.output = self.init_output()

        current_iter = 1
        good = ~np.isnan(y_meas)

        if self.save_output:
            self.output['y_meas'] = y_meas
            self.output['y_meas_error'] = y_meas_error
            self.output['apriori'] = copy.copy(species.get_atmospheric_profile())
            self.output['W'] = self.W
            self.output['tanalts'] = np.asarray(species.tanalts)
            # TODO: figure out a better way to get the y coordinates
            self.output['profile_alts'] = np.arange(0.5, 100.5)

        while current_iter <= self.maxiter:

            logging.info(' MART ' + species.ident + ' iteration '+str(current_iter)+' of '+str(self.maxiter))
            ret_inst.calculate_radiance()
            
            y_ret,y_ret_error = species.measurement_vector( ret_inst )

            if self.W is None:
                alpha = y_meas / y_ret
                current_residual = y_meas-y_ret
            else:
                # get rid of bad measurements and rescale weights
                self.W[:,~good] = 0.0

                np.seterr(divide='ignore',invalid='ignore') # nans dealt with right below
                self.W /= np.sum(self.W,axis=1)[:,np.newaxis]
                np.seterr(divide=None,invalid=None)

                self.W[np.isnan(self.W)] = 0.0

                # set nans to zero so we can do the calculation
                ratio = y_meas / y_ret
                ratio[np.isnan(ratio)] = 0.0
                # compute the update factor
                alpha = np.dot( self.W, ratio )
                alpha[alpha==0] = np.nan
                # alpha = alpha[good]
                dy = (y_meas - y_ret) # Store measurement vector difference to compute residual without NaNs
                dy[np.isnan(dy)] = 0
                current_residual = np.sum(self.W.dot(dy)**2)

            if self.save_output:
                self.output['y_ret'].append(y_ret)
                self.output['profile'].append(copy.copy(species.get_atmospheric_profile()))
                self.output['residual'].append(current_residual)
                self.output['rad_ret'].append(ret_inst.rad)

            logging.info('\tResidual: ' + str(np.linalg.norm(current_residual) / len(current_residual)))

            species.update_profile(alpha, mode='multiplicative')
            current_iter += 1
