import logging as logging
from retrieval.method.lsq import LeastSquares
from scipy import sparse
from copy import copy
import numpy as np


class Rodgers(LeastSquares):
    """
    Method based on the optimal estimation method outlined in 'Inverse Methods for Atmospheric Sounding'.  The core
    of the retrieval is the iterative step

    .. math::

        x_{i+1} = x_i + (S_a^{-1} + R^T R + K^T S_{\epsilon} K + \lambda diag(K^T S_{\epsilon} K) )^{-1} [K^T S_{\epsilon}^{-1} (y - F(x_i)) - (S_a^{-1} + R^T R)(x - x_a)]

    Where :math:`y` is the measurement vector of length :math:`m`, :math:`x` is the state vector of length :math:`n`,
    :math:`S_a` is the :math:`n \\times n` apriori covariance, :math:`R` is a regularization matrix with leading dimension
    :math:`n`, :math:`\\lambda` is a LM style damping factor, :math:`K=dy/dx` is the Jacobian matrix of size :math:`n \\times m`, and :math:`S_\epsilon` is the
    measurement covariance.  Every iteration we calculate the statistic,

    .. math::

        \chi^2 = (y - f(x_i))^T S_\epsilon^{-1} (y - f(x_i))

    which at the postulated minimum tends to :math:`m`.  Therefore the quantity :math:`\chi^2 / m` is recorded,
    if the retrieval is correct this quantity should tend to unity.

    In this formalism we can also calculate the solution covariance,

    .. math::

        S = (S_a^{-1} + K^T S_\epsilon K)^{-1}

    and the averaging kernel,

    .. math::

        A = S K^T S_\epsilon^{-1} K.
    """
    def __init__(self, do_sparse_linalg=False, damping=0, calc_averaging_kernel=False, stop_on_convergence=False):
        self.maxiter = 5

        # TODO: Change this.  I think the OMPS retrieval needs this somewhere
        self.type = 'LM'

        # TODO: In the future this flag will be used to switch between sparse and not sparse linear algebra
        self.do_sparse_linalg = do_sparse_linalg
        self.calc_averaging_kernel = calc_averaging_kernel

        # Tikhonov matrix (R)
        self.gamma = None

        # Solution Covariance
        self.S = None

        # Averaging Kernel
        self.A = None

        # Weighting matrix, instead of minimizing |y - f(x)| we minimize W|y-f(x)|
        self.W = None
        self.user_defined_weight = False

        # LM Style damping weight
        self.damping = damping

        # Convergence statistic
        self.chi_squared = None

        # Apriori
        self.apriori_set = False
        self.x_a = None
        self.apriori_covariance = None
        self.output = dict()

        self.stop_on_convergence = stop_on_convergence
        self.do_error_analysis = True

    def needs_jacobian(self):
        """
        Returns
        -------
        bool
            True
        """
        return True

    def set_apriori(self, x, covariance=None):
        """
        Sets the apriori and its covariance prior to starting the retrieval.

        Parameters
        ----------
        x: np.ndarray
            Apriori value for the state vector, length :math:`n`
        covariance: np.ndarray
            Diagonal elements of the covariance for x, length :math:`n`
        """
        self.x_a = x
        self.apriori_set = True
        self.apriori_covariance = covariance

    def retrieve(self, y_meas, y_meas_error, ret_inst, species):
        """
        Performs the retrieval.
        """
        # Only use measurements that are not nan
        good = ~np.isnan(y_meas)
        y_meas = y_meas[good]
        # Start by creating the constant matrices needed for the retrieval
        # Measurement covariance
        S_e = sparse.csc_matrix(sparse.diags(y_meas_error[good]**2))

        inv_S_e = sparse.linalg.inv(S_e)
        if self.user_defined_weight:
            if len(np.shape(self.W)) == 1:
                W = np.diag(self.W[good])**2
            else:
                W = sparse.diags(self.W.data[0][good]**2, 0)
            inv_S_e *= W
            # ensure inv_S_e is sparse
            inv_S_e = sparse.csc_matrix(inv_S_e)

        converged = False
        current_iter = 1

        self.output['residual'] = []
        while (current_iter <= self.maxiter) and (not converged):
            logging.info(' Rodgers ' + species.ident + ' iteration ' + str(current_iter) + ' of ' + str(self.maxiter))
            # Perform the radiative transfer calculations for the current iteration
            ret_inst.calculate_radiance()
            y_ret, y_ret_error = species.measurement_vector(ret_inst, do_error=False)
            K = species.measurement_vector_jacobian(ret_inst)

            if current_iter == 1:
                # At first iteration we need to set up the apriori information.
                inv_S_a = None
                if self.apriori_set is False:
                    logging.info('Rodgers: Apriori not set, using initial species state')
                    self.x_a = copy(species.get_x())
                    self.apriori_set = True

                if self.apriori_covariance is not None:
                    inv_S_a = sparse.csc_matrix(sparse.diags(1/self.apriori_covariance))

                if self.gamma is not None:
                    R = self.gamma
                    if inv_S_a is None:
                        inv_S_a = sparse.csc_matrix(R.T @ R)
                    else:
                        inv_S_a += sparse.csc_matrix(R.T @ R)

                if inv_S_a is None:
                    # This means that no regularization or apriori covariance was given, in this case just make the
                    # covariance really large, therefore it will not affect the retrieval
                    inv_S_a = sparse.csc_matrix(sparse.diags(np.zeros_like(self.x_a)))

            # Remove data where the measurements were nan
            K = K[good, :]
            y_ret = y_ret[good]
            y_ret_error = y_ret_error[good]

            # We write the iterative equation in the form A*(delta x) = B
            KSK = K.T @ inv_S_e @ K
            x_i = copy(species.get_x())
            if sparse.issparse(KSK):
                A = sparse.csc_matrix(inv_S_a + KSK + self.damping * np.diag(np.diag(KSK.toarray())))
            else:
                A = sparse.csc_matrix(inv_S_a + KSK + self.damping * np.diag(np.diag(KSK)))
            A_without_lm = sparse.csc_matrix(inv_S_a + KSK)
            B = sparse.csc_matrix(K.T @ inv_S_e @ (y_meas - y_ret) - inv_S_a @ (x_i - self.x_a))

            # Solve densely
            try:
                delta_x = np.linalg.solve(A.toarray(), B.toarray().flatten())
            except np.linalg.LinAlgError:
                logging.info('Matrix was close to singular, using lstsq to solve instead')
                delta_x = np.linalg.lstsq(A.toarray(), B.toarray().flatten())[0]
            try:
                delta_x_without_lm = np.linalg.solve(A_without_lm.toarray(), B.toarray().flatten())
            except np.linalg.LinAlgError:
                logging.info('Matrix was close to singular, using lstsq to solve instead')
                delta_x_without_lm = np.linalg.lstsq(A_without_lm.toarray(), B.toarray().flatten())[0]
            delta_x = np.real(delta_x)

            species.update_profile(delta_x)
            current_iter += 1

            # Convergence criteria
            d = (y_meas - y_ret).T @ inv_S_e @ (y_meas - y_ret)
            d_linear = (y_meas - y_ret - K @ delta_x_without_lm).T @ inv_S_e @ (y_meas - y_ret - K @ delta_x_without_lm)

            chi_squared = float(d / len(y_meas))
            chi_squared_linear = float(d_linear / len(y_meas))

            logging.info('\tchi squared: ' + str(chi_squared))
            logging.info('\tchi squared linear: ' + str(chi_squared_linear))

            self.output['residual'].append([chi_squared])
            if chi_squared / chi_squared_linear < 1.02:
                logging.info('Solution has Converged, stopping retrieval')
                if self.stop_on_convergence:
                    break

        self.output['profile'] = copy(species.get_atmospheric_profile())
        self.output['residual'] = np.array(self.output['residual']).flatten()

        if self.do_error_analysis:
            logging.info('Doing error analysis for ' + str(species.ident))
            # Calculate the solution covariance and averaging kernels

            try:
                S = np.linalg.inv(A_without_lm.toarray())
            except np.linalg.LinAlgError:
                logging.info("Covariance matrix appears to be singular, using pinv instead")
                S = np.linalg.pinv(A_without_lm.toarray())

            if sparse.issparse(K):
                G = S @ (K.T.dot(inv_S_e).toarray())
            else:
                G = S @ K.T @ inv_S_e.toarray()
            if self.calc_averaging_kernel:
                self.A = G @ K
                self.output['averaging_kernel'] = self.A
            else:
                self.output['averaging_kernel'] = None

            self.output['solution_covariance'] = S

            self.output['measurement_error_covariance'] = G @ (S_e.dot(G.T))

        self.chi_squared = chi_squared


