import numpy as np
import scipy as sp
from scipy import sparse

# Calculates the Tikhonov matrix for a standard first order derivative 
# operator
def first_order_deriv(x):
    gamma = np.zeros( (np.size(x)-1,np.size(x)) )
    
    for idx in range(np.size(x)-1):
        diff = x[idx+1] - x[idx]
        gamma[idx,idx] = 1/diff
        gamma[idx,idx+1] = -1/diff
    
    return gamma


# Note: probably only works for evenly spaced input
def weighted_fft(x):
    n = np.size(x)
    
    gamma = sp.fft( sp.eye(n) )
    
    # Weight the results to prefer low frequency signals
    for idx in range(n):
        gamma[:,idx] *= n
    
    return gamma


def two_dim_horizontal_second_deriv(numangle, numalt, factor=1):
    """
    Calculates the second derivatvie Tikhonov regularization matrix for a two dimensional uniform grid.
    The matrix is calculated assuming that the measurement vector is constructed with altitude being the leading
    dimension

    Parameters
    ----------
    numangle : scalar
        The number of angular grid points
    numalt : scalar
        The number of altitude grid points
    factor : scalar or vector length numangle, optional
        If scalar, the resulting matrix is multiplied by this value.  If a vector of length numangle, then each angular
        level is multiplied by the corresponding value. Default is 1
    """

    n = numalt*numangle
    gamma = sparse.lil_matrix((n, n))
    for idx in range(numalt, n-numalt):
        if np.shape(factor) != ():
            horiz = int(idx / numalt)
            mfactor = factor[horiz]
        else:
            mfactor = factor
        gamma[idx, idx-numalt] = -1/4 * mfactor
        gamma[idx, idx] = 1/2 * mfactor
        gamma[idx, idx+numalt] = -1/4 * mfactor

    # Convert gamma to CSR format for faster matrix operations
    gamma = gamma.asformat("csr")
    return gamma


def two_dim_vertical_second_deriv(numangle, numalt, factor=1):
    """
    Calculates the second derivatvie Tikhonov regularization matrix for a two dimensional uniform grid.
    The matrix is calculated assuming that the measurement vector is constructed with altitude being the leading
    dimension

    Parameters
    ----------
    numangle : scalar
        The number of angular grid points
    numalt : scalar
        The number of altitude grid points
    factor : scalar or length numalt, optional
        If scalar, the resulting matrix is multiplied by this value.  If a vector of length numalt, then each altitude
        level is multiplied by its corresponding factor.  Default is 1
    """

    n = numalt*numangle
    gamma = sparse.lil_matrix((n, n))
    for idangle in range(0, numangle):
        for idalt in range(1, numalt-1):
            if np.shape(factor) != ():
                mfactor = factor[idalt]
            else:
                mfactor = factor
            idx = idangle*numalt + idalt
            gamma[idx, idx-1] = -1/4 * mfactor
            gamma[idx, idx] = 1/2 * mfactor
            gamma[idx, idx+1] = -1/4 * mfactor

    # Convert gamma to CSR format for faster matrix operations
    gamma = gamma.asformat("csr")
    return gamma
