class Species(object):
    """
    Base class which handles calculating species specific quantities needed for the retrieval.
    """
    def configure_instrument_and_method(self, ret_inst, meas_inst, method):
        """
        Called before the retrieval starts.  This is where method specific things can be done, such as adding
        regularization to the retrieval.  The retrieval instrument also must be set up here, usually this involves
        setting the wavelengths, and if necessary the weighting function grids.
        """
        raise NotImplementedError()

    def measurement_vector(self, inst):
        """
        Calculates the measurement vector from an instrument.  We can rely on the fact that configure_instrument_and_method
        has been called before this so we have some information about the instrument.  This is only called AFTER
        calculate_radiance has been done on the instrument
        """
        raise NotImplementedError()

    def measurement_vector_jacobian(self, inst):
        """
        Calculates the jacobian of the measurment vector.  Once again we can rely on the fact that the instrument
        has been configured properly.  Usually here we use analytic weighting functions stored in the instrument.
        """
        raise NotImplementedError()

    def update_profile(self, delta, mode):
        """
        Updates the internally stored profile.  Usually called after each retrieval iteration.  It is the species
        responsibility to update the atmosphere used in the radiative transfer calculation.
        """
        raise NotImplementedError()

    def get_atmospheric_profile(self):
        """
        Gets the internally stored profile.  Usually this is done to output a result or to store the current state
        to go back to it later.
        """
        raise NotImplementedError()

    def set_atmospheric_profile(self, profile):
        """
        Sets the internally stored profile.   Most retrieval methods will use update_profile afterwards to set the
        profile, so this is usually called to retreat back to a known better state.
        """
        raise NotImplementedError()

    def get_x(self):
        """
        Gets 'x' defined by the jacobian J = dy/dx
        """
        raise NotImplementedError()

    def output_data(self):
        """
        The output of the species, usually called after the final iteration of the retrieval and then stored.
        """
        raise NotImplementedError()


