import copy as copy
import os
import logging

import numpy as np
from scipy import io
from scipy import signal

import retrieval.species as species
import radtran.geometry as geo
import radtran.climatologies.utility as clim_util

class AerosolLimbProfile(species.Species):
    """
    Handles the retrieval specifics for the aerosol species including configuration
    of retrieval method, measurement vector and jacobian calculations and updating
    of atmoshperic state with new aerosol profile.
    
    The aerosol measurement vector is a long wavelength radiance normalized by high altitude, a
    second wavelength, and a modelled measurement vector assuming zero aerosol
    loading.
    
    The high altitude normalization helps to remove dependence on characterization
    of upwelling radiation and absolute calibration
    
    The wavelength calibration decreases sensitivity to the background atmosphere
    and particle size. However it may decrease senstivity to aerosols as well
    depending on scattering geometries.
    
    The normalization by a 'clean atmosphere' measurement vector is not neccessary
    for Jacobian based methods, but helps to determine the best reference height for
    normalization.

    Parameters
    ----------
    atmosphere : :class:`~radtran.atmosphere.Atmosphere`
        Holds the atmospheric state, including climatologie, optical properties
        and species identifiers (GUIDs)
    """
    
    def __init__(self, atmosphere):
        """
        """
        self.ident = 'limb_aerosol'
        
        #TODO: setup triplets from a file
        self.triplet = []
        self.triplet.append( ([745.67], [1]))

        self.ext_wavelength = 745.67
        self.profile_alts,self.profile = atmosphere.get_user_defined( self.ident )
        
        self.normalt     = [(np.array( [25000.0,45000.0]))]
        self.ref_idx     = [None] * len(self.triplet) #np.zeros(len(self.triplet), dtype=np.int)
        self.good_alts   = None
        self.y_ray       = None
        self.wavel_nm    = None        
        self.atmosphere  = atmosphere
        self.refaltrange = 0.0
        self.max_alt     = 40000
        self.min_alt     = 5000
        self.max_val     = 1000
        self.max_ext_val = 0.01

        self.min_step_size = 0.5 #if step size is < 0.1% solution has converged
        #use weighting functions for jacobians
        self.use_wf = True
        self.jacobian_scatter_order = 20

        self.aer_clim = atmosphere.user_optprop[self.ident] #use this to get the aerosol cross section
        
    def get_required_wavelengths(self):
        """
        Returns
        -------
        wavelength (1d array):
            array of wavelengths suggested by the retrieval
        """
        #find the unique wavelengths used by the triplets
        wavel_nm = []
        for triplet in self.triplet:
            wavel_nm.append( triplet[0] )
        wavel_nm = [item for sublist in wavel_nm for item in sublist]
        wavel_nm = np.unique(np.array(wavel_nm))
        return wavel_nm
        
    def configure_instrument_and_method( self, ret_inst, meas_inst, method ):
        """
        Configures the retrieval method and retrieval instrument with appropriate
        values. Uses information from the measurement instrument to do this, such
        as the measured wavelengths.
        
        Parameters
        ----------
        ret_inst : :class:`~instruments.instrument.Instrument`
            The instrument used to generate simulated level one data for the retrieval
            
        ret_inst : :class:`~instruments.instrument.Instrument`
            The instrument used to generate the 'real' measurements. Used to generate the
            measurement vectors, y
            
        method : :class:`~retrieval.lsq.LeastSquares`
            least squares minimization technique. Must support weighting matrix.
        """
        # Get the wavelengths we need for our triplet
        # and store it as an np array
        temp_wavel = self.get_required_wavelengths()
        
        #TODO: allow variable wavelengths over exposures
        #if the wavelengths change with exposure just take the first ones.
        if isinstance(meas_inst.wavel_nm, list):
            meas_wavel = meas_inst.wavel_nm[0]
        else:
            meas_wavel = meas_inst.wavel_nm
            
        #find the closest wavelengths in the meas_inst and use those
        self.wavel_nm = copy.copy(temp_wavel)
        for i in range(len(temp_wavel)):
            idx = np.argmin(np.abs(temp_wavel[i] - meas_wavel))
            self.wavel_nm[i] = meas_wavel[idx]        

        ret_inst.set_wavelengths( self.wavel_nm )
        
        #get the tangent altitude of the measurements
        los, obs = ret_inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )
    
        #calculate the molecular-only aerosol vector
        #save the initial state of aerosol and radiance
        [alts,aer] = self.atmosphere.get_user_defined(self.ident)
        temp_rad = copy.copy(ret_inst.rad)
        #set aerosols to zero and calculate the radiance with no aerosol
        self.atmosphere.modify_user_defined(self.ident, np.zeros(len(aer)))
        ret_inst.calculate_radiance()
        #calculate the measurement vector with no aerosol
        self.y_ray, dy_ray = self.measurement_vector_unnorm( ret_inst, do_error=False )
        #reset the atmosphere and instrument
        self.atmosphere.modify_user_defined(self.ident, aer)
        ret_inst.rad = temp_rad
        
        if method.type == 'LM':
            W = self.lm_weighting_matrix( ret_inst )
        elif method.type == 'MART':
            W = self.mart_weighting_matrix( ret_inst )
        
        method.W = W
        method.user_defined_weight = True
        
    def measurement_vector( self, inst, do_error=False ):
        """
        Generates the measurement vector
        
        Parameters
        ----------        
        inst : :class:`~instruments.instrument.Instrument`
            Instrument that provides level 1 used to generate the measurement vector.
        
        do_error : (optional)
            Whether or not to compute the error on the measurement vector. If false,
            the error is returned as zeros.
            
        Returns
        -------
        y : 1d array
            The measurement vectors stacked into a one dimensional array.
        """
        y, dy = self.measurement_vector_unnorm( inst, do_error )

        #normalize by rayleigh
        y = y-self.y_ray

        #find the altitude normalization value, yref, and normalize y
        for idx,triplet in enumerate(self.triplet):
            good_idx = (np.asarray(self.tanalts)>self.normalt[idx][0]) & (np.asarray(self.tanalts)<self.normalt[idx][1]) & (~np.isnan(y[idx]))
            #if it hasnt been done yet find the normalization point of the measurement vector (the altitude chosen by y_meas)
            if self.ref_idx[idx] is None:
                y_ref_range       =  y[idx, good_idx]
                min_idx = np.argmin( y_ref_range )
                if do_error:
                    # good    = y_ref_range == y_ref_range[min_idx]
                    good    = y_ref_range <= y_ref_range[min_idx] + dy[idx, good_idx]   #find altitudes in reference range
                else:
                    good    = y_ref_range == y_ref_range[min_idx]
                self.ref_idx[idx] = np.arange(0,len(self.tanalts))[good_idx][good]

            #update measurement error to include normalization altitudes
            if do_error:
                dy[idx] += np.sqrt( np.sum(dy[idx, self.ref_idx[idx]]**2) )/len(self.ref_idx[idx])

            #find the minimum value and normalize and normalize by this
            yref = np.nanmean(y[idx,self.ref_idx[idx]] )
            ref_alt = np.min(np.asarray(self.tanalts)[self.ref_idx[idx]])
            y[idx,:] = y[idx,:]-yref

            #get rid of values above normalization point and below min alt
            y[idx,np.asarray(self.tanalts) >= ref_alt-self.refaltrange] = np.nan
            y[idx,np.asarray(self.tanalts) < self.min_alt]              = np.nan
            # y[idx,y[idx,:] < 0] = np.nan

        #find the measurements used in the retrieval
        if self.good_alts is None:
             temp_good = np.zeros(np.shape(y))
             for idx in range(0,len(self.triplet)):
                 ref_alt = np.min(np.asarray(self.tanalts)[self.ref_idx[idx]])
                 temp_good[idx,:] = (np.asarray(self.tanalts) < ref_alt-self.refaltrange) & (np.asarray(self.tanalts)>self.min_alt) & (~np.isnan(y[idx]))

             #if any measuerment is good use the altitude
             self.good_alts = np.any(temp_good,0)
             self.max_alt   = np.max(np.asarray(self.tanalts)[self.good_alts])
             self.min_alt   = np.min(np.asarray(self.tanalts)[self.good_alts])

        #flatten to 1-d and return
        return y.flatten(), dy.flatten()
        
    def measurement_vector_unnorm( self, inst, do_error=False ):
        """
        Generates the measurment vector unnormalized by high altitude and rayleigh

        Returns
        -------
        y : 2d array
            An array of height x num meas vectors
        dy : 2d array
            An array the same size as y, containing the measurement vector error
        """     
        
        rad = np.transpose(inst.rad)
        if do_error:
            try:
                rad_error = np.transpose(inst.rad_error)
            except AttributeError:
                logging.info("Instrument has no error attribute, assuming 1%")
                rad_error = rad * 0.01
        else:
            rad_error = np.zeros(np.shape(rad))
            
        # rad is now [wavel, los]
        y  = np.zeros((len(self.triplet), np.shape(rad)[1]))
        dy = np.zeros((len(self.triplet), np.shape(rad)[1]))
        # y is now [triplet, los]

        for idx,triplet in enumerate(self.triplet):
            for wavel, coeff in zip(triplet[0],triplet[1]):
                
                #find the closest wavelength and get the radiance
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))    
                radI = rad[wave_idx,:]
                errI = rad_error[wave_idx,:]

                #find the maximum altitude
                alt_idx = np.argmin(np.abs(self.max_alt-np.asarray(self.tanalts)))
                radR = radI[alt_idx]
                errR = errI[alt_idx]
                
                #compute the measurement vector
                y[idx,:] += coeff * np.log(radI)# - coeff * np.log(radR)

                #TODO: update now that normalization has changed
                dy[idx,:]+= (coeff * errI/radI)**2 #sum in quadrature

        dy = np.sqrt(dy)
        return y, dy
    
    def measurement_vector_jacobian(self, inst):
        """
        Computes the measurement vector jacobian using either analytic or numerical
        perturbation methods.
        
        Returns
        -------
        J : 2d array
            The Jacobian of the measurement vector. Elemenent J[i,j] give the 
            sensitivity of the measurement vector y[i], to the perturbations of
            the atmospheric state x[j], or dy[j]/dx[i].
        """
        #use either weighting functions or numerical perturbation to compute the jacobian
        if self.use_wf:
            J = self.measurement_vector_jacobian_wf( inst )
        else:
            J = self.measurement_vector_jacobian_num( inst )
            
        return J
    
    def measurement_vector_jacobian_wf( self, inst ):
        """
        Computes the measurement vector jacobian using analytical methods from the
        HR sasktran model.
        """
        wf =  inst.wf 
        rad = np.transpose(inst.rad)
        # wf is [wavel, los, pert]
        # Rad is [wavel, los]
        # Jacobian has to be [triplet*los,pert]
        J = np.zeros( (len(self.triplet), np.shape(wf)[1], np.shape(wf)[2]) )
        for idx,triplet in enumerate(self.triplet):
            for wavel, coeff in zip(triplet[0],triplet[1]):
                
                #find the weighting function and radiance for this triplet wavelength
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))
                wfI = wf[wave_idx,:,:]
                radI = rad[wave_idx,:]
                
                #compute the jacobian element
                J[idx,:,:] += coeff * wfI / radI[:,np.newaxis]# - coeff * wfNorm / radR
                
        # Combine the first two axis in J
        J = np.reshape( J.flatten(), (len(self.triplet)*np.shape(wf)[1], np.shape(wf)[2]) )

        #include the fact that we scale everything above the highest retrieved alt
        # high_alt_idx = np.max(np.where(self.good_alts))
        # J[:,high_alt_idx] = np.sum(J[:,high_alt_idx::],axis=1)

        #TODO: change this to handle more than one measurment vector
        J = J[:,self.good_alts]
        return J
        
    def measurement_vector_jacobian_num( self, inst ):
        """
        Computes the unnormalized measurement vector jacobian using simple numerical forward
        difference method.
        """
        #store the initial state
        temp_rad    = copy.copy(inst.rad)
        num_sctr    = copy.copy(inst.radtran.options.optmap['numordersofscatter'])
        alts, aer_0 = copy.copy(self.atmosphere.get_user_defined(self.ident))
        
        #find altitudes to perturb and initialize J
        pert_alts   = np.asarray(self.tanalts)[self.good_alts]
        
        #set scatter order for Jacobians
        inst.set_num_orders_of_scatter( self.jacobian_scatter_order )
        inst.calculate_radiance()
        y_0,temp = self.measurement_vector_unnorm( inst )
            
        J = np.zeros((np.size(y_0),np.size(pert_alts)))
        
        for i in range(0,len(pert_alts)):
            
            #perturb aerosol at an altitude
            idx           = np.argmin( np.abs(pert_alts[i]-alts))
            aer_1         = copy.copy(aer_0)
            aer_1[idx]    = aer_1[idx]*1.01
            self.atmosphere.modify_user_defined(self.ident, aer_1)
                
            #calculate new measuerment vector
            inst.calculate_radiance()
            y_1,temp = self.measurement_vector_unnorm( inst )
            inst.rad = temp_rad
            
            #find numerical jacobian
            J[:,i] = (y_1-y_0)/(aer_1[idx]-aer_0[idx])
            
        #reset variables
        inst.rad = temp_rad
        self.atmosphere.modify_user_defined(self.ident, aer_0)
        inst.set_num_orders_of_scatter( num_sctr )
        
        return J
        
    def update_profile( self, delta, mode='additive' ):
        """
        Update the aerosol profile in the atmosphere at based on the recommendations
        given by delta. 
        
        Parameters
        ----------
        delta : 1d array
            The suggested numerical change of the profile. Must be the same length
            as the tangent altitudes.
        """

        #interpolate the profile onto the tangent altitudes
        ret_profile = np.interp(np.asarray(self.tanalts)[self.good_alts], self.profile_alts, self.profile)

        #delta = delta[good_alts]

        #interpolate delta onto retrieval grid
        if mode == 'additive':
            delta = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], (ret_profile+delta)/ret_profile)
        else:
            delta = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], delta[self.good_alts])

        min_alt     = np.min(np.asarray(self.tanalts)[self.good_alts])
        max_alt     = np.max(np.asarray(self.tanalts)[self.good_alts])
        min_alt_idx = np.min(np.where(self.profile_alts > min_alt))
        max_alt_idx = np.max(np.where(self.profile_alts < max_alt))+1
        delta[self.profile_alts > max_alt] = delta[max_alt_idx]
        delta[self.profile_alts < min_alt] = delta[min_alt_idx]

        #update profile (dont allow it change by more than a factor of 5/20)
        delta[delta<0.2]    = 0.2
        delta[delta>20.0]   = 20.0
        self.profile *= delta
        self.profile[self.profile>self.max_val] = self.max_val
        self.atmosphere.modify_user_defined( self.ident, self.profile )
        
    def get_atmospheric_profile(self):
        alt, x = self.atmosphere.get_user_defined(self.ident)
        x = clim_util.number_density_to_extinction(alt, x, self.ext_wavelength, self.aer_clim)
        return copy.copy(x)
        
    def set_atmospheric_profile(self, profile):

        # get the altitude grid
        alt, x_last = self.atmosphere.get_user_defined(self.ident)

        # convert to number density
        x = clim_util.extinction_to_number_density(alt, profile, self.ext_wavelength, self.aer_clim)

        self.profile = x
        self.atmosphere.modify_user_defined(self.ident, x)

    def get_x(self):
        profile = self.get_atmospheric_profile()
        pert_alts = np.asarray(self.tanalts)[self.good_alts]
        x = np.interp(pert_alts, self.profile_alts, profile)
        return x
        
    def output_data( self ):
        """
        :return: the profile in units of extinction per km
        """
        alt,x = self.atmosphere.get_user_defined( self.ident )
        x     = clim_util.number_density_to_extinction(alt, x, self.ext_wavelength, self.aer_clim)
        return x

    def output_data_bounded( self ):

        min_alt = np.min(np.asarray(self.tanalts)[self.good_alts])
        max_alt = np.max(np.asarray(self.tanalts)[self.good_alts])
        p = self.output_data()
        #get first alt above 'saturated' aerosol
        if np.any( p > self.max_ext_val):
            min_idx = np.max(np.where(p > self.max_ext_val)[0])+1
        else:
            min_idx = 0
        p[(self.profile_alts > max_alt) | (self.profile_alts < min_alt)] = np.nan
        p[ 0:min_idx] = np.nan
        #return the ozone profile
        return p
        
    def lm_weighting_matrix( self, inst ):
        
        #get the tangent altitude of the measurements
        los, obs = inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )
        
        #initialize weighting matrix to zero
        W = np.zeros((len(self.tanalts),len(self.triplet) ))  
        
        #Give every measuerment vector equal weight
        for i in range(0,len(self.tanalts)):
            for j in range(0,len(self.triplet)):
                W[i,j] = 1
                
        #reshape for lm retrieval
        return np.reshape(W.transpose(),(np.size(W)))

    def mart_weighting_matrix(self, inst ):

        #get the tangent altitude of the measurements
        los, obs = inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )

        #load the weighting matrix from file
        dir_path = os.path.dirname( os.path.abspath(__file__) )
        wik = io.loadmat(dir_path + r'/../utility/data/aerosol/wik.mat')['wik']
        wij = io.loadmat(dir_path + r'/../utility/data/aerosol/wij.mat')['wij']

       # set up J matrix for line of sight weighting
        n = len(self.tanalts)
        m = len(wij)
        J = np.eye(n,n)
        J = signal.convolve2d(J,np.fliplr(wij))[:,len(wij)-1::]
        J = J/(np.sum(J,axis=1)[:,np.newaxis])

        # set up L matrices for spectral weighting and build H
        H = np.zeros((len(self.tanalts), 0))
        for k in range(np.shape(wik)[1]-1):#k=1:obj.numnlets
            L = np.diag(np.interp(self.tanalts, wik[:,0]*1000,wik[:,k+1]))
            H = np.hstack((H, np.dot(L,J)))

        np.seterr(divide='ignore',invalid='ignore') #deal with nans in a second
        H = H/(np.sum(H,axis=1)[:,np.newaxis])
        np.seterr(divide=None,invalid=None)
        H[np.isnan(H)] = 0.0
        return H

    def check_convergence(self, y_ret, previous_y_ret):

        converged = False

        if previous_y_ret is not None:
            step_perc = np.abs(y_ret[self.good_alts] - previous_y_ret[self.good_alts]) / y_ret[self.good_alts] * 100
            if np.all( step_perc < self.min_step_size):
                converged = True

        return converged