import copy as copy
import os
import logging

import numpy as np
from scipy import io
from scipy import signal

import retrieval.species as species
import radtran.geometry as geo
import radtran.climatologies.utility as clim_util
from retrieval.species_aerosol import AerosolLimbProfile


class AerosolLimbProfileBimodal(AerosolLimbProfile):

    def __init__(self,atmosphere):
        """
        Parameters
        ----------
        atmosphere : :class:`~utility.sasktran.atmosphere.Atmosphere`
            Holds the atmospheric state, including climatologie, optical properties
            and species identifiers (GUIDs)
        """
        self.ident_1 = 'limb_aerosol_1'
        self.ident_2 = 'limb_aerosol_2'

        self.aer_clim_1 = atmosphere.user_optprop[self.ident_1]
        self.aer_clim_2 = atmosphere.user_optprop[self.ident_2]

        #TODO: setup triplets from a file
        self.triplet = []
        self.triplet.append( ([745.67], [1]))

        self.ext_wavelength = 745.67
        #TODO: setup normalization based on measurement noise
        self.normalt = np.array( [[25000.0,40000.0]])

        self.profile_alts,self.profile_1 = atmosphere.get_user_defined( self.ident_1 )
        self.profile_alts,self.profile_2 = atmosphere.get_user_defined( self.ident_2 )

        x_1 = clim_util.number_density_to_extinction(self.profile_alts, self.profile_1,self.ext_wavelength,self.aer_clim_1)
        x_2 = clim_util.number_density_to_extinction(self.profile_alts, self.profile_2,self.ext_wavelength,self.aer_clim_2)
        self.mode_fraction = x_1/(x_2 + x_1)
        self.mode_fraction[np.isnan(self.mode_fraction) | np.isinf(self.mode_fraction)] = 1.0
        self.profile       = x_1 + x_2

        self.normalt = [(np.array( [25000.0,45000.0]))]
        self.ref_idx     = [None] * len(self.triplet) #np.zeros(len(self.triplet), dtype=np.int)
        self.good_alts   = None
        self.y_ray       = None
        self.wavel_nm    = None
        self.atmosphere  = atmosphere
        self.refaltrange = 0.0
        self.max_alt     = 40000
        self.min_alt     = 5000
        self.max_val     = 0.01

        self.min_step_size = 0.5 #if step size is < 0.1% solution has converged
        #use weighting functions for jacobians
        self.use_wf = True
        self.jacobian_scatter_order = 20


    def configure_instrument_and_method( self, ret_inst, meas_inst, method ):
        """
        Configures the retrieval method and retrieval instrument with appropriate
        values. Uses information from the measurement instrument to do this, such
        as the measured wavelengths.

        Parameters
        ----------
        ret_inst : :class:`~instruments.instrument.Instrument`
            The instrument used to generate simulated level one data for the retrieval

        ret_inst : :class:`~instruments.instrument.Instrument`
            The instrument used to generate the 'real' measurements. Used to generate the
            measurement vectors, y

        method : :class:`~retrieval.lsq.LeastSquares`
            least squares minimization technique. Must support weighting matrix.
        """
        # Get the wavelengths we need for our triplet
        # and store it as an np array
        temp_wavel = self.get_required_wavelengths()

        #TODO: allow variable wavelengths over exposures
        #if the wavelengths change with exposure just take the first ones.
        if isinstance(meas_inst.wavel_nm, list):
            meas_wavel = meas_inst.wavel_nm[0]
        else:
            meas_wavel = meas_inst.wavel_nm

        #find the closest wavelengths in the meas_inst and use those
        self.wavel_nm = copy.copy(temp_wavel)
        for i in range(len(temp_wavel)):
            idx = np.argmin(np.abs(temp_wavel[i] - meas_wavel))
            self.wavel_nm[i] = meas_wavel[idx]

        ret_inst.set_wavelengths( self.wavel_nm )

        #get the tangent altitude of the measurements
        los, obs = ret_inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )

        #calculate the molecular-only aerosol vector
        #save the initial state of aerosol and radiance
        [alts,aer_1] = self.atmosphere.get_user_defined(self.ident_1)
        [alts,aer_2] = self.atmosphere.get_user_defined(self.ident_2)
        temp_rad = copy.copy(ret_inst.rad)
        #set aerosols to zero and calculate the radiance with no aerosol
        self.atmosphere.modify_user_defined(self.ident_1, np.zeros(len(aer_1)))
        self.atmosphere.modify_user_defined(self.ident_2, np.zeros(len(aer_2)))
        ret_inst.calculate_radiance()
        #calculate the measurement vector with no aerosol
        self.y_ray, dy_ray = self.measurement_vector_unnorm( ret_inst, do_error=False )
        #reset the atmosphere and instrument
        self.atmosphere.modify_user_defined(self.ident_1, aer_1)
        self.atmosphere.modify_user_defined(self.ident_2, aer_2)
        ret_inst.rad = temp_rad

        if method.type == 'LM':
            W = self.lm_weighting_matrix( ret_inst )
            print('only mart is currently supported why are you using this')
        elif method.type == 'MART':
            W = self.mart_weighting_matrix( ret_inst )

        method.W = W
        method.user_defined_weight = True

    def update_profile( self, delta, mode='additive' ):
        """
        Update the aerosol profile in the atmosphere at based on the recommendations
        given by delta.

        Parameters
        ----------
        delta : 1d array
            The suggested numerical change of the profile. Must be the same length
            as the tangent altitudes.
        """

        #interpolate the profile onto the tangent altitudes
        ret_profile = np.interp(np.asarray(self.tanalts)[self.good_alts], self.profile_alts, self.profile)

        #delta = delta[good_alts]

        #interpolate delta onto retrieval grid
        if mode == 'additive':
            delta = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], (ret_profile+delta)/ret_profile)
        else:
            delta = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], delta[self.good_alts])

        min_alt     = np.min(np.asarray(self.tanalts)[self.good_alts])
        max_alt     = np.max(np.asarray(self.tanalts)[self.good_alts])
        min_alt_idx = np.min(np.where(self.profile_alts > min_alt))
        max_alt_idx = np.max(np.where(self.profile_alts < max_alt))+1
        delta[self.profile_alts > max_alt] = delta[max_alt_idx]
        delta[self.profile_alts < min_alt] = delta[min_alt_idx]

        #update profile (dont allow it change by more than a factor of 5/20)
        delta[delta<0.2]    = 0.2
        delta[delta>20.0]   = 20.0
        self.profile *= delta   #total extinction


        self.profile[self.profile>self.max_val] = self.max_val

        #convert extinction to number density
        self.set_x(self.profile)

    def get_atmospheric_profile(self):
        alt,x_1=self.atmosphere.get_user_defined( self.ident_1 )
        alt,x_2=self.atmosphere.get_user_defined( self.ident_2 )

        x_1 = clim_util.number_density_to_extinction(alt,x_1,self.ext_wavelength,self.aer_clim_1)
        x_2 = clim_util.number_density_to_extinction(alt,x_2,self.ext_wavelength,self.aer_clim_2)

        x = x_1 + x_2   #total extinction

        return copy.copy(x)

    def set_atmospheric_profile(self, profile):
        """
        Set the number of density of profiles given the total extinction

        :param x: total extinction
        """
        alt,temp=self.atmosphere.get_user_defined( self.ident_1 )

        self.profile = profile

        #extinction profiles
        x_1  = profile * self.mode_fraction
        x_2  = profile * (1 - self.mode_fraction)

        #number density profiles
        x_1 = clim_util.extinction_to_number_density(alt,x_1 * 1e-5,self.ext_wavelength,self.aer_clim_1)
        x_2 = clim_util.extinction_to_number_density(alt,x_2 * 1e-5 ,self.ext_wavelength,self.aer_clim_2)

        x_1[np.isnan(x_1)] = 0.0
        x_2[np.isnan(x_2)] = 0.0

        self.atmosphere.modify_user_defined( self.ident_1, x_1 )
        self.atmosphere.modify_user_defined( self.ident_2, x_2 )

    def output_data( self ):
        """
        :return: the profile in units of extinction per km
        """
        alt,x_1 = self.atmosphere.get_user_defined( self.ident_1 )
        alt,x_2 = self.atmosphere.get_user_defined( self.ident_2 )

        x_1     = clim_util.number_density_to_extinction(alt, x_1, self.ext_wavelength,self.aer_clim_1)
        x_2     = clim_util.number_density_to_extinction(alt, x_2, self.ext_wavelength,self.aer_clim_2)

        x = x_1 + x_2
        return x