import copy as copy

import numpy as np

import radtran.geometry as geo
import radtran.climatologies.utility as clim_util
import retrieval.species_aerosol as species_aerosol
import sasktranif.sasktranif as skif


class AerosolPartSizeLimbProfile(species_aerosol.AerosolLimbProfile):
    def __init__(self, atmosphere, apriori):
        
        self.ident = 'limb_aerosol'
        
        self.triplet = []
        #self.triplet.append( ([526.0,   470.00], [1,-1]))
        self.triplet.append( ([675.0,   470.00], [1,-1]))
        self.triplet.append( ([745.677, 470.00], [1,-1]))
        self.triplet.append( ([869.0,   470.00], [1,-1]))
        
        self.profile_alts,self.profile = atmosphere.get_user_defined( self.ident )
        
        #TODO: setup normalization based on measurement noise
        self.normalt        = np.array( [[40000.0,40000.0]])
        self.normalt_range  = np.array( [25000.0,45000.0])
        self.ref_idx        = np.zeros(len(self.triplet))
        self.good_alts      = None
        self.y_ray          = None
        self.wavel_nm       = None
        
        self.atmosphere     = atmosphere
        self.refaltrange    = 1500.000
        self.max_alt        = 40000
        self.min_alt        = 8000
        
        self.max_n          = 1000
        self.max_rg         = 0.15
        self.min_rg         = 0.06
        self.max_ext        = 1e-2
        
        self.const_ext_wavel = 745.677  #wavelength at which extinction is kept constant between particle size changes
        
        #initialize particle size
        self.aer_optprop   = atmosphere.user_optprop['limb_aerosol']
        self.mode_width    = apriori['modewidth_profile']
        self.mode_radius   = apriori['moderadius_profile']
        self.partsize_alts = apriori['particlesize_heights']
        
        self.particlesize = skif.ISKClimatology('USERDEFINED_PROFILE')
        self.particlesize.SetProperty('Heights', self.partsize_alts)
        self.particlesize.SetPropertyUserDefined( skif.SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS, self.mode_radius.tolist() )
        self.particlesize.SetPropertyUserDefined( skif.SKCLIMATOLOGY_LOGNORMAL_MODEWIDTH,          self.mode_width.tolist() )
        self.particlesize.SetProperty('DoLogInterpolation',0)
        self.particlesize.SetProperty('DoPiecewiseLinear',1)
        
        self.aer_optprop.SetProperty('SetParticleSizeClimatology', self.particlesize )
        
    def set_mode_radius( self, mode_radius ):
                
        #get the previous extinction
        ext_old = clim_util.number_density_to_extinction(self.profile_alts, self.profile, self.const_ext_wavel, self.aer_optprop )
        
        #update the particle size
        self.particlesize.SetPropertyUserDefined( skif.SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS, mode_radius.tolist() )
        
        #get the new extinction
        ext_new = clim_util.number_density_to_extinction(self.profile_alts, self.profile, self.const_ext_wavel, self.aer_optprop )
        
        #update number density so that extinction remains unchanged
        for i in range(len(ext_old)):
            if ext_new[i]==0:
                self.profile[i] = 0
            else:
                self.profile[i] *= ext_old[i] / ext_new[i]
            
        self.atmosphere.modify_user_defined( self.ident, self.profile )
        ext_new = clim_util.number_density_to_extinction(self.profile_alts, self.profile, self.const_ext_wavel, self.aer_optprop )
        
        #update the internal climatology
        self.mode_radius = mode_radius
        
    def get_mode_radius( self ):
        
        return self.mode_radius

    def measurement_vector_jacobian( self, inst ):
        
        #store the initial state
        temp_rad    = copy.copy(inst.rad)
        num_sctr    = copy.copy(inst.radtran.options.optmap['numordersofscatter'])
        alts, aer_0 = copy.copy(self.atmosphere.get_user_defined(self.ident))
        rg_0        = copy.copy( self.mode_radius )
        #find altitudes to perturb and initialize J
        pert_alts   = np.asarray(self.tanalts)[self.good_alts]
        
        #do single scatter jacobians
        inst.set_num_orders_of_scatter( 1 )
        inst.calculate_radiance()
        y_0,temp = self.measurement_vector_unnorm( inst )
        J = np.zeros((np.size(y_0),2*np.size(pert_alts)))
        
        #compute Jacobian elements with respect to number density
        for i in range(0,len(pert_alts)):
            
            #perturb aerosol at an altitude
            idx           = np.argmin( np.abs(pert_alts[i]-alts))
            aer_1         = copy.copy(aer_0)
            aer_1[idx]    = aer_1[idx]*1.01
            self.atmosphere.modify_user_defined(self.ident, aer_1)
                
            #calculate new measuerment vector
            inst.calculate_radiance()
            y_1,temp = self.measurement_vector_unnorm( inst )

            #find numerical jacobian
            J[:,i] = (y_1.flatten()-y_0.flatten())/(aer_1[idx]-aer_0[idx])
        
        self.atmosphere.modify_user_defined(self.ident, aer_0)
        
        #compute Jacobian elements with respect to particle radius (keeping 
        #extinction the same)
        for i in range(0,len(pert_alts)):
            
            #perturb particle size at an altitude
            idx           = np.argmin( np.abs(pert_alts[i]-alts))
            rg_1         = copy.copy(rg_0)
            rg_1[idx]    = rg_1[idx]+0.002   #increase size by 2 nm
            self.set_mode_radius( rg_1 )
                                
            #calculate new measuerment vector
            inst.calculate_radiance()
            y_1,temp = self.measurement_vector_unnorm( inst )
            
            #find numerical jacobian
            J[:,i+len(pert_alts)] = (y_1.flatten()-y_0.flatten())/(rg_1[idx]-rg_0[idx])
            
            #reset particle size and extinction
            self.set_mode_radius( rg_0 )
        
        #reset variables
        inst.rad = temp_rad
        self.atmosphere.modify_user_defined(self.ident, aer_0)
        self.set_mode_radius( rg_0 )
        inst.set_num_orders_of_scatter( num_sctr )
        
        return J
        
    def update_profile( self, delta ):

        #break up delta into rg and number density
        delta_n  = np.reshape(delta, (-1,2))[:,0]
        delta_rg = np.reshape(delta, (-1,2))[:,1]
        
        #interpolate the profile onto the tangent altitudes
        ret_profile_n  = np.interp(np.asarray(self.tanalts)[self.good_alts], self.profile_alts, self.profile)
        ret_profile_rg = np.interp(np.asarray(self.tanalts)[self.good_alts], self.profile_alts, self.mode_radius)
        
        #interpolate delta onto retrieval grid        
        delta_n  = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], (ret_profile_n+delta_n)/ret_profile_n)
        delta_rg = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], (ret_profile_rg+delta_rg)/ret_profile_rg)
        
        #update profile (dont allow it change by more than a factor of 5)
        delta_n[delta_n<0.2]    = 0.2
        delta_n[delta_n>10.0]   = 10.0    
        self.profile            *= delta_n
        self.mode_radius        *= delta_rg
        
        #set profile limits and new mode radius
        self.mode_radius[self.mode_radius>self.max_rg] = self.max_rg
        self.mode_radius[self.mode_radius<self.min_rg] = self.min_rg
        self.set_mode_radius(self.mode_radius)
        
        #set extinction limits and number density
        ext = clim_util.number_density_to_extinction(self.profile_alts,self.profile, self.const_ext_wavel, self.aer_optprop)
        nmax = clim_util.extinction_to_number_density(np.array([10000.0]), np.array([self.max_ext])/1e5,self.const_ext_wavel, self.aer_optprop )
        self.profile[ ext > self.max_ext] = nmax
        self.atmosphere.modify_user_defined( self.ident, self.profile )

    def get_atmospheric_profile( self ):
        alt,x=self.atmosphere.get_user_defined( self.ident )
        x = np.hstack((x, self.mode_radius))
        return x
        
    def set_atmospheric_profile( self, profile):
        pass #self.atmosphere.modify_user_defined( self.ident, x )
        
    def output_data( self ):
        return self.profile
        
    def weighting_matrix( self, inst ):
        
        #get the tangent altitude of the measurements
        los, obs = inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )
        
         
        W = np.zeros((len(self.tanalts),len(self.triplet) ))        
        #interpolate onto tangent altitudes
        for i in range(0,len(self.tanalts)):
            for j in range(0,len(self.triplet)):
                W[i,j] = 1
                
        #reshape for lm retrieval
        return np.reshape(W.transpose(),(np.size(W)))