import copy as copy
import logging
import numpy as np

import retrieval.species as species
import radtran.geometry as geo


class AlbedoLimbProfile(species.Species):
    def __init__(self, atmosphere, init_albedo=0.3):
        
        self.ident = 'limb_albedo'
        
        #TODO: setup triplets from a file albedo is computed as the longest wavelength
        #of the triplet
        self.triplet = []
        self.triplet.append( ([745.67], [1]))
        
        #TODO: setup normalization based on measurement noise
        self.normalt = [np.array( [25000.0,45000.0])]
        self.ref_idx     = [None] * len(self.triplet)
        self.ref_alt     = np.zeros(len(self.triplet))
        self.good_alts   = None
        self.y_ray       = None
        self.wavel_nm    = None
        self.atmosphere  = atmosphere
        self.refaltrange = 1500.000
        self.max_alt     = 40000
        self.min_alt     = 8000
        
        self.albedo_wavel = self.triplet[0][0]
        self.albedo = init_albedo
        self.normalize_albedo = False

        self.min_step_size = 0.1
        
    def get_required_wavelengths(self):
        #find the unique wavelengths used by the triplets
        wavel_nm = []
        for triplet in self.triplet:
            wavel_nm.append( triplet[0] )
        wavel_nm = [item for sublist in wavel_nm for item in sublist]
        wavel_nm = np.unique(np.array(wavel_nm))
        return wavel_nm
        
    def configure_instrument_and_method( self, ret_inst, meas_inst, method ):
        
        # Get the wavelengths we need for our triplet
        # and store it as an np array
        temp_wavel = self.get_required_wavelengths()
        
        #TODO: allow variable wavelengths over exposures
        #if the wavelengths change with exposure just take the first ones.
        if isinstance(meas_inst.wavel_nm, list):
            meas_wavel = meas_inst.wavel_nm[0]
        else:
            meas_wavel = meas_inst.wavel_nm
            
        #find the closest wavelengths in the meas_inst and use those
        self.wavel_nm = copy.copy(temp_wavel)
        for i in range(len(temp_wavel)):
            idx = np.argmin(np.abs(temp_wavel[i] - meas_wavel))
            self.wavel_nm[i] = meas_wavel[idx]        
            
        ret_inst.set_wavelengths( self.wavel_nm )
        
        #get the tangent altitude of the measurements
        los, obs = ret_inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )
    
        #calculate the molecular-only aerosol vector
        #save the initial state of aerosol and radiance
        aer_type = None
        try:
            [alts,aer] = self.atmosphere.get_user_defined('limb_aerosol')
            aer_type = 'unimodal'
        except:
            pass
        try:
            [alts,aer_1] = self.atmosphere.get_user_defined('limb_aerosol_1')
            [alts,aer_2] = self.atmosphere.get_user_defined('limb_aerosol_2')
            aer_type = 'bimodal'
        except:
            pass
        if aer_type is None:
            raise Exception('could not determine aerosol background')
        temp_rad = copy.copy(ret_inst.rad)
        #set aerosols to zero and calculate the radiance with no aerosol
        if aer_type == 'unimodal':
            self.atmosphere.modify_user_defined('limb_aerosol', np.zeros(len(aer)))
        elif aer_type == 'bimodal':
            self.atmosphere.modify_user_defined('limb_aerosol_1', np.zeros(len(aer_1)))
            self.atmosphere.modify_user_defined('limb_aerosol_2', np.zeros(len(aer_2)))
        ret_inst.calculate_radiance()
        #calculate the measurement vector with no aerosol
        self.y_ray, dy_ray = self.measurement_vector_aer_unnorm( ret_inst, do_error=False )

        if self.normalize_albedo:
            self.y_0, dy_0     = self.measurement_vector_zero_albedo( ret_inst, do_error=False )
        else:
            self.y_0 = 0.0

        #reset the atmosphere and instrument
        if aer_type == 'unimodal':
            self.atmosphere.modify_user_defined('limb_aerosol', aer)
        elif aer_type == 'bimodal':
            self.atmosphere.modify_user_defined('limb_aerosol_1', aer_1)
            self.atmosphere.modify_user_defined('limb_aerosol_2', aer_2)

        ret_inst.rad = temp_rad

        if method.type == 'LM':
            W = self.lm_weighting_matrix( ret_inst )
            method.user_defined_weight = False
        elif method.type == 'MART':
            W = self.mart_weighting_matrix( ret_inst )

            method.W = W
            method.user_defined_weight = True

    def measurement_vector( self, inst, do_error=False ):

        if self.ref_idx[0] is None:
            self.measurement_vector_aer( inst, do_error )

        y = np.zeros(len(self.triplet))
        for idx,triplet in enumerate(self.triplet):
            wavel = triplet[0][0]  #get albedo wavelength
            #find the closest wavelength and get the radiance
            wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))
            rad = np.transpose(inst.rad)
            if self.normalize_albedo:
                y[idx] = np.array( [ np.mean(rad[wave_idx, self.ref_idx[idx]] ) ] ) - np.array( [ np.mean(self.y_0[idx, self.ref_idx[idx]] ) ] )
            else:
                y[idx] = np.array([np.mean(rad[wave_idx, self.ref_idx[idx]])])
        #TODO: set error properly
        dy = 0

        return y, dy

    def measurement_vector_zero_albedo(self, inst, do_error):

        albedo = copy.copy(self.atmosphere.albedo)

        self.atmosphere.albedo = 0.0
        inst.calculate_radiance()

        y = np.zeros((len(self.triplet),inst.rad.shape[0]))
        for idx,triplet in enumerate(self.triplet):
            wavel = triplet[0][0]  #get albedo wavelength
            #find the closest wavelength and get the radiance
            wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))
            rad = np.transpose(inst.rad)
            y[idx,:] = rad[wave_idx]

        self.atmosphere.albedo = albedo

        return y, y*0.0

    def measurement_vector_aer( self, inst, do_error=False ):
        """
        Generates the measurement vector

        Parameters
        ----------
        inst : :class:`~instruments.instrument.Instrument`
            Instrument that provides level 1 used to generate the measurement vector.

        do_error : (optional)
            Whether or not to compute the error on the measurement vector. If false,
            the error is returned as zeros.

        Returns
        -------
        y : 1d array
            The measurement vectors stacked into a one dimensional array.
        """
        #calculate the measurement vector unnormalized by altitude and rayleigh
        y, dy = self.measurement_vector_aer_unnorm( inst, do_error )

        #normalize by rayleigh
        y = y-self.y_ray

        #find the altitude normalization value, yref, and normalize y
        for idx,triplet in enumerate(self.triplet):
            good_idx = (np.asarray(self.tanalts)>self.normalt[idx][0]) & (np.asarray(self.tanalts)<self.normalt[idx][1]) & (~np.isnan(y[idx]))
            #if it hasnt been done yet find the normalization point of the measurement vector (the altitude chosen by y_meas)
            if self.ref_idx[idx] is None:
                y_ref_range       =  y[idx, good_idx]
                min_idx = np.argmin( y_ref_range )
                if do_error:
                    # good    = y_ref_range == y_ref_range[min_idx]
                    good    = y_ref_range <= y_ref_range[min_idx] + dy[idx, good_idx]   #find altitudes in reference range
                else:
                    good    = y_ref_range == y_ref_range[min_idx]
                self.ref_idx[idx] = np.arange(0,len(self.tanalts))[good_idx][good]

            #update measurement error to include normalization altitudes
            if do_error:
                dy[idx] += np.sqrt( np.sum(dy[idx, self.ref_idx[idx]]**2) )/len(self.ref_idx[idx])

            #find the minimum value and normalize and normalize by this
            yref = np.nanmean(y[idx,self.ref_idx[idx]] )
            ref_alt = np.min(np.asarray(self.tanalts)[self.ref_idx[idx]])
            y[idx,:] = y[idx,:]-yref

            #get rid of values above normalization point and below min alt
            y[idx,np.asarray(self.tanalts) >= ref_alt-self.refaltrange] = np.nan
            y[idx,np.asarray(self.tanalts) < self.min_alt]              = np.nan

        #find the measurements used in the retrieval
        if self.good_alts is None:
             temp_good = np.zeros(np.shape(y))
             for idx in range(0,len(self.triplet)):
                 ref_alt = np.min(np.asarray(self.tanalts)[self.ref_idx[idx]])
                 temp_good[idx,:] = np.logical_and(np.asarray(self.tanalts) < ref_alt-self.refaltrange, np.asarray(self.tanalts)>self.min_alt)

             #if any measuerment is good use the altitude
             self.good_alts = np.any(temp_good,0)
             self.max_alt   = np.max(np.asarray(self.tanalts)[self.good_alts])
             self.min_alt   = np.min(np.asarray(self.tanalts)[self.good_alts])

        #flatten to 1-d and return
        return y.flatten(), dy.flatten()

    def measurement_vector_aer_unnorm( self, inst, do_error=False ):
        """
        Generates the measurment vector unnormalized by high altitude and rayleigh

        Returns
        -------
        y : 2d array
            An array of height x num meas vectors
        dy : 2d array
            An array the same size as y, containing the measurement vector error
        """

        rad = np.transpose(inst.rad)
        if do_error:
            rad_error = np.transpose(inst.rad_error)
        else:
            rad_error = np.zeros(np.shape(rad))

        # rad is now [wavel, los]
        y  = np.zeros((len(self.triplet), np.shape(rad)[1]))
        dy = np.zeros((len(self.triplet), np.shape(rad)[1]))
        # y is now [triplet, los]

        for idx,triplet in enumerate(self.triplet):
            for wavel, coeff in zip(triplet[0],triplet[1]):

                #find the closest wavelength and get the radiance
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))
                radI = rad[wave_idx,:]
                errI = rad_error[wave_idx,:]

                #find the maximum altitude
                alt_idx = np.argmin(np.abs(self.max_alt-np.asarray(self.tanalts)))
                radR = radI[alt_idx]
                errR = errI[alt_idx]

                #compute the measurement vector
                y[idx,:] += coeff * np.log(radI)# - coeff * np.log(radR)

                #TODO: update now that normalization has changed
                dy[idx,:]+= (coeff * errI/radI)**2 #sum in quadrature

        dy = np.sqrt(dy)
        return y, dy
        
    def measurement_vector_jacobian( self, inst ):
        
        y_0 = self.measurement_vector( inst )[0]
        temp_rad = copy.copy(inst.rad)
        
        a_0 = self.atmosphere.albedo
        a_1 = self.atmosphere.albedo+0.01
        self.atmosphere.albedo = a_1
        inst.calculate_radiance()
        y_1 = self.measurement_vector( inst )[0]
        inst.rad = temp_rad
        
        self.set_atmospheric_profile( a_0 )
        J = np.zeros((1,1))
        J[0,0] = (y_1-y_0)/(a_1-a_0)
        return J
        
    def update_profile( self, delta, mode='additive' ):

        #update the albedo
        if mode=='additive':
            self.atmosphere.albedo += delta[0]
        else:
            self.atmosphere.albedo *= delta[0]
        
        #albedo is a double so dont use logical indexing
        if self.atmosphere.albedo > 1:
            self.atmosphere.albedo = 1
        if self.atmosphere.albedo < 0:
            self.atmosphere.albedo = 0
        
        
    def get_atmospheric_profile( self ):
        return self.atmosphere.albedo
        
    def set_atmospheric_profile(self, profile):
        self.atmosphere.albedo = profile
        
    def output_data( self ):
        return self.atmosphere.albedo
        
    def lm_weighting_matrix( self, inst ):
        
        #get the tangent altitude of the measurements
        los, obs = inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )
        
         
        W = np.zeros((len(self.tanalts),len(self.triplet) ))        
        #interpolate onto tangent altitudes
        for i in range(0,len(self.tanalts)):
            for j in range(0,len(self.triplet)):
                W[i,j] = 1
                
        #reshape for lm retrieval
        # return np.reshape(W.transpose(),(np.size(W)))
        return np.array([[1]])

    def mart_weighting_matrix( self, inst ):
        return None

    def check_convergence(self, y_ret, previous_y_ret):

        converged = False

        if previous_y_ret is not None:
            step_perc = np.abs(y_ret - previous_y_ret) / y_ret * 100
            if np.all( step_perc < self.min_step_size):
                converged = True

        return converged