from retrieval.species_no2 import NO2LimbProfile
import os
import logging

import matplotlib.pyplot as plt
import numpy as np
from scipy import io
from scipy import signal
import radtran.geometry as geo
import copy

class NO2LimbProfileDOAS(NO2LimbProfile):
    def __init__(self, atmosphere):
        self.ident = 'limb_no2'

        self.wavel_range = np.array([434.7,476.72])

        #This defines the normalization altitudes used by the measurement vector
        #Any altitudes within the range [begin, end] are used.
        self.normalt = []
        self.normalt.append( ([50000.0,70000.0]))

        #The actual wavelengths that are used in the retrieval.
        self.wavel_nm = None

        #get the ozone profile from the atmosphere
        self.profile_alts,self.profile = atmosphere.get_user_defined( self.ident )
        self.atmosphere = atmosphere

        #dont use values between the first normalization point and refaltrange below
        self.refaltrange = 0.000

        #only retrieve up to a maximum of max_alt meters. However, if no measurement
        #vectors exist below this point retrieve only to the highest valid measurement
        #vector
        self.max_alt     = 48000

        #set the lower bound of the retrieval
        self.min_alt     = 8000

        #compute the jacobians using analytic weighting functions (True) or
        #numerically (False)
        self.use_wf = True
        self.jacobian_scatter_order = 10

        self.good_alts   = None
        self.poly_order  = 5        # 4 --> a_0 + a_1*x + a_2*x^2 + a_3*x^3
        self.optprop_o3  = None
        self.optprop_no2 = None

        self.optprop_o3  = atmosphere.user_optprop['limb_ozone']
        self.optprop_no2 = atmosphere.user_optprop['limb_no2']
        self.background_clim = atmosphere.backgroundclim

        self.doas_meas_vec = True
        self.tanalts = None
        self.is_setup = False

        self.ms_target_wavel = [ 430.92,  432.52,  435.12,  437.78,  439.28,  441.82,  444.84,  446.66,
                             448.06,  451.3,   452.44,  454.06,  454.82,  456.66,  457.9,   459.18,
                             460.44,  461.94,  463.22,  465.78,  467.1 ,  472.6,   474.04,  478.2  ]


    def get_required_wavelengths(self):
        return self.wavel_range

    def configure_instrument_and_method( self, ret_inst, meas_inst, method ):
        """
        Sets up the retrieval method, and retrieval instrument. Uses wavelengths
        from the measurement instrument.
        """

        self.is_setup = True
        #setup wavelengths for the retrieval
        temp_wavel = self.get_required_wavelengths()
        if isinstance(meas_inst.wavel_nm, list):
            meas_wavel = meas_inst.wavel_nm[0]
        else:
            meas_wavel = meas_inst.wavel_nm
        self.wavel_nm = meas_wavel[(meas_wavel > temp_wavel[0]) & (meas_wavel < temp_wavel[1])]
        ret_inst.set_wavelengths( self.wavel_nm )
        #get the tangent altitude of the measurements
        los, obs = ret_inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )
        self.lat, self.lon,temp = geo.lat_lon_alts( self.tp )
        self.tanalts = np.asarray(self.tanalts)

        wavel_idx = np.ones(len(self.ms_target_wavel))
        for idx,wt in enumerate(self.ms_target_wavel):
            wavel_idx[idx] = np.argmin(np.abs(wt - self.wavel_nm))

        wavel_idx = np.unique(wavel_idx).astype(int)

        #set the multiscatter wavelengths if were using them
        try:
            ret_inst.radtran.set_wavelengths(self.wavel_nm)
            ret_inst.radtran.set_multiscatter_wavelength_indicies(wavel_idx)
        except AttributeError:
           logging.info('could not set multiscatter wavelengths')

        #all measurements used equally
        if method.type == 'LM':
            W = self.lm_weighting_matrix( ret_inst )
        elif method.type == 'MART':
            W = self.mart_weighting_matrix(ret_inst )

        method.W = W
        method.user_defined_weight = True

        #setup the cross section used in the doas fitting
        temp_wavel = self.get_required_wavelengths()
        good_wavel = (meas_wavel > temp_wavel[0]) & (meas_wavel < temp_wavel[1])

        self.mjd = meas_inst.mjd

    def measurement_vector( self, inst, do_error=False ):
        """
        Compute the ozone measurement vector

        Parameters
        ----------
        inst : instrument class
            Instance of an instrument, either real or simulated that contains
            radiances.

        Return
        ------
        y: 1d array
            1 dimensional array containing the stacked measuremrent vectors
        dy: 1d array
            1 dimensional array containing the stacked measuremrent vector errors
        """

        if not self.is_setup:
            self.is_setup = True
            temp_wavel = self.get_required_wavelengths()
            if isinstance(inst.wavel_nm, list):
                meas_wavel = inst.wavel_nm[0]
            else:
                meas_wavel = inst.wavel_nm

            self.wavel_nm = meas_wavel[(meas_wavel > temp_wavel[0]) & (meas_wavel < temp_wavel[1])]
            los, obs = inst.get_meas_geometry()
            self.tp, self.tanalts = geo.tangent_points( obs, los )
            self.lat, self.lon,temp = geo.lat_lon_alts( self.tp )
            self.tanalts = np.asarray(self.tanalts)
            self.mjd = inst.mjd

        rad = np.transpose(inst.rad)
        if do_error:
            rad_error = np.transpose(inst.rad_error)
        else:
            rad_error = np.zeros(np.shape(rad))

        # rad is now [wavel, los]
        y  = np.zeros((1, np.shape(rad)[1]))
        dy = np.zeros((1, np.shape(rad)[1]))
        # y is now [triplet, los]

        if isinstance(inst.wavel_nm, list):
            meas_wavel = inst.wavel_nm[0]   #assumes wavelength doesnt change between lines of sight
        else:
            meas_wavel = inst.wavel_nm

        temp_wavel = self.get_required_wavelengths()
        good_wavel = (meas_wavel > temp_wavel[0]) & (meas_wavel < temp_wavel[1])

        ref_alts = (self.tanalts > self.normalt[0][0]) & (self.tanalts < self.normalt[0][1])
        if np.sum(ref_alts) > 1:
            rad_ref = np.nanmean(rad[good_wavel][:,ref_alts] / np.nanmean(rad[good_wavel][:,ref_alts],axis=0),axis=1)
        else:
            rad_ref = rad[good_wavel, ref_alts] / np.nanmean(rad[good_wavel, ref_alts])

        norm_rad = np.log(rad[good_wavel,:] / rad_ref[:,np.newaxis])
        # norm_rad = rad[good_wavel,:] / rad_ref[:,np.newaxis]
        y[0], dy[0] = self.regression_fit(self.wavel_nm, norm_rad)

        if self.good_alts is None:
            below_min = self.tanalts < self.min_alt
            y[0,below_min] = np.nan
            above_norm = self.tanalts > self.normalt[0][0] - self.refaltrange
            y[0,above_norm] = np.nan

            np.seterr(divide='ignore',invalid='ignore')
            y[0,np.abs(y[0,:]) < dy[0,:]] = np.nan
            np.seterr(divide='ignore',invalid='ignore')

            #set all values above first nan to nan
            nogood = (self.tanalts > 30000) & (np.isnan(y[0]))
            if np.any(nogood):
                first_nogood = np.where(nogood)[0][0]
                y[0,first_nogood::] = np.nan

            self.good_alts = np.any(~np.isnan(y), axis=0)
        else:
            y[0, ~self.good_alts] = np.nan

        return y.flatten(), dy.flatten()

    def measurement_vector_jacobian_wf(self,inst):

        rad = np.transpose(inst.rad)
        #setup the cross section used in the doas fitting
        if isinstance(inst.wavel_nm, list):
            meas_wavel = inst.wavel_nm[0]
        else:
            meas_wavel = inst.wavel_nm

        temp_wavel = self.get_required_wavelengths()
        good_wavel = (meas_wavel > temp_wavel[0]) & (meas_wavel < temp_wavel[1])

        ref_alts = (self.tanalts > self.normalt[0][0]) & (self.tanalts < self.normalt[0][1])
        if np.sum(ref_alts) > 1:
            rad_ref = np.nanmean(rad[good_wavel][:,ref_alts] / np.nanmean(rad[good_wavel][:,ref_alts],axis=0),axis=1)
        else:
            rad_ref = rad[good_wavel, ref_alts] / np.nanmean(rad[good_wavel, ref_alts])

        wf =  inst.wf
        # wf is [wavel, los, pert]
        # Rad is [wavel, los]
        # Jacobian has to be [wavel*los,pert]
        J = np.zeros( (np.shape(wf)[1], np.shape(wf)[2]) )

        for idx in range(0, len(self.tanalts)):

            norm_rad = wf[good_wavel,:,idx] / rad[good_wavel,idx][:,np.newaxis]
            # norm_rad = wf[good_wavel,:,idx] / rad_ref[:,np.newaxis]
            y,dy = self.regression_fit(self.wavel_nm, norm_rad)
            J[:,idx] = y

        J = J[:,self.good_alts]

        return J

    def measurement_vector_jacobian_num(self,inst):

        """
        Computes the measurement vector Jacobian through numerical perturbations.
        This is both slower and less accurate (assuming single scatter) than the
        analytic method, but is included incase the radiative transfer model does
        not support analytic jacobians.
        """

        #store the initial state
        temp_rad    = copy.copy(inst.rad)
        num_sctr    = copy.copy(inst.radtran.options.optmap['numordersofscatter'])
        alts, no2_0 = copy.copy(self.atmosphere.get_user_defined(self.ident))

        #find altitudes to perturb and initialize J
        pert_alts   = self.tanalts[self.good_alts]
        pert_alts   = np.concatenate(([pert_alts[0]-np.diff(pert_alts)[0]], pert_alts, [pert_alts[-1]+np.diff(pert_alts)[-1]]))
        #do single scatter jacobians
        inst.set_num_orders_of_scatter( self.jacobian_scatter_order )
        inst.calculate_radiance()
        y_0,temp = self.measurement_vector( inst )

        J = np.zeros((np.size(y_0),np.size(pert_alts)-2))

        num_wf = np.zeros(inst.wf.shape)
        for i in range(1,len(pert_alts)-1):

            #interpolate increase between lines of sight
            pert = np.ones(len(pert_alts))
            pert[i] = 1.01
            no2_1 = np.interp(alts,pert_alts,pert)*no2_0
            no2_diff = np.interp(pert_alts[i], alts, no2_1-no2_0)
            self.atmosphere.modify_user_defined(self.ident, no2_1)

            #calculate new measuerment vector
            inst.calculate_radiance()
            y_1,temp = self.measurement_vector( inst )
            inst.rad = temp_rad

            #find numerical jacobian
            J[:,i-1] = (y_1-y_0)/no2_diff

        #reset variables
        inst.rad = temp_rad
        self.atmosphere.modify_user_defined(self.ident, no2_0)
        inst.set_num_orders_of_scatter( num_sctr )

        return J

    def regression_fit(self, wavel, norm_rad):

        #regression polynomials
        x = wavel
        tx = np.zeros((0,len(wavel)))
        for p in range(self.poly_order):
            tx = np.vstack(( tx, (x-np.mean(x))**p ))
        tx[0,:] = 1.0

        xsec_o3  = np.zeros((1,len(wavel)))
        xsec_no2 = np.zeros((1,len(wavel)))
        y  = np.zeros(np.shape(norm_rad)[1])
        dy = np.zeros(np.shape(norm_rad)[1])

        self.optprop_o3.SetAtmosphericState(self.background_clim)
        self.optprop_no2.SetAtmosphericState(self.background_clim)
        rad_pred = np.zeros(norm_rad.shape)*np.nan
        rad_meas = np.zeros(norm_rad.shape)*np.nan
        for idx in range(np.shape(norm_rad)[1]):

            #don't regression fit nans
            good = ~np.isnan(norm_rad[:,idx])

            #set the location in the atmosphere to the tangent point so that we get the
            #cross sections at the correct temperature
            self.optprop_o3.SetLocation(  [self.lat[idx], self.lon[idx], self.tanalts[idx], self.mjd[idx]] )
            self.optprop_no2.SetLocation( [self.lat[idx], self.lon[idx], self.tanalts[idx], self.mjd[idx]] )

            #get the cross sections at each wavelength (CrossSectionsArray currently broken)
            for widx,w in enumerate(wavel):
                xsec_o3[0,widx] = self.optprop_o3.CalculateCrossSections(1e7/w)[1]
                xsec_no2[0,widx]= self.optprop_no2.CalculateCrossSections(1e7/w)[1]

            #create mean centered regression matrix
            t = np.vstack((tx,xsec_o3,xsec_no2))
            for i in range(1,t.shape[0]):
                t[i,:] = (t[i,:]-np.mean(t[i,:]))/np.std(t[i,:])

            #compute the regression coefficients and select the one for no2
            # multiply by -1 since no2 absorbs (decreases radiance)
            # b = np.dot( np.dot( np.linalg.inv(np.dot( t, t.T )), t ), norm_rad[:,idx])
            b, res, rank, s = np.linalg.lstsq(t[:,good].T, norm_rad[good,idx])
            # rad_pred[:,idx] = np.dot(b[-1], t[-1,:])
            # rad_meas[:,idx] = norm_rad[:,idx] - np.dot(b[0:-1],t[0:-1,:])
            rad_pred[:,idx] = np.dot(b,t)
            y[idx] = -b[-1]
            #one-sided 95% confidence interval on the regression coefficient
            dy[idx] = 0.5*1.96*np.sqrt(  np.linalg.inv(np.dot(t[:,good],t[:,good].T))[-1,-1]*(1/(len(wavel[good])-self.poly_order-2))*np.sum((rad_pred[good,idx] - norm_rad[good,idx])**2) )

        return y, dy

    def lm_weighting_matrix( self, inst ):

        #setup the cross section used in the doas fitting
        if isinstance(inst.wavel_nm, list):
            meas_wavel = inst.wavel_nm[0]
        else:
            meas_wavel = inst.wavel_nm

        temp_wavel = self.get_required_wavelengths()
        good_wavel = (meas_wavel > temp_wavel[0]) & (meas_wavel < temp_wavel[1])

        n = len(self.tanalts)
        # m = np.sum(good_wavel)
        m = 1

        return np.ones(n*m)

    def mart_weighting_matrix(self, inst ):
        """
        Compute the relative weights of the measurements for MART fitting.
        """
        #load the weighting matrix from file
        dir_path = os.path.dirname( os.path.abspath(__file__) )
        wik = io.loadmat(dir_path + r'/../utility/data/no2/wik.mat')['wik']
        wij = io.loadmat(dir_path + r'/../utility/data/no2/wij.mat')['wij']

        # set up J matrix for line of sight weighting
        n = len(self.tanalts)
        m = len(wij)
        J = np.eye(n,n)
        J = signal.convolve2d(J,np.fliplr(wij))[:,len(wij[0])-1::]
        J = J/(np.sum(J,axis=1)[:,np.newaxis])

        # set up L matrices for spectral weighting and build H
        H = np.zeros((len(self.tanalts), 0))
        for k in range(np.shape(wik)[1]-1):#k=1:obj.numnlets
            L = np.diag(np.interp(self.tanalts, wik[:,0]*1000,wik[:,k+1]))
            H = np.hstack((H, np.dot(L,J)))

        H = H/(np.sum(H,axis=1)[:,np.newaxis])
        H[np.isnan(H)] = 0.0

        return H