import copy as copy
import os
import logging

import numpy as np
from scipy import io
from scipy import signal

import retrieval.species as species
import radtran.geometry as geo


class OzoneLimbProfile(species.Species):
    """
    Handles the retrieval specifics for the ozone species including configuration
    of retrieval method, measurement vector and jacobian calculations and updating
    of atmoshperic state with new ozone profile.
    
    The ozone measurement vector are triplets normalized by high altitude to help
    decrease senstivity to albedo and instrument calibration.
    """
    
    def __init__(self, atmosphere):
        
        self.ident = 'limb_ozone'
        
        #This defines the doublets/triplets used to create measurement vectors.
        #first row is the wavelengths, second row is the weights, so for the first
        #doublet we would have:
        #
        #   y = -1*log( Radiance at 292.43 ) + 1*log( Radiance at 292.43 )
        #
        #definitions can contain as many doublets/triplets/quads/etc as desired,
        #the only conidtion being that each wavelength must have a weight and the
        #weights should sum to zero (although this is not checked explicitly).
        self.triplet = []
        self.triplet.append( ([292.43, 350.31], [-1,1]))
        self.triplet.append( ([302.17, 350.31], [-1,1]))
        self.triplet.append( ([306.06, 350.31], [-1,1]))
        self.triplet.append( ([310.70, 350.31], [-1,1]))
        self.triplet.append( ([315.82, 350.31], [-1,1]))
        self.triplet.append( ([322.00, 350.31], [-1,1]))
        self.triplet.append( ([331.09, 350.31], [-1,1]))
        self.triplet.append( ([543.84, 602.39, 678.85],[0.5,-1,0.5]) )
        #NOTE: if a weighting matrix is being used in the retrieval code a weight
        #should exist for each triplet.
        
        #This defines the normalization altitudes used by each measurement vector
        #Any altitudes within the range [begin, end] are used.
        #self.normalt should be the same length as self.triplet    
        self.normalt = []
        self.normalt.append( ([59000.0,64000.0]))
        self.normalt.append( ([55000.0,60000.0]))
        self.normalt.append( ([51000.0,56000.0]))
        self.normalt.append( ([48000.0,53000.0]))
        self.normalt.append( ([46000.0,51000.0]))
        self.normalt.append( ([42000.0,47000.0]))
        self.normalt.append( ([39000.0,44000.0]))
        self.normalt.append( ([30000.0,35000.0]))

        self.wik_file = '/../utility/data/ozone/wik_lm.mat'
        
        #The actual wavelengths that are used in the retrieval.
        self.wavel_nm = None
        
        #get the ozone profile from the atmosphere
        self.profile_alts,self.profile = atmosphere.get_user_defined( self.ident )
        self.atmosphere = atmosphere
        
        #dont use values between the first normalization point and refaltrange below
        self.refaltrange = 0.000
        
        #only retrieve up to a maximum of max_alt meters. However, if no measurement
        #vectors exist below this point retrieve only to the highest valid measurement
        #vector
        self.max_alt     = 64000
        
        #set the lower bound of the retrieval
        self.min_alt     = 10000
        
        #compute the jacobians using analytic weighting functions (True) or
        #numerically (False)
        self.use_wf = True

        #convergence criteria - if measurement vector changes by < x% at all good
        #altitudes it has converged
        self.min_step_size = 0.5

        # Initialize good altitudes, indexes of normalization altitudes
        self.good_alts = None
        self.normalt_idx = None

        #if computing the numeric jacobian use this many orders of scatter
        self.jacobian_scatter_order = 1
        
    def get_required_wavelengths(self):
        
        """
        Calculates the suggested wavelengths for the retrieval
        
        Returns
        -------
        wavel : numpy array
            Wavelengths suggested for the retrieval
        """
    
        #find the unique wavelengths used by the triplets
        wavel_nm = []
        for triplet in self.triplet:
            wavel_nm.append( triplet[0] )
        wavel_nm = [item for sublist in wavel_nm for item in sublist]
        wavel_nm = np.unique(np.array(wavel_nm))
        return wavel_nm
        
    def configure_instrument_and_method( self, ret_inst, meas_inst, method ):
        """
        Sets up the retrieval method, and retrieval instrument. Uses wavelengths
        from the measurement instrument.
        """
        
        # Get the wavelengths we need for our triplet
        # and store it as an np array
        temp_wavel = self.get_required_wavelengths()


        if isinstance(meas_inst.wavel_nm, list):
            meas_wavel = meas_inst.wavel_nm[0]
        else:
            meas_wavel = meas_inst.wavel_nm
            
        #find the closest wavelengths in the meas_inst and use those
        self.wavel_nm = copy.copy(temp_wavel)
        for i in range(len(temp_wavel)):
            idx = np.argmin(np.abs(temp_wavel[i] - meas_wavel))
            self.wavel_nm[i] = meas_wavel[idx]
            
        ret_inst.set_wavelengths( self.wavel_nm )
        
        # Get the tangent altitudes of the measurements
        los, obs = ret_inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )

        if method.type == 'LM':
            W = self.lm_weighting_matrix( ret_inst, self.measurement_vector(meas_inst)[0] )
        elif method.type == 'MART':
            W = self.mart_weighting_matrix(ret_inst, self.measurement_vector(meas_inst)[0])
        
        method.W = W
        method.user_defined_weight = True
            
    def measurement_vector( self, inst, do_error=False ):
        """
        Compute the ozone measurement vector
        
        Parameters
        ----------
        inst : instrument class
            Instance of an instrument, either real or simulated that contains
            radiances.
        
        Return
        ------
        y: 1d array
            1 dimensional array containing the stacked measuremrent vectors
        dy: 1d array
            1 dimensional array containing the stacked measuremrent vector errors
        """
        rad = np.transpose(inst.rad)
        if do_error:
            rad_error = np.transpose(inst.rad_error)
        else:
            rad_error = np.zeros(np.shape(rad))
            
        # rad is now [wavel, los]
        y  = np.zeros((len(self.triplet), np.shape(rad)[1]))
        dy = np.zeros((len(self.triplet), np.shape(rad)[1]))
        # y is now [triplet, los]


        # Find the normalization altitude(s) for the measurement vectors
        # This calculation is done only once with the measurement instrument
        # We're then sure that the normalts will be the same for each instrument throughout the retrieval
        # It's done with the measurement instrument to get rid of the eventual nan in its radiance (radiation hit etc...)
        if self.normalt_idx is None:

            # Initialize the list of indexes
            self.normalt_idx = list()

            # Loop on the triplet(s)
            for idx,triplet in enumerate(self.triplet):

                # First we determine an array of indexes where the altitudes are between the boundaries of self.normalt[idx]
                self.normalt_idx.append(np.nonzero( np.logical_and(np.asarray(self.tanalts) > self.normalt[idx][0],np.asarray(self.tanalts) < self.normalt[idx][1]) )[0])

                # Loop inside the triplet on wavelengths
                for wavel in triplet[0]:

                    # For each wavelength the radiance at the normalization altitude is checked for nans.
                    # If a nan is found the corresponding altitude is removed from the array
                    # The final normalization altitude(s) array for a measurement vector is the intersection of each
                    # array found per wavelength
                    # With that we assure that the normalization altitudes are the same for:
                    #   - 2 wavelengths in the same triplet
                    #   - The measurement instrument and the retrieval instrument
                    if self.normalt_idx[idx].size != 0:
                        # Find the closest wavelength and get the radiance
                        wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))
                        radI = rad[wave_idx,:]

                        # Remove the nans found at this wavelength
                        self.normalt_idx[idx] = self.normalt_idx[idx][~np.isnan(radI[self.normalt_idx[idx]])]

                # Throw a warning if no normalization altitude could be found
                if self.normalt_idx[idx].size == 0:
                    logging.warning("\tNo normalization altitude could be found for ozone measurement vector n°{:d}.".format(idx+1))
                    logging.warning("\tThere seems to be no good altitude(s) between {:2.3f}m and {:2.3f}m for this scan.".format(self.normalt[idx][0],self.normalt[idx][1]))
        # End normalization altitude(s) calculation

        # Calculate the measurement vectors
        # Loop on the triplets
        for idx,triplet in enumerate(self.triplet):

            # Measurement vector and error are nan if no normalization altitude was found
            if (self.normalt_idx[idx].size == 0):
                y[idx,:]  = np.nan
                dy[idx,:] = np.nan
            else:
                # Loop inside the triplet on wavelength and coefficient
                for wavel, coeff in zip(triplet[0],triplet[1]):

                    # Find the closest wavelength and get the radiance
                    wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))
                    radI = rad[wave_idx,:]
                    errI = rad_error[wave_idx,:]

                    # Normalization radiance and error are nan if all the normalization radiances are nan
                    if np.all(np.isnan(radI[self.normalt_idx[idx]])):
                        radR = np.nan
                        errR = np.nan
                    else:
                        # Compute the normalization radiance and error
                        radR = np.nanmean(radI[self.normalt_idx[idx]])
                        errR = np.nanmean(errI[self.normalt_idx[idx]])

                    # Compute the measurement vector and the error.
                    # Zero and negative values in the radiance are naned out first
                    radI[np.nan_to_num(radI) <= 1e-16] = np.nan
                    y[idx,:] += coeff * np.log(radI) - coeff * np.log(radR)
                    dy[idx,:]+= (coeff * errI/radI)**2 + (coeff * errR/radR)**2 #sum in quadrature

                # Remove points above 1st reference altitude
                y[idx,np.asarray(self.tanalts) >= self.tanalts[self.normalt_idx[idx][0]] - self.refaltrange] = np.nan
                dy[idx,np.asarray(self.tanalts) >= self.tanalts[self.normalt_idx[idx][0]] - self.refaltrange] = np.nan

            # Remove the points below minimum altitude
            below_min = np.asarray(self.tanalts) < self.min_alt
            y[idx,below_min] = np.nan

        # Check the measurement vectors that are all nan and throw a warning
        if np.sum(np.all(np.isnan(y),axis=1)) > 0:
            logging.warning("\t{:d} ozone measurement vector(s) are entirely nan, results unreliable.".format(np.sum(np.all(np.isnan(y),axis=1))))

        # Compute the final error
        dy = np.sqrt(dy)

        # The "good altitudes" are those where at least one of the measurement vectors is not NaN
        # It is set only once when using the measurement instrument to calculate the measurement vector y_meas
        if self.good_alts is None:
            self.good_alts = np.any(~np.isnan(y), axis=0)
        return y.flatten(), dy.flatten()

        
    def measurement_vector_jacobian( self, inst ):
        
        if self.use_wf:
            J = self.measurement_vector_jacobian_wf( inst )
        else:
            J = self.measurement_vector_jacobian_num( inst )
            
        return J
        
    def measurement_vector_jacobian_wf( self, inst ):
        """
        Computes the measurement vector Jacobian using weighting functions from the HR model
        """
        wf =  inst.wf 
        rad = np.transpose(inst.rad)
        # wf is [wavel, los, pert]
        # Rad is [wavel, los]
        # Jacobian has to be [triplet*los,pert]
        J = np.zeros( (len(self.triplet), np.shape(wf)[1], np.shape(wf)[2]) )
        for idx,triplet in enumerate(self.triplet):
            for wavel, coeff in zip(triplet[0],triplet[1]):
                
                #find the weighting function and radiance for this triplet wavelength
                wave_idx = np.argmin(np.abs(np.unique(inst.wavel_nm)-wavel))
                wfI = wf[wave_idx,:,:]
                radI = rad[wave_idx,:]
                
                #get the normalization radiance and weighting function
                # alt_idx     = np.nonzero( np.logical_and(np.asarray(self.tanalts) > self.normalt[idx][0],np.asarray(self.tanalts) < self.normalt[idx][1])   )[0]
                # radR = np.nanmean(radI[alt_idx])
                # wfNorm = np.nanmean(wfI[alt_idx,:])

                #compute the jacobian element
                J[idx,:,:] += coeff * wfI / radI[:,np.newaxis]#- coeff * wfNorm / radR

        # Combine the first two axis in J
        J = np.reshape( J.flatten(), (len(self.triplet)*np.shape(wf)[1], np.shape(wf)[2]) )
        
        J = J[:,self.good_alts]
        
        return J
        
    def measurement_vector_jacobian_num( self, inst ):
        """
        Computes the measurement vector Jacobian through numerical perturbations.
        This is both slower and less accurate (assuming single scatter) than the 
        analytic method, but is included incase the radiative transfer model does
        not support analytic jacobians.
        """
        #store the initial state
        temp_rad    = copy.copy(inst.rad)
        num_sctr    = copy.copy(inst.radtran.options.optmap['numordersofscatter'])
        alts, o3_0 = copy.copy(self.atmosphere.get_user_defined(self.ident))
        
        #find altitudes to perturb and initialize J
        pert_alts   = np.asarray(self.tanalts)[self.good_alts]
        pert_alts   = np.concatenate(([pert_alts[0]-(pert_alts[1]-pert_alts[0])], pert_alts, [pert_alts[-1]+(pert_alts[-1]-pert_alts[-2])]))

        #do single scatter jacobians
        inst.set_num_orders_of_scatter( self.jacobian_scatter_order )
        inst.calculate_radiance()
        y_0,temp = self.measurement_vector( inst )
        
        J = np.zeros((np.size(y_0),np.size(pert_alts)-2))

        for i in range(1,len(pert_alts)-1):

            pert_tanalts     = np.ones(len(pert_alts))
            pert_tanalts[i] *= 1.01
            o3_1             = o3_0 * np.interp(self.profile_alts,pert_alts,pert_tanalts)
            do3              = np.interp(pert_alts[i], self.profile_alts, o3_0)*0.01
            self.atmosphere.modify_user_defined(self.ident, o3_1)

            #calculate new measuerment vector
            inst.calculate_radiance()
            y_1,temp = self.measurement_vector( inst )
            inst.rad = temp_rad
            
            #find numerical jacobian
            J[:,i-1] = (y_1-y_0)/do3

        #reset variables
        inst.rad = temp_rad
        self.atmosphere.modify_user_defined(self.ident, o3_0)
        inst.set_num_orders_of_scatter( num_sctr )
        
        return J
        
    def update_profile( self, delta, mode='additive' ):
        """
        Updates the ozone profile used by the radiative transfer model based on
        recommended step size, delta. Additional constraints can be imposed on delta,
        such as only using certain altitudes, what to do above/below retrieval limits,
        and maximum step size.
        
        Parameters
        ----------
        delta : 1d array
            An array of numbers on the retrieval grid (defined by the jacobian calculation)
            in units of particles per cm^3.
        
        """
        
        #interpolate the profile onto the good tangent altitudes. Above/below limts
        #the values are extrapolated
        # ret_profile = np.interp(np.asarray(self.tanalts)[self.good_alts], self.profile_alts, self.profile)
        
        #interpolate delta onto retrieval grid
        # delta = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], (ret_profile+delta)/ret_profile)

        #interpolate the delta onto the retrieval grid
        ret_profile = np.interp(np.asarray(self.tanalts)[self.good_alts], self.profile_alts, self.profile)

        if mode == 'additive':
            delta = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts], (ret_profile+delta)/ret_profile)
        else:
            delta = np.interp(self.profile_alts, np.asarray(self.tanalts)[self.good_alts & ~np.isnan(delta)], delta[self.good_alts & ~np.isnan(delta)])

        min_alt     = np.min(np.asarray(self.tanalts)[self.good_alts])
        max_alt     = np.max(np.asarray(self.tanalts)[self.good_alts])
        min_alt_idx = np.min(np.where(self.profile_alts > min_alt))
        max_alt_idx = np.max(np.where(self.profile_alts < max_alt))+1
        delta[self.profile_alts > max_alt] = delta[max_alt_idx]
        delta[self.profile_alts < min_alt] = delta[min_alt_idx]

        #update profile (dont allow it change by more than a factor of 5/20)
        delta[delta<0.2]    = 0.2
        delta[delta>20.0]   = 20.0
        self.profile *= delta
        # self.profile[self.profile>self.max_val] = self.max_val
        self.atmosphere.modify_user_defined( self.ident, self.profile )
        
    def get_atmospheric_profile(self):
        
        # query the atmospheric state for the ozone density
        alt, x=self.atmosphere.get_user_defined(self.ident)
        return copy.copy(x)

    def set_atmospheric_profile(self, profile):
        # set the ozone density
        self.atmosphere.modify_user_defined(self.ident, profile)

    def get_x(self):
        profile = self.get_atmospheric_profile()
        pert_alts = np.asarray(self.tanalts)[self.good_alts]
        x = np.interp(pert_alts, self.profile_alts, profile)
        return x

    def get_retrieval_alts( self ):

        return self.profile_alts
        
    def output_data( self ):

        #return the ozone profile
        return self.profile

    def output_data_bounded( self ):

        min_alt = np.min(np.asarray(self.tanalts)[self.good_alts])
        max_alt = np.max(np.asarray(self.tanalts)[self.good_alts])
        p = copy.copy(self.profile)
        p[(self.profile_alts > max_alt) | (self.profile_alts < min_alt)] = np.nan
        #return the ozone profile
        return p
        
    def lm_weighting_matrix( self, inst, y_meas ):
        """
        Compute the relative weights of the measurements for least squares fitting.
        """
        #get the tangent altitude of the measurements
        los, obs = inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )
        
         
        W = np.zeros((len(self.tanalts),len(self.triplet) ))

        #load the weighting matrix from file
        dir_path = os.path.dirname( os.path.abspath(__file__) )        
        wik = io.loadmat(dir_path + self.wik_file)['wik']
        
        #interpolate the weighting matrix onto tangent altitudes
        for i in range(0,len(self.tanalts)):
            for j in range(0,len(self.triplet)):
                W[i,j] = np.interp( self.tanalts[i], wik[:,0]*1000.0, np.double(wik[:,j+1])  )
                
        #reshape for lm retrieval
        # W = np.reshape(W.transpose(),(np.size(W)))
        W = np.reshape(W.transpose(),(np.size(W))) / y_meas
        W[np.isnan(W)] = 0.0
        
        return W

    def mart_weighting_matrix(self, inst, y_meas):
        """
        Compute the relative weights of the measurements for MART fitting.
        """
        #get the tangent altitude of the measurements
        los, obs = inst.get_meas_geometry()
        self.tp, self.tanalts = geo.tangent_points( obs, los )

        #load the weighting matrix from file
        dir_path = os.path.dirname( os.path.abspath(__file__) )
        wik = io.loadmat(dir_path + r'/../utility/data/ozone/wik_lm.mat')['wik']
        wij = io.loadmat(dir_path + r'/../utility/data/ozone/wij.mat')['wij']

        # set up J matrix for line of sight weighting
        n = len(self.tanalts)
        m = len(wij[0])
        J = np.eye(n,n)
        J = signal.convolve2d(J,np.fliplr(wij))[:,m-1::]
        J = J/(np.sum(J,axis=1)[:,np.newaxis])

        # set up L matrices for spectral weighting and build H
        H = np.zeros((len(self.tanalts), 0))
        for k in range(np.shape(wik)[1]-1):#k=1:obj.numnlets
            L = np.diag(np.interp(self.tanalts, wik[:,0]*1000,wik[:,k+1]))
            H = np.hstack((H, np.dot(L,J)))

        np.seterr(divide='ignore',invalid='ignore') #we'll deal with nan's right away
        H = H/(np.sum(H,axis=1)[:,np.newaxis])
        np.seterr(divide=None,invalid=None)
        H[np.isnan(H)] = 0.0
        return H

    def check_convergence(self, y_ret, previous_y_ret):

        converged = False

        if previous_y_ret is not None:
            good = ~np.isnan(y_ret)
            step_perc = np.abs(y_ret[good] - previous_y_ret[good]) / y_ret[good] * 100
            if np.all( step_perc < self.min_step_size):
                converged = True

        return converged