import copy as copy

import numpy as np
import matplotlib.pyplot as plt

import retrieval.core as retrieval
import retrieval.instruments.limb_profiler as instrument
import radtran.scan as scan
import radtran.atmosphere as atmosphere
import radtran.climatologies.ozone as ozone
import radtran.climatologies.utility as clim_util
import radtran.options as options
import sasktranif.sasktranif as skif
import retrieval.method.lsq as lsq
import retrieval.species_o3 as species
import radtran.sasktran as radtran



# Set up the geometry for both instruments
# Scan isn't really part of the retrieval package so we have to do some 
# conversions here for now

latitude = 40
longitude = 20
saa = 60
sza = 80
mjd = 54372.0
tp_alt = 25

sim_scan = scan.Scan()
sim_scan.make_scan_saa( sza, saa, latitude, longitude, np.arange(7.5,60.5,1), mjd, tp_alt )
los = sim_scan.look
obs = sim_scan.obs

mjd_list = [0,]*len(los)
for i in range(len(los)):
    mjd_list[i] = np.array([mjd], np.double)

# TODO: The species should be setting wavel_nm in the ret_inst
wavel_nm_ret = np.array([292.43,302.17,306.06,310.7,315.82,322.00,331.09,350.31,543.84,602.39,678.85])

inst_sim = instrument.LimbProfiler_scan( wavel_nm_ret, los, None,  obs, mjd_list )
inst_sim.initialize_sensors_and_optics()

inst_ret = instrument.LimbProfiler_scan( wavel_nm_ret, los, None,  obs, mjd_list )
inst_ret.initialize_sensors_and_optics()

#setup atmospheric state for simulator
atmo_sim = atmosphere.Atmosphere()
atmo_sim.add_species( skif.ISKClimatology('ECMWF'),
                  skif.ISKOpticalProperty('Rayleigh'),
                  skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3 )
clim_lab, optprop_o3, guid_o3 = ozone.labow()
atmo_sim.set_wf_species( optprop_o3 )
atmo_sim.albedo = 0.3

#setup true and apriori ozone profiles
apriori_alts    = np.arange(500.0, 101500.0, 1000)
apriori_profile = clim_util.height_profile_from_clim(clim_lab, guid_o3, apriori_alts, latitude, longitude, mjd )
plt.plot(apriori_profile,apriori_alts, color=(0.7,0.4,0.2))

true_profile = apriori_profile*1.1
atmo_sim.add_user_defined( guid_o3, optprop_o3, apriori_alts, copy.copy(true_profile), 'limb_ozone' )

#setup retrieved atmospheric state
atmo_ret = atmosphere.Atmosphere()
atmo_ret.add_species( skif.ISKClimatology('ECMWF'),
                  skif.ISKOpticalProperty('Rayleigh'),
                  skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3 )
atmo_ret.add_user_defined( guid_o3, optprop_o3, apriori_alts, copy.copy(apriori_profile), 'limb_ozone' )
atmo_ret.set_wf_species( optprop_o3 )
atmo_ret.albedo = 0.3

atmo_ret.get_user_defined('limb_ozone')
# And radiative transfer models
ret_options = options.HROptions()
ret_options.add_option('calcwf', 1 )
ret_options.add_option('numordersofscatter', 1)
ret_options.add_option('setsun', sim_scan.sun.tolist() )
ret_options.add_option('wfprecision', 0)

sim_options = options.HROptions()
sim_options.add_option('calcwf', 1 )
sim_options.add_option('numordersofscatter', 1)
sim_options.add_option('setsun', sim_scan.sun.tolist() )
sim_options.add_option('wfprecision', 0)

sktran_sim = radtran.SasktranHR( atmo_sim, sim_options )
sktran_ret = radtran.SasktranHR( atmo_ret, ret_options )


inst_sim.setup_radtran_model(sktran_sim)
inst_ret.setup_radtran_model(sktran_ret)
inst_sim.calculate_radiance()

#inst_sim.rad[32,0] = np.nan
#inst_sim.rad[20,4] = np.nan


ret = retrieval.RetrievalLimb( inst_sim, inst_ret )
method = lsq.LevenbergMarquardt(maxiter=5, damping=0.1, dampingpower = 10, save_output=True)
ret_species = species.OzoneLimbProfile( atmo_ret )
output = ret.retrieve_species( ret_species, method )
lsq_output = method.output

#make some plots
plt.subplot(1,2,1)
plt.plot(apriori_profile,apriori_alts, color=(0.7,0.4,0.2))
plt.plot(true_profile, apriori_alts, color='black')

for i in range(0,len(lsq_output['profile'])):
    plt.plot(lsq_output['profile'][i],apriori_alts,color=(0.2, 0.4, 0.7,0.5))

plt.plot(lsq_output['profile'][-1],apriori_alts,color=(0.2, 0.4, 0.7))

plt.subplot(1,2,2)

for i in range(8):
    plt.plot( np.reshape(lsq_output['y_meas'],(8,-1))[i,:], ret_species.tanalts, color='black' )

for i in range(0,len(lsq_output['y_ret'])):
    for j in range(8):
        plt.plot( np.reshape(lsq_output['y_ret'][i],(8,-1))[j,:], ret_species.tanalts, color=(0.2, 0.4, 0.7,0.5) )
        
for i in range(8):
        plt.plot( np.reshape(lsq_output['y_ret'][-1],(8,-1))[i,:], ret_species.tanalts, color=(0.2, 0.4, 0.7) )