import numpy as np
from scipy import interpolate


def binned_data(xs, ys, vals, x_grid, y_grid):
    """
    Bins data specified on coordinates.  For each value pair (x,y,val) the closest grid point
    in (x_grid, y_grid) is found.  Grid points with no data are set to np.nan

    Parameters
    ----------
    xs : numpy array
        The x values the data is specified on
    ys : numpy array
        The y values the data is specified on
    vals : numpy array
        The values
    x_grid : numpy array
        The desired x grid to bin on
    y_grid : numpy array
        The desired y grid to bin on

    Returns
    -------
    Binned data : numpy array
        Size (len(y_grid), len(x_grid))
    """
    binned_vals = np.zeros((len(y_grid), len(x_grid)))
    num_in_bin = np.zeros((len(y_grid), len(x_grid)))
    for idx, (x, y) in enumerate(zip(xs, ys)):
        x_idx = np.argmin(np.abs(x-x_grid))
        y_idx = np.argmin(np.abs(y-y_grid))

        binned_vals[y_idx, x_idx] += vals[idx]
        num_in_bin[y_idx, x_idx] += 1

    num_in_bin[num_in_bin == 0] = 1
    binned_vals /= num_in_bin
    binned_vals[binned_vals == 0] = np.nan

    return binned_vals


def smooth(xs, ys, vals, factor=4):
    """
    Smooths data already on a uniform grid.

    Parameters
    ----------
    xs : numpy array
        x values for the input uniform grid
    ys : numpy array
        y values for the input uniform grid
    vals : numpy array
        Size (len(ys), len(xs)) the values to smooth
    factor : scalar
        The grid factor to scale by.  Both the x and y grids are increased
        by this factor.

    Returns
    -------
    new_x : numpy array
        The new x grid the data is specified on.  Size is (len(xs)*factor)
    new_y : numpy array
        The new y grid the data is specified on. Size is (len(ys)*factor)
    grid_vals : numpy array
        The smoothed values.  Shape is (len(new_y), len(new_x))
    """
    new_x = np.linspace(np.min(xs), np.max(xs), len(xs)*factor)
    new_y = np.linspace(np.min(ys), np.max(ys), len(ys)*factor)

    return new_x, new_y, interpolate.interp2d(xs, ys, vals)(new_x, new_y)
