import numpy as np
import scipy.signal as sp
import sasktranif.sasktranif as skif

def convolved_solar_spectrum(wavelengths, psf_wavelengths, psf, spectrum='SAO2010'):
    """
    Parameters
    -----------
    wavelengths: numpy array
        An array of wavelenghts in nanometers at which the solar spectrum will 
        be calculated
    
    psf_wavelengths: numpy array
        An array of wavelengths in nanometers at which the point spread function 
        is specified. The output wavelenghts are linearly interpolated from this
        array.
        
    psf: numpy array
        An array of wavelengths in nanometers defining the spectral point spread 
        function. The PSF is assumed be gaussian with a standard deviation given by psf
        
    spectrum: 'string'
        options argument specifying which solar spectrum to use. Currenlty supported 
        options are 'SAO2010' and ''.
        Default is 'SAO2010'.
    
    Returns
    -------
    convovled_irradiance: numpy.array
        The solar spectrum in units of photons/m^2/sec/nm at the resolution specified
        
        
    Examples
    --------
    Get the solar irradiance
    
    >>> import utility.solar_spectrum as sol
    >>> import numpy as np
    >>> wavel = np.array([300, 400])
    >>> psf_wavel = np.array([250, 500])
    >>> psf = np.array([0.35, 3])
    >>> irr = sol.convolved_solar_spectrum(wavel, psf_wavel, psf)
    """
    #Multiply by the radiance by the solar spectrum
    sun = skif.ISKSolarSpectrum( spectrum )
    num_std = 4
    conv_irrad = np.zeros(len(wavelengths))
            
    for j in range(0,len(wavelengths)):
                
        std = np.interp(wavelengths[j], psf_wavelengths, psf)
        wavel = np.arange( wavelengths[j]-num_std*std, wavelengths[j]+num_std*std, 0.01)
        gaussian = np.exp(-(wavel-wavelengths[j])**2/(std**2 * 2))
                
        #get the solar irradiance at each wavelength
        ok = False
        irradiance = np.zeros(len(wavel))
        for i in range(len(wavel)):
            ok, irradiance[i] = sun.Irradiance( wavel[i] )
                                
        if ok:
            conv_irrad[j] = sp.fftconvolve( irradiance, gaussian, 'valid' )/np.sum(gaussian)
        else:
            conv_irrad[j] = -1
                
    return conv_irrad